#include <stdio.h>
#include <stdlib.h>

#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spPluginP.h>

#include <sp/spDIBitmapP.h>


#define SP_NUM_BMP_TYPE 1
#define SP_BMP_TYPE_DEFAULT 0

typedef struct _spBmpPluginInstance
{
    char type[SP_MAX_LINE];
    int type_index;

    FILE *fp;
    DWORD bfSize, bfOffBits;
    BITMAPV5HEADER header;
    
    spBool write_flag;

    spDIBitmapInfoInternal dibinfo_internal;
    spDIBitmapInfo file_dibinfo;
    unsigned long file_strict_mask;
} *spBmpPluginInstance;

static spBool spIsSupportedByPluginBmp(const char *filename)
{
    spBool supported = SP_FALSE;

    spDebug(80, "spIsSupportedByPluginBmp", "filename = %s\n", filename);
    
    if (spGetBaseName(filename) == filename && filename[0] == '.') {
	if (strcaseeq(filename, ".bmp")) {
	    supported = SP_TRUE;
	}
    } else {
        FILE *fp;
        DWORD bfSize, bfOffBits;
        BITMAPV5HEADER header;
        spDIBitmapInfo dibinfo;

	if ((fp = spOpenFile(filename, "rb")) != NULL) {
            if (spReadBmpFileHeader(&bfSize, &bfOffBits, fp) > 0
                && spReadBmpV5Header(&header, bfOffBits, fp) > 0
                && spGetFileDIBInfoFromBmpV5HeaderToRead(&header, &dibinfo)) {
                supported = SP_TRUE;
            }
	    spCloseFile(fp);
	}
    }
    
    spDebug(80, "spIsSupportedByPluginBmp", "supported = %d\n", supported);
    
    return supported;
}

static spBool spInitPluginBmp(const char *lang)
{
    /*spSetDebugLevel(10);*/

    return SP_TRUE;
}

static spBool spFreePluginBmp(void)
{

    return SP_TRUE;
}

static void *spInitPluginInstanceBmp(const char *lang)
{
    spBmpPluginInstance pinstance;
    
    pinstance = xalloc(1, struct _spBmpPluginInstance);
    memset(pinstance, 0, sizeof(struct _spBmpPluginInstance));
    pinstance->type_index = SP_BMP_TYPE_DEFAULT;
    
    return (void *)pinstance;
}

static spBool spFreePluginInstanceBmp(void *instance)
{
    xfree(instance);
    return SP_TRUE;
}

static spBool spSetPluginFileTypeBmp(void *instance, int index)
{
    spBmpPluginInstance pinstance = (spBmpPluginInstance)instance;

    if (index == 0) {
	pinstance->type_index = index;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

static int spGetPluginFileTypeBmp(void *instance)
{
    spBmpPluginInstance pinstance = (spBmpPluginInstance)instance;
    
    return pinstance->type_index;
}

static const char *spGetPluginBestSuffixBmp(void *instance)
{
    return ".bmp";
}

static unsigned long spGetPluginDIBCapsBmp(void *instance)
{
    return SP_DI_BITMAP_CAPS_DIRECT_ACCESS | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_3 | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_4
	| SP_DI_BITMAP_CAPS_READ | SP_DI_BITMAP_CAPS_WRITE
	| SP_DI_BITMAP_CAPS_UPSIDE_UP | SP_DI_BITMAP_CAPS_UPSIDE_DOWN 
	| SP_DI_BITMAP_CAPS_NONINVERSE_FORMAT | SP_DI_BITMAP_CAPS_INVERSE_FORMAT
	| SP_DI_BITMAP_CAPS_ALPHA | SP_DI_BITMAP_CAPS_LAST_ALPHA;
}

static spBool spClosePluginBmp(void *instance)
{
    spBmpPluginInstance pinstance = (spBmpPluginInstance)instance;

    if (pinstance->fp != NULL) {
	/* close file */
	spCloseFile(pinstance->fp);
	pinstance->fp = NULL;
    }
    
    return SP_TRUE;
}

static spPluginError spOpenPluginBmp(void *instance, const char *filename, const char *mode)
{
    const char *mode2;
    spBmpPluginInstance pinstance = (spBmpPluginInstance)instance;

    if (mode[0] == 'w') {
	mode2 = "wb";
    } else {
	mode2 = "rb";
    }
    
    /* open file */
    if (strnone(filename) || (pinstance->fp = spOpenFile(filename, mode2)) == NULL) {
	spDebug(10, "spOpenPluginBmp", "can't open file: %s\n", filename);
	return SP_PLUGIN_ERROR_OPEN;
    }
    spDebug(80, "spOpenPluginBmp", "filename = %s, mode = %s\n", filename, mode);

    if (mode[0] == 'w') {
	pinstance->write_flag = SP_TRUE;
    } else {
        spDIBitmapInfo dibinfo;
        
        if (spReadBmpFileHeader(&pinstance->bfSize, &pinstance->bfOffBits, pinstance->fp) <= 0
            || spReadBmpV5Header(&pinstance->header, pinstance->bfOffBits, pinstance->fp) <= 0
            || spGetFileDIBInfoFromBmpV5HeaderToRead(&pinstance->header, &dibinfo) == SP_FALSE) {
	    spCloseFile(pinstance->fp);
	    pinstance->fp = NULL;
	    spDebug(10, "spOpenPluginBmp", "can't read bmp header\n");
	    return SP_PLUGIN_ERROR_OPEN;
        }
	
	pinstance->write_flag = SP_FALSE;
    }

    spDebug(80, "spOpenPluginBmp", "done\n");
    
    return SP_PLUGIN_ERROR_SUCCESS;
}

static spBool spSetPluginOtherInfoBmp(void *instance, const char *id, char *data)
{
    return SP_FALSE;
}

static spBool spGetPluginOtherInfoBmp(void *instance, const char *id, char *data)
{
    return SP_FALSE;
}

static spBool spSetPluginFileDIBInfoBmp(void *instance, spDIBitmapInfo *dibinfo, unsigned long *strict_mask)
{
    spBool flag;
    spBmpPluginInstance pinstance = (spBmpPluginInstance)instance;
    
    if (pinstance->write_flag == SP_FALSE) return SP_FALSE;
    
    spDebug(80, "spSetPluginFileDIBInfoBmp", "bit_size = %ld, pixel_format = %lx\n",
	    dibinfo->bit_size, dibinfo->pixel_format);

    flag = spSetFileDIBInfoToBmpV5HeaderToWrite(&pinstance->header, dibinfo, strict_mask);
    
    memcpy(&pinstance->file_dibinfo, dibinfo, sizeof(spDIBitmapInfo));
    pinstance->file_strict_mask = *strict_mask;

    spDebug(100, "spSetPluginFileDIBInfoBmp", "done\n");
    
    return flag;
}

static spBool spSetPluginDIBInfoBmp(void *instance, spDIBitmapInfoInternal *dibinfo)
{
    if (dibinfo->bit_size != 8) {
        return SP_FALSE;
    }
    if ((dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK)
        || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_SINGLE_RGB_MASK)
        || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_FLOAT_MASK)
        || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK)
        || (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK)
        ) {
        spDebug(10, "spSetPluginDIBInfoBmp", "unsupported pixel format: %lx\n", dibinfo->pixel_format);
        return SP_FALSE;
    }

    if (dibinfo->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)  {
        if (dibinfo->pixel_stride != 4) {
            spDebug(10, "spSetPluginDIBInfoBmp", "pixel_stride (%ld) must be 4 for alpha included file\n", dibinfo->pixel_stride);
            return SP_FALSE;
        }
    } else {
        if (!(dibinfo->pixel_stride == 3 || dibinfo->pixel_stride == 4)) {
            spDebug(10, "spSetPluginDIBInfoBmp", "pixel_stride (%ld) must be 3 or 4\n", dibinfo->pixel_stride);
            return SP_FALSE;
        }
    }
        
    spDebug(80, "spSetPluginDIBInfoBmp", "done\n");
    
    return SP_TRUE;
}
    
static spBool spGetPluginFileDIBInfoBmp(void *instance, spDIBitmapInfo *dibinfo)
{
    spBmpPluginInstance pinstance = (spBmpPluginInstance)instance;

    if (pinstance->write_flag) return SP_FALSE;

    spDebug(80, "spGetPluginFileDIBInfoBmp", "in\n");

    return spGetFileDIBInfoFromBmpV5HeaderToRead(&pinstance->header, dibinfo);
}

static spBool spGetPluginDIBInfoBmp(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned long *strict_mask)
{
    spBmpPluginInstance pinstance = (spBmpPluginInstance)instance;

    spDebug(80, "spGetPluginDIBInfoBmp", "in\n");

    return spGetDIBInfoFromBmpV5Header(&pinstance->header, dibinfo, strict_mask, pinstance->write_flag);
}

static spBool spReadPluginDIBDataBmp(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned char *data)
{
    spBmpPluginInstance pinstance = (spBmpPluginInstance)instance;
    
    spDebug(80, "spReadPluginDIBDataBmp", "in\n");

    if (spReadBmpV5Data(&pinstance->header, dibinfo->pixel_format, dibinfo->line_stride,
                        dibinfo->upside_down, data, pinstance->fp) <= 0) {
        return SP_FALSE;
    }
    
    spDebug(80, "spReadPluginDIBDataBmp", "done\n");
    
    return SP_TRUE;
}

static spBool spWritePluginDIBDataBmp(void *instance, spDIBitmapInfoInternal *dibinfo, unsigned char *data)
{
    spBmpPluginInstance pinstance = (spBmpPluginInstance)instance;

    spDebug(80, "spWritePluginDIBDataBmp", "in\n");

    if (spWriteBmpFileHeader(pinstance->bfSize, pinstance->bfOffBits, pinstance->fp) <= 0
        || spWriteBmpV5Header(&pinstance->header, pinstance->bfOffBits, pinstance->fp) <= 0) {
        return SP_FALSE;
    }
    
    if (spWriteBmpV5Data(&pinstance->header, dibinfo->pixel_format, dibinfo->line_stride,
                         dibinfo->upside_down, data, pinstance->fp) <= 0) {
        return SP_FALSE;
    }
    
    spDebug(80, "spWritePluginDIBDataBmp", "done\n");
    
    return SP_TRUE;
}

static const char *sp_bmp_file_type_list[] = {
    "bmp",
    NULL,
};
static const char *sp_bmp_file_desc_list[] = {
    "BMP",
    NULL,
};
static const char *sp_bmp_file_filter_list[] = {
    "*.bmp",
    NULL,
};
    
static spDIBitmapPluginRec sp_dibitmap_plugin_bmp = {
    NULL,
    NULL,

    SP_PLUGIN_DIBITMAP,
    "bmp",
    1,
    SP_PLUGIN_PRIORITY_HIGHEST,
    SP_PLUGIN_CAPS_THREAD_SAFE,
    spInitPluginBmp,
    spFreePluginBmp,
    "BMP I/O Plugin",
    "BMP I/O Plugin Version 0.1\nCopyright(C) Hideki Banno",
    
    spInitPluginInstanceBmp,
    spFreePluginInstanceBmp,
    NULL,
    NULL,

    SP_PLUGIN_DEVICE_FILE,
    sp_bmp_file_type_list,
    sp_bmp_file_desc_list,
    sp_bmp_file_filter_list,
    spIsSupportedByPluginBmp,
    spSetPluginFileTypeBmp,  
    spGetPluginFileTypeBmp,
    0L,
    NULL,
    NULL,
    spGetPluginBestSuffixBmp,
    
    spGetPluginDIBCapsBmp,
	
    spOpenPluginBmp,              
    spClosePluginBmp,
	
    spSetPluginFileDIBInfoBmp,
    spSetPluginDIBInfoBmp,
    spGetPluginFileDIBInfoBmp,
    spGetPluginDIBInfoBmp,
    spSetPluginOtherInfoBmp,   
    spGetPluginOtherInfoBmp,
	
    spReadPluginDIBDataBmp,          
    spWritePluginDIBDataBmp,
	
    NULL,
    NULL,
};

#if 1
spBool spAddBmpDIBitmapPlugin(void)
{
    return spAddPluginRecList((spPluginRec *)&sp_dibitmap_plugin_bmp, "bmp");
}

spBool spRemoveBmpDIBitmapPlugin(void)
{
    return spRemovePluginRecList((spPluginRec *)&sp_dibitmap_plugin_bmp);
}
#else
#include <sp/spPluginMain.h>

spPluginExport spPluginRec *spGetPluginRec(void)
{
    return (spPluginRec *)&sp_dibitmap_plugin_bmp;
}
#endif
