/*
 *	spButton.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spContainer.h>
#include <sp/spToolItem.h>

#include <sp/spFrameP.h>
#include <sp/spMenuItemP.h>
#include <sp/spButtonP.h>

static spParamTable sp_button_param_tables[] = {
    {SppSet, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spButton, button.set), SP_FALSE_STRING},
    {SppDefaultButton, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spButton, button.default_flag), SP_FALSE_STRING},
    {SppCancelButton, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spButton, button.cancel_flag), SP_FALSE_STRING},
    {SppRadioEnd, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spButton, button.radio_end), SP_FALSE_STRING},
    {SppRadioGroupParent, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.subgroup_parent), NULL},
    {SppDestructiveButton, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spButton, button.destructive_flag), SP_FALSE_STRING},
};

spButtonClassRec SpButtonClassRec = {
    /* spObjectClassPart */
    {
	SpButton,
	(spObjectClass)&SpLabelClassRec,
	sizeof(spButtonRec),
	spArraySize(sp_button_param_tables),
	sp_button_param_tables,
	spButtonPartInit,
	spButtonPartFree,
	SP_FALSE,
	NULL,
	NULL,
	spButtonCreate,
	NULL,
	spButtonSetParams,
	spButtonGetParams,
    },
    /* spComponentClassPart */
    {
	SP_TRUE,
	SP_FALSE,
	SP_FALSE,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    },
    /* spPrimitiveClassPart */
    {
	0,
    },
    /* spLabelClassPart */
    {
	0,
    },
    /* spButtonClassPart */
    {
	0,
    },
};

spComponentClass SpButtonClass = (spComponentClass)&SpButtonClassRec;

int spGetMnemonic(const char *label, char *text)
{
    int i, j;
    int len;
    int c = NUL;
#ifndef SP_SUPPORT_MNEMONIC
    int prev_c = NUL;
#endif

    len = (int)strlen(label);
    spDebug(40, "spGetMnemonic", "strlen = %d\n", len);
    
    for (i = 0, j = 0; i < len; i++) {
	if (label[i] == '&' && label[i + 1] == '&') {
	    i++;
	    text[j] = label[i];
	    j++;
	} else if (label[i] == '&') {
	    c = label[i + 1];
#ifndef SP_SUPPORT_MNEMONIC
	    if (prev_c == '(' && c != NUL && label[i + 2] == ')') {
		j--;
		i += 2;
	    }
#endif
	} else {
	    text[j] = label[i];
	    j++;
	}
#ifndef SP_SUPPORT_MNEMONIC
	prev_c = label[i];
#endif
    }
    text[j] = NUL;

#ifdef MNEMONIC_LOWER
    if (isupper(c)) {
	c = tolower(c);
    }
#endif
    
    spDebug(30, NULL, "label = %s, mnemonic = %c, text = %s\n", label, c, text);
    
    return c;
}

spBool spGetNonMnemonicTitle(spComponent component, char *label)
{
    spBool flag;
    
    if (component == NULL) return SP_FALSE;
    
    strcpy(label, "");
    if (!strnone(SpComponentPart(component).title)) {
	spGetMnemonic(SpComponentPart(component).title, label);
    }
    if (strnone(label)) {
	strcpy(label, (!strnone(SpGetName(component)) ? SpGetName(component) : ""));
	flag = SP_FALSE;
    } else {
	flag = SP_TRUE;
    }
    spDebug(80, "spGetNonMnemonicTitle", "label = %s\n", label);

    return flag;
}

void spButtonPartInit(spObject object)
{
    spComponent component = (spComponent)object;
    
    SpComponentPart(component).spacing_flag = SP_TRUE;
    SpButtonPart(component).set = SP_FALSE;
    SpButtonPart(component).default_flag = SP_FALSE;
    SpButtonPart(component).cancel_flag = SP_FALSE;
    SpButtonPart(component).radio_end = SP_FALSE;
    SpButtonPart(component).destructive_flag = SP_FALSE;

    return;
}

void spButtonPartFree(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == SpFramePart(SpGetWindow(component)).default_button) {
	SpFramePart(SpGetWindow(component)).default_button = NULL;
    }
    if (component == SpFramePart(SpGetWindow(component)).cancel_button) {
	SpFramePart(SpGetWindow(component)).cancel_button = NULL;
    }
    
    return;
}

spBool spButtonCreate(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    if (SpButtonPart(component).default_flag == SP_TRUE) {
	SpFramePart(SpGetWindow(component)).default_button = component;
    }
    if (SpButtonPart(component).cancel_flag == SP_TRUE) {
	SpFramePart(SpGetWindow(component)).cancel_button = component;
    }
    
    spPrimitiveSetDefaultSize(component);
    if (SpComponentPart(component).disable_initial_scaling == SP_FALSE)
	spPrimitiveScaleDefaultSizeArch(component);
    spButtonCreateArch(component);
    spShowToolTip(component);
	
    if (spIsToggleButton(component) == SP_TRUE) {
	spSetToggleState(component, SpButtonPart(component).set);
    }
    
    if (SpComponentPart(component).call_func != NULL) {
	spDebug(40, "spButtonCreate", "add callback\n");
	if (spIsToggleButton(component) == SP_TRUE) {
	    spAddCallback(component, SP_VALUE_CHANGED_CALLBACK,
			  SpComponentPart(component).call_func,
			  SpComponentPart(component).call_data);
	} else {
	    spAddCallback(component, SP_ACTIVATE_CALLBACK,
			  SpComponentPart(component).call_func,
			  SpComponentPart(component).call_data);
	}
    }
    
    return SP_TRUE;
}

spBool spButtonSetParams(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    spButtonSetParamsArch(component);

    if (spIsToggleButton(component) == SP_TRUE) {
	spSetToggleStateArch(component);
    }
    
    return SP_TRUE;
}

spBool spButtonGetParams(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;

    if (spIsToggleButton(component) == SP_TRUE) {
	spGetToggleStateArch(component);
    }
    
    return SP_TRUE;
}

spBool spIsButton(spComponent component)
{
    return spIsSubClass(component, SpButton);
}

spBool spIsToggleButton(spComponent component)
{
    if (spIsSubClass(component, SpToggleButton) == SP_TRUE
	|| spIsSubClass(component, SpCheckBox) == SP_TRUE
	|| spIsSubClass(component, SpRadioButton) == SP_TRUE
	|| spIsSubClass(component, SpCheckBoxMenuItem) == SP_TRUE
	|| spIsSubClass(component, SpRadioButtonMenuItem) == SP_TRUE
	|| spIsSubClass(component, SpCheckToolItem) == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spIsRadioButton(spComponent component)
{
    if (spIsSubClass(component, SpRadioButton) == SP_TRUE
	|| spIsSubClass(component, SpRadioButtonMenuItem) == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spComponent spCreatePushButton(spComponent parent, const char *name, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsCreated(parent) == SP_FALSE || spIsContainer(parent) == SP_FALSE)
	return NULL;
    
    va_start(argp, name);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spCreateComponentArg(SpButtonClass, SpPushButton, name, parent, args, num_arg);
}

spComponent spCreateCheckBox(spComponent parent, const char *name, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsCreated(parent) == SP_FALSE || spIsContainer(parent) == SP_FALSE)
	return NULL;
    
    va_start(argp, name);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spCreateComponentArg(SpButtonClass, SpCheckBox, name, parent, args, num_arg);
}

spComponent spCreateRadioButton(spComponent parent, const char *name, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsCreated(parent) == SP_FALSE || spIsContainer(parent) == SP_FALSE)
	return NULL;
    
    va_start(argp, name);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spCreateComponentArg(SpButtonClass, SpRadioButton, name, parent, args, num_arg);
}

spBool spSetToggleState(spComponent component, spBool set)
{
    if (spIsToggleButton(component) == SP_FALSE) {
	if (spIsToggleButton(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return SP_FALSE;
	}
    }

    if (spIsToolItem(component) == SP_TRUE) {
	return spSetToolItemToggleState(component, set);
    } else {
	SpButtonPart(component).set = set;
	spDebug(50, "spSetToggleState", "%s: set: %d\n",
		spGetName(component), SpButtonPart(component).set);
	spSetToggleStateArch(component);
    }

    return SP_TRUE;
}

spBool spGetToggleState(spComponent component, spBool *set)
{
    if (spIsToggleButton(component) == SP_FALSE) {
	if (spIsToggleButton(spGetComponent(component)) == SP_TRUE) {
	    component = spGetComponent(component);
	} else {
	    return SP_FALSE;
	}
    }
    
    if (spIsToolItem(component) == SP_TRUE) {
	return spGetToolItemToggleState(component, set);
    } else {
	spGetToggleStateArch(component);
	if (set != NULL) *set = SpButtonPart(component).set;
    }
    
    return SP_TRUE;
}

spComponent spGetRadioGroupPrev(spComponent component)
{
    spComponent prev_radio = NULL;
    
    if ((SpComponentPart(component).subgroup_prev != NULL)
	&& (SpComponentPart(component).subgroup_id
	    == SpComponentPart(SpComponentPart(component).subgroup_prev).subgroup_id)) {
	prev_radio = SpComponentPart(component).subgroup_prev;
    }

    return prev_radio;
}

spComponent spGetRadioGroupNext(spComponent component)
{
    spComponent next_radio = NULL;
    
    if ((SpComponentPart(component).subgroup_next != NULL)
	&& (SpComponentPart(component).subgroup_id
	    == SpComponentPart(SpComponentPart(component).subgroup_next).subgroup_id)) {
	next_radio = SpComponentPart(component).subgroup_next;
    }

    return next_radio;
}

spComponent spGetRadioGroupFirst(spComponent component)
{
    spComponent temp;
    spComponent first = NULL;
    
    temp = SpComponentPart(component).subgroup_prev;
    while (temp != NULL) {
	if (SpComponentPart(temp).subgroup_id != SpComponentPart(component).subgroup_id) {
	    break;
	}
	first = temp;

	temp = SpComponentPart(temp).subgroup_prev;
    }

    return first;
}

spComponent spGetRadioGroupLast(spComponent component)
{
    spComponent temp;
    spComponent first = NULL;
    
    temp = SpComponentPart(component).subgroup_next;
    while (temp != NULL) {
	if (SpComponentPart(temp).subgroup_id != SpComponentPart(component).subgroup_id) {
	    break;
	}
	first = temp;

	temp = SpComponentPart(temp).subgroup_next;
    }

    return first;
}

void spAddRadioGroup(spComponent component, spBool radio_end)
{
    spComponent parent;
    spComponent child;
    
    if (component == NULL) return;

    if (SpComponentPart(component).subgroup_parent == NULL) {
	parent = SpGetParent(component);
    } else {
	parent = SpComponentPart(component).subgroup_parent;
    }
    SpComponentPart(component).subgroup_parent = parent;
    SpComponentPart(component).subgroup_id = SpComponentPart(parent).subgroup_id;
    
    if (SpComponentPart(parent).subgroup_child == NULL) {
	SpComponentPart(parent).subgroup_child = component;
	SpPrimitivePart(component).index = SP_RADIO_GROUP_START;
    } else {
	child = SpComponentPart(parent).subgroup_child;
	while (child != NULL) {
	    if (SpComponentPart(child).subgroup_next == NULL) {
		SpComponentPart(child).subgroup_next = component;
		SpComponentPart(component).subgroup_prev = child;

		if (radio_end == SP_TRUE) {
		    SpPrimitivePart(component).index = SP_RADIO_GROUP_END;
		    SpComponentPart(parent).subgroup_id++;
		} else if (SpPrimitivePart(child).index == SP_RADIO_GROUP_END) {
		    SpPrimitivePart(component).index = SP_RADIO_GROUP_START;
		} else {
		    SpPrimitivePart(component).index = SP_RADIO_GROUP_MEMBER;
		}
		break;
	    }
	    child = SpComponentPart(child).subgroup_next;
	}
    }
    
    return;
}

spBool spCheckRadioButton(spComponent component)
{
    spComponent temp;
    
    if (spIsCreated(component) == SP_FALSE ||
	spIsRadioButton(component) == SP_FALSE) {
	spDebug(80, "spCheckRadioButton", "not radio button\n");
	return SP_FALSE;
    }

    /* set state for previous radio buttons */
    temp = SpComponentPart(component).subgroup_prev;
    while (temp != NULL) {
	if (SpComponentPart(temp).subgroup_id != SpComponentPart(component).subgroup_id) {
	    break;
	}
	if (spIsRadioButton(temp) == SP_TRUE) {
	    spSetToggleState(temp, SP_FALSE);
	}

	temp = SpComponentPart(temp).subgroup_prev;
    }

    /* set state for next radio buttons */
    temp = SpComponentPart(component).subgroup_next;
    while (temp != NULL) {
	if (SpComponentPart(temp).subgroup_id != SpComponentPart(component).subgroup_id) {
	    break;
	}
	if (spIsRadioButton(temp) == SP_TRUE) {
	    spSetToggleState(temp, SP_FALSE);
	}

	temp = SpComponentPart(temp).subgroup_next;
    }

    spSetToggleState(component, SP_TRUE);
    
    spDebug(80, "spCheckRadioButton", "done\n");
    
    return SP_TRUE;
}

void spCheckRadioButtonCB(spComponent component, void *data)
{
    spCheckRadioButton(component);
    return;
}

int spGetButtonDefaultHeight(spBool margin_flag)
{
    int height = SP_DEFAULT_BUTTON_HEIGHT;
    
    if (margin_flag == SP_TRUE) {
	height += (2 * SP_DEFAULT_MARGIN);
    }
    
    return height;
}
