/*
 *	spComponent.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spTopLevel.h>
#include <sp/spFrame.h>
#include <sp/spDialog.h>
#include <sp/spToolBar.h>
#include <sp/spStatusBar.h>
#include <sp/spCanvas.h>
#include <sp/spDraw.h>
#include <sp/spPluginCanvas.h>

#include <sp/spTopLevelP.h>
#include <sp/spComponentP.h>
#include <sp/spFrameP.h>
#include <sp/spCanvasP.h>
#include <sp/spContainerP.h>
#include <sp/spTabBoxP.h>

#if defined(MACOS)
#pragma import on
#endif
extern spTopLevel sp_toplevel;
#if defined(MACOS)
#pragma import off
#endif

static spComponents sp_components = NULL;

static spParamTable sp_component_param_tables[] = {
    {SppTitle, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.title), ""},
    {SppDescription, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.description), ""},
    {SppHelpPath, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.help_path), ""},
    {SppSenseLevel, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.sense_level), SP_INIT_SENSE_LEVEL_STRING},
    {SppGroupId, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.group_id), SP_INIT_GROUP_ID_STRING},
    {SppUserData, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.user_data), NULL},
    {SppOrientation, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.orientation), SP_VERTICAL_STRING},
    {SppSpacingOn, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.spacing_flag), NULL},
    {SppVisible, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.visible_flag), SP_TRUE_STRING},
    {SppWidth, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.width), "0"},
    {SppHeight, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.height), "0"},
    {SppInitialWidth, SP_CREATE_ACCESS,
	 spOffset(spComponent, component.require_width), "0"},
    {SppInitialHeight, SP_CREATE_ACCESS,
	 spOffset(spComponent, component.require_height), "0"},
    {SppMarginWidth, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.margin_width), "0"},
    {SppMarginHeight, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.margin_height), "0"},
    {SppSpacing, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.spacing), "0"},
    {SppCallbackFunc, SP_CREATE_ACCESS,
	 spOffset(spComponent, component.call_func), NULL},
    {SppCallbackData, SP_CREATE_ACCESS,
	 spOffset(spComponent, component.call_data), NULL},
    {SppPercentageWidth, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.percentage_width), NULL},
    {SppPercentageHeight, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.percentage_height), NULL},
    {SppDisableInitialScaling, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spComponent, component.disable_initial_scaling), NULL},
};

spComponentClassRec SpComponentClassRec = {
    /* spObjectClassPart */
    {
	SpComponent,
	NULL,
	sizeof(spComponentRec),
	spArraySize(sp_component_param_tables),
	sp_component_param_tables,
	spComponentPartInit,
	spComponentPartFree,
	SP_FALSE,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    },
    /* spComponentClassPart */
    {
	SP_FALSE,
	SP_FALSE,
	SP_FALSE,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,

	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    },
};

spComponentClass SpComponentClass = &SpComponentClassRec;

spBool spSetParamsArg(spComponent component, spArg *args, int num_arg)
{
    return spSetObjectParamsArg((spObject)component, args, num_arg);
}

spBool spGetParamsArg(spComponent component, spArg *args, int num_arg)
{
    return spGetObjectParamsArg((spObject)component, args, num_arg);
}
    
spBool spSetParams(spComponent component, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (component == NULL) return SP_FALSE;
    
    va_start(argp, component);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spSetParamsArg(component, args, num_arg);
}
    
spBool spGetParams(spComponent component, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (component == NULL) return SP_FALSE;
    
    va_start(argp, component);
    /*spGetArgs(argp, args, num_arg);*/
    spGetArgsForOutput(argp, args, num_arg);
    va_end(argp);

    return spGetParamsArg(component, args, num_arg);
}

static void spComponentsAlloc(void)
{
    if (sp_components == NULL) {
	sp_components = xalloc(1, struct _spComponents);
	
	sp_components->sense_level_init = SP_FALSE;
	/*sp_components->current_component_id = 0;*/
	sp_components->current_component_id = 1;
	sp_components->destroy_component_id = -1;
	sp_components->child = NULL;
	sp_components->last_child = NULL;
    }

    return;
}

static long spUseCurrentComponentId(void)
{
    long component_id;
    
    component_id = sp_components->current_component_id;
    sp_components->current_component_id++;
    
    return component_id;
}

spComponent spCreateComponentArg(spComponentClass component_class, const char *class_name,
				 const char *name, spComponent parent, 
				 spArg *args, int num_arg)
{
    spComponent component;

    if (component_class == NULL)
	return NULL;
    
    spDebug(100, "spCreateComponentArg", "in\n");
    
    if ((component = (spComponent)spAllocObject((spObjectClass)component_class,
						class_name, name)) == NULL) {
	return NULL;
    }

    spDebug(100, "spCreateComponentArg", "spAllocObject done\n");
    
    spComponentsAlloc();
    if (parent != NULL) {
	SpComponentPart(component).window = SpComponentPart(parent).window;
	if (SpComponentPart(parent).component != NULL) {
	    SpComponentPart(component).parent = SpComponentPart(parent).component;
	} else {
	    SpComponentPart(component).parent = parent;
	}
	SpComponentPart(component).prev_component = SpParentComponentPart(component).last_child;
	if (SpComponentPart(component).prev_component != NULL) 
	    SpGetNextComponent(SpComponentPart(component).prev_component) = component;
	if (SpParentComponentPart(component).child == NULL)
	    SpParentComponentPart(component).child = component;
	SpParentComponentPart(component).last_child = component;
    } else {
	SpComponentPart(component).window = NULL;
	SpComponentPart(component).parent = NULL;
	SpComponentPart(component).prev_component = NULL;
    }
    if (spIsWindow(component) == SP_TRUE && spIsPrimitive(component) == SP_TRUE) {
	spDebug(80, "spCreateComponentArg", "setting for window\n");
	SpComponentPart(component).window = component;
	if (parent == NULL) {
	    SpComponentPart(component).prev_component = sp_components->last_child;
	    if (SpComponentPart(component).prev_component != NULL)
		SpGetNextComponent(SpComponentPart(component).prev_component) = component;
	    if (sp_components->child == NULL)
		sp_components->child = component;
	    sp_components->last_child = component;
	}
    }
    SpComponentPart(component).component = NULL;
    SpComponentPart(component).component_id = spUseCurrentComponentId();
    SpComponentPart(component).group_id = SP_INIT_GROUP_ID;
    SpComponentPart(component).sense_level = SP_INIT_SENSE_LEVEL;

    spDebug(100, "spCreateComponentArg", "original: width = %d, height = %d\n",
	    SpComponentPart(component).width, SpComponentPart(component).height);
    
    if (spSetParamsArg(component, args, num_arg) == SP_FALSE) {
	spDebug(50, "spCreateComponentArg", "spSetParamsArg error\n");
	spFreeComponent(component);
	return NULL;
    }

    spDebug(100, "spCreateComponentArg", "updated: width = %d, height = %d\n",
	    SpComponentPart(component).width, SpComponentPart(component).height);
    spDebug(80, "spCreateComponentArg", "spSetParamsArg done: class_name = %s\n",
	    SpGetObjectClass(component)->object.class_name);
    
    return component;
}

void spFreeComponent(spComponent component)
{
    spFreeObject((spObject)component);
    return;
}

static void destroyComponent(spComponent component, spBool can_quit)
{
    spBool quit_flag = SP_FALSE;
    
    if (spIsCreated(component) == SP_FALSE
	|| spGetDestroyedComponentId() == SpGetComponentId(component)) {
	spDebug(30, "spDestroyComponent", "already destroyed\n");
	return;
    }
    
    sp_components->destroy_component_id = SpGetComponentId(component);
    
    spDebug(30, "spDestroyComponent", "in: %s, id = %ld\n", spGetName(component), sp_components->destroy_component_id);
    
    if (spIsWindow(component) == SP_TRUE) {
	if (spIsLastWindow(component) == SP_TRUE || spIsSubClass(component, SpMainFrame) == SP_TRUE) {
	    quit_flag = SP_TRUE;
	}
	spPopdownWindow(component);
    }
    if (spIsFrame(component) == SP_TRUE
	&& spIsTaskTrayVisible(component) == SP_TRUE) {
	spShowTaskTray(component, SP_FALSE);
    }

    spDestroyObject((spObject)component);
    
    if (can_quit == SP_TRUE && quit_flag == SP_TRUE) {
	spDebug(30, "spDestroyComponent", "last window, call spQuit\n");
	spQuit(0);
    }
    
    spDebug(30, "spDestroyComponent", "done\n");
    
    return;
}

void spDestroyComponent(spComponent component)
{
    destroyComponent(component, SP_TRUE);

    return;
}

static void destroyWindow(spComponent component, spBool can_quit)
{
    if (component == NULL) return;
    
    destroyComponent(SpGetWindow(component), can_quit);
    
    return;
}

void spMapComponent(spComponent component)
{
    spComponentClass component_class;
    
    if (spIsCreated(component) == SP_FALSE) return;
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.map != NULL) {
	    spDebug(40, "spMapComponent", "in\n");
	    component_class->component.map(component);
	    SpComponentPart(component).visible_flag = SP_TRUE;
	    
	    if (SpComponentPart(component).geometry_flag == SP_FALSE
		&& (SpGetComponentClass(component)->component.resize_flag == SP_TRUE
		    || spIsStatusBar(component) == SP_TRUE
		    /*|| spIsToolBar(component) == SP_TRUE*/)) {
		spAdjustComponentSize(SpGetParent(component));
	    }
	    spDebug(40, "spMapComponent", "done\n");
	    break;
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }

    return;
}

void spUnmapComponent(spComponent component)
{
    spComponentClass component_class;
    
    if (spIsCreated(component) == SP_FALSE) return;
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.unmap != NULL) {
	    spDebug(40, "spUnmapComponent", "in\n");
	    SpComponentPart(component).prev_width = -1;
	    SpComponentPart(component).prev_height = -1;
	    SpComponentPart(component).current_width = -1;
	    SpComponentPart(component).current_height = -1;
	    SpComponentPart(component).client_width = -1;
	    SpComponentPart(component).client_height = -1;
	    component_class->component.unmap(component);
	    SpComponentPart(component).visible_flag = SP_FALSE;
	    if (SpComponentPart(component).geometry_flag == SP_FALSE
		&& (SpGetComponentClass(component)->component.resize_flag == SP_TRUE
		    || spIsStatusBar(component) == SP_TRUE
		    /*|| spIsToolBar(component) == SP_TRUE*/)) {
		spAdjustComponentSize(SpGetParent(component));
	    }
	    spDebug(40, "spUnmapComponent", "done\n");
	    break;
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }
    
    return;
}

void spComponentCallbackFunc(spCallback *callback)
{
    spComponent component;
    spComponentClass component_class;
    
    spDebug(100, "spComponentCallbackFunc", "in\n");
    
    if (callback == NULL || callback->component == NULL) {
	spDebug(100, "spComponentCallbackFunc", "invalid input\n");
	return;
    }

    component = callback->component;
    spDebug(100, "spComponentCallbackFunc", "component = %ld\n", (long)component);

    while (1) {
	SpComponentPart(component).call_reason = SP_CR_NONE;
	
	component_class = SpGetComponentClass(component);
	while (component_class != NULL) {
	    if (component_class->component.callback_func != NULL) {
		spDebug(100, "spComponentCallbackFunc", "call class callback_func, component = %ld\n",
			(long)component);
		if ((SpComponentPart(component).call_reason =
		     component_class->component.callback_func(component)) == SP_CR_NONE) {
		    return;
		} else {
		    break;
		}
	    }

	    component_class = (spComponentClass)component_class->object.super_class;
	}

	if (SpGetPropagateComponent(component) != NULL
	    && callback->propagate == SP_TRUE) {
	    component = SpGetPropagateComponent(component);
	    spDebug(100, "spComponentCallbackFunc", "callback propagated to %ld\n", (long)component);
	} else {
	    break;
	}
    }

    spDebug(100, "spComponentCallbackFunc", "call callback->call_func, component = %ld\n", (long)component);
    callback->call_func(component, callback->call_data);
    spDebug(80, "spComponentCallbackFunc", "done\n");
    
    return;
}

spCallbackReason spGetCallbackReason(spComponent component)
{
    if (component == NULL) return SP_CR_NONE;
    
    return SpComponentPart(component).call_reason;
}

spBool spAddCallback(spComponent component, spCallbackType call_type,
		     spCallbackFunc call_func, void *call_data)
{
    spComponentClass component_class;
    
    if (component == NULL) return SP_FALSE;
    
    spDebug(100, "spAddCallback", "call_type = %lx\n", call_type);
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.add_callback != NULL) {
	    return component_class->component.add_callback(component, SP_FALSE, call_type,
							   call_func, call_data);
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }

    return SP_FALSE;
}

spBool spAddPropagateCallback(spComponent component, spCallbackType call_type,
			      spCallbackFunc call_func, void *call_data)
{
    spComponentClass component_class;
    
    if (component == NULL) return SP_FALSE;
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.add_callback != NULL) {
	    return component_class->component.add_callback(component, SP_TRUE, call_type,
							   call_func, call_data);
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }

    return SP_FALSE;
}

spBool spRemoveCallback(spComponent component, spCallbackType call_type,
			spCallbackFunc call_func, void *call_data)
{
    spComponentClass component_class;
    
    if (component == NULL) return SP_FALSE;
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.remove_callback != NULL) {
	    return component_class->component.remove_callback(component, call_type,
							      call_func, call_data);
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }

    return SP_FALSE;
}

spBool spGetCallbackMousePosition(spComponent component, int *x, int *y)
{
    spComponentClass component_class;
    
    if (spIsCreated(component) == SP_FALSE) return SP_FALSE;
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.get_callback_mouse_position != NULL) {
	    return component_class->component.get_callback_mouse_position(component, x, y);
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }

    return SP_FALSE;
}

spBool spGetCallbackWheelValue(spComponent component, double *delta_x, double *delta_y)
{
    if (spIsCreated(component) == SP_FALSE
	|| spIsSubClass(component, SpCanvas) == SP_FALSE) return SP_FALSE;
    
    return spPrimitiveGetCallbackWheelValueArch(component, delta_x, delta_y);
}

spBool spGetCallbackKeySym(spComponent component, spKeySym *key_sym)
{
    spComponentClass component_class;
    
    if (spIsCreated(component) == SP_FALSE
	|| key_sym == NULL) return SP_FALSE;
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.get_callback_key_sym != NULL) {
	    return component_class->component.get_callback_key_sym(component, key_sym);
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }

    return SP_FALSE;
}

int spGetCallbackKeyString(spComponent component,
                           char *buf, int buf_size, spBool *overflow)
{
    spComponentClass component_class;
    
    if (spIsCreated(component) == SP_FALSE
	|| buf == NULL || buf_size <= 0) return -1;
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.get_callback_key_string != NULL) {
	    return component_class->component.get_callback_key_string(component, buf,
								      buf_size, overflow);
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }

    return -1;
}

spBool spGetModifierKeyMask(spComponent component, spModifierMask *mask)
{
    spComponent window;
    
    if (spIsCreated(component) == SP_FALSE || mask == NULL) return SP_FALSE;

    window = SpGetWindow(component);

    if (spIsCreated(window) == SP_FALSE
	|| spIsFrame(window) == SP_FALSE) return SP_FALSE;
    
    if (spIsVisible(window) == SP_TRUE) {
	if (spGetFrameModifierKeyMaskArch(window, mask) == SP_TRUE) {
	    return SP_TRUE;
	}
    }
    
    return SP_FALSE;
}

spBool spOverwritePrompt(spComponent component, const char *filename)
{
    static char buf[SP_MAX_MESSAGE];
    
    sprintf(buf, SP_OVERWRITE_PROMPT_MESSAGE, filename);
    
    if (spCreateMessageBox(component,
			   SP_OVERWRITE_PROMPT_TITLE,
			   buf,
			   SppDialogType, /*SP_QUESTION_DIALOG,*/SP_WARNING_DIALOG,
			   SppMessageBoxButtonType, SP_MB_YES_NO,
			   NULL) == SP_DR_YES) {
	return SP_TRUE;
    }

    return SP_FALSE;
}

spBool spSaveChangesPrompt(spComponent component, const char *filename)
{
    static char buf[SP_MAX_MESSAGE];
    
    sprintf(buf, SP_SAVE_CHANGES_PROMPT_MESSAGE, filename);
    
    if (spCreateMessageBox(component,
			   SP_SAVE_CHANGES_PROMPT_TITLE,
			   buf,
			   SppDialogType, SP_QUESTION_DIALOG,
			   SppMessageBoxButtonType, SP_MB_YES_NO,
			   NULL) == SP_DR_YES) {
	return SP_TRUE;
    }

    return SP_FALSE;
}

spBool spQuitPrompt(spComponent component)
{
    spBool flag = SP_TRUE;

    spDebug(50, "spQuitPrompt", "exit_prompt = %d\n",
	    SpTopLevelPart(sp_toplevel).exit_prompt);
    
    if (SpTopLevelPart(sp_toplevel).exit_prompt == SP_TRUE) {
	if (spCreateMessageBox(/*SpGetWindow(component)*/NULL, SP_QUIT_QUESTION_TITLE,
			       SP_QUIT_QUESTION_MESSAGE,
			       SppDialogType, SP_QUESTION_DIALOG,
			       SppMessageBoxButtonType, SP_MB_YES_NO,
			       NULL) == SP_DR_YES) {
	    flag = SP_TRUE;
	} else {
	    flag = SP_FALSE;
	}
    }

    spDebug(10, "spQuitPrompt", "flag = %d\n", flag);
    
    return flag;
}

void spQuitCB(spComponent component, spBool no_prompt)
{
    spDebug(10, "spQuitCB", "in\n");

    if (no_prompt == SP_TRUE
	|| spQuitPrompt(component) == SP_TRUE) {
	destroyWindow(component, SP_FALSE);
	spDebug(10, "spQuitCB", "quit\n");
	spQuit(0);
    }
    
    spDebug(10, "spQuitCB", "not quit\n");
    
    return;
}

spBool spSetComponentSize(spComponent component, int width, int height)
{
    spComponentClass component_class;
    
    if (spIsCreated(component) == SP_FALSE) return SP_FALSE;

    spDebug(80, "spSetComponentSize", "width = %d, height = %d\n", width, height);
    
    if (width <= 0 && height <= 0) return SP_FALSE;
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.set_size != NULL) {
	    if (width <= 0) {
		width = SpComponentPart(component).current_width;
	    }
	    if (height <= 0) {
		height = SpComponentPart(component).current_height;
	    }
	    SpComponentPart(component).client_width = -1;
	    SpComponentPart(component).client_height = -1;
	    spDebug(80, "spSetComponentSize", "call set_size: width = %d, height = %d\n", width, height);
	    if (component_class->component.set_size(component, width, height) == SP_TRUE) {
		if (spIsWindow(component) == SP_TRUE) {
		    SpComponentPart(component).prev_width = -1;
		    SpComponentPart(component).prev_height = -1;
		} else {
		    SpComponentPart(component).prev_width = SpComponentPart(component).current_width;
		    SpComponentPart(component).prev_height = SpComponentPart(component).current_height;
		}
		SpComponentPart(component).current_width = width;
		SpComponentPart(component).current_height = height;

		if (spIsFrame(component) == SP_TRUE) {
		    spDebug(80, "spSetComponentSize", "for frame: call spAdjustComponentSize\n");
		    spAdjustComponentSize(component);
		}
		spDebug(80, "spSetComponentSize", "done\n");
		return SP_TRUE;
	    }
	    break;
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }
    
    spDebug(50, "spSetComponentSize", "failed\n");
    
    return SP_FALSE;
}

spBool spSetSize(spComponent component, int width, int height)
{
    return spSetComponentSize(component, width, height);
}

spBool spSetRequireSize(spComponent component, int width, int height)
{
    if (spIsCreated(component) == SP_FALSE) return SP_FALSE;

    SpComponentPart(component).require_width = width;
    SpComponentPart(component).require_height = height;

    return SP_TRUE;
}

spBool spGetComponentSize(spComponent component, int *width, int *height)
{
    spComponentClass component_class;
    
    if (spIsCreated(component) == SP_FALSE) return SP_FALSE;

    spDebug(30, "spGetComponentSize", "current width = %d, current height = %d\n",
	    SpComponentPart(component).current_width, SpComponentPart(component).current_height);
    
    if (SpComponentPart(component).current_width >= 0
	&& SpComponentPart(component).current_height >= 0) {
	if (width != NULL) *width = SpComponentPart(component).current_width;
	if (height != NULL) *height = SpComponentPart(component).current_height;
	return SP_TRUE;
    }
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.get_size != NULL) {
	    if (component_class->component.get_size(component,
						    &SpComponentPart(component).current_width,
						    &SpComponentPart(component).current_height) == SP_TRUE) {
		if (width != NULL) *width = SpComponentPart(component).current_width;
		if (height != NULL) *height = SpComponentPart(component).current_height;
		return SP_TRUE;
	    }
	    break;
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }
    
#if 1
    SpComponentPart(component).current_width = -1;
    SpComponentPart(component).current_height = -1;
#endif
    
    return SP_FALSE;
}

spBool spGetClientSize(spComponent component, int *width, int *height)
{
    spComponentClass component_class;
    
    if (spIsCreated(component) == SP_FALSE) return SP_FALSE;

    spDebug(30, "spGetClientSize", "%s: client_width = %d, client_height = %d\n",
	    SpGetClassName(component),
	    SpComponentPart(component).client_width,
	    SpComponentPart(component).client_height);
    
    if (SpComponentPart(component).client_width >= 0
	&& SpComponentPart(component).client_height >= 0) {
	if (width != NULL) *width = SpComponentPart(component).client_width;
	if (height != NULL) *height = SpComponentPart(component).client_height;
	return SP_TRUE;
    }
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.get_client_size != NULL) {
	    if (component_class->component.get_client_size(component,
							   &SpComponentPart(component).client_width,
							   &SpComponentPart(component).client_height) == SP_TRUE) {
		spDebug(30, "spGetClientSize", "%s: get_client_size: width = %d, height = %d\n",
			SpGetClassName(component),
			SpComponentPart(component).client_width,
			SpComponentPart(component).client_height);
    
		if (spIsWindow(component) == SP_TRUE) {
		    SpComponentPart(component).current_width = SpComponentPart(component).client_width;
		    SpComponentPart(component).current_height = SpComponentPart(component).client_height;
		}
		
		if (width != NULL) *width = SpComponentPart(component).client_width;
		if (height != NULL) *height = SpComponentPart(component).client_height;
		return SP_TRUE;
	    }
	    break;
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }

#if 1
    spDebug(30, "spGetClientSize", "%s: client_size reset\n", SpGetClassName(component));
    
    SpComponentPart(component).client_width = -1;
    SpComponentPart(component).client_height = -1;
#endif
    
    return SP_FALSE;
}

spBool spGetSize(spComponent component, int *width, int *height)
{
    spDebug(80, "spGetSize", "in: %s, container_flag = %d, create_flag = %d\n",
	    spGetName(component), SpGetComponentClass(component)->component.container_flag,
	    SpObjectPart(component).create_flag);
    
    if (spIsContainer(component) == SP_TRUE || spIsDrawable(component) == SP_TRUE
	|| spIsPluginCanvas(component) == SP_TRUE) {
	return spGetClientSize(component, width, height);
    } else {
	return spGetComponentSize(component, width, height);
    }
}

spBool spSetSensitive(spComponent component, spBool flag)
{
    spComponentClass component_class;
    
    if (spIsCreated(component) == SP_FALSE) return SP_FALSE;
    
    component_class = SpGetComponentClass(component);
    
    while (component_class != NULL) {
	if (component_class->component.set_sensitive != NULL) {
	    SpComponentPart(component).sensitive_flag = flag;
	    return component_class->component.set_sensitive(component, flag);
	}

	component_class = (spComponentClass)component_class->object.super_class;
    }

    return SP_FALSE;
}

static void setSenseLevel(spComponent component, long group_id, long level, spBool init_flag)
{
    spComponent next;

    if (spIsCreated(component) == SP_FALSE) return;

    if ((group_id < 0 && SpComponentPart(component).group_id >= 0)
	|| SpComponentPart(component).group_id == group_id) {
	spDebug(80, "setSenseLevel", "%s: component group_id = %ld, group_id = %ld\n",
		spGetName(component), SpComponentPart(component).group_id, group_id);
	spDebug(80, "setSenseLevel", "component level = %ld, level = %ld\n", SpComponentPart(component).sense_level, level);
    
	SpComponentPart(component).current_sense_level = level;
	if (SpComponentPart(component).sense_level <= level) {
	    if (init_flag == SP_FALSE) {
		spSetSensitive(component, SP_TRUE);
	    }
	} else {
	    spSetSensitive(component, SP_FALSE);
	}
    }
    
    next = SpGetChild(component);
    while (next != NULL) {
	/* set child sense level */
	setSenseLevel(next, group_id, level, init_flag);
	next = SpGetNextComponent(next);
    }
	
    spDebug(100, "setSenseLevel", "%s: done\n", spGetName(component));
    
    return;
}

spBool spSetWindowSenseLevel(spComponent component, long group_id, long level)
{
    if (spIsCreated(component) == SP_FALSE) {
	return SP_FALSE;
    }

    level = MAX(0, level);

    setSenseLevel(SpGetWindow(component), group_id, level, SP_FALSE);
    
    return SP_TRUE;
}

static spBool setCurrentSenseLevel(long group_id, long level, spBool init_flag)
{
    spComponent next;
    
    if (sp_components == NULL)
	return SP_FALSE;

    spDebug(80, "setCurrentSenseLevel", "group = %ld, level = %ld\n", group_id, level);
    
    next = sp_components->child;
    while (next != NULL) {
	if (init_flag == SP_FALSE || SpComponentPart(next).current_sense_level < 0) {
	    setSenseLevel(SpGetWindow(next), group_id, MAX(level, 0), init_flag);
	    /*spSetWindowSenseLevel(next, group_id, level);*/
	}
	    
	next = SpGetNextComponent(next);
    }
    if (init_flag == SP_TRUE) {
	sp_components->sense_level_init = SP_TRUE;
    }

    spDebug(80, "setCurrentSenseLevel", "done\n");
    
    return SP_TRUE;
}

spBool spSetCurrentSenseLevel(long group_id, long level)
{
    return setCurrentSenseLevel(group_id, level, SP_FALSE);
}

static spBool getSenseLevel(spComponent component, long group_id, long *level)
{
    spComponent next;

    if (spIsCreated(component) == SP_FALSE) {
	return SP_FALSE;
    }

    if (SpComponentPart(component).group_id == group_id) {
	*level = SpComponentPart(component).current_sense_level;
	return SP_TRUE;
    }
	
    next = SpGetChild(component);
    while (next != NULL) {
	/* get child sense level */
	if (getSenseLevel(SpComponentPart(next).child, group_id, level) == SP_TRUE) {
	    return SP_TRUE;
	}
	next = SpGetNextComponent(next);
    }

    return SP_FALSE;
}

spBool spGetWindowSenseLevel(spComponent component, long group_id, long *level)
{
    if (spIsCreated(component) == SP_FALSE || level == NULL) {
	return SP_FALSE;
    }

    return getSenseLevel(SpGetWindow(component), group_id, level);
}

spBool spSetSenseLevel(spComponent component, long level)
{
    if (component != NULL) {
	SpComponentPart(component).sense_level = level;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spGetSenseLevel(spComponent component, long *level)
{
    if (component != NULL && level != NULL) {
	*level = SpComponentPart(component).sense_level;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

void spSetInitSenseLevel(void)
{
    if (sp_components != NULL) {
	setCurrentSenseLevel(-1, 0, SP_TRUE);
	spDebug(80, "spSetInitSenseLevel", "done\n");
    }

    return;
}

spBool spSetUserData(spComponent component, void *data)
{
    if (component != NULL) {
	SpComponentPart(component).user_data = data;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

void *spGetUserData(spComponent component)
{
    if (component != NULL) {
	return SpComponentPart(component).user_data;
    } else {
	return NULL;
    }
}

const char *spGetName(spComponent component)
{
    if (component != NULL) {
#if 1
	spComponent parent;
	char *name;

	name = NULL;
	parent = component;
	while (parent != NULL) {
	    name = SpGetName(parent);
	    if (name != NULL && name[0] != '_') {
		break;
	    }
	    parent = SpGetParent(parent);
	}

	if (name != NULL && name[0] == '_') {
	    name = SpGetName(component);
	}
	
	return name;
#else 
	return SpGetName(component);
#endif
    } else {
	return NULL;
    }
}

const char *spGetTitle(spComponent component)
{
    const char *title;
    static const char *default_title = "";
    
    if (component != NULL) {
	title = SpGetTitle(component);
	if (strnone(title)) {
	    title = SpGetName(component);
	    if (strnone(title)) {
		title = default_title;
	    }
	}
    } else {
	title = NULL;
    }

    return title;
}

long spGetComponentId(spComponent component)
{
    if (component != NULL) {
	return SpGetComponentId(component);
    } else {
	return -1;
    }
}

spComponent spGetComponent(spComponent component)
{
    if (component != NULL) {
	return SpComponentPart(component).component;
    } else {
	return NULL;
    }
}

spComponent spGetParent(spComponent component)
{
    if (component != NULL) {
	return SpGetParent(component);
    } else {
	return NULL;
    }
}

spComponent spGetChild(spComponent component)
{
    if (component != NULL) {
	return SpGetChild(component);
    } else if (sp_components != NULL) {
	return sp_components->child;
    } else {
	return NULL;
    }
}

spComponent spGetLastChild(spComponent component)
{
    if (component != NULL) {
	return SpGetLastChild(component);
    } else if (sp_components != NULL) {
	return sp_components->last_child;
    } else {
	return NULL;
    }
}

spComponent spGetNextComponent(spComponent component)
{
    if (component != NULL) {
	return SpGetNextComponent(component);
    } else {
	return NULL;
    }
}

spComponent spGetPrevComponent(spComponent component)
{
    if (component != NULL) {
	return SpGetPrevComponent(component);
    } else {
	return NULL;
    }
}

spComponent spGetPrevVisibleComponent(spComponent component)
{
    component = SpGetPrevComponent(component);
    
    while (component != NULL) {
	if (SpComponentPart(component).visible_flag || spIsTabItem(component) == SP_TRUE) {
	    return component;
	}
	
	component = SpGetPrevComponent(component);
    }

    return NULL;
}

spComponent spGetNextVisibleComponent(spComponent component)
{
    component = SpGetNextComponent(component);
    
    while (component != NULL) {
	if (SpComponentPart(component).visible_flag || spIsTabItem(component) == SP_TRUE) {
	    return component;
	}
	
	component = SpGetNextComponent(component);
    }

    return NULL;
}

spComponent spGetWindow(spComponent component)
{
    if (component != NULL) {
	return SpGetWindow(component);
    } else {
	return NULL;
    }
}

void *spGetArchWindow(spComponent component)
{
    spComponent window;

    if (component == NULL) return NULL;

    window = SpGetWindow(component);
    
    if (spIsFrame(window) == SP_TRUE) {
	return spGetArchFrame(window);
    } else {
	return NULL;
    }
}

long spGetDestroyedComponentId(void)
{
    if (sp_components == NULL) return 0;
    
    return sp_components->destroy_component_id;
}

spBool spIsCreated(spComponent component)
{
    return spIsObjectCreated((spObject)component);
}

spBool spIsVisible(spComponent component)
{
    if (component == NULL) return SP_FALSE;
    
    return SpComponentPart(component).visible_flag;
}

spBool spIsSensitive(spComponent component)
{
    if (component == NULL) return SP_FALSE;
    
    return SpComponentPart(component).sensitive_flag;
}

spBool spIsResizable(spComponent component)
{
    if (component == NULL) return SP_FALSE;

#if 1
    if (spIsTabItem(component) == SP_TRUE) {
	return SP_TRUE;
    }
#endif

    if (SpGetComponentClass(component)->component.resize_flag == SP_TRUE
	&& SpComponentPart(component).visible_flag == SP_TRUE) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spIsSpacingOn(spComponent component)
{
    if (component == NULL) return SP_FALSE;
    
    return SpComponentPart(component).spacing_flag;
}

spBool spEqClass(spComponent component, const char *class_name)
{
    spBool flag = SP_FALSE;

    if (component == NULL) return SP_FALSE;

    if (!strnone(SpObjectPart(component).class_name)) {
	if (streq(SpObjectPart(component).class_name, class_name)) {
	    flag = SP_TRUE;
	}
    } else {
	if (streq(SpGetComponentClass(component)->object.class_name, class_name)) {
	    flag = SP_TRUE;
	}
    }
    
    return flag;
}

spBool spIsSubClass(spComponent component, const char *class_name)
{
    spBool flag = SP_FALSE;
    spComponentClass component_class;
    
    if (component == NULL) return SP_FALSE;

    if (streq(SpObjectPart(component).class_name, class_name)) {
	flag = SP_TRUE;
    } else if (streq(SpGetComponentClass(component)->object.class_name, class_name)) {
	flag = SP_TRUE;
    }

    if (flag == SP_FALSE) {
	component_class = SpGetComponentClass(component);
    
	while (component_class != NULL) {
	    if (streq(component_class->object.class_name, class_name)) {
		flag = SP_TRUE;
		break;
	    }
	    
	    component_class = (spComponentClass)component_class->object.super_class;
	}
    }

    return flag;
}

spBool spIsComponentType(spComponent component, const char *class_name)
{
    spBool flag = SP_FALSE;
    spComponentClass component_class;
    
    if (component == NULL) return SP_FALSE;

    flag = spIsSubClass(component, class_name);

    if (flag == SP_FALSE) {
	component_class = SpGetComponentClass(component);
    
	while (component_class != NULL) {
	    if (component_class->component.is_component_type != NULL) {
		flag = component_class->component.is_component_type(component, class_name);
	    }
	    
	    if (flag == SP_TRUE) {
		break;
	    }
	    
	    component_class = (spComponentClass)component_class->object.super_class;
	}
    }

    return flag;
}

void spComponentPartInit(spObject object)
{
    spComponent component = (spComponent)object;
    
    SpComponentPart(component).component_id = 0;
    SpComponentPart(component).current_sense_level = -1;
    SpComponentPart(component).component = NULL;
    SpComponentPart(component).window = NULL;
    SpComponentPart(component).parent = NULL;
    SpComponentPart(component).child = NULL;
    SpComponentPart(component).last_child = NULL;
    SpComponentPart(component).next_component = NULL;
    SpComponentPart(component).prev_component = NULL;
    SpComponentPart(component).propagate_component = NULL;
    SpComponentPart(component).subgroup_child = NULL;
    SpComponentPart(component).subgroup_id = 0;
    SpComponentPart(component).subgroup_parent = NULL;
    SpComponentPart(component).subgroup_next = NULL;
    SpComponentPart(component).subgroup_prev = NULL;
    SpComponentPart(component).x = 0;
    SpComponentPart(component).y = 0;
    SpComponentPart(component).current_width = -1;
    SpComponentPart(component).current_height = -1;
    SpComponentPart(component).prev_width = -1;
    SpComponentPart(component).prev_height = -1;
    SpComponentPart(component).require_width = 0;
    SpComponentPart(component).require_height = 0;
    SpComponentPart(component).client_width = -1;
    SpComponentPart(component).client_height = -1;
    SpComponentPart(component).percentage_width = -1;
    SpComponentPart(component).percentage_height = -1;
    SpComponentPart(component).border_width = 0;
    SpComponentPart(component).top_offset = 0;
    SpComponentPart(component).bottom_offset = 0;
    SpComponentPart(component).left_offset = 0;
    SpComponentPart(component).right_offset = 0;
    SpComponentPart(component).margin_bottom = 0;
    SpComponentPart(component).geometry_flag = SP_FALSE;
    SpComponentPart(component).spacing_flag = SP_FALSE;
    SpComponentPart(component).visible_flag = SP_TRUE;
    SpComponentPart(component).sensitive_flag = SP_TRUE;
    SpComponentPart(component).focus_flag = SP_FALSE;
    SpComponentPart(component).disable_initial_scaling = SP_FALSE;
    SpComponentPart(component).call_reason = SP_CR_NONE;
    
    SpComponentPart(component).title = NULL;
    SpComponentPart(component).description = NULL;
    SpComponentPart(component).help_path = NULL;
    SpComponentPart(component).sense_level = SP_INIT_SENSE_LEVEL;
    SpComponentPart(component).group_id = SP_INIT_GROUP_ID;
    SpComponentPart(component).user_data = NULL;
    SpComponentPart(component).orientation = SP_VERTICAL;
    SpComponentPart(component).width = 0;
    SpComponentPart(component).height = 0;
    SpComponentPart(component).margin_width = 0;
    SpComponentPart(component).margin_height = 0;
    SpComponentPart(component).spacing = 0;
    SpComponentPart(component).call_func = NULL;
    SpComponentPart(component).call_data = NULL;

    return;
}

void spComponentPartFree(spObject object)
{
    spComponent component = (spComponent)object;
    spComponent child, next_child;

    if (SpComponentPart(component).parent != NULL
	&& SpParentComponentPart(component).subgroup_child == component) {
	SpParentComponentPart(component).subgroup_child = SpComponentPart(component).subgroup_next;
    }
    if (SpComponentPart(component).subgroup_prev != NULL) {
	SpComponentPart(SpComponentPart(component).subgroup_prev).subgroup_next
	    = SpComponentPart(component).subgroup_next;
    }
    if (SpComponentPart(component).subgroup_next != NULL) {
	SpComponentPart(SpComponentPart(component).subgroup_next).subgroup_prev
	    = SpComponentPart(component).subgroup_prev;
    }

    child = SpComponentPart(component).child;
    while (child != NULL) {
	next_child = SpGetNextComponent(child);
	if (next_child != NULL) SpGetPrevComponent(next_child) = NULL;
	spFreeComponent(child);
	child = next_child;
    }
    spDebug(60, "spComponentPartFree", "free children done\n");
    
    if (SpComponentPart(component).parent != NULL
	&& SpParentComponentPart(component).child == component) {
	SpParentComponentPart(component).child = SpGetNextComponent(component);
    } else if (sp_components->child == component) {
	sp_components->child = SpGetNextComponent(component);
    }
    if (SpComponentPart(component).parent != NULL
	&& SpParentComponentPart(component).last_child == component) {
	SpParentComponentPart(component).last_child = SpGetPrevComponent(component);
    } else if (sp_components->last_child == component) {
	sp_components->last_child = SpGetPrevComponent(component);
    }
    if (SpGetPrevComponent(component) != NULL) {
	SpGetNextComponent(SpGetPrevComponent(component)) = SpGetNextComponent(component);
    }
    if (SpGetNextComponent(component) != NULL) {
	SpGetPrevComponent(SpGetNextComponent(component)) = SpGetPrevComponent(component);
    }
    
    if (SpComponentPart(component).title != NULL) {
	xfree(SpComponentPart(component).title);
	SpComponentPart(component).title = NULL;
    }
    if (SpComponentPart(component).description != NULL) {
	xfree(SpComponentPart(component).description);
	SpComponentPart(component).description = NULL;
    }
    if (SpComponentPart(component).help_path != NULL) {
	xfree(SpComponentPart(component).help_path);
	SpComponentPart(component).help_path = NULL;
    }
    
    spDebug(80, "spComponentPartFree", "done\n");
	
    return;
}

#if !defined(ANDROID) /*&& !defined(IPHONE)*/
static void setComponentsSize(spComponent component)
{
    spComponent child;
    
    if (spIsCreated(component) == SP_FALSE
	|| (spIsResizable(component) == SP_FALSE && spIsStatusBar(component) == SP_FALSE)
	|| SpComponentPart(component).geometry_flag == SP_TRUE) return;

    SpComponentPart(component).geometry_flag = SP_TRUE;
    SpComponentPart(component).client_width = -1;
    SpComponentPart(component).client_height = -1;
    
    spDebug(60, "setComponentsSize", "%s: width = %d, height = %d\n",
	    spGetTitle(component), SpComponentPart(component).width, SpComponentPart(component).height);
    spDebug(50, "setComponentsSize", "%s: current size = (%d, %d), previous size = (%d, %d)\n",
	    spGetTitle(component), SpComponentPart(component).current_width, SpComponentPart(component).current_height,
	    SpComponentPart(component).prev_width, SpComponentPart(component).prev_height);

    if (spIsFrame(component) == SP_FALSE
	|| ((SpComponentPart(component).prev_width < 0
	     && SpComponentPart(component).prev_height < 0)
	    && SpFramePart(component).window_type != SP_FULLSCREEN_WINDOW)) {
	spDebug(80, "setComponentsSize", "%s: call spSetSize\n", spGetTitle(component));
	spSetSize(component, SpComponentPart(component).current_width,
		  SpComponentPart(component).current_height);
	if (spIsTabBox(component) == SP_TRUE) {
	    if (SpComponentPart(component).width > 0) {
		SpComponentPart(component).width = SpComponentPart(component).current_width;
	    }
	    if (SpComponentPart(component).height > 0) {
		SpComponentPart(component).height = SpComponentPart(component).current_height;
	    }
	}
    }
    
    child = SpComponentPart(component).child;
    while (child != NULL) {
	setComponentsSize(child);
	child = SpGetNextComponent(child);
    }
    
    SpComponentPart(component).geometry_flag = SP_FALSE;

    spDebug(80, "setComponentsSize", "%s: done\n", spGetTitle(component));
    
    return;
}

static spBool isRequireWidthValid(spComponent component)
{
    spComponent window;

    if (SpComponentPart(component).require_width == 0) {
	return SP_FALSE;
    }

    window = SpGetWindow(component);
    if (window == NULL || (SpComponentPart(window).prev_width >= 0)) {
	return SP_FALSE;
    } else {
	return SP_TRUE;
    }
}

static spBool isRequireHeightValid(spComponent component)
{
    spComponent window;

    if (SpComponentPart(component).require_height == 0) {
	return SP_FALSE;
    }

    window = SpGetWindow(component);
    if (window == NULL || (SpComponentPart(window).prev_height >= 0)) {
	return SP_FALSE;
    } else {
	return SP_TRUE;
    }
}

static spBool isMarginRequired(spComponent component, spComponent child)
{
    spBool flag;
    spComponent parent;
    
    if (spIsSpacingOn(child) == SP_TRUE) {
	if (spIsSubClass(component, SpCanvas) == SP_TRUE) {
	    flag = SpCanvasPart(component).border_on;
	} else if (spIsSubClass(component, SpContainer) == SP_TRUE) {
	    parent = SpGetParent(component);
	    if (parent == NULL
		|| spIsSubClass(parent, SpTabBox) == SP_TRUE) {
		flag = SP_TRUE;
	    } else if (SpContainerPart(component).border_on == SP_TRUE
		       || spIsSpacingOn(component) == SP_FALSE) {
		flag = SP_TRUE;
	    } else {
		flag = SP_FALSE;
	    }
	} else if (spIsSpacingOn(component) == SP_FALSE
		   || spIsWindow(component) == SP_TRUE) {
	    flag = SP_TRUE;
	} else {
	    flag = SP_FALSE;
	}
    } else {
	flag = SP_FALSE;
    }

    spDebug(100, "isMarginRequired", "component = %s, flag = %d\n",
	    SpGetClassName(component), flag);
    
    return flag;
}

static int getOptimumHeight(spComponent component)
{
    int flag;
    int h, hc;
    int spacing;
    int margin;
    int height = 0;
    int total_height = -1;
    spComponent child = NULL, prev_child = NULL;

    if (spIsResizable(component) == SP_FALSE) return 0;
    
    child = SpGetChild(component);
    
    if (child != NULL) {
	if (spIsTabBox(component) == SP_FALSE && SpComponentPart(component).orientation == SP_VERTICAL) {
	    flag = 0;
	    while (1) {
		if (spIsResizable(child) == SP_TRUE) {
		    if (spIsToolBar(child) == SP_FALSE && flag == 0) {
			height += SpComponentPart(component).border_width;
			
			if (isMarginRequired(component, child) == SP_TRUE) {
			    height += MAX(SpComponentPart(component).margin_height,
					  SpComponentPart(child).top_offset);
			} else if (SpComponentPart(child).top_offset > 0) {
			    height += SpComponentPart(child).top_offset;
			}
			flag = 1;
		    }

		    SpComponentPart(child).y = height;

		    if (SpComponentPart(child).height <= 0
			|| (SpComponentPart(child).percentage_height > 0 && SpComponentPart(child).require_height)) {
			if (total_height > 0) {
			    if (SpComponentPart(child).percentage_height > 0 && SpComponentPart(child).require_height) {
			        SpComponentPart(child).current_height
				    = (int)spRound((double)total_height * (double)SpComponentPart(child).percentage_height / 100.0);
				if (prev_child == NULL) {
				    SpComponentPart(child).current_height -= SpComponentPart(child).y;
				}
				if (spIsSpacingOn(child) == SP_TRUE) {
				    if (spIsSpacingOn(prev_child) == SP_TRUE) {
					SpComponentPart(child).current_height -= SpComponentPart(component).spacing / 2;
				    }
				    if (spIsSpacingOn(spGetNextVisibleComponent(child)) == SP_TRUE) {
					SpComponentPart(child).current_height -= SpComponentPart(component).spacing / 2;
				    }
				}
			    } else {
				SpComponentPart(child).current_height
				    = total_height - SpComponentPart(child).y + SpComponentPart(child).height;
			    }
			    
			    SpComponentPart(child).current_height -= SpComponentPart(component).border_width;
			    SpComponentPart(child).current_height -= SpComponentPart(component).margin_bottom;

			    if (spGetNextVisibleComponent(child) == NULL) {
				margin = 0;
				
				if (isMarginRequired(component, child) == SP_TRUE) {
				    margin = MAX(SpComponentPart(component).margin_height, SpComponentPart(child).bottom_offset);
				} else if (spIsCreated(child) == SP_TRUE && SpComponentPart(child).bottom_offset > 0) {
				    margin = SpComponentPart(child).bottom_offset;
				}
				if (margin > -SpComponentPart(child).height) {
				    SpComponentPart(child).current_height -= margin - MAX(-SpComponentPart(child).height, 0);
				}
			    }
			    
			    height += SpComponentPart(child).current_height;

			    if (spIsContainer(child) == SP_TRUE
				&& spIsCanvas(child) == SP_FALSE) {
				getOptimumHeight(child);
			    }
			} else {
			    spacing = 0;
			    if (SpComponentPart(child).percentage_height > 0 && SpComponentPart(child).require_height) {
				hc = (int)spRound((double)SpComponentPart(component).current_height * (double)SpComponentPart(child).percentage_height / 100.0);

				if (prev_child == NULL) {
				    hc -= SpComponentPart(child).y;
				}
				if (spIsSpacingOn(child) == SP_TRUE) {
				    if (spIsSpacingOn(prev_child) == SP_TRUE) {
					spacing += SpComponentPart(component).spacing / 2;
				    }
				    if (spIsSpacingOn(spGetNextVisibleComponent(child)) == SP_TRUE) {
					spacing += SpComponentPart(component).spacing / 2;
				    }
				}
			    } else {
				hc = SpComponentPart(component).current_height
				    - SpComponentPart(child).y + SpComponentPart(child).height;
			    }

			    spacing += SpComponentPart(component).border_width;
			    spacing += SpComponentPart(component).margin_bottom;
			    
			    if (spGetNextVisibleComponent(child) == NULL) {
				margin = 0;
				
				if (isMarginRequired(component, child) == SP_TRUE) {
				    margin = MAX(SpComponentPart(component).margin_height, SpComponentPart(child).bottom_offset);
				} else if (spIsCreated(child) == SP_TRUE && SpComponentPart(child).bottom_offset > 0) {
				    margin = SpComponentPart(child).bottom_offset;
				}
				if (margin > -SpComponentPart(child).height) {
				    spacing += margin - MAX(-SpComponentPart(child).height, 0);
				}
			    }

			    hc -= spacing;
			    
			    /* update current height */
			    if (SpComponentPart(component).prev_height >= 0
				|| (SpGetParent(component) != NULL
				    && SpComponentPart(SpGetParent(component)).prev_height >= 0)) {
				SpComponentPart(child).current_height = hc;

				if (spIsContainer(child) == SP_TRUE
				    && spIsCanvas(child) == SP_FALSE) {
				    getOptimumHeight(child);
				}
			    } else {
				if (isRequireHeightValid(child) == SP_TRUE) {
				    SpComponentPart(child).current_height
					= MAX(hc, SpComponentPart(child).require_height);
				} else {
				    SpComponentPart(child).current_height = hc;
				}
				if (spIsContainer(child) == SP_TRUE
				    && spIsCanvas(child) == SP_FALSE) {
				    SpComponentPart(child).current_height
					= MAX(SpComponentPart(child).current_height,
					      getOptimumHeight(child));
				}
			    }

			    height += SpComponentPart(child).current_height;
			    if (SpComponentPart(child).percentage_height > 0 && SpComponentPart(child).require_height) {
			    } else {
				total_height = height - SpComponentPart(child).height + spacing;
			    }
			}
		    } else {
			SpComponentPart(child).current_height = SpComponentPart(child).height; 

			if (spIsContainer(child) == SP_TRUE
			    && spIsCanvas(child) == SP_FALSE) {
			    h = getOptimumHeight(child);
			    if (h > SpComponentPart(child).current_height) {
				SpComponentPart(child).current_height = h;
			    }
			}
			
			height += SpComponentPart(child).current_height;
		    }
		    spDebug(50, "getOptimumHeight",
			    "y = %d\n", SpComponentPart(child).y);
		    prev_child = child;
		}

		child = spGetNextVisibleComponent(child);
	    
		if (child == NULL) {
		    margin = 0;

                    if (prev_child != NULL) {
                        if (isMarginRequired(component, prev_child) == SP_TRUE) {
                            margin = MAX(SpComponentPart(component).margin_height, SpComponentPart(prev_child).bottom_offset);
                        } else if (spIsCreated(prev_child) == SP_TRUE && SpComponentPart(prev_child).bottom_offset > 0) {
                            margin = SpComponentPart(prev_child).bottom_offset;
                        }
                        if (margin > -SpComponentPart(prev_child).height) {
                            height += margin - MAX(-SpComponentPart(prev_child).height, 0);
                        }
                    }
		    
		    height += SpComponentPart(component).margin_bottom;
		    height += SpComponentPart(component).border_width;
		    
		    break;
		} else if (spIsResizable(child) == SP_TRUE && flag == 1) {
		    if (spIsSpacingOn(prev_child) == SP_TRUE
			|| spIsSpacingOn(child) == SP_TRUE) {
			height += MAX(SpComponentPart(component).spacing,
				      SpComponentPart(child).top_offset);
		    } else if (SpComponentPart(child).top_offset > 0) {
			height += SpComponentPart(child).top_offset;
		    }
		}
	    }
	} else /*if (SpComponentPart(component).orientation == SP_HORIZONTAL)*/ {
	    if (spIsFrameToolBarVisible(component) == SP_TRUE) {
		if (SpGetComponentClass(SpFramePart(component).tool_bar)->component.resize_flag == SP_TRUE) {
		    SpComponentPart(SpFramePart(component).tool_bar).current_height
			= getOptimumHeight(SpFramePart(component).tool_bar);
		}
		spDebug(80, "getOptimumHeight", "tool bar height = %d\n",
			SpComponentPart(SpFramePart(component).tool_bar).current_height);
	    }
	    while (child != NULL) {
		if (spIsResizable(child) == SP_TRUE && spIsToolBar(child) == SP_FALSE) {
		    if (spIsFrameToolBarVisible(component) == SP_TRUE
			&& SpGetComponentClass(SpFramePart(component).tool_bar)->component.resize_flag == SP_TRUE) {
			h = SpComponentPart(SpFramePart(component).tool_bar).height;
			spDebug(50, "getOptimumHeight", "%s: tool bar height = %d\n",
				spGetTitle(child), h);
		    } else {
			h = 0;
		    }

		    SpComponentPart(child).y = h;
		    if (isMarginRequired(component, child) == SP_TRUE) {
			SpComponentPart(child).y += MAX(SpComponentPart(component).margin_height,
						       SpComponentPart(child).top_offset);
		    } else if (SpComponentPart(child).top_offset > 0) {
			SpComponentPart(child).y += SpComponentPart(child).top_offset;
		    }
		    SpComponentPart(child).y += SpComponentPart(component).border_width;
		    h = SpComponentPart(child).y;
		    
		    if (SpComponentPart(child).height <= 0
			|| (SpComponentPart(child).percentage_height > 0 && SpComponentPart(child).require_height)) {
			if (SpComponentPart(child).percentage_height > 0 && SpComponentPart(child).require_height) {
			    hc = (int)spRound((double)SpComponentPart(component).current_height * (double)SpComponentPart(child).percentage_height / 100.0);
			} else {
			    hc = SpComponentPart(component).current_height
				- SpComponentPart(child).y + SpComponentPart(child).height;
			}
			spDebug(80, "getOptimumHeight", "%s: hc = %d, parent current height = %d\n",
				spGetTitle(child), hc, SpComponentPart(component).current_height);

			hc -= SpComponentPart(component).border_width;

			margin = 0;
			
			if (isMarginRequired(component, child) == SP_TRUE) {
			    margin = MAX(SpComponentPart(component).margin_height, SpComponentPart(child).bottom_offset);
			} else if (spIsCreated(child) == SP_TRUE && SpComponentPart(child).bottom_offset > 0) {
			    margin = SpComponentPart(child).bottom_offset;
			}
			if (margin > -SpComponentPart(child).height) {
			    hc -= margin - MAX(-SpComponentPart(child).height, 0);
			}
			
			hc -= SpComponentPart(component).margin_bottom;
		    
			if (SpComponentPart(component).prev_height >= 0
			    || (SpGetParent(component) != NULL
				&& SpComponentPart(SpGetParent(component)).prev_height >= 0)) {
			    SpComponentPart(child).current_height = MAX(hc, 0);
			    spDebug(80, "getOptimumHeight", "%s prev_height >= 0: hc = %d, current height = %d\n",
				    spGetTitle(child), hc, SpComponentPart(child).current_height);

			    if (spIsContainer(child) == SP_TRUE
				&& spIsCanvas(child) == SP_FALSE) {
				getOptimumHeight(child);
			    }
			} else {
			    if (isRequireHeightValid(child) == SP_TRUE) {
				SpComponentPart(child).current_height
				    = MAX(hc, SpComponentPart(child).require_height);
				spDebug(80, "getOptimumHeight", "%s require_height used: hc = %d, require_height = %d\n",
					spGetTitle(child), hc, SpComponentPart(child).require_height);
			    } else {
				SpComponentPart(child).current_height = hc;
				spDebug(80, "getOptimumHeight", "%s hc is used: hc = %d\n", spGetTitle(child), hc);
			    }
			    if (spIsContainer(child) == SP_TRUE
				&& spIsCanvas(child) == SP_FALSE) {
				SpComponentPart(child).current_height
				    = MAX(SpComponentPart(child).current_height,
					  getOptimumHeight(child));
			    }
			}
		    } else {
			SpComponentPart(child).current_height = SpComponentPart(child).height;

			if (spIsContainer(child) == SP_TRUE
			    && spIsCanvas(child) == SP_FALSE) {
			    getOptimumHeight(child);
			}
		    }
		    h += SpComponentPart(child).current_height;
		    spDebug(80, "getOptimumHeight", "%s: h = %d, current height = %d\n",
			    spGetTitle(child), h, SpComponentPart(child).current_height);
		    
		    margin = 0;
		    
		    if (isMarginRequired(component, child) == SP_TRUE) {
			margin = MAX(SpComponentPart(component).margin_height, SpComponentPart(child).bottom_offset);
		    } else if (spIsCreated(child) == SP_TRUE && SpComponentPart(child).bottom_offset > 0) {
			margin = SpComponentPart(child).bottom_offset;
		    }
		    if (margin > -SpComponentPart(child).height) {
			h += margin - MAX(-SpComponentPart(child).height, 0);
		    }
		    
		    h += SpComponentPart(component).margin_bottom;
		    h += SpComponentPart(component).border_width;
		    
		    height = MAX(height, h);
		    spDebug(80, "getOptimumHeight", "%s: computed height = %d\n",
			    spGetTitle(child), height);
		}
		
		prev_child = child;
		child = spGetNextVisibleComponent(child);
	    }
	}
    } else {
	if (SpComponentPart(component).prev_height >= 0 || isRequireHeightValid(component) == SP_FALSE) {
	    height = SpComponentPart(component).height;
	} else {
	    spDebug(50, "getOptimumHeight",
		    "height = %d, require_height = %d\n",
		    SpComponentPart(component).height,
		    SpComponentPart(component).require_height);
	    height = MAX(SpComponentPart(component).height,
			 SpComponentPart(component).require_height);
	}
    }

    if (total_height < 0) {
	total_height = height;
    }

    spDebug(50, "getOptimumHeight", "done: %s, total_height = %d\n",
	    spGetTitle(component), total_height);
    
    return total_height;
}

static int getOptimumWidth(spComponent component)
{
    int flag;
    int w, wc;
    int spacing;
    int margin;
    int width = 0;
    int total_width = -1;
    spComponent child = NULL, prev_child = NULL;

    if (spIsResizable(component) == SP_FALSE) return 0;
    
    child = SpGetChild(component);

    if (child != NULL) {
	if (spIsTabBox(component) == SP_FALSE && SpComponentPart(component).orientation == SP_HORIZONTAL) {
	    flag = 0;
	    while (1) {
		if (spIsToolBar(child) == SP_TRUE) {
		    SpComponentPart(child).x = 0;
		    SpComponentPart(child).current_width = SpComponentPart(component).current_width;
		    getOptimumWidth(child);
		} else if (spIsResizable(child) == SP_TRUE) {
		    if (flag == 0) {
			width += SpComponentPart(component).border_width;
			
			if (isMarginRequired(component, child) == SP_TRUE) {
			    width += MAX(SpComponentPart(component).margin_width,
					 SpComponentPart(child).left_offset);
			} else if (SpComponentPart(child).left_offset > 0) {
			    width += SpComponentPart(child).left_offset;
			}
			flag = 1;
		    }
	    
		    SpComponentPart(child).x = width;
		    
		    if (SpComponentPart(child).width <= 0 
			|| (SpComponentPart(child).percentage_width > 0 && SpComponentPart(child).require_width)) {
			if (total_width > 0) {
			    if (SpComponentPart(child).percentage_width > 0 && SpComponentPart(child).require_width) {
			        SpComponentPart(child).current_width
				    = (int)spRound((double)total_width * (double)SpComponentPart(child).percentage_width / 100.0);
				if (prev_child == NULL) {
				    SpComponentPart(child).current_width -= SpComponentPart(child).x;
				}
				if (spIsSpacingOn(child) == SP_TRUE) {
				    if (spIsSpacingOn(prev_child) == SP_TRUE) {
					SpComponentPart(child).current_width -= SpComponentPart(component).spacing / 2;
				    }
				    if (spIsSpacingOn(spGetNextVisibleComponent(child)) == SP_TRUE) {
					SpComponentPart(child).current_width -= SpComponentPart(component).spacing / 2;
				    }
				}
			    } else {
			        SpComponentPart(child).current_width
				    = total_width - SpComponentPart(child).x + SpComponentPart(child).width;
			    }
			
			    SpComponentPart(child).current_width -= SpComponentPart(component).border_width;
			
			    if (spGetNextVisibleComponent(child) == NULL) {
				margin = 0;
				
				if (isMarginRequired(component, child) == SP_TRUE) {
				    margin = MAX(SpComponentPart(component).margin_width, SpComponentPart(child).right_offset);
				} else if (spIsCreated(child) == SP_TRUE && SpComponentPart(child).right_offset > 0) {
				    margin = SpComponentPart(child).right_offset;
				}
				if (margin > -SpComponentPart(child).width) {
				    SpComponentPart(child).current_width -= margin - MAX(-SpComponentPart(child).width, 0);
				}
			    }
			    
			    width += SpComponentPart(child).current_width;

			    if (spIsContainer(child) == SP_TRUE
				&& spIsCanvas(child) == SP_FALSE) {
				getOptimumWidth(child);
			    }
			} else {
			    spacing = 0;
			    if (SpComponentPart(child).percentage_width > 0 && SpComponentPart(child).require_width) {
				wc = (int)spRound((double)SpComponentPart(component).current_width * (double)SpComponentPart(child).percentage_width / 100.0);

				if (prev_child == NULL) {
				    wc -= SpComponentPart(child).x;
				}
				if (spIsSpacingOn(child) == SP_TRUE) {
				    if (spIsSpacingOn(prev_child) == SP_TRUE) {
					spacing += SpComponentPart(component).spacing / 2;
				    }
				    if (spIsSpacingOn(spGetNextVisibleComponent(child)) == SP_TRUE) {
					spacing += SpComponentPart(component).spacing / 2;
				    }
				}
			    } else {
				wc = SpComponentPart(component).current_width
				    - SpComponentPart(child).x + SpComponentPart(child).width;
			    }
			
			    spacing += SpComponentPart(component).border_width;
			    
			    if (spGetNextVisibleComponent(child) == NULL) {
				margin = 0;
				
				if (isMarginRequired(component, child) == SP_TRUE) {
				    margin = MAX(SpComponentPart(component).margin_width, SpComponentPart(child).right_offset);
				} else if (spIsCreated(child) == SP_TRUE && SpComponentPart(child).right_offset > 0) {
				    margin = SpComponentPart(child).right_offset;
				}
				if (margin > -SpComponentPart(child).width) {
				    spacing += margin - MAX(-SpComponentPart(child).width, 0);
				}
			    }
			    
			    wc -= spacing;

			    /* update current width */
			    if (SpComponentPart(component).prev_width >= 0
				|| (SpGetParent(component) != NULL
				    && SpComponentPart(SpGetParent(component)).prev_width >= 0)) {
				SpComponentPart(child).current_width = wc;
			    
				if (spIsContainer(child) == SP_TRUE
				    && spIsCanvas(child) == SP_FALSE) {
				    getOptimumWidth(child);
				}
			    } else {
				if (isRequireWidthValid(child) == SP_TRUE) {
				    SpComponentPart(child).current_width
					= MAX(wc, SpComponentPart(child).require_width);
				} else {
				    SpComponentPart(child).current_width = wc;
				}
				if (spIsContainer(child) == SP_TRUE
				    && spIsCanvas(child) == SP_FALSE) {
				    SpComponentPart(child).current_width
					= MAX(SpComponentPart(child).current_width,
					      getOptimumWidth(child));
				}
			    }

			    width += SpComponentPart(child).current_width;
			    if (SpComponentPart(child).percentage_width > 0 && SpComponentPart(child).require_width) {
			    } else {
				total_width = width - SpComponentPart(child).width + spacing;
			    }
			}
		    } else {
			SpComponentPart(child).current_width = SpComponentPart(child).width; 

			if (spIsContainer(child) == SP_TRUE
			    && spIsCanvas(child) == SP_FALSE) {
			    w = getOptimumWidth(child);
			    if (w > SpComponentPart(child).current_width) {
				SpComponentPart(child).current_width = w;
			    }
			}
			
			width += SpComponentPart(child).current_width;
		    }
		    spDebug(50, "getOptimumWidth",
			    "x = %d\n", SpComponentPart(child).x);
		    prev_child = child;
		}

		child = spGetNextVisibleComponent(child);
	    
		if (child == NULL) {
		    margin = 0;

                    if (prev_child != NULL) {
                        if (isMarginRequired(component, prev_child) == SP_TRUE) {
                            margin = MAX(SpComponentPart(component).margin_width, SpComponentPart(prev_child).right_offset);
                        } else if (spIsCreated(prev_child) == SP_TRUE && SpComponentPart(prev_child).right_offset > 0) {
                            margin = SpComponentPart(prev_child).right_offset;
                        }
                        if (margin > -SpComponentPart(prev_child).width) {
                            width += margin - MAX(-SpComponentPart(prev_child).width, 0);
                        }
                    }
		    
		    width += SpComponentPart(component).border_width;
		    
		    break;
		} else if (spIsResizable(child) == SP_TRUE && flag == 1) {
		    if (spIsSpacingOn(prev_child) == SP_TRUE
			|| spIsSpacingOn(child) == SP_TRUE) {
			width += MAX(SpComponentPart(component).spacing,
				     SpComponentPart(child).left_offset);
		    } else if (SpComponentPart(child).left_offset > 0) {
			width += SpComponentPart(child).left_offset;
		    }
		}
	    }
	} else /*if (SpComponentPart(component).orientation == SP_VERTICAL)*/ {
	    width = 0;
	    while (child != NULL) {
		if (spIsResizable(child) == SP_TRUE) {
		    w = 0;
		    if (isMarginRequired(component, child) == SP_TRUE) {
			SpComponentPart(child).x = MAX(SpComponentPart(component).margin_width,
						       SpComponentPart(child).left_offset);
		    } else if (SpComponentPart(child).left_offset > 0) {
			SpComponentPart(child).x = SpComponentPart(child).left_offset;
		    } else {
			SpComponentPart(child).x = 0;
		    }
		    SpComponentPart(child).x += SpComponentPart(component).border_width;
		    w += SpComponentPart(child).x;
		    
		    if (SpComponentPart(child).width <= 0
			|| (SpComponentPart(child).percentage_width > 0 && SpComponentPart(child).require_width)) {
			if (SpComponentPart(child).percentage_width > 0 && SpComponentPart(child).require_width) {
			    wc = (int)spRound((double)SpComponentPart(component).current_width * (double)SpComponentPart(child).percentage_width / 100.0);
			} else {
			    wc = SpComponentPart(component).current_width
				- SpComponentPart(child).x + SpComponentPart(child).width;
			}

			wc -= SpComponentPart(component).border_width;

			margin = 0;
			
			if (isMarginRequired(component, child) == SP_TRUE) {
			    margin = MAX(SpComponentPart(component).margin_width, SpComponentPart(child).right_offset);
			} else if (spIsCreated(child) == SP_TRUE && SpComponentPart(child).right_offset > 0) {
			    margin = SpComponentPart(child).right_offset;
			}
			if (margin > -SpComponentPart(child).width) {
			    wc -= margin - MAX(-SpComponentPart(child).width, 0);
			}
			
			spDebug(50, "getOptimumWidth", "%s: wc = %d, require_width = %d, parent prev_width = %d, parent border = %d\n",
				spGetTitle(child), wc, SpComponentPart(child).require_width, SpComponentPart(component).prev_width,
				SpComponentPart(component).border_width);
		    
			if (SpComponentPart(component).prev_width >= 0
			    || (SpGetParent(component) != NULL
				&& SpComponentPart(SpGetParent(component)).prev_width >= 0)) {
			    SpComponentPart(child).current_width = MAX(wc, 0);

			    if (spIsContainer(child) == SP_TRUE
				&& spIsCanvas(child) == SP_FALSE) {
				getOptimumWidth(child);
			    }
			} else {
			    if (isRequireWidthValid(child) == SP_TRUE) {
				SpComponentPart(child).current_width
				    = MAX(wc, SpComponentPart(child).require_width);
			    } else {
				SpComponentPart(child).current_width = wc;
			    }
			    if (spIsContainer(child) == SP_TRUE
				&& spIsCanvas(child) == SP_FALSE) {
				SpComponentPart(child).current_width
				    = MAX(SpComponentPart(child).current_width,
					  getOptimumWidth(child));
			    }
			}
		    } else {
			SpComponentPart(child).current_width = SpComponentPart(child).width; 

			if (spIsContainer(child) == SP_TRUE
			    && spIsCanvas(child) == SP_FALSE) {
			    getOptimumWidth(child);
			}
		    }
		    w += SpComponentPart(child).current_width;

		    margin = 0;
		    
		    if (isMarginRequired(component, child) == SP_TRUE) {
			margin = MAX(SpComponentPart(component).margin_width, SpComponentPart(child).right_offset);
		    } else if (spIsCreated(child) == SP_TRUE && SpComponentPart(child).right_offset > 0) {
			margin = SpComponentPart(child).right_offset;
		    }
		    if (margin > -SpComponentPart(child).width) {
			w += margin - MAX(-SpComponentPart(child).width, 0);
		    }
		    
		    w += SpComponentPart(component).border_width;
		    
		    spDebug(50, "getOptimumWidth", "%s: w = %d, width = %d, current_width = %d, prev_width = %d\n",
			    spGetTitle(child), w, width,
			    SpComponentPart(child).current_width, SpComponentPart(child).prev_width);
		    width = MAX(width, w);
		}
		
		prev_child = child;
		child = spGetNextVisibleComponent(child);
	    }
	}
    } else {
	if (SpComponentPart(component).prev_width >= 0 || isRequireWidthValid(component) == SP_FALSE) {
	    width = SpComponentPart(component).width;
	} else {
	    width = MAX(SpComponentPart(component).width,
			SpComponentPart(component).require_width);
	}
    }

    if (total_width < 0) {
	total_width = width;
    }

    return total_width;
}

static spBool getOptimumSize(spComponent component)
{
    int i;
    
    if (spIsResizable(component) == SP_FALSE) return SP_TRUE;

    spDebug(50, "getOptimumSize", "%s, prev_width = %d, prev_height = %d, current_width = %d, current_height = %d\n",
	    spGetTitle(component), SpComponentPart(component).prev_width, SpComponentPart(component).prev_height,
	    SpComponentPart(component).current_width, SpComponentPart(component).current_height);
    
    if (SpComponentPart(component).current_width >= 0 || SpComponentPart(component).current_height >= 0) {
	if (spIsWindow(component) == SP_TRUE) {
	    SpComponentPart(component).prev_width = SpComponentPart(component).current_width;
	    SpComponentPart(component).prev_height = SpComponentPart(component).current_height;
	    SpComponentPart(component).current_width = -1;
	    SpComponentPart(component).current_height = -1;
	    SpComponentPart(component).client_width = -1;
	    SpComponentPart(component).client_height = -1;
	    spGetSize(component, NULL, NULL);
	}
	
	for (i = 0; i < /*3*/4; i++) {
	    getOptimumWidth(component);
	    getOptimumHeight(component);
	}
    } else {
	for (i = 0; i < /*3*/4; i++) {
	    SpComponentPart(component).current_width = getOptimumWidth(component);
	    SpComponentPart(component).current_height = getOptimumHeight(component);
	    spDebug(100, "getOptimumSize", "loop %d: width = %d, height = %d, require_width = %d, require_height = %d\n",
		    i, SpComponentPart(component).current_width, SpComponentPart(component).current_height,
		    SpComponentPart(component).require_width, SpComponentPart(component).require_height);
	}
	spDebug(50, "getOptimumSize", "width = %d, height = %d\n",
                SpComponentPart(component).current_width,
		SpComponentPart(component).current_height);
    }

    if (spIsWindow(component) == SP_TRUE) {
	if (spIsVisible(SpFramePart(component).status_bar) == SP_TRUE) {
	    SpComponentPart(SpFramePart(component).status_bar).x = 0;
	    SpComponentPart(SpFramePart(component).status_bar).y =
		SpComponentPart(component).current_height;
	    SpComponentPart(SpFramePart(component).status_bar).current_width
		= SpComponentPart(component).current_width;
	    SpComponentPart(SpFramePart(component).status_bar).current_height
		= spScaleY(SP_DEFAULT_STATUS_BAR_HEIGHT);
	}
    }

    return SP_TRUE;
}
#endif

void spAdjustComponentSize(spComponent component)
{
#if !defined(ANDROID) /*&& !defined(IPHONE)*/ /*&& !defined(GTK3)*/
    if (spIsCreated(component) == SP_TRUE
	&& SpComponentPart(component).geometry_flag == SP_FALSE) {
	spDebug(50, "spAdjustComponentSize", "start\n");
	
	getOptimumSize(component);
	setComponentsSize(component);

	spDebug(50, "spAdjustComponentSize", "done: current_width = %d, current_height = %d\n",
		SpComponentPart(component).current_width, SpComponentPart(component).current_height);
    }
#endif
    return;
}

spBool spIsWindow(spComponent component)
{
    if (component == NULL) return SP_FALSE;
    
    return SpGetComponentClass(component)->component.window_flag;
}

void spDestroyWindow(spComponent component)
{
    destroyWindow(component, SP_TRUE);
    return;
}

void spDestroyAllWindow(void)
{
    spComponent window, next_window;

    spDebug(80, "spDestroyAllWindow", "in\n");
    
    window = spGetWindow(spGetChild(NULL));
    
    while (window != NULL) {
	next_window = spGetNextWindow(window, SP_TRUE);
	if (next_window == window) {
	    next_window = NULL;
	}
	spDebug(80, "spDestroyAllWindow", "title = %s\n", spGetTitle(window));
	destroyWindow(window, SP_FALSE);
	window = next_window;
    }
    
    spDebug(80, "spDestroyAllWindow", "done\n");
    
    return;
}

void spCloseWindowPrompt(spComponent component, spBool prompt)
{
    spBool flag = SP_TRUE;
    spComponent window;
    
    if (spIsCreated(component) == SP_FALSE) return;

    window = SpGetWindow(component);
    
    if (spIsLastWindow(window) == SP_TRUE
	|| spIsSubClass(window, SpMainFrame) == SP_TRUE) { /* 2000/7/20 added */
	if (prompt == SP_TRUE) {
	    spDebug(30, "spCloseWindowPrompt", "exit_prompt = %d\n",
		    SpTopLevelPart(sp_toplevel).exit_prompt);
	    if (SpTopLevelPart(sp_toplevel).exit_prompt == SP_TRUE) {
		if (spCreateMessageBox(/*window*/NULL, SP_QUIT_QUESTION_TITLE,
				       SP_QUIT_QUESTION_MESSAGE,
				       SppDialogType, SP_QUESTION_DIALOG,
				       SppMessageBoxButtonType, SP_MB_YES_NO,
				       NULL) != SP_DR_YES) {
		    flag = SP_FALSE;
		}
	    }
	}

	if (flag == SP_TRUE) {
	    /*SpComponentPart(window).visible_flag = SP_FALSE;*/
	    spDebug(50, "spCloseWindowPrompt", "quit program\n");
	    destroyComponent(window, SP_FALSE);
	    spQuit(0);
	}
    } else {
	spDebug(80, "spCloseWindowPrompt", "not last window\n");
	
	if (prompt == SP_TRUE) {
	    spDebug(50, "spCloseWindowPrompt", "close_prompt = %d\n",
		    SpTopLevelPart(sp_toplevel).close_prompt);
	    if (SpTopLevelPart(sp_toplevel).close_prompt == SP_TRUE) {
		if (spCreateMessageBox(window, SP_CLOSE_QUESTION_TITLE,
				       SP_CLOSE_QUESTION_MESSAGE,
				       SppDialogType, SP_QUESTION_DIALOG,
				       SppMessageBoxButtonType, SP_MB_YES_NO,
				       NULL) == SP_DR_YES) {
		    flag = SP_TRUE;
		} else {
		    flag = SP_FALSE;
		}
	    }
	}
	
	if (flag == SP_TRUE) {
	    /*SpComponentPart(window).visible_flag = SP_FALSE;*/
	    spDestroyComponent(window);
	}
    }
    
    spDebug(80, "spCloseWindowPrompt", "done\n");
    
    return;
}

void spCloseWindow(spComponent component)
{
    spCallbackReason reason = SP_CR_NONE;
    
    if (spIsCreated(component) == SP_FALSE) return;

    reason = spGetCallbackReason(component);
    spDebug(80, "spCloseWindow", "reason = %d\n", reason);
    spCloseWindowPrompt(component, (reason == SP_CR_ACTIVATE ? SP_TRUE : SP_FALSE));
    
    return;
}

static void popupWindow(spComponent component, spBool move_cursor)
{
    spComponent window;
    
    if (spIsCreated(component) == SP_FALSE) return;

    window = SpGetWindow(component);

    if (spIsCreated(window) == SP_FALSE
	|| spIsFrame(window) == SP_FALSE) return;
    
    SpComponentPart(window).visible_flag = SP_TRUE;
    spPopupFrameArch(window, move_cursor);
    
    return;
}

void spPopupWindow(spComponent component)
{
    popupWindow(component, SP_TRUE);
    return;
}

void spPopdownWindow(spComponent component)
{
    spComponent window;
    
    if (spIsCreated(component) == SP_FALSE) return;

    spDebug(30, "spPopdownWindow", "component = %s\n", spGetName(component));
    
    window = SpGetWindow(component);
    
    if (spIsCreated(window) == SP_FALSE
	|| spIsFrame(window) == SP_FALSE
	|| spIsVisible(window) == SP_FALSE) {
	spDebug(30, "spPopdownWindow", "created = %d, visible = %d\n",
		spIsCreated(window), SpComponentPart(component).visible_flag);
	spDebug(30, "spPopdownWindow", "popdown unnecessary\n");
	return;
    }

    /*SpComponentPart(window).visible_flag = SP_FALSE;*/
    spPopdownFrameArch(window);
    SpComponentPart(window).visible_flag = SP_FALSE;
    
    spDebug(30, "spPopdownWindow", "done\n");
    
    return;
}

void spMapWindow(spComponent component)
{
    popupWindow(component, SP_FALSE);
    return;
}

void spUnmapWindow(spComponent component)
{
    spPopdownWindow(component);
    return;
}

void spRaiseWindow(spComponent component)
{
    spComponent window;
    
    if (spIsCreated(component) == SP_FALSE) return;

    window = SpGetWindow(component);

    if (spIsCreated(window) == SP_FALSE
	|| spIsFrame(window) == SP_FALSE
	|| SpComponentPart(window).visible_flag == SP_FALSE) return;
    
    spRaiseFrameArch(window);
    
    return;
}

void spLowerWindow(spComponent component)
{
    spComponent window;
    
    if (spIsCreated(component) == SP_FALSE) return;

    window = SpGetWindow(component);

    if (spIsCreated(window) == SP_FALSE
	|| spIsFrame(window) == SP_FALSE
	|| SpComponentPart(window).visible_flag == SP_FALSE) return;
    
    spLowerFrameArch(window);
    
    return;
}

spBool spIsTaskTrayVisible(spComponent component)
{
    spComponent window;
    
    if (spIsCreated(component) == SP_FALSE) return SP_FALSE;

    window = SpGetWindow(component);

    if (spIsCreated(window) == SP_FALSE
	|| spIsFrame(window) == SP_FALSE) return SP_FALSE;

    return SpFramePart(window).task_tray_visible;
}

spBool spShowTaskTray(spComponent component, spBool flag)
{
    spComponent window;
    
    if (spIsCreated(component) == SP_FALSE) return SP_FALSE;

    window = SpGetWindow(component);

    if (spIsCreated(window) == SP_FALSE
	|| spIsMainFrame(window) == SP_FALSE) return SP_FALSE;

    if (spShowTaskTrayArch(window, flag) == SP_TRUE) {
	SpFramePart(window).task_tray_visible = flag;
	return SP_TRUE;
    }

    return SP_FALSE;
}

static void clearPrevSize(spComponent component, spBool child_only)
{
    spComponent next;

    if (component == NULL) {
	return;
    }

    next = SpGetChild(component);
    while (next != NULL) {
	clearPrevSize(next, SP_FALSE);
	next = SpGetNextComponent(next);
    }

    SpComponentPart(component).geometry_flag = SP_FALSE;

    if (child_only == SP_FALSE) {
	SpComponentPart(component).prev_width = -1;
	SpComponentPart(component).prev_height = -1;
    }
    
    if (spIsFrame(component) == SP_FALSE || SpFramePart(component).window_type != SP_FULLSCREEN_WINDOW) {
	SpComponentPart(component).current_width = -1;
	SpComponentPart(component).current_height = -1;
    }

#if defined(GTK3)
    if (spIsPrimitive(component) == SP_TRUE) {
	if (spIsSubClass(component, SpLabel) == SP_FALSE && SpPrimitiveArch(component).top_widget != NULL) {
	    if (spIsCanvas(component) == SP_TRUE) {
		gtk_widget_set_size_request(SpPrimitiveArch(component).widget, 0, 0);
	    }
	    gtk_widget_set_size_request(SpPrimitiveArch(component).top_widget, 0, 0);
	} else {
	    gtk_widget_set_size_request(SpPrimitiveArch(component).widget, 0, 0);
	}
    }
#endif
    
    return;
}

void spAdjustWindowSize(spComponent component, spBool reset)
{
    spComponent window;
    
    spDebug(80, "spAdjustWindowSize", "in\n");
    
    if (spIsCreated(component) == SP_FALSE) return;

    window = SpGetWindow(component);

    if (spIsCreated(window) == SP_FALSE) return;

    if (reset == SP_TRUE) {
	clearPrevSize(window, SP_FALSE);
    }
    spAdjustComponentSize(window);

    return;
}

spBool spAllocateSize(spComponent component, int width, int height)
{
    spComponent window;
    
    spDebug(50, "spAllocateSize", "in: width = %d, height = %d\n", width, height);
    
    if (spIsCreated(component) == SP_FALSE || (width <= 0 && height <= 0)) {
	spDebug(50, "spAllocateSize", "return SP_FALSE\n");
	return SP_FALSE;
    }

    window = SpGetWindow(component);

    if (spIsCreated(window) == SP_FALSE) return SP_FALSE;

#if 1
    SpComponentPart(window).prev_width = -1;
    SpComponentPart(window).prev_height = -1;
    SpComponentPart(window).current_width = -1;
    SpComponentPart(window).current_height = -1;
    SpComponentPart(window).client_width = -1;
    SpComponentPart(window).client_height = -1;
#endif
	    
    SpComponentPart(component).client_width = -1;
    SpComponentPart(component).client_height = -1;
    if (width > 0) {
	SpComponentPart(component).prev_width = SpComponentPart(component).current_width;
	SpComponentPart(component).current_width = width;
	if (spIsCanvas(component) == SP_TRUE
	    || spIsContainer(component) == SP_FALSE) {
	    SpComponentPart(component).require_width = width;
	}
    }
    if (height > 0) {
	SpComponentPart(component).prev_height = SpComponentPart(component).current_height;
	SpComponentPart(component).current_height = height;
	if (spIsCanvas(component) == SP_TRUE
	    || spIsContainer(component) == SP_FALSE) {
	    SpComponentPart(component).require_height = height;
	}
    }

    /*clearPrevSize(window, SP_TRUE);*/
    spAdjustComponentSize(window);

    spDebug(50, "spAllocateSize", "done\n");
    
    return SP_TRUE;
}

spBool spSetWindowPosition(spComponent component, int x, int y)
{
    spComponent window;
    
    if (spIsCreated(component) == SP_FALSE) return SP_FALSE;

    window = SpGetWindow(component);

    if (spIsCreated(window) == SP_FALSE
	|| spIsFrame(window) == SP_FALSE
	|| SpFramePart(window).window_type == SP_FULLSCREEN_WINDOW) return SP_FALSE;

    spDebug(80, "spSetWindowPosition", "x = %d, y = %d\n", x, y);
    
    SpComponentPart(window).x = x;
    SpComponentPart(window).y = y;
    if (spIsVisible(window) == SP_TRUE) {
	SpFramePart(window).position_set_flag = SP_TRUE;
	return spSetFramePositionArch(window, x, y);
    }
    
    return SP_TRUE;
}

spBool spGetWindowPosition(spComponent component, int *x, int *y)
{
    spComponent window;
    
    if (spIsCreated(component) == SP_FALSE) return SP_FALSE;

    window = SpGetWindow(component);

    if (spIsCreated(window) == SP_FALSE
	|| spIsFrame(window) == SP_FALSE) return SP_FALSE;
    
    if (spIsVisible(window) == SP_TRUE) {
	if (spGetFramePositionArch(window,
				   &SpComponentPart(window).x,
				   &SpComponentPart(window).y) == SP_TRUE) {
	    if (x != NULL) *x = SpComponentPart(window).x;
	    if (y != NULL) *y = SpComponentPart(window).y;
	    return SP_TRUE;
	}
    }
    
    return SP_FALSE;
}

void spDestroyWindowCB(spComponent component, void *data)
{
    spDestroyWindow(component);
    return;
}

void spCloseWindowCB(spComponent component, void *data)
{
    spDebug(80, "spCloseWindowCB", "in\n");
    spCloseWindowPrompt(component, SP_TRUE);
    return;
}

void spPopupWindowCB(spComponent component, void *data)
{
    spPopupWindow(component);
    return;
}

void spPopdownWindowCB(spComponent component, void *data)
{
    spPopdownWindow(component);
    return;
}

spComponent spGetNextWindow(spComponent component, spBool dialog_flag)
{
    spComponent current, next;

    if (sp_components == NULL
	|| spIsWindow(component) == SP_FALSE) return NULL;

    next = NULL;
    current = component;
    while (1) {
	if ((next = SpGetNextComponent(current)) == NULL) {
	    if (SpGetParent(current) != NULL) {
		next = SpGetChild(SpGetParent(current));
	    } else {
		next = sp_components->child;
	    }
	}
	
	if (!(spIsFrame(next) == SP_TRUE && SpFramePart(next).internal_flag == SP_TRUE)) {
	    if (next == NULL
		|| (dialog_flag == SP_TRUE && spIsWindow(next) == SP_TRUE)
		|| (spIsDialog(next) == SP_FALSE && spIsWindow(next) == SP_TRUE)) {
		break;
	    }
	} 
	current = next;
    }
    
    return next;
}

spComponent spGetPrevWindow(spComponent component, spBool dialog_flag)
{
    spComponent current, prev;

    if (sp_components == NULL
	|| spIsWindow(component) == SP_FALSE) return NULL;

    prev = NULL;
    current = component;
    while (1) {
	if ((prev = SpGetPrevComponent(current)) == NULL) {
	    if (SpGetParent(current) != NULL) {
		prev = SpGetLastChild(SpGetParent(current));
	    } else {
		prev = sp_components->last_child;
	    }
	}

	if (!(spIsFrame(prev) == SP_TRUE && SpFramePart(prev).internal_flag == SP_TRUE)) {
	    if (prev == NULL
		|| (dialog_flag == SP_TRUE && spIsWindow(prev) == SP_TRUE)
		|| (spIsDialog(prev) == SP_FALSE && spIsWindow(prev) == SP_TRUE)) {
		break;
	    }
	} 
	current = prev;
    }
    
    return prev;
}

void spPopupNextWindow(spComponent component, spBool dialog_flag)
{
    spComponent next = NULL;
    
    next = spGetNextWindow(SpGetWindow(component), dialog_flag);
    if (next != NULL) spPopupWindow(next);
    
    return;
}

void spPopupPrevWindow(spComponent component, spBool dialog_flag)
{
    spComponent prev = NULL;
    
    prev = spGetPrevWindow(SpGetWindow(component), dialog_flag);
    if (prev != NULL) spPopupWindow(prev);
    
    return;
}

spBool spIsLastWindow(spComponent component)
{
    int flag = 0;
    long num_window = 0;
    long num_dialog = 0;
    spComponent child;

    if (sp_components == NULL || spIsCreated(component) == SP_FALSE)
	return SP_FALSE;

    if (spIsWindow(component) == SP_FALSE) {
	component = SpGetWindow(component);
	if (spIsCreated(component) == SP_FALSE)
	    return SP_FALSE;
    }
    if (spIsSubClass(component, SpDialog) == SP_TRUE) {
        return SP_FALSE;
    }

    child = sp_components->child;

    while (child != NULL) {
#if 0
	if (!(spIsFrame(child) == SP_TRUE && SpFramePart(child).internal_flag == SP_TRUE)
	    && spIsCreated(child) == SP_TRUE) {
	    if (child == component) {
		flag = 1;
	    }
	    ++num_window;
	    if (spIsDialog(child) == SP_TRUE && spIsVisible(child) == SP_FALSE) {
		++num_dialog;
	    }
	}
#else
	if (!(spIsFrame(child) == SP_TRUE && SpFramePart(child).internal_flag == SP_TRUE)
	    && spIsVisible(child) == SP_TRUE) {
	    if (child == component) {
		flag = 1;
	    }
	    ++num_window;
	}
#endif
	
	child = SpGetNextComponent(child);
    }

    spDebug(50, "spIsLastWindow", "flag = %d, num_window = %ld, num_dialog = %ld\n",
	    flag, num_window, num_dialog);

    if (flag && num_window - num_dialog <= 1)
	return SP_TRUE;
    else
	return SP_FALSE;
}

spBool spIsAncestor(spComponent component, spComponent target)
{
    spComponent next;
    
    if (component == NULL || target == NULL) return SP_FALSE;

    spDebug(100, "spIsAncestor", "component = %ld, spGetComponent = %ld, target= %ld\n",
            (long)component, (long)spGetComponent(component), (long)target);
    
    if (component == target || component == spGetComponent(target)) return SP_TRUE;
    
    next = SpGetParent(component);
    while (next != NULL) {
	if (next == target || next == spGetComponent(target)) {
	    return SP_TRUE;
	}
	next = SpGetParent(next);
    }
    
    return SP_FALSE;
}

spBool spIsDescendant(spComponent component, spComponent target)
{
    spComponent next;
    
    if (component == NULL || target == NULL) return SP_FALSE;

    if (component == target || spGetComponent(component) == target) return SP_TRUE;
    
    next = SpGetChild(component);
    while (next != NULL) {
	if (next == target || spGetComponent(next) == target
            || spIsDescendant(next, target) == SP_TRUE) {
	    return SP_TRUE;
	}
	next = SpGetNextComponent(next);
    }
    
    return SP_FALSE;
}

int spCountChildren(spComponent component)
{
    int i;
    spComponent child;

    child = spGetChild(component);

    for (i = 0; child != NULL; i++) {
	child = SpGetNextComponent(child);
    }

    return i;
}

void *spGetInternalCallbackData(spComponent component)
{
    if (component != NULL) {
	return SpComponentPart(component).call_data;
    } else {
	return NULL;
    }
}
