/*
 *	spContainer_Win.c
 */

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spTabBox.h>

#include <sp/spContainerP.h>

#include <uxtheme.h>
#include <vssym32.h>

extern spTopLevel sp_toplevel;

static void drawContainer(spComponent container, HDC hdc, spBool null_flag)
{
    RECT rect;
    HPEN prev_pen;

    if (spIsVisible(container) == SP_FALSE
	|| SpPrimitiveArch(SpGetWindow(container)).realize_flag == SP_FALSE) return;

    GetClientRect(SpPrimitiveArch(container).hwnd, &rect);
	
    prev_pen = SelectObject(hdc, SpTopLevelArch(sp_toplevel).null_pen);

    /* draw container edge */
    if (null_flag == SP_TRUE) {
	HBRUSH brush;
	HGDIOBJ prev_obj;
	
	spDebug(80, "drawContainer", "call FillRect\n");
	brush = GetSysColorBrush(COLOR_BTNFACE);
	prev_obj = SelectObject(hdc, brush);
	FillRect(hdc, &rect, brush);
	SelectObject(hdc, prev_obj);
    } else {
	int title_len;
	int x, y;
	const char *title;
	SIZE size;
	RECT str_rect;
	HFONT prev_font;
	COLORREF color, prev_color;
	HANDLE htheme = NULL;
	spBool drawn_flag = SP_FALSE;
	
	if (SpContainerPart(container).border_on == SP_TRUE) {
	    if (spIsOSWinXPLater() == SP_TRUE) {
		if ((htheme = OpenThemeData(SpPrimitiveArch(container).hwnd, L"BUTTON")) == NULL) {
		    spDebug(80, "drawContainer", "OpenThemeData failed\n");
		}
	    }
	    
	    /* draw title */
	    title = spGetTitle(container);
	    if (SpContainerPart(container).title_on == SP_TRUE && !strnone(title)) {
		title_len = (int)strlen(title);
		
		prev_font = SelectObject(hdc, SpTopLevelArch(sp_toplevel).sys_font);
		GetTextExtentPoint32(hdc, title, title_len, &size);
		
		color = 0L;
		
		x = spScaleX(SP_CONTAINER_TITLE_LEFT_OFFSET);
		y = spScaleY(2);
		rect.top = spScaleY(3) + size.cy / 2;
		/*rect.top = spScaleY(SP_CONTAINER_TITLE_TOP_OFFSET);*/

		SetBkMode(hdc, TRANSPARENT);
		/*SelectObject(hdc, GetStockObject(BLACK_PEN));*/
		prev_color = SetTextColor(hdc, color);

		str_rect.left = x;
		str_rect.right = x + size.cx;
		str_rect.top = y;
		str_rect.bottom = y + size.cy;
		
		if (htheme != NULL) {
		    int nchar;
		    WCHAR *wstr;

		    nchar = MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED, (LPSTR)title, -1, NULL, 0);
		    wstr = xalloc((nchar + 1) * 2, WCHAR);

		    if (MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED,
					    (LPSTR)title, -1, wstr, nchar) > 0) {
			DrawThemeText(htheme, hdc, BP_GROUPBOX, GBS_NORMAL, wstr, nchar - 1,
				      DT_SINGLELINE | DT_CENTER | DT_VCENTER, 0, &str_rect);
		    }

		    xfree(wstr);
		} else {
		    TextOut(hdc, x, y, title, title_len);
		}
		
		SelectObject(hdc, prev_font);
		SetTextColor(hdc, prev_color);

		InflateRect(&str_rect, spScaleX(2), spScaleY(2));
		ExcludeClipRect(hdc, str_rect.left, str_rect.top, str_rect.right, str_rect.bottom);
	    }
	    
	    if (htheme != NULL) {
		spDebug(80, "drawContainer", "call DrawThemeBackground\n");
		DrawThemeBackground(htheme, hdc, BP_GROUPBOX, GBS_NORMAL, &rect, NULL);
		CloseThemeData(htheme);
		drawn_flag = SP_TRUE;
	    }

	    if (drawn_flag == SP_FALSE) {
		spDebug(80, "drawContainer", "normal DrawEdge\n");
		DrawEdge(hdc, &rect, EDGE_ETCHED, BF_RECT);
	    }
	}
    }

    SelectObject(hdc, prev_pen);

    spDebug(80, "drawContainer", "done\n");
    
    return;
}

void spDrawContainerWin(spComponent component, spBool null_flag)
{
    HDC hdc;

    hdc = GetDC(SpPrimitiveArch(component).hwnd);
    drawContainer(component, hdc, null_flag);
    ReleaseDC(SpPrimitiveArch(component).hwnd, hdc);
    
    spDebug(80, "spDrawContainerWin", "done\n");

    return;
}

static void drawContainerCB(spComponent component, void *data)
{
    PAINTSTRUCT paintstruct;
    spComponent container = (spComponent)data;

    if (SpPrimitiveArch(component).message == WM_PAINT) {
	spBeginPaintWin(component, &paintstruct);
	drawContainer(container, paintstruct.hdc, SP_FALSE);
	spEndPaintWin(component, &paintstruct);
    }

    return;
}

void spContainerCreateArch(spComponent component)
{
    DWORD dwStyle = 0;
    DWORD dwExStyle = WS_EX_CONTROLPARENT;
    HWND hparent;
    
    if (SpContainerPart(component).border_on == SP_TRUE) {
	SpComponentPart(component).border_width = spScaleX(2);
    }

    dwStyle |= WS_CHILD | WS_CLIPSIBLINGS | WS_CLIPCHILDREN | WS_GROUP;
    if (spIsVisible(component) == SP_TRUE) {
	dwStyle |= WS_VISIBLE;
    }
    if (spIsTabBox(SpGetParent(component)) == SP_TRUE) {
	hparent = SpParentPrimitiveArch(SpGetParent(component)).hwnd;
    } else {
	hparent = SpParentPrimitiveArch(component).hwnd;
    }

    if (SpContainerPart(component).title_on == SP_TRUE) {
    } else if (SpContainerPart(component).border_on == SP_TRUE) {
	dwExStyle |= WS_EX_CLIENTEDGE;
    }

    SpPrimitiveArch(component).hwnd =
	CreateWindowEx(dwExStyle,
		       SP_WINDOW_CLASS_NAME,
		       (!strnone(SpGetName(component)) ? SpGetName(component) : ""),
		       dwStyle,
		       SpComponentPart(component).x, SpComponentPart(component).y,
		       SpComponentPart(component).current_width,
		       SpComponentPart(component).current_height,
		       hparent,
		       (HMENU)SpComponentPart(component).component_id,
		       SpTopLevelArch(sp_toplevel).hThisInst,
		       NULL);
    
    if (spIsTabBox(SpGetParent(component)) == SP_TRUE) {
	SetWindowPos(SpParentPrimitiveArch(component).hwnd,
		     SpPrimitiveArch(component).hwnd,
		     0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE | SWP_NOACTIVATE);
    }

    if (SpContainerPart(component).title_on == SP_TRUE) {
	SpComponentPart(component).border_width = spScaleX(2);
	SpContainerPart(component).title_height = spScaleY(SP_CONTAINER_TITLE_TOP_OFFSET - 4);
	SpComponentPart(component).top_offset += spScaleY(SP_CONTAINER_TITLE_TOP_OFFSET);;
    }
    
    spAddCallback(component, SP_EXPOSE_CALLBACK, drawContainerCB, component);
    
    return;
}

void spContainerSetParamsArch(spComponent component)
{
    return;
}
