#include <sp/spBaseLib.h>
#include <sp/spComponentLib.h>
#include <sp/spCanvasP.h>
#include <sp/spGraphicsP.h>
#include <sp/spPrimitiveP.h>
#include <sp/spDrawP.h>

#include <sp/spDIBitmapP.h>

extern spTopLevel sp_toplevel;

spBool spCheckBasicDIBitmapStrictMaskGtk(spDIBitmapInfoInternal *info, unsigned long strict_mask)
{
    if (((strict_mask & SP_DI_BITMAP_STRICT_GRAYSCALE_STATE)
	 && (info->pixel_format & SP_DI_PIXEL_FORMAT_GRAYSCALE_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_YUV_STATE)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_YUV_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_UPSIDE_DOWN_STATE)
	    && info->upside_down != SP_FALSE)
	|| ((strict_mask & SP_DI_BITMAP_STRICT_BIT_SIZE)
	    && info->bit_size != 8)) {
	spDebug(10, "spCheckBasicDIBitmapStrictMaskGtk", "initial check failed\n");
	return SP_FALSE;
    }

#if GTK_CHECK_VERSION(2,0,0)
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    if (((strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
         && !(((BYTE_ORDER == BIG_ENDIAN) && !(info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK))
              || ((BYTE_ORDER != BIG_ENDIAN) && (info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK))))
        || ((strict_mask & SP_DI_BITMAP_STRICT_ALPHA_ORDER)
            && !(((BYTE_ORDER == BIG_ENDIAN) && (info->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK))
                 || ((BYTE_ORDER != BIG_ENDIAN) && (info->pixel_format & SP_DI_PIXEL_FORMAT_LAST_ALPHA_MASK))))
        || ((strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
	    && info->pixel_stride != 4)) {
	return SP_FALSE;
    }
#else
    if ((strict_mask & SP_DI_BITMAP_STRICT_ALPHA_ORDER)
        && (info->pixel_format & SP_DI_PIXEL_FORMAT_FIRST_ALPHA_MASK)) {
	return SP_FALSE;
    }
    if ((strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
	&& (info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK)) {
	return SP_FALSE;
    }
#endif
#else
    if (((strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
	 && (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_PIXEL_ORDER)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_INVERSE_MASK))
	|| ((strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
	    && info->pixel_stride != 3)) {
	return SP_FALSE;
    }
#endif
    
    spDebug(100, "spCheckBasicDIBitmapStrictMaskGtk", "done\n");
    
    return SP_TRUE;
}

spBool spUpdateDIBitmapInfoArch(spDIBitmapInfoInternal *info, unsigned long *strict_mask)
{
    if (spCheckBasicDIBitmapStrictMaskGtk(info, *strict_mask) == SP_FALSE) {
	return SP_FALSE;
    }
    
    info->upside_down = SP_FALSE;
    info->bit_size = 8;
    info->pixel_stride = 3;
    
#if GTK_CHECK_VERSION(2,0,0)
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
	info->pixel_format =
#if BYTE_ORDER == BIG_ENDIAN
            SP_DI_PIXEL_FORMAT_ARGB
#else
            SP_DI_PIXEL_FORMAT_BGRA
#endif
            | SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK;
    } else {
#if BYTE_ORDER == BIG_ENDIAN
	info->pixel_format = SP_DI_PIXEL_FORMAT_RGB;
#else
	info->pixel_format = SP_DI_PIXEL_FORMAT_BGR;
#endif
    }
    info->pixel_stride = 4;
#else
    if (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) {
	info->pixel_format = SP_DI_PIXEL_FORMAT_RGBA;
	info->pixel_stride = 4;
    } else {
	info->pixel_format = SP_DI_PIXEL_FORMAT_RGB;
    }
#endif
#else
    info->pixel_format = SP_DI_PIXEL_FORMAT_RGB;
#endif
    
    *strict_mask |= (SP_DI_BITMAP_STRICT_ALL & ~SP_DI_BITMAP_STRICT_ALPHA_STATE);
    
    return SP_TRUE;
}

void *spCreateDIBitmapArch(spDIBitmapInfoInternal *info, unsigned long strict_mask)
{
    void *bitmap;
    unsigned long strict_mask2;

    spDebug(100, "spCreateDIBitmapArch", "strict_mask = %lx\n", strict_mask);
    
    strict_mask2 = strict_mask;

    if (spUpdateDIBitmapInfoArch(info, &strict_mask2) == SP_FALSE) {
	spDebug(10, "spCreateDIBitmapArch", "spUpdateDIBitmapInfoArch failed\n");
	return NULL;
    }
    
#if GTK_CHECK_VERSION(2,0,0)
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    bitmap = cairo_image_surface_create((info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)
					? CAIRO_FORMAT_ARGB32 : CAIRO_FORMAT_RGB24,
					info->width, info->height);
    info->data = cairo_image_surface_get_data((cairo_surface_t *)bitmap);
    info->line_stride = cairo_image_surface_get_stride((cairo_surface_t *)bitmap);
    info->pixel_stride = info->line_stride / info->width;
    spDebug(80, "spCreateDIBitmapArch", "reference count = %d\n", cairo_surface_get_reference_count((cairo_surface_t *)bitmap));
#else
    bitmap = gdk_pixbuf_new(GDK_COLORSPACE_RGB, (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) ? TRUE : FALSE, 8,
			    info->width, info->height);
    info->data = (unsigned char *)gdk_pixbuf_get_pixels((GdkPixbuf *)bitmap);
    info->line_stride = gdk_pixbuf_get_rowstride((GdkPixbuf *)bitmap);
    memset(info->data, 0, info->line_stride * info->height);
#endif
#else /* GTK_CHECK_VERSION(2,0,0) */
    info->line_stride = info->width * info->pixel_stride;
    bitmap = xalloc(info->line_stride * info->height, char);
    memset(bitmap, 0, info->line_stride * info->height);
    info->data = (unsigned char *)bitmap;
#endif /* !GTK_CHECK_VERSION(2,0,0) */
    spDebug(80, "spCreateDIBitmapArch", "width = %ld, height = %ld, pixel_stride = %ld, line_stride = %ld\n",
	    info->width, info->height, info->pixel_stride, info->line_stride);

    return bitmap;
}

spBool spDestroyDIBitmapArch(void *bitmap)
{
    spDebug(80, "spDestroyDIBitmapArch", "in\n");
    
    if (bitmap != NULL) {
#if GTK_CHECK_VERSION(2,0,0)
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
	cairo_surface_destroy((cairo_surface_t *)bitmap);
#else
	g_object_unref(bitmap);
#endif
#else
	xfree(bitmap);
#endif
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}


#if defined(GTK3)
spBool spLockDIBitmapArch(spDIBitmapInfoInternal *info, void *bitmap)
{
#if defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    cairo_surface_flush((cairo_surface_t *)bitmap);
#endif
    return SP_TRUE;
}

spBool spUnlockDIBitmapArch(spDIBitmapInfoInternal *info, void *bitmap)
{
#if defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    cairo_surface_mark_dirty((cairo_surface_t *)bitmap);
#endif
    return SP_TRUE;
}
#endif

#if !GTK_CHECK_VERSION(2,0,0)
static spBool sp_gdk_rgb_init_done = SP_FALSE;
#endif

spBool spCopyDIBitmapToImageArch(spDIBitmapInfoInternal *info, void *bitmap, spComponent dest, int src_x, int src_y,
				 int width, int height, int dest_x, int dest_y)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpPrimitiveArch(dest).pixmap_cr == NULL) return SP_FALSE;

    spDebug(80, "spCopyDIBitmapToImageArch", "src_x = %d, src_y = %d, width = %d, height = %d, dest_x = %d, dest_y = %d\n",
	    src_x, src_y, width, height, dest_x, dest_y);
    
#if defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    spDebug(80, "spCopyDIBitmapToImageArch", "reference count = %d\n", cairo_surface_get_reference_count((cairo_surface_t *)bitmap));
    cairo_set_source_surface(SpPrimitiveArch(dest).pixmap_cr,
			     (cairo_surface_t *)bitmap, dest_x - src_x, dest_y - src_y);
#else
    gdk_cairo_set_source_pixbuf(SpPrimitiveArch(dest).pixmap_cr,
				(GdkPixbuf *)bitmap, dest_x - src_x, dest_y - src_y);
#endif
    spDebug(80, "spCopyDIBitmapToImageArch", "set source done\n");
    
    cairo_rectangle(SpPrimitiveArch(dest).pixmap_cr, dest_x, dest_y, width, height);
    cairo_fill(SpPrimitiveArch(dest).pixmap_cr);
#elif GTK_CHECK_VERSION(2,0,0)
    gdk_pixbuf_render_to_drawable((GdkPixbuf *)bitmap,
				  SpPrimitiveArch(dest).pixmap,
				  SpTopLevelArch(sp_toplevel).fg_gc,
				  src_x, src_y,
				  dest_x, dest_y,
				  width, height,
				  GDK_RGB_DITHER_NONE, 0, 0);
    
#else
    if (sp_gdk_rgb_init_done == SP_FALSE) {
	gdk_rgb_init();
	sp_gdk_rgb_init_done = SP_TRUE;
    }
	
    gdk_draw_rgb_image(SpPrimitiveArch(dest).pixmap,
		       SpTopLevelArch(sp_toplevel).fg_gc,
		       src_x, src_y, width, height, GDK_RGB_DITHER_NONE,
		       bitmap, info->line_stride);
#endif

    spDebug(80, "spCopyDIBitmapToImageArch", "done\n");
    
    return SP_TRUE;
}

void *spCopyImageToDIBitmapArch(spComponent src, spDIBitmapInfoInternal *info, void *bitmap, int src_x, int src_y,
				int width, int height, int dest_x, int dest_y)
{
#if GTK_CHECK_VERSION(3,0,0)
    spBool flag = SP_FALSE;
#if defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    cairo_t *cr;
    cairo_surface_t *surface = NULL;
    
    surface = (cairo_surface_t *)bitmap;

    if ((cr = cairo_create(surface)) != NULL) {
	cairo_set_source_surface(cr, SpPrimitiveArch(src).pixmap, dest_x - src_x, dest_y - src_y);
	cairo_rectangle(cr, dest_x, dest_y, width, height);
	cairo_fill(cr);
	cairo_destroy(cr);
	flag = SP_TRUE;
    } else {
        spDebug(10, "spCopyImageToDIBitmapArch", "cairo_create failed\n");
    }
#else
    GdkPixbuf *src_pixbuf;

    if ((src_pixbuf = gdk_pixbuf_get_from_surface(SpPrimitiveArch(src).pixmap,
                                                  src_x, src_y, width, height)) != NULL) {
        gdk_pixbuf_copy_area(src_pixbuf, 0, 0, width, height, (GdkPixbuf *)bitmap, dest_x, dest_y);
        g_object_unref(src_pixbuf);
	flag = SP_TRUE;
    } else {
        spDebug(10, "spCopyImageToDIBitmapArch", "gdk_pixbuf_get_from_surface failed\n");
    }
#endif

    if (flag) {
	return bitmap;
    } else {
	return NULL;
    }
#elif GTK_CHECK_VERSION(2,0,0)
    bitmap = gdk_pixbuf_get_from_drawable((GdkPixbuf *)bitmap,
					  SpPrimitiveArch(src).pixmap,
					  gdk_colormap_get_system(),
					  src_x, src_y,
					  dest_x, dest_y,
					  width, height);
    return bitmap;
#else
    GdkImage *image;
    GdkColormap *cm;
#if !GTK_CHECK_VERSION(2,0,0)
    GdkColorContext *cc;
#endif
    int x, y;
    guchar *rgb_buf;
    GdkColor color;

#if GTK_CHECK_VERSION(2,0,0)
    image = gdk_drawable_get_image(SpPrimitiveArch(src).pixmap, src_x, src_y, width, height);
#else
    image = gdk_image_get(SpPrimitiveArch(src).pixmap, src_x, src_y, width, height);
#endif
    if (image == NULL) {
	return NULL;
    }
    cm = gdk_colormap_get_system();
#if !GTK_CHECK_VERSION(2,0,0)
    cc = gdk_color_context_new(gdk_visual_get_system(), cm);
#endif
    rgb_buf = (unsigned char *)bitmap;
	
    for (y = 0; y < height; y++) {
	for (x = 0; x < width; x++) {
	    color.pixel = gdk_image_get_pixel(image, x, y);
#if GTK_CHECK_VERSION(2,0,0)
	    gdk_colormap_query_color(cm, color.pixel, &color);
#else
	    gdk_color_context_query_color(cc, &color);
#endif
	    *rgb_buf++ = color.red / 256;
	    *rgb_buf++ = color.green / 256;
	    *rgb_buf++ = color.blue / 256;
	}
    }
#if !GTK_CHECK_VERSION(2,0,0)
    gdk_color_context_free(cc);
#endif

    return bitmap;
#endif
}

typedef struct _spDIBitmapInstanceArch {
    int type_index;
#if GTK_CHECK_VERSION(2,0,0)
    char *filename;
    GdkPixbuf *pixbuf;
#endif
} *spDIBitmapInstanceArch;

void *spInitDIBitmapInstanceArch(const char *name)
{
    spDIBitmapInstanceArch dibinstance = NULL;
    
    dibinstance = xalloc(1, struct _spDIBitmapInstanceArch);
    memset(dibinstance, 0, sizeof(struct _spDIBitmapInstanceArch));
    dibinstance->type_index = -1;
    
    spDebug(80, "spInitDIBitmapInstanceArch", "done\n");
    
    return dibinstance;
}

spBool spFreeDIBitmapInstanceArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    xfree(dibinstance);
    
    spDebug(80, "spFreeDIBitmapInstanceArch", "done\n");
    
    return SP_TRUE;
}

int spGetDIBitmapPriorityArch(void *instance)
{
    return SP_PLUGIN_PRIORITY_MIDDLE;
}

unsigned long spGetDIBitmapCapsArch(void *instance)
{
    unsigned long caps;
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    caps = SP_DI_BITMAP_CAPS_DIRECT_ACCESS
	| SP_DI_BITMAP_CAPS_COPY_IMAGE | SP_DI_BITMAP_CAPS_COPY_IMAGE_STRETCHABLE;

#if GTK_CHECK_VERSION(2,0,0)
    caps |= SP_DI_BITMAP_CAPS_PIXEL_STRIDE_4 | SP_DI_BITMAP_CAPS_READ | SP_DI_BITMAP_CAPS_WRITE
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
        /*| SP_DI_BITMAP_CAPS_READ_USE_BITMAP_ARGUMENT*/ /*| SP_DI_BITMAP_CAPS_WRITE_USE_BITMAP_ARGUMENT*/
#else
        | SP_DI_BITMAP_CAPS_PIXEL_STRIDE_3
#endif
        ;
        
    if (dibinstance->type_index == 0 || dibinstance->type_index == 3) {
	caps |= 
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
	    SP_DI_BITMAP_CAPS_ALPHA_PREMULTIPLIED | SP_DI_BITMAP_CAPS_COPY_ALPHA
#else
	    SP_DI_BITMAP_CAPS_ALPHA
#endif
	    ;
    }
#else /* GTK_CHECK_VERSION(2,0,0) */
    caps |= SP_DI_BITMAP_CAPS_PIXEL_STRIDE_3;
#endif /* !GTK_CHECK_VERSION(2,0,0) */
    
    return caps;
}

#if GTK_CHECK_VERSION(2,0,0)
static const char *sp_dibitmap_file_type_list_arch[] = {
    "png",
    "jpeg",
    "bmp",
    "tiff",
    NULL,
};

static const char *sp_dibitmap_file_desc_list_arch[] = {
    "PNG (GDK-PixBuf)",
    "JPEG (GDK-PixBuf)",
    "BMP (GDK-PixBuf)",
    "TIFF (GDK-PixBuf)",
    NULL,
};

static const char *sp_dibitmap_file_filter_list_arch[] = {
    "*.png",
    "*.jpg",
    "*.bmp",
    "*.tif",
    NULL,
};
#endif

const char **spGetDIBitmapFileTypeListArch(void *instance)
{
#if GTK_CHECK_VERSION(2,0,0)
    return sp_dibitmap_file_type_list_arch;
#else
    return NULL;
#endif
}

const char **spGetDIBitmapFileDescListArch(void *instance)
{
#if GTK_CHECK_VERSION(2,0,0)
    return sp_dibitmap_file_desc_list_arch;
#else
    return NULL;
#endif
}

const char **spGetDIBitmapFileFilterListArch(void *instance)
{
#if GTK_CHECK_VERSION(2,0,0)
    return sp_dibitmap_file_filter_list_arch;
#else
    return NULL;
#endif
}

#if GTK_CHECK_VERSION(2,0,0)
static int getTypeIndexFromSuffix(const char *suffix)
{
    int type_index = -1;

    if (strcaseeq(suffix, ".png")) {
	type_index = 0;
    } else if (strcaseeq(suffix, ".jpg") || strcaseeq(suffix, ".jpeg")) {
	type_index = 1;
    } else if (strcaseeq(suffix, ".bmp")) {
	type_index = 2;
    } else if (strcaseeq(suffix, ".tif") || strcaseeq(suffix, ".tiff")) {
	type_index = 3;
    }
    
    return type_index;
}
#endif

spBool spIsDIBitmapSupportedArch(void *instance, const char *filename)
{
#if GTK_CHECK_VERSION(2,0,0)
    spBool supported = SP_FALSE;

    if (spGetBaseName(filename) == filename && filename[0] == '.') {
	if (getTypeIndexFromSuffix(filename) >= 0) {
	    supported = SP_TRUE;
	}
    } else {
	GError *error = NULL;
	GdkPixbuf *pixbuf;
	
	if ((pixbuf = gdk_pixbuf_new_from_file(filename, &error)) != NULL) {
	    supported = SP_TRUE;
	    g_object_unref(pixbuf);
	} else {
	    supported = SP_FALSE;
	}
    }
    
    return supported;
#else
    return SP_FALSE;
#endif
}

spBool spSetDIBitmapFileTypeArch(void *instance, int index)
{
#if GTK_CHECK_VERSION(2,0,0)
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (index >= 0 && index <= 3) {
	dibinstance->type_index = index;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
#else
    return SP_FALSE;
#endif
}

int spGetDIBitmapFileTypeArch(void *instance)
{
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    return dibinstance->type_index;
}

const char *spGetDIBitmapBestSuffixArch(void *instance)
{
#if GTK_CHECK_VERSION(2,0,0)
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (dibinstance->type_index == 0) {
	return ".png";
    } else if (dibinstance->type_index == 1) {
	return ".jpg";
    } else if (dibinstance->type_index == 2) {
	return ".bmp";
    } else if (dibinstance->type_index == 3) {
	return ".tif";
    } else {
	return NULL;
    }
#else
    return NULL;
#endif
}

spBool spOpenDIBitmapFileArch(void *instance, const char *filename, const char *mode)
{
#if GTK_CHECK_VERSION(2,0,0)
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (mode[0] == 'w') {
	if (dibinstance->type_index < 0) {
	    if ((dibinstance->type_index = getTypeIndexFromSuffix(spGetSuffix(filename))) < 0) {
		return SP_FALSE;
	    }
	}
	dibinstance->filename = strclone(filename);
	dibinstance->pixbuf = NULL;
	
	return SP_TRUE;
    } else {
	GError *error = NULL;
	
	dibinstance->filename = NULL;
	dibinstance->type_index = -1;
	
	if ((dibinstance->pixbuf = gdk_pixbuf_new_from_file(filename, &error)) == NULL) {
	    return SP_FALSE;
	} else {
	    dibinstance->type_index = getTypeIndexFromSuffix(filename);

	    return SP_TRUE;
	}
    }
#else
    return SP_FALSE;
#endif
}

spBool spCloseDIBitmapFileArch(void *instance)
{
#if GTK_CHECK_VERSION(2,0,0)
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (dibinstance->pixbuf != NULL) {
	g_object_unref(dibinstance->pixbuf);
	dibinstance->pixbuf = NULL;
    }
    if (dibinstance->filename != NULL) {
	xfree(dibinstance->filename);
	dibinstance->filename = NULL;
    }
    dibinstance->type_index = -1;
#endif
    
    return SP_TRUE;
}

spBool spGetDIBitmapInfoArch(void *instance, spDIBitmapInfoInternal *info, unsigned long *strict_mask)
{
#if GTK_CHECK_VERSION(2,0,0)
    int rowstride;
    int n_channels;
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    spDebug(100, "spGetDIBitmapInfoArch", "in\n");
    
    if (dibinstance->pixbuf == NULL) {
	spDebug(10, "spGetDIBitmapInfoArch", "no pixbuf\n");
	return SP_FALSE;
    }
    
    if (spCheckBasicDIBitmapStrictMaskGtk(info, *strict_mask) == SP_FALSE) {
	spDebug(10, "spGetDIBitmapInfoArch", "spCheckBasicDIBitmapStrictMaskGtk failed: strict_mask = %lx\n", *strict_mask);
	return SP_FALSE;
    }

#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    n_channels = 4;
#else
    n_channels = gdk_pixbuf_get_n_channels(dibinstance->pixbuf);
#endif
    if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
	&& info->pixel_stride != n_channels) {
	spDebug(10, "spGetDIBitmapInfoArch", "wrong n_channels: %d\n", n_channels);
	return SP_FALSE;
    }

#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    rowstride = n_channels * gdk_pixbuf_get_width(dibinstance->pixbuf);
#else
    rowstride = gdk_pixbuf_get_rowstride(dibinstance->pixbuf);
#endif
    if ((*strict_mask & SP_DI_BITMAP_STRICT_PIXEL_STRIDE)
	&& info->line_stride != rowstride) {
	spDebug(10, "spGetDIBitmapInfoArch", "wrong row_stride: %d\n", rowstride);
	return SP_FALSE;
    }
    
    if (gdk_pixbuf_get_has_alpha(dibinstance->pixbuf)) {
	if ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)) {
	    return SP_FALSE;
	}
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
	if ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
	    && !(info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK)) {
	    return SP_FALSE;
	}
	info->pixel_format =
#if BYTE_ORDER == BIG_ENDIAN
            SP_DI_PIXEL_FORMAT_ARGB
#else
            SP_DI_PIXEL_FORMAT_BGRA
#endif
            | SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK;
#else
	if ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_PREMULTIPLIED_STATE)
	    && (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_PREMULTIPLIED_MASK)) {
	    return SP_FALSE;
	}
	info->pixel_format = SP_DI_PIXEL_FORMAT_RGBA;
#endif
    } else {
	if ((*strict_mask & SP_DI_BITMAP_STRICT_ALPHA_STATE)
	    && !(info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK)) {
	    spDebug(10, "spGetDIBitmapInfoArch", "alpha mask exists though file doesn't include alpha\n");
	    return SP_FALSE;
	}
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
        if (BYTE_ORDER == BIG_ENDIAN) {
            info->pixel_format = SP_DI_PIXEL_FORMAT_NRGB;
        } else {
            info->pixel_format = SP_DI_PIXEL_FORMAT_BGRN;
        }
#else
	info->pixel_format = SP_DI_PIXEL_FORMAT_RGB;
#endif
    }
    info->upside_down = SP_FALSE;
    info->bit_size = 8;

    spDebug(80, "spGetDIBitmapInfoArch", "rowstride = %d, n_channels = %d\n", rowstride, n_channels);
    
    info->line_stride = rowstride;
    info->pixel_stride = n_channels;
    info->width = gdk_pixbuf_get_width(dibinstance->pixbuf);
    info->height = gdk_pixbuf_get_height(dibinstance->pixbuf);

    *strict_mask |= (SP_DI_BITMAP_STRICT_ALL & ~SP_DI_BITMAP_STRICT_ALPHA_STATE);
    
    spDebug(100, "spGetDIBitmapInfoArch", "done\n");
    
    return SP_TRUE;
#else
    return SP_FALSE;
#endif
}

void *spReadDIBitmapArch(void *instance, spDIBitmapInfoInternal *info, void *bitmap)
{
#if GTK_CHECK_VERSION(2,0,0)
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    cairo_t *cr;
    cairo_surface_t *surface = NULL;
#endif
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;
    
    if (dibinstance->pixbuf == NULL) {
	return SP_FALSE;
    }

#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    if (info->pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
        spDIPixelFormat pixel_format = 0L;
        
        pixel_format = SP_DI_PIXEL_FORMAT_MASK & info->pixel_format;

        if (info->pixel_stride != 4
            || !(((BYTE_ORDER == BIG_ENDIAN)
                  && ((pixel_format == SP_DI_PIXEL_FORMAT_PRGB) || (pixel_format == SP_DI_PIXEL_FORMAT_NRGB)))
                 || ((BYTE_ORDER != BIG_ENDIAN)
                     && ((pixel_format == SP_DI_PIXEL_FORMAT_BGRP) || (pixel_format == SP_DI_PIXEL_FORMAT_BGRN))))) {
            spDebug(80, "spReadDIBitmapArch", "nonnative case: not supported\n");
            return SP_FALSE;
        }

        memset(info->data, 0, info->width * 4 * info->height);
        surface = cairo_image_surface_create_for_data(info->data,
                                                      (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) ? CAIRO_FORMAT_ARGB32 : CAIRO_FORMAT_RGB24,
                                                      info->width, info->height,
                                                      info->width * 4);
        
    } else {
        surface = (cairo_surface_t *)bitmap;
    }

    if (surface != NULL) {
	if ((cr = cairo_create(surface)) != NULL) {
	    gdk_cairo_set_source_pixbuf(cr, dibinstance->pixbuf, 0, 0);
	    cairo_rectangle(cr, 0, 0, info->width, info->height);
	    cairo_fill(cr);
	    cairo_destroy(cr);
	}
        if (surface != (cairo_surface_t *)bitmap) {
            cairo_surface_destroy(surface);
        }
    } else {
        spDebug(10, "spReadDIBitmapArch", "error: NULL surface\n");
    }
#else
    memcpy(info->data, gdk_pixbuf_get_pixels(dibinstance->pixbuf), info->line_stride * info->height);
#endif
    
    return bitmap;
#else
    return NULL;
#endif
}

spBool spWriteDIBitmapArch(void *instance, spDIBitmapInfoInternal *info, void *bitmap)
{
#if GTK_CHECK_VERSION(2,0,0)
    char *suffix;
    char *format = NULL;
    GdkPixbuf *pixbuf = NULL;
    GError *error = NULL;
    spBool flag = SP_FALSE;
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    cairo_surface_t *image_surface = NULL;
#endif
    spDIBitmapInstanceArch dibinstance = (spDIBitmapInstanceArch)instance;

    if (strnone(dibinstance->filename)) return SP_FALSE;

    suffix = spGetSuffix(dibinstance->filename);
    if (dibinstance->type_index < 0 && suffix == NULL) {
	return SP_FALSE;
    }
    
    if (dibinstance->type_index == 0 || strcaseeq(suffix, ".png")) {
	format = "png";
	dibinstance->type_index = 0;
    } else if (dibinstance->type_index == 1 || strcaseeq(suffix, ".jpg") || strcaseeq(suffix, ".jpeg")) {
	format = "jpeg";
	dibinstance->type_index = 1;
    } else if (dibinstance->type_index == 2 || strcaseeq(suffix, ".bmp")) {
	format = "bmp";
	dibinstance->type_index = 2;
    } else if (dibinstance->type_index == 3 || strcaseeq(suffix, ".tif") || strcaseeq(suffix, ".tiff")) {
	format = "tiff";
	dibinstance->type_index = 1;
    } else {
	dibinstance->type_index = -1;
	return SP_FALSE;
    }
    spDebug(50, "spWriteDIBitmapArch", "type_index = %d\n", dibinstance->type_index);

#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    if (info->pixel_format & SP_DI_BITMAP_TYPE_NONNATIVE_MASK) {
        spDIPixelFormat pixel_format = 0L;
        pixel_format = SP_DI_PIXEL_FORMAT_MASK & info->pixel_format;
        spDebug(100, "spWriteDIBitmapArch",
                "nonnative case: width = %ld, height = %ld, pixel_stride = %ld, line_stride = %ld, bitmap = %ld\n",
                info->width, info->height, info->pixel_stride, info->line_stride, (long)bitmap);
        if (info->pixel_stride != 4
            || !(((BYTE_ORDER == BIG_ENDIAN)
                  && ((pixel_format == SP_DI_PIXEL_FORMAT_PRGB) || (pixel_format == SP_DI_PIXEL_FORMAT_NRGB)))
                 || ((BYTE_ORDER != BIG_ENDIAN)
                     && ((pixel_format == SP_DI_PIXEL_FORMAT_BGRP) || (pixel_format == SP_DI_PIXEL_FORMAT_BGRN))))) {
            spDebug(10, "spWriteDIBitmapArch", "not supported\n");
            return SP_FALSE;
        }
        image_surface = cairo_image_surface_create_for_data(bitmap,
                                                            (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) ? CAIRO_FORMAT_ARGB32 : CAIRO_FORMAT_RGB24,
                                                            info->width, info->height,
                                                            info->line_stride);
        if (image_surface != NULL) {
            pixbuf = gdk_pixbuf_get_from_surface(image_surface, 0, 0, info->width, info->height);
        } else {
            spDebug(50, "spWriteDIBitmapArch", "cairo_image_surface_create_for_data failed\n");
        }
    } else {
        pixbuf = gdk_pixbuf_get_from_surface((cairo_surface_t *)bitmap, 0, 0, info->width, info->height);
    }
#else
    pixbuf = gdk_pixbuf_new_from_data(info->data,
				      GDK_COLORSPACE_RGB, (info->pixel_format & SP_DI_PIXEL_FORMAT_ALPHA_MASK) ? TRUE : FALSE, 8,
				      info->width, info->height, info->line_stride,
				      NULL, NULL);
#endif

    if (pixbuf != NULL) {
	if (gdk_pixbuf_save(pixbuf, dibinstance->filename, format, &error, NULL)) {
	    flag = SP_TRUE;
	}
    
	g_object_unref(pixbuf);
    }
    
#if GTK_CHECK_VERSION(3,0,0) && defined(SP_DIBITMAP_USE_CAIRO_IMAGE)
    if (image_surface != NULL) {
        cairo_surface_destroy(image_surface);
    }
#endif
    
    return flag;
#else
    return SP_FALSE;
#endif
}
