/*
 *	spDraw.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spDialog.h>
#include <sp/spImage.h>

#include <sp/spCanvasP.h>
#include <sp/spDrawP.h>
#include <sp/spGraphicsP.h>
#include <sp/spDrawingPluginP.h>
#include <sp/spDialogBoxP.h>
#include <sp/spPluginCanvasP.h>

spBool spCopyImage(spComponent src, spComponent dest,
		   int src_x, int src_y, int width, int height,
		   int dest_x, int dest_y)
{
    if (spIsDrawable(src) == SP_FALSE) return SP_FALSE;
    
    if (spIsPluginCanvas(dest) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(dest) == SP_FALSE
	    || SpPluginCanvasDrawingRec(dest)->copy_image == NULL) return SP_FALSE;

	return SpPluginCanvasDrawingRec(dest)->copy_image(SpPluginCanvasInstance(dest), spGetImageHandleArch(src),
							  src_x, src_y, width, height,
							  dest_x, dest_y, width, height);
    } else {
	if (spIsDrawable(dest) == SP_FALSE) return SP_FALSE;

	return spCopyImageArch(src, dest, src_x, src_y, width, height, dest_x, dest_y);
    }
}

spBool spStretchImageDefault(spComponent src, spComponent dest,
			     int src_x, int src_y, int src_width, int src_height,
			     int dest_x, int dest_y, int dest_width, int dest_height)
{
    spDIBitmap dibitmap;
    spDIBitmapInfo dibinfo;
    spBool flag = SP_FALSE;
    
    dibinfo.upside_down = SP_FALSE;
    dibinfo.bit_size = 8;
    dibinfo.pixel_format = SP_DI_PIXEL_FORMAT_RGB;
    dibinfo.width = src_width;
    dibinfo.height = src_height;

    if ((dibitmap = spCreateDIBitmap(&dibinfo)) != NULL) {
	spLockDIBitmap(dibitmap);
	if (spCopyImageToDIBitmap(src, dibitmap, 
				  src_x, src_y, src_width, src_height,
				  dest_x, dest_y) == SP_TRUE) {
	    flag = spStretchDIBitmapToImage(dibitmap, dest,
					    src_x, src_y, src_width, src_height,
					    dest_x, dest_y, dest_width, dest_height);
	}
	spUnlockDIBitmap(dibitmap);
	
	spDestroyDIBitmap(dibitmap);
    }

    return flag;
}

spBool spStretchImage(spComponent src, spComponent dest,
		      int src_x, int src_y, int src_width, int src_height,
		      int dest_x, int dest_y, int dest_width, int dest_height)
{
    if (spIsDrawable(src) == SP_FALSE) return SP_FALSE;
    
    if (spIsPluginCanvas(dest) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(dest) == SP_FALSE
	    || SpPluginCanvasDrawingRec(dest)->copy_image == NULL) return SP_FALSE;

	return SpPluginCanvasDrawingRec(dest)->copy_image(SpPluginCanvasInstance(dest), spGetImageHandleArch(src),
							  src_x, src_y, src_width, src_height,
							  dest_x, dest_y, src_width, src_height);
    } else {
	if (spIsDrawable(dest) == SP_FALSE) return SP_FALSE;

	return spStretchImageArch(src, dest, src_x, src_y, src_width, src_height,
				  dest_x, dest_y, dest_width, dest_height);
    }
}

void spFillRectangle(spComponent component, spGraphics graphics,
		     int x, int y, int width, int height)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->fill_rectangle == NULL) return;
    
	SpPluginCanvasDrawingRec(component)->fill_rectangle(SpPluginCanvasInstance(component),
							    graphics, x, y, width, height);
    } else {
	if (spIsDrawable(component) == SP_FALSE) return;

	spFillRectangleArch(component, graphics, x, y, width, height);
    }
    
    return;
}

void spDrawRectangle(spComponent component, spGraphics graphics,
		     int x, int y, int width, int height)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->draw_rectangle == NULL) return;
    
	SpPluginCanvasDrawingRec(component)->draw_rectangle(SpPluginCanvasInstance(component),
							    graphics, x, y, width, height);
    } else {
	if (spIsDrawable(component) == SP_FALSE) return;

	spDrawRectangleArch(component, graphics, x, y, width, height);
    }
    
    return;
}

double spConvertArcAngleToSquareArc(double yfactor, double angle)
{
    double radian;
    double cos_value, sin_value;
    double o_angle;

    radian = PI * angle / 180.0;
    cos_value = cos(radian);
    
    if (cos_value == 0.0) return angle;

    sin_value = sin(radian);
    o_angle = 180.0 * atan2(sin_value / yfactor, cos_value) / PI;
    if (angle > 0.0 && o_angle < 0.0) {
	o_angle += 360.0;
    } else if (angle < 0.0 && o_angle > 0.0) {
	o_angle -= 360.0;
    }

    spDebug(80, "spConvertArcAngleToSquareArc", "angle = %f, o_angle = %f\n", angle, o_angle);
    
    return o_angle;
}

void spFillArc(spComponent component, spGraphics graphics,
	       int x, int y, int width, int height, double angle1, double angle2)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->fill_arc == NULL) return;
    
	SpPluginCanvasDrawingRec(component)->fill_arc(SpPluginCanvasInstance(component),
						      graphics, x, y, width, height, angle1, angle2);
    } else {
	if (spIsDrawable(component) == SP_FALSE) return;

	spFillArcArch(component, graphics, x, y, width, height, angle1, angle2);
    }
    
    return;
}

void spDrawArc(spComponent component, spGraphics graphics,
	       int x, int y, int width, int height, double angle1, double angle2)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->draw_arc == NULL) return;
    
	SpPluginCanvasDrawingRec(component)->draw_arc(SpPluginCanvasInstance(component),
						      graphics, x, y, width, height, angle1, angle2);
    } else {
	if (spIsDrawable(component) == SP_FALSE) return;

	spDrawArcArch(component, graphics, x, y, width, height, angle1, angle2);
    }
    
    return;
}

void spDrawLine(spComponent component, spGraphics graphics,
		int x1, int y1, int x2, int y2)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->draw_line == NULL) return;
    
	SpPluginCanvasDrawingRec(component)->draw_line(SpPluginCanvasInstance(component),
						       graphics, x1, y1, x2, y2);
    } else {
	if (spIsDrawable(component) == SP_FALSE) return;

	spDrawLineArch(component, graphics, x1, y1, x2, y2);
    }

    return;
}

void spDrawPoint(spComponent component, spGraphics graphics,
		 int x, int y)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->draw_point == NULL) return;
    
	SpPluginCanvasDrawingRec(component)->draw_point(SpPluginCanvasInstance(component),
							graphics, x, y);
    } else {
	if (spIsDrawable(component) == SP_FALSE) return;

	spDrawPointArch(component, graphics, x, y);
    }

    return;
}

void spDrawString(spComponent component, spGraphics graphics,
		  int x, int y, const char *string)
{
    if (strnone(string)) return;
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->draw_string == NULL) return;
    
	SpPluginCanvasDrawingRec(component)->draw_string(SpPluginCanvasInstance(component),
							 graphics, x, y, string);
    } else {
	if (spIsDrawable(component) == SP_FALSE) return;

	spDrawStringArch(component, graphics, x, y, string);
    }

    return;
}

spBool spGetStringExtent(spComponent component, spGraphics graphics, const char *string,
			 int *x, int *y, int *width, int *height, int *next_x)
{
    if (strnone(string)) return SP_FALSE;
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE) return SP_FALSE;

	if (SpPluginCanvasDrawingRec(component)->get_string_extent == NULL) {
	    return spGetStringExtentArch(NULL, graphics, string, x, y, width, height, next_x);
	}
    
	return SpPluginCanvasDrawingRec(component)->get_string_extent(SpPluginCanvasInstance(component),
								      graphics, string, x, y, width, height, next_x);
    } else {
	if (spIsDrawable(component) == SP_FALSE) return SP_FALSE;
	
	return spGetStringExtentArch(component, graphics, string,
				     x, y, width, height, next_x);
    }
}

int spGetNumPolygonPoint(spPolygon polygon)
{
    if (polygon == NULL) return 0;

    return polygon->num_point;
}

spPolygon spCreatePolygon(int max_point)
{
    spPolygon polygon;

    polygon = xalloc(1, struct _spPolygon);
    polygon->max_point = max_point;
    polygon->num_point = 0;
    polygon->end_flag = SP_FALSE;
    polygon->px = 0;
    polygon->py = 0;
    polygon->move_dx = 0;
    polygon->move_dy = 0;
    polygon->instance = NULL;
    if (spCreatePolygonArch(polygon) == SP_FALSE) {
	xfree(polygon);
	return NULL;
    }

    return polygon;
}

spBool spDestroyPolygon(spPolygon polygon)
{
    spBool flag;
    
    if (polygon == NULL) return SP_FALSE;

    flag = SP_TRUE;
    if (spDestroyPluginPolygonInstance(polygon) == SP_FALSE) {
	flag = SP_FALSE;
    }
    if (spDestroyPolygonArch(polygon) == SP_FALSE) {
	flag = SP_FALSE;
    }
    xfree(polygon);

    return flag;
}

spBool spStartPolygonPoint(spPolygon polygon, int x, int y)
{
    spBool flag;
    
    if (polygon == NULL) return SP_FALSE;

    if ((flag = spStartPolygonPointArch(polygon, x, y)) == SP_TRUE) {
	polygon->num_point = 1;
	polygon->px = x;
	polygon->py = y;
    }

    return flag;
}

spBool spAddPolygonPoint(spPolygon polygon, int x, int y)
{
    spBool flag;
    
    if (polygon == NULL || polygon->num_point < 1
	|| polygon->num_point + 1 > polygon->max_point) return SP_FALSE;

    if ((flag = spAddPolygonPointArch(polygon, x, y)) == SP_TRUE) {
	polygon->px = x;
	polygon->py = y;
	polygon->num_point++;
    }

    return flag;
}

spBool spAddPolygonRelativePoint(spPolygon polygon, int dx, int dy)
{
    spBool flag;
    
    if (polygon == NULL || polygon->num_point < 1
	|| polygon->num_point + 1 > polygon->max_point) return SP_FALSE;

    if ((flag = spAddPolygonRelativePointArch(polygon, dx, dy)) == SP_TRUE) {
	polygon->px += dx;
	polygon->py += dy;
	polygon->num_point++;
    }

    return flag;
}

spBool spEndPolygonPoint(spPolygon polygon)
{
    spBool flag;
    
    if (polygon == NULL || polygon->num_point <= 1) return SP_FALSE;

    if ((flag = spEndPolygonPointArch(polygon)) == SP_TRUE) {
	polygon->end_flag = SP_TRUE;
    }

    return flag;
}

spBool spGetPolygonPoint(spPolygon polygon, int index, int *x, int *y, spBool before_move)
{
    int tx, ty;
    
    if (polygon == NULL || index < 0 || index >= polygon->num_point) return SP_FALSE;

    if (spGetPolygonPointArch(polygon, index, &tx, &ty) == SP_TRUE) {
	if (before_move == SP_TRUE) {
	    tx -= polygon->move_dx;
	    ty -= polygon->move_dy;
	}
	if (x != NULL) *x = tx;
	if (y != NULL) *y = ty;
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spMovePolygon(spPolygon polygon, int dx, int dy)
{
    if (polygon == NULL || polygon->end_flag == SP_FALSE) return SP_FALSE;

    spDebug(50, "spMovePolygon", "dx = %d, dy = %d\n", dx, dy);
    
    if (spMovePolygonArch(polygon, dx, dy) == SP_TRUE) {
	polygon->move_dx += dx;
	polygon->move_dy += dy;

	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spFillPolygon(spComponent component, spGraphics graphics, spPolygon polygon)
{
    void *polygon_instance;
    
    if (polygon == NULL || polygon->end_flag == SP_FALSE) return SP_FALSE;
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->fill_polygon == NULL) return SP_FALSE;

	spDebug(50, "spFillPolygon", "call spGetPluginCanvasPolygonInstance\n");
	if ((polygon_instance = spGetPluginCanvasPolygonInstance(component, polygon)) == NULL) {
	    return SP_FALSE;
	}
    
	spDebug(50, "spFillPolygon", "call fill_polygon\n");
	return SpPluginCanvasDrawingRec(component)->fill_polygon(SpPluginCanvasInstance(component),
								 graphics, polygon_instance);
    } else {
	if (spIsDrawable(component) == SP_FALSE) return SP_FALSE;

	return spFillPolygonArch(component, graphics, polygon);
    }
}

spBool spDrawPolygon(spComponent component, spGraphics graphics, spPolygon polygon)
{
    void *polygon_instance;
    
    if (polygon == NULL || polygon->end_flag == SP_FALSE) return SP_FALSE;

    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->draw_polygon == NULL) return SP_FALSE;
    
	if ((polygon_instance = spGetPluginCanvasPolygonInstance(component, polygon)) == NULL) {
	    return SP_FALSE;
	}
    
	return SpPluginCanvasDrawingRec(component)->draw_polygon(SpPluginCanvasInstance(component),
								 graphics, polygon_instance);
    } else {
	if (spIsDrawable(component) == SP_FALSE) return SP_FALSE;

	return spDrawPolygonArch(component, graphics, polygon);
    }
}

spBool spOpenDrawingSetupDialogArg(spComponent component, spArg *args, int num_arg)
{
    spBool flag;
    
    if (spIsPluginCanvas(component) == SP_TRUE
        && (SpPluginCanvasDrawingRec(component)->open_setup_dialog != NULL
            || SpPluginCanvasDrawingRec(component)->set_page_info != NULL)) {
	spDebug(80, "spOpenDrawingSetupDialog", "component is plugin canvas\n");
	if (spIsPluginCanvasCreated(component) == SP_FALSE) return SP_FALSE;

	flag = SP_FALSE;
	
	if (SpPluginCanvasDrawingRec(component)->open_setup_dialog == NULL) {
            spPaperSize paper_size = SP_PAPER_SIZE_A4;
            spPaperOrientation paper_orientation = SP_PAPER_ORIENTATION_PORTRAIT;
            double left_margin_mm = SP_PAGE_SETUP_DEFAULT_HORIZONTAL_MARGIN_MM;
            double right_margin_mm = SP_PAGE_SETUP_DEFAULT_HORIZONTAL_MARGIN_MM;
            double top_margin_mm = SP_PAGE_SETUP_DEFAULT_VERTICAL_MARGIN_MM;
            double bottom_margin_mm = SP_PAGE_SETUP_DEFAULT_VERTICAL_MARGIN_MM;
            
            if (SpPluginCanvasDrawingRec(component)->get_page_info != NULL) {
                SpPluginCanvasDrawingRec(component)->get_page_info(SpPluginCanvasInstance(component), &paper_size, &paper_orientation,
                                                                   &left_margin_mm, &right_margin_mm, &top_margin_mm, &bottom_margin_mm);
            }
            
	    flag = spGetPageSetup(SpGetParent(component), &paper_size, &paper_orientation,
                                  &left_margin_mm, &right_margin_mm, &top_margin_mm, &bottom_margin_mm, NULL);
            if (flag) {
                SpPluginCanvasDrawingRec(component)->set_page_info(SpPluginCanvasInstance(component), paper_size, paper_orientation,
                                                                   left_margin_mm, right_margin_mm, top_margin_mm, bottom_margin_mm);
            }
	} else {
	    flag = SpPluginCanvasDrawingRec(component)->open_setup_dialog(SpPluginCanvasInstance(component),
									  args, num_arg);
	}

	if (flag == SP_TRUE) {
	    SpComponentPart(component).client_width = -1;
	    SpComponentPart(component).client_height = -1;
	    spDebug(80, "spOpenDrawingSetupDialog", "client size initialized\n");
	}
	return flag;
    } else {
	return SP_FALSE;
    }
}

spBool spOpenDrawingSetupDialog(spComponent component, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;

    va_start(argp, component);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    return spOpenDrawingSetupDialogArg(component, args, num_arg);
}

spBool spSetPageRange(spComponent component, int start_page, int end_page, int max_page)
{
    spBool flag = SP_FALSE;
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
        SpPluginCanvasPart(component).num_page = MAX(max_page, SpPluginCanvasPart(component).num_page);
        spDebug(80, "spSetPageRange", "num_page = %d\n", SpPluginCanvasPart(component).num_page);
        
	if (SpPluginCanvasDrawingRec(component)->set_page_range != NULL) {
            flag = SpPluginCanvasDrawingRec(component)->set_page_range(SpPluginCanvasInstance(component),
                                                                       start_page, end_page, max_page);
        }
        
        return flag;
    } else if (max_page == 1) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spGetPageRange(spComponent component, int *start_page, int *end_page)
{
    int sp, ep;

    sp = ep = 1;
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (SpPluginCanvasDrawingRec(component)->get_page_range != NULL) {
            return SpPluginCanvasDrawingRec(component)->get_page_range(SpPluginCanvasInstance(component), &sp, &ep);
        }
    }

    if (start_page != NULL) *start_page = sp;
    if (end_page != NULL) *end_page = ep;

    return SP_TRUE;
}

spBool spOpenDrawingTargetDialogArg(spComponent component, spArg *args, int num_arg)
{
    spBool flag;
    
    spDebug(80, "spOpenDrawingTargetDialogArg", "in\n");
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasCreated(component) == SP_FALSE) return SP_FALSE;
	spDebug(80, "spOpenDrawingTargetDialogArg", "num_arg = %d\n", num_arg);

	flag = SP_FALSE;
	
	    
	if (SpPluginCanvasDrawingRec(component)->start_drawing == NULL) {
	    SpPluginCanvasPart(component).drawing_started = SP_TRUE;
	}
	
	if (SpPluginCanvasDrawingRec(component)->set_page_range != NULL) {
            int num_page;
            num_page = MAX(SpPluginCanvasPart(component).num_page, 1);
            spDebug(80, "spOpenDrawingTargetDialogArg", "num_page = %d\n", num_page);
            SpPluginCanvasDrawingRec(component)->set_page_range(SpPluginCanvasInstance(component),
                                                                1, num_page, num_page);
        }
        
	if (SpPluginCanvasDrawingRec(component)->open_target_dialog == NULL) {
	    char *filename;
	    int file_filter_index;
	    char **file_desc_list = NULL;
	    char **file_filter_list = NULL;
	    
            spDebug(80, "spOpenDrawingTargetDialogArg", "case of SpPluginCanvasDrawingRec(component)->open_target_dialog == NULL\n");
            
	    if (SpPluginCanvasDrawingRec(component)->open_target_file == NULL) {
                spDebug(80, "spOpenDrawingTargetDialogArg", "error: SpPluginCanvasDrawingRec(component)->open_target_file == NULL\n");
  	        return SP_FALSE;
	    }

	    file_desc_list = SpPluginCanvasDrawingRec(component)->file_desc_list;
	    file_filter_list = SpPluginCanvasDrawingRec(component)->file_filter_list;
	    file_filter_index = 0;
	    
	    filename = xspGetSaveFileName(spGetParent(component), NULL,
					  SppOverwritePrompt, SP_TRUE,
					  SppFileTypes, file_desc_list,
					  SppFileFilters, file_filter_list,
					  SppFileFilterIndex, &file_filter_index,
					  NULL);
	    
	    if (SpPluginCanvasDrawingRec(component)->set_file_type != NULL) {
		SpPluginCanvasDrawingRec(component)->set_file_type(SpPluginCanvasInstance(component), file_filter_index);
	    }
	    
	    if (filename != NULL) {
		flag = SpPluginCanvasDrawingRec(component)->open_target_file(SpPluginCanvasInstance(component), filename);
		xfree(filename);
	    }

	} else {
	    flag = SpPluginCanvasDrawingRec(component)->open_target_dialog(SpPluginCanvasInstance(component),
									   args, num_arg);
	    spDebug(80, "spOpenDrawingTargetDialogArg", "SpPluginCanvasDrawingRec(component)->open_target_dialog result: flag = %d\n", flag);
	}

	if (flag == SP_TRUE) {
	    SpComponentPart(component).client_width = -1;
	    SpComponentPart(component).client_height = -1;
	    spDebug(80, "spOpenDrawingTargetDialogArg", "client size initialized\n");
	} else {
	    if (SpPluginCanvasDrawingRec(component)->start_drawing == NULL) {
		SpPluginCanvasPart(component).drawing_started = SP_FALSE;
	    }
	}
		    
	return flag;
    } else {
	return SP_FALSE;
    }
}

spBool spOpenDrawingTargetDialog(spComponent component, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;

    va_start(argp, component);
    spGetArgs(argp, args, num_arg);
    va_end(argp);
    
    return spOpenDrawingTargetDialogArg(component, args, num_arg);
}

spBool spCloseDrawingTarget(spComponent component)
{
    spBool flag;
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (SpPluginCanvasDrawingRec(component)->close_target == NULL) return SP_TRUE;
	
	flag = SpPluginCanvasDrawingRec(component)->close_target(SpPluginCanvasInstance(component));
        SpPluginCanvasPart(component).current_page = 0;

        return flag;
    } else {
	return SP_TRUE;
    }
}

unsigned long spGetDrawingCaps(spComponent component)
{
    unsigned long caps = 0L;

    if (component != NULL && spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawable(component) == SP_FALSE
            || SpPluginCanvasDrawingRec(component)->get_drawing_caps == NULL) return 0L;
	
	caps = SpPluginCanvasDrawingRec(component)->get_drawing_caps(SpPluginCanvasInstance(component));
    } else {
	caps = spGetDrawingCapsArch(component);
    }
    
    return caps;
}

unsigned long spGetGraphicsModeCaps(spComponent component)
{
    unsigned long caps = 0L;

    if (component != NULL && spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawable(component) == SP_FALSE
            || SpPluginCanvasDrawingRec(component)->get_graphics_mode_caps == NULL) return 0L;
	
	caps = SpPluginCanvasDrawingRec(component)->get_graphics_mode_caps(SpPluginCanvasInstance(component));
    } else {
	caps = spGetGraphicsModeCapsArch(component);
    }
    
    return caps;
}

unsigned long spGetSystemDrawingCaps(void)
{
    return spGetDrawingCapsArch(NULL);
}

unsigned long spGetSystemGraphicsModeCaps(void)
{
    return spGetGraphicsModeCapsArch(NULL);
}

spBool spGetPhysicalSize(spComponent component, double *width_mm, double *height_mm)
{
    double w, h;
    spBool flag;
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawable(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->get_physical_size == NULL) return SP_FALSE;

	w = h = 0.0;
	flag = SpPluginCanvasDrawingRec(component)->get_physical_size(SpPluginCanvasInstance(component),
								      &w, &h);
	if (flag == SP_TRUE) {
	    if (width_mm != NULL) *width_mm = w;
	    if (height_mm != NULL) *height_mm = h;
	}
	
	return flag;
    } else {
	return SP_FALSE;
    }
}

spBool spSetScalingFactor(spComponent component, double xfactor, double yfactor)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawable(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->set_scaling_factor == NULL) return SP_FALSE;

	if (SpPluginCanvasDrawingRec(component)->set_scaling_factor(SpPluginCanvasInstance(component),
								    xfactor, yfactor) == SP_TRUE) {
	    SpComponentPart(component).current_width = -1;
	    SpComponentPart(component).current_height = -1;
	    SpComponentPart(component).client_width = -1;
	    SpComponentPart(component).client_height = -1;
	    return SP_TRUE;
	}
    }
    
    return SP_FALSE;
}

spBool spSetNumPage(spComponent component, int num_page)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
        if (num_page >= 2 && !(SpPluginCanvasDrawingRec(component)->caps & SP_DRAWING_PLUGIN_CAPS_MULTIPLE_PAGES)) {
	    spDebug(80, "spSetNumPage", "multi-page not supported\n");
            return SP_FALSE;
        }

        SpPluginCanvasPart(component).num_page = MAX(num_page, 1);
        spDebug(80, "spSetNumPage", "num_page = %d\n", SpPluginCanvasPart(component).num_page);
        return SP_TRUE;
    } else if (num_page == 1) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

int spGetNumPage(spComponent component)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
        return SpPluginCanvasPart(component).num_page;
    } else if (spIsDrawable(component) == SP_TRUE) {
        return 1;
    } 
    
    return -1;
}

spBool spStartDrawing(spComponent component, const char *job_name)
{
    spBool flag;
    int page;
    spBool current_page_set = SP_FALSE;
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (SpPluginCanvasDrawingRec(component)->start_drawing == NULL) return SP_TRUE;
	
	if (spIsPluginCanvasDrawable(component) == SP_FALSE) return SP_FALSE;

	SpPluginCanvasPart(component).drawing_started = SP_TRUE;
	if (SpPluginCanvasPart(component).current_page >= 1
            && SpPluginCanvasPart(component).current_page <= SpPluginCanvasPart(component).num_page) {
            current_page_set = SP_TRUE;
        } else {
            SpPluginCanvasPart(component).current_page = 1;
        }
	
	flag = SpPluginCanvasDrawingRec(component)->start_drawing(SpPluginCanvasInstance(component),
								  job_name);
        spDebug(30, "spStartDrawing", "start_drawing result: %d\n", flag);
        
	if (flag == SP_FALSE) {
	    /* failed */
	    SpPluginCanvasPart(component).drawing_started = SP_FALSE;
	    SpPluginCanvasPart(component).current_page = 0;
	} else {
            if (!(SpPluginCanvasDrawingRec(component)->caps & SP_DRAWING_PLUGIN_CAPS_CALLBACK_DRAW)
                && SpPluginCanvasPart(component).draw_func != NULL) {
                int end_page;

                page = 1;
                end_page = MAX(SpPluginCanvasPart(component).num_page, 1);
                
                if (SpPluginCanvasDrawingRec(component)->get_page_range != NULL) {
                    SpPluginCanvasDrawingRec(component)->get_page_range(SpPluginCanvasInstance(component),
                                                                        &page, &end_page);
                    spDebug(30, "spStartDrawing", "get_page_range result: start = %d, end = %d\n",
                            page, end_page);
                    if (current_page_set == SP_TRUE && page == SP_PAGE_RANGE_CURRENT) {
                        page = end_page = SpPluginCanvasPart(component).current_page;
                    } else if (page <= 0) {
                        page = 1;
                        end_page = MAX(SpPluginCanvasPart(component).num_page, 1);
                    } else {
                        end_page = MIN(end_page, SpPluginCanvasPart(component).num_page);
                    }
                }

                if (end_page >= 2) {
                    if (!(spGetDrawingCaps(component) & SP_DRAWING_CAPS_MULTIPLE_PAGES)) {
                        page = 1;
                        end_page = 1;
                    }
                }
                
                spDebug(30, "spStartDrawing", "start = %d, end = %d, num_page = %d\n",
                        page, end_page, SpPluginCanvasPart(component).num_page);
                
                if (SpComponentPart(component).client_width == -1
                    && SpComponentPart(component).client_height == -1) {
                    SpComponentPart(component).call_reason = SP_CR_RESIZE;
                } else {
                    SpComponentPart(component).call_reason = SP_CR_EXPOSE;
                }
                
                for (; page <= end_page; page++) {
                    SpPluginCanvasPart(component).current_page = page;

                    spDebug(30, "spStartDrawing", "call draw_func of spPluginCanvas: page = %d\n", page);
                    SpPluginCanvasPart(component).draw_func(component, SpPluginCanvasPart(component).draw_data);

                    if (page == end_page) {
                        break;
                    } else if (SpPluginCanvasDrawingRec(component)->forward_page(SpPluginCanvasInstance(component)) == SP_FALSE) {
                        flag = SP_FALSE;
                        break;
                    }
                }
                
                SpComponentPart(component).call_reason = SP_CR_NONE;
                SpPluginCanvasPart(component).drawing_started = SP_FALSE;
            } else {
                spDebug(30, "spStartDrawing", "skip draw_func of spPluginCanvas\n");
            }
        }
	
	return flag;
    } else {
	return spRedrawCanvas(component);
    }
}

spBool spEndDrawing(spComponent component)
{
    spDebug(80, "spEndDrawing", "in\n");
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
        /*SpPluginCanvasPart(component).drawing_started = SP_FALSE;*/
        
	if (SpPluginCanvasDrawingRec(component)->end_drawing == NULL) return SP_TRUE;
	
	if (spIsPluginCanvasDrawable(component) == SP_FALSE) return SP_FALSE;
	
	return SpPluginCanvasDrawingRec(component)->end_drawing(SpPluginCanvasInstance(component));
    } else {
	return SP_TRUE;
    }
}

spBool spForwardPage(spComponent component)
{
    spBool flag;
    
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawingStarted(component) == SP_FALSE
	    || SpPluginCanvasDrawingRec(component)->forward_page == NULL) return SP_FALSE;
	
	flag = SpPluginCanvasDrawingRec(component)->forward_page(SpPluginCanvasInstance(component));
	
	if (flag == SP_TRUE) {
	    SpPluginCanvasPart(component).current_page++;
	}
	
	return flag;
    } else {
	return SP_FALSE;
    }
}

spBool spSetCurrentPage(spComponent component, int current_page)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
        SpPluginCanvasPart(component).current_page = current_page;
        return SP_TRUE;
    } else if (current_page <= 1) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

int spGetCurrentPage(spComponent component)
{
    if (spIsPluginCanvas(component) == SP_TRUE) {
	if (spIsPluginCanvasDrawable(component) == SP_FALSE) return -1;

	if (SpPluginCanvasPart(component).drawing_started == SP_TRUE) {
            return SpPluginCanvasPart(component).current_page;
        }
    } else if (spIsDrawable(component) == SP_TRUE) {
        return 1;
    }
    
    return -1;
}

spBool spIsPrintCanvas(spComponent component)
{
    if (spIsPluginCanvas(component) == SP_TRUE
	&& strcaseeq(SpGetName(component), SP_DRAWING_PLUGIN_PRINTER)) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spComponent spCreatePrintCanvas(spComponent parent, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;

    va_start(argp, parent);
    spGetArgs(argp, args, num_arg);
    va_end(argp);
    
    return spCreatePluginCanvasArg(parent, SP_DRAWING_PLUGIN_PRINTER, args, num_arg);
}
