/*
 *	spDraw_Mac.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spPrimitiveP.h>
#include <sp/spCanvasP.h>
#include <sp/spDrawP.h>
#include <sp/spGraphicsP.h>

extern spBool spPathToFSSpecMac(const char *filename, FSSpec *fsspec);

spBool spOpenPictFromFSSpecMac(FSSpec *pfsspec, GraphicsImportComponent *pgic, Rect *prect)
{
    Rect rect;
    spBool flag = SP_FALSE;
    
    if (GetGraphicsImporterForFile(pfsspec, pgic) == noErr) {
	if (prect != NULL) {
	    if (GraphicsImportGetNaturalBounds(*pgic, &rect) == noErr) {
		OffsetRect(&rect, -rect.left, -rect.top);
		*prect = rect;
		flag = SP_TRUE;
	    }
	} else {
	    flag = SP_TRUE;
	}

	if (flag == SP_TRUE) {
	    return SP_TRUE;
	}
	
	CloseComponent(*pgic);
    }

    return SP_FALSE;
}
    
spBool spDrawImportedPictToLockedGWorldMac(GWorldPtr gworld, Rect *prect, GraphicsImportComponent *pgic)
{
    spBool flag = SP_FALSE;

    if (GraphicsImportSetGWorld(*pgic, gworld, 0L) == noErr) {
	if (GraphicsImportSetBoundsRect(*pgic, prect) == noErr) {
	    if (GraphicsImportDraw(*pgic) == noErr) {
		flag = SP_TRUE;
	    }
	}
    }

    return flag;
}

spBool spGetFileTypeFromGraphicsImportComponentMac(GraphicsImportComponent *pgic, OSType *pFileType)
{
    OSType fileType = 0;
    QTAtomContainer container;
    QTAtom atom;
    long dataSize;
    char mimeStr[SP_MAX_LINE];
    Ptr atomData;
    
    if (GraphicsImportGetMIMETypeList(*pgic, &container) != noErr) {
	return SP_FALSE;
    }
    
    if ((atom = QTFindChildByID(container, 0, 'mime', 1, NULL)) != 0) {
	if (QTGetAtomDataPtr(container, atom, &dataSize, (Ptr *)&atomData) == noErr) {
	    BlockMove(atomData, mimeStr, dataSize);
	    mimeStr[dataSize] = NUL;
	    spDebug(10, "spGetFileTypeFromGraphicsImportComponentMac", "mimeStr = %s\n", mimeStr);
	    
	    if (streq(mimeStr, "image/x-photoshop")) { 
		fileType = '8BPS';
	    } else if (streq(mimeStr, "image/x-bmp")) {
		fileType = 'BMPf';
	    } else if (streq(mimeStr, "image/pict")) {
		fileType = 'PICT';
	    } else if (streq(mimeStr, "image/jpeg")) {
		fileType = 'JPEG';
	    } else if (streq(mimeStr, "image/tiff")) {
		fileType = 'TIFF';
	    } else if (streq(mimeStr, "image/gif")) {
		fileType = 'GIFf';
	    } else if (streq(mimeStr, "image/png")) {
		fileType = 'PNGf';
	    }
	    
	    if (fileType) {
		*pFileType = fileType;
		return SP_TRUE;
	    }
	}
    } 

    return SP_FALSE;
}

GWorldPtr spLoadPictToGWorldMac(const char *basename, Rect *prect)
{
    char filename[SP_MAX_PATHNAME];
    FSSpec fsspec;
    GWorldPtr gworld = nil;
    Rect rect;

    spStrCopy(filename, sizeof(filename), basename);
    
    if (spExists(filename) == SP_FALSE) {
	spStrCat(filename, sizeof(filename), ".pct");
	if (spExists(filename) == SP_FALSE) {
	    spReplaceSuffix(filename, ".pict");
	    if (spExists(filename) == SP_FALSE) {
		return nil;
	    }
	}
    }

    if (spPathToFSSpecMac(filename, &fsspec) == SP_TRUE) {
	GraphicsImportComponent gic;

	if (spOpenPictFromFSSpecMac(&fsspec, &gic, &rect) == SP_TRUE) {
	    if (NewGWorld(&gworld,
			  0, &rect, NULL, NULL, 0) == noErr) {
		PixMapHandle pixmap;
		    
#if TARGET_API_MAC_CARBON
		LockPortBits(gworld);
#endif
		if ((pixmap = GetGWorldPixMap(gworld)) != NULL) {
		    LockPixels(pixmap);
		    spDrawImportedPictToLockedGWorldMac(gworld, &rect, &gic);
		    UnlockPixels(pixmap);
		}
		    
#if TARGET_API_MAC_CARBON
		UnlockPortBits(gworld);
#endif
		    
		if (prect != NULL) *prect = rect;
	    }
	    
	    CloseComponent(gic);
	}
    }

    return gworld;
}

spBool spWritePictFromGWorldMac(GWorldPtr gworld, GraphicsExportComponent *pgec, FSSpec *pfsspec, OSType type)
{
    spBool flag = SP_FALSE;
    PixMapHandle pixmap;
	
    FSpDelete(pfsspec);
	
    if ((pixmap = GetGWorldPixMap(gworld)) != NULL) {
	LockPixels(pixmap);
	if (GraphicsExportSetInputGWorld(*pgec, gworld) == noErr) {
	    if (GraphicsExportSetOutputFile(*pgec, pfsspec) == noErr) {
		GraphicsExportSetCompressionQuality(*pgec, codecLosslessQuality);
		GraphicsExportSetOutputFileTypeAndCreator(*pgec, type, 0);
		if (GraphicsExportDoExport(*pgec, NULL) == noErr) {
		    flag = SP_TRUE;
		}
	    }
	}
	UnlockPixels(pixmap);
    }
	
    return flag;
}

PixMapHandle spGetLockedPixmapMac(spComponent component)
{
    PixMapHandle pixmap = nil;
    
#if TARGET_API_MAC_CARBON
#if !defined(SP_GWORLD_PIXMAP_ALWAYS_LOCK)
    LockPortBits(SpPrimitiveArch(component).gworld);
#endif
#endif
    
    if ((pixmap = GetGWorldPixMap(SpPrimitiveArch(component).gworld)) != nil) {
#if 1 || !defined(SP_GWORLD_PIXMAP_ALWAYS_LOCK)
	if (GetPixBaseAddr(pixmap) == NULL || LockPixels(pixmap) == false) {
	    Rect rect;
	
	    spDebug(10, "spGetLockedPixmapMac", "LockPixels error: client size = %d, %d\n",
		    SpComponentPart(component).client_width, SpComponentPart(component).client_height);
	
	    SetRect(&rect, 0, 0, SpComponentPart(component).client_width,
		    SpComponentPart(component).client_height);
	
	    if (UpdateGWorld(&SpPrimitiveArch(component).gworld,
			     0, &rect, NULL, NULL, clipPix) == gwFlagErr) {
		spDebug(10, "spGetLockedPixmapMac", "UpdateGWorld error\n");
		pixmap = nil;
	    } else {
		if (LockPixels(pixmap) == false) {
		    pixmap = nil;
		}
	    }
	}
#endif
    }

    if (pixmap == nil) {
#if TARGET_API_MAC_CARBON
#if !defined(SP_GWORLD_PIXMAP_ALWAYS_LOCK)
	UnlockPortBits(SpPrimitiveArch(component).gworld);
#endif
#endif
	
	return nil;
    } else {
	return pixmap;
    }
}

void spUnlockPixmapMac(spComponent component, PixMapHandle pixmap)
{
    if (pixmap != nil) {
#if !defined(SP_GWORLD_PIXMAP_ALWAYS_LOCK)
	UnlockPixels(pixmap);
#endif
    }
    
#if TARGET_API_MAC_CARBON
#if !defined(SP_GWORLD_PIXMAP_ALWAYS_LOCK)
    UnlockPortBits(SpPrimitiveArch(component).gworld);
#endif
#endif

    return;
}

static RGBColor sp_original_fg_rgb;
static RGBColor sp_original_bg_rgb;

void spGetOriginalRGBMac(void)
{
    GetForeColor(&sp_original_fg_rgb);
    GetBackColor(&sp_original_bg_rgb);
    return;
}

void spSetOriginalRGBMac(void)
{
    RGBForeColor(&sp_original_fg_rgb);
    RGBBackColor(&sp_original_bg_rgb);

    return;
}

void spSetNormalRGBMac(void)
{
    RGBColor black_rgb = {0, 0, 0};
    RGBColor white_rgb = {65535, 65535, 65535};
    
    RGBForeColor(&black_rgb);
    RGBBackColor(&white_rgb);

    return;
}

void spSetBackgroundRGBMac(void)
{
    RGBColor white_rgb = {65535, 65535, 65535};
    RGBColor gray_rgb = {56576, 56576, 56576};

    if (spGetAppearanceVersionMac() < 0x00000101 || spIsAquaMac() == SP_TRUE) {
	RGBBackColor(&white_rgb);
    } else {
	RGBBackColor(&gray_rgb);
    }

    return;
}

void spSetDeactivateRGBMac(void)
{
    RGBColor rgb = {32768, 32768, 32768};
    RGBColor white_rgb = {65535, 65535, 65535};
    
    RGBForeColor(&rgb);
    RGBBackColor(&white_rgb);

    return;
}

void spSetGraphicsMac(spGraphics graphics)
{
    if (graphics != NULL) {
	PenSize(SpGraphicsPart(graphics).line_width,
		SpGraphicsPart(graphics).line_width);
	PenMode(SpGraphicsArch(graphics).mode);
	
	RGBForeColor(&SpGraphicsArch(graphics).fg_rgb);
	RGBBackColor(&SpGraphicsArch(graphics).bg_rgb);
    }

    return;
}

void spUnsetGraphicsMac(spGraphics graphics)
{
    PenNormal();
    spSetOriginalRGBMac();
    
    return;
}

spBool spBeginDrawMac(spComponent component, spGraphics graphics,
		      PixMapHandle *current_pixmap, GWorldPtr *save_gwp, GDHandle *save_gdh)
{
    
    if (spIsDrawable(component) == SP_FALSE) return SP_FALSE;

    spLockControlMutexMac();
    spLockDrawMutexMac();
    spLockPortMutexMac();
    
    spGetOriginalGWorldMac(save_gwp, save_gdh);
    SetGWorld(SpPrimitiveArch(component).gworld, NULL);
    
    spGetOriginalRGBMac();

    *current_pixmap = spGetLockedPixmapMac(component);

    spSetGraphicsMac(graphics);
    
    return SP_TRUE;
}

spBool spEndDrawMac(spComponent component, spGraphics graphics,
		    PixMapHandle current_pixmap, GWorldPtr save_gwp, GDHandle save_gdh)
{
    spUnsetGraphicsMac(graphics);
    
    spUnlockPixmapMac(component, current_pixmap);
    
    SetGWorld(save_gwp, save_gdh);
    
    spUnlockPortMutexMac();
    spUnlockDrawMutexMac();
    spUnlockControlMutexMac();
    
    return SP_TRUE;
}

spBool spIsAlphaSupported(spComponent component)
{
    return SP_FALSE;
}

spBool spIsDrawable(spComponent component)
{
    if (spIsPrimitive(component) == SP_TRUE
	&& SpPrimitiveArch(component).gworld != NULL) {
	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

void *spGetImageHandleArch(spComponent component)
{
    return SpPrimitiveArch(component).gworld;
}

spBool spStretchImageArch(spComponent src, spComponent dest,
			  int src_x, int src_y, int src_width, int src_height,
			  int dest_x, int dest_y, int dest_width, int dest_height)
{
    Rect src_rect;
    Rect dest_rect;
    PixMapHandle src_pixmap;
    PixMapHandle dest_pixmap;

    spLockControlMutexMac();
    spLockDrawMutexMac();
    spLockPortMutexMac();

    SetRect(&src_rect, src_x, src_y, src_x + src_width, src_y + src_height);
    SetRect(&dest_rect, dest_x, dest_y, dest_x + dest_width, dest_y + dest_height);
    
    /* get pixmap */
    if ((src_pixmap = spGetLockedPixmapMac(src)) != nil) {
	if ((dest_pixmap = spGetLockedPixmapMac(dest)) != nil) {
#ifdef SP_COPYBITS_USE_SET_GWORLD
	    GWorldPtr save_gwp;
	    GDHandle save_gdh;
	    
	    spGetOriginalGWorldMac(&save_gwp, &save_gdh);
	    SetGWorld(SpPrimitiveArch(dest).gworld, NULL);
#endif
    
	    /* set normal RGB */
	    spGetOriginalRGBMac();
	    spSetNormalRGBMac();
    
	    /* copy pixmap */
	    CopyBits((BitMap *)*src_pixmap, (BitMap *)*dest_pixmap,
		     &src_rect, &dest_rect, srcCopy, NULL);

	    /* restore RGB */
	    spSetOriginalRGBMac();
#ifdef SP_COPYBITS_USE_SET_GWORLD
	    SetGWorld(save_gwp, save_gdh);
#endif
	    
	    spUnlockPixmapMac(dest, dest_pixmap);
	}
	spUnlockPixmapMac(src, src_pixmap);
    }
    
    spUnlockPortMutexMac();
    spUnlockDrawMutexMac();
    spUnlockControlMutexMac();
    
    return SP_TRUE;
}

spBool spCopyImageArch(spComponent src, spComponent dest,
		       int src_x, int src_y, int width, int height,
		       int dest_x, int dest_y)
{
    return spStretchImageArch(src, dest, src_x, src_y, width, height,
			      dest_x, dest_y, width, height);
}

static char *getDash(spLineType type, int *num_dashp)
{
    static char dash_line[] = {9, 4};
    static char dot_line[] = {2, 4};
    static char dash_dot_line[] = {9, 4, 2, 4};
    static char dash_dot_dot_line[] = {9, 4, 2, 4, 2, 4};
    char *dash;
    int num_dash;

    switch (type) {
      case SP_LINE_DASH:
	dash = dash_line;
	num_dash = spArraySize(dash_line);
	break;
      case SP_LINE_DOT:
	dash = dot_line;
	num_dash = spArraySize(dot_line);
	break;
      case SP_LINE_DASH_DOT:
	dash = dash_dot_line;
	num_dash = spArraySize(dash_dot_line);
	break;
      case SP_LINE_DASH_DOT_DOT:
	dash = dash_dot_dot_line;
	num_dash = spArraySize(dash_dot_dot_line);
	break;
      default:
	return NULL;
    }

    if (num_dashp != NULL) {
	*num_dashp = num_dash;
    }
    
    return dash;
}

static int drawDashLine(RGBColor rgb, spLineType type, int doffset,
			int x1, int y1, int x2, int y2)
{
    double theta;
    int num_dash;
    int px, py;
    int cx, cy;
    int distance;
    int sumdash;
    int prev_sumdash;
    int dindex;
    char *dash;

    if (x1 == x2 && y1 == y2) {
	SetCPixel(x1, y1, &rgb);
	return 0;
    }
    if ((dash = getDash(type, &num_dash)) == NULL) {
	return 0;
    }

    distance = sqrt(SQUARE(x2 - x1) + SQUARE(y2 - y1));
    theta = atan2((double)(y2 - y1), (double)(x2 - x1));

    px = x1; py = y1;

    dindex = doffset;
    sumdash = 0;
    prev_sumdash = 0;
    for (;;) {
	sumdash += dash[dindex];
	if (sumdash > distance) {
	    if (prev_sumdash < distance) {
		sumdash = distance;
	    } else {
		break;
	    }
	}
	
	cx = (int)spRound((double)sumdash * cos(theta)) + x1;
	cy = (int)spRound((double)sumdash * sin(theta)) + y1;

	if (dindex % 2 == 0) {
	    MoveTo(px, py);
	    LineTo(cx, cy);
	}

	px = cx; py = cy;
	prev_sumdash = sumdash;
	
	++dindex;
	if (dindex >= num_dash) dindex = 0;
    }

    return dindex;
}

static void drawDashRectangle(RGBColor rgb, spLineType type, 
			      int x, int y, int width, int height)
{
    int doffset;

    doffset = drawDashLine(rgb, type, 0, x, y, x + width, y);
    doffset = drawDashLine(rgb, type, doffset, x + width, y, x + width, y + height);
    doffset = drawDashLine(rgb, type, doffset, x + width, y + height, x, y + height);
    doffset = drawDashLine(rgb, type, doffset, x, y + height, x, y);
    
    return;
}

void spDrawRectangleMac(spGraphics graphics, int x, int y, int width, int height, spBool fill)
{
    Rect rect;
    
    if (fill == SP_TRUE) {
	SetRect(&rect, x, y, x + width, y + height);
	PaintRect(&rect);
    } else {
	if (graphics == NULL || SpGraphicsPart(graphics).line_type == SP_LINE_SOLID) {
	    SetRect(&rect, x, y, x + width, y + height);
	    FrameRect(&rect);
	} else {
	    drawDashRectangle(SpGraphicsArch(graphics).fg_rgb,
			      SpGraphicsPart(graphics).line_type, 
			      x, y, width, height);
	}
    }
    
    return;
}

void spFillRectangleArch(spComponent component, spGraphics graphics,
			 int x, int y, int width, int height)
{
    PixMapHandle current_pixmap;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    
    if (spBeginDrawMac(component, graphics, &current_pixmap, &save_gwp, &save_gdh) == SP_TRUE) {
	spDrawRectangleMac(graphics, x, y, width, height, SP_TRUE);
	
	spEndDrawMac(component, graphics, current_pixmap, save_gwp, save_gdh);
    }
    
    return;
}

void spDrawRectangleArch(spComponent component, spGraphics graphics,
			 int x, int y, int width, int height)
{
    PixMapHandle current_pixmap;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    
    if (spBeginDrawMac(component, graphics, &current_pixmap, &save_gwp, &save_gdh) == SP_TRUE) {
	spDrawRectangleMac(graphics, x, y, width, height, SP_FALSE);
	
	spEndDrawMac(component, graphics, current_pixmap, save_gwp, save_gdh);
    }
    
    return;
}

static void drawArc(int x, int y, int width, int height, double angle1, double angle2)
{
    Rect rect;
    double angle1c;
    
    SetRect(&rect, x, y, x + width, y + height);
    angle1c = -angle1 + 90.0;
    FrameArc(&rect, (short)spRound(angle1c), -(short)spRound(angle2));
    
    return;
}

static int drawDashArc(spLineType type, int doffset,
		       int x, int y, int width, int height, double angle1, double angle2)
{
    int num_dash;
    int dindex;
    double total_angle;
    double current_angle;
    double angle_size;
    char *dash;

    if ((dash = getDash(type, &num_dash)) == NULL) {
	return 0;
    }

    total_angle = angle1 + angle2;

    dindex = doffset;
    current_angle = angle1;
    for (;;) {
	if (current_angle >= total_angle) {
	    break;
	}
	
	angle_size = spRound(360.0 * (double)dash[dindex] / (double)width / PI);
	if (current_angle + angle_size > total_angle) {
	    angle_size = total_angle - current_angle;
	}

	if (dindex % 2 == 0) {
	    drawArc(x, y, width, height, current_angle, angle_size);
	}

	current_angle += angle_size;

	++dindex;
	if (dindex >= num_dash) dindex = 0;
    }

    return dindex;
}

void spDrawArcMac(spGraphics graphics,
		  int x, int y, int width, int height, double angle1, double angle2, spBool fill)
{
    Rect rect;
    
    if (fill == SP_TRUE) {
	SetRect(&rect, x, y, x + width, y + height);
	PaintArc(&rect, (short)spRound(-angle1 + 90), -(short)spRound(angle2));
    } else {
	if (graphics == NULL || SpGraphicsPart(graphics).line_type == SP_LINE_SOLID) {
	    drawArc(x, y, width, height, angle1, angle2);
	} else {
	    drawDashArc(SpGraphicsPart(graphics).line_type, 0,
			x, y, width, height, angle1, angle2);
	}
    }
    
    return;
}

void spFillArcArch(spComponent component, spGraphics graphics,
		   int x, int y, int width, int height, double angle1, double angle2)
{
    PixMapHandle current_pixmap;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    
    if (spBeginDrawMac(component, graphics, &current_pixmap, &save_gwp, &save_gdh) == SP_TRUE) {
	spDrawArcMac(graphics, x, y, width, height, angle1, angle2, SP_TRUE);

	spEndDrawMac(component, graphics, current_pixmap, save_gwp, save_gdh);
    }
    
    return;
}

void spDrawArcArch(spComponent component, spGraphics graphics,
		   int x, int y, int width, int height, double angle1, double angle2)
{
    PixMapHandle current_pixmap;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    
    if (spBeginDrawMac(component, graphics, &current_pixmap, &save_gwp, &save_gdh) == SP_TRUE) {
	spDrawArcMac(graphics, x, y, width, height, angle1, angle2, SP_FALSE);

	spEndDrawMac(component, graphics, current_pixmap, save_gwp, save_gdh);
    }
    
    return;
}

void spDrawLineMac(spGraphics graphics, int x1, int y1, int x2, int y2)
{
    if (graphics == NULL || SpGraphicsPart(graphics).line_type == SP_LINE_SOLID) {
	MoveTo(x1, y1);
	LineTo(x2, y2);
    } else {
	drawDashLine(SpGraphicsArch(graphics).fg_rgb,
		     SpGraphicsPart(graphics).line_type, 0,
		     x1, y1, x2, y2);
    }
    
    return;
}


void spDrawLineArch(spComponent component, spGraphics graphics,
		    int x1, int y1, int x2, int y2)
{
    PixMapHandle current_pixmap;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    
    if (spBeginDrawMac(component, graphics, &current_pixmap, &save_gwp, &save_gdh) == SP_TRUE) {
	spDrawLineMac(graphics, x1, y1, x2, y2);

	spEndDrawMac(component, graphics, current_pixmap, save_gwp, save_gdh);
    }
    
    return;
}

void spDrawPointMac(spGraphics graphics, int x, int y)
{
    RGBColor rgb;
    
    if (graphics != NULL) {
	rgb = SpGraphicsArch(graphics).fg_rgb;
    } else {
	rgb.red = 0; rgb.green = 0; rgb.blue = 0;
    }
    SetCPixel(x, y, &rgb);
    
    return;
}

void spDrawPointArch(spComponent component, spGraphics graphics, int x, int y)
{
    PixMapHandle current_pixmap;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    
    if (spBeginDrawMac(component, graphics, &current_pixmap, &save_gwp, &save_gdh) == SP_TRUE) {
	spDrawPointMac(graphics, x, y);
	
	spEndDrawMac(component, graphics, current_pixmap, save_gwp, save_gdh);
    }
    
    return;
}

void spSetFontIdMac(short font_id, int style, int size)
{
    TextFont(font_id);
    TextFace(style);
    TextSize(size);
    return;
}

void spSetFontMac(spGraphics graphics)
{
    if (graphics != NULL) {
	spDebug(50, "setFont", "font_id = %d, font_style = %d, font_size = %d\n",
		SpGraphicsArch(graphics).font_id,
		SpGraphicsArch(graphics).font_style,
		SpGraphicsArch(graphics).font_size);
	
	spSetFontIdMac(SpGraphicsArch(graphics).font_id,
		       SpGraphicsArch(graphics).font_style,
		       SpGraphicsArch(graphics).font_size);
    }
    return;
}

void spDrawStringMac(spGraphics graphics, int x, int y, const char *string)
{
    spAlignment alignment;
    int x_shift = 0;
    
    if (SpGraphicsPart(graphics).string_rotation != 0) {
	alignment = SpGraphicsPart(graphics).rotation_failed_alignment;
    } else {
	alignment = SpGraphicsPart(graphics).string_alignment;
    }
    if (alignment == SP_ALIGNMENT_CENTER || alignment == SP_ALIGNMENT_END) {
	int text_width;
	    
	text_width = TextWidth(string, 0, strlen(string));
	    
	if (alignment == SP_ALIGNMENT_CENTER) {
	    x_shift = text_width / 2;
	} else {
	    x_shift = text_width;
	}
    }
	
    spSetFontMac(graphics);
    MoveTo(x - x_shift, y);
    DrawText(string, 0, strlen(string));
	
    return;
}

void spDrawStringArch(spComponent component, spGraphics graphics,
		      int x, int y, const char *string)
{
    PixMapHandle current_pixmap;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    
    if (spBeginDrawMac(component, graphics, &current_pixmap, &save_gwp, &save_gdh) == SP_TRUE) {
	spDrawStringMac(graphics, x, y, string);
	
	spEndDrawMac(component, graphics, current_pixmap, save_gwp, save_gdh);
    }
    
    return;
}

void spGetStringExtentMac(spGraphics graphics, const char *string,
			  int *x, int *y, int *width, int *height, int *next_x)
{
    int text_width;
    FontInfo font_info;
    
    spSetFontMac(graphics);

    GetFontInfo(&font_info);
    text_width = TextWidth(string, 0, strlen(string));
    
    if (x != NULL) {
	*x = -1;
    }
    if (y != NULL) {
	*y = -font_info.ascent - 1;
    }
    if (width != NULL) {
	*width = text_width + 2;
    }
    if (height != NULL) {
	*height = font_info.ascent + font_info.descent + 2;
    }
    if (next_x != NULL) {
	*next_x = text_width;
    }
    
    return;
}

spBool spGetStringExtentArch(spComponent component, spGraphics graphics, const char *string,
			     int *x, int *y, int *width, int *height, int *next_x)
{
    GWorldPtr save_gwp;
    GDHandle save_gdh;

    spLockControlMutexMac();
    spLockDrawMutexMac();
    spLockPortMutexMac();
    
    spGetOriginalGWorldMac(&save_gwp, &save_gdh);
    spSetDummyGWorldMac();

    spGetStringExtentMac(graphics, string, x, y, width, height, next_x);
    
    SetGWorld(save_gwp, save_gdh);
    
    spUnlockPortMutexMac();
    spUnlockDrawMutexMac();
    spUnlockControlMutexMac();
    
    return SP_TRUE;
}

spBool spCreatePolygonArch(spPolygon polygon)
{
    SpPolygonArch(polygon).hpoly = OpenPoly();
    SpPolygonArch(polygon).start_x = 0;
    SpPolygonArch(polygon).start_y = 0;
    SpPolygonArch(polygon).last_x = 0;
    SpPolygonArch(polygon).last_y = 0;
    SpPolygonArch(polygon).points = xalloc(polygon->max_point + 1, Point);
    
    return SP_TRUE;
}

spBool spDestroyPolygonArch(spPolygon polygon)
{
    xfree(SpPolygonArch(polygon).points);
    KillPoly(SpPolygonArch(polygon).hpoly);
    return SP_TRUE;
}

spBool spStartPolygonPointArch(spPolygon polygon, int x, int y)
{
    MoveTo(x, y);
    SpPolygonArch(polygon).start_x = SpPolygonArch(polygon).last_x = x;
    SpPolygonArch(polygon).start_y = SpPolygonArch(polygon).last_y = y;
    SpPolygonArch(polygon).points[0].h = x;
    SpPolygonArch(polygon).points[0].v = y;

    return SP_TRUE;
}

spBool spAddPolygonPointArch(spPolygon polygon, int x, int y)
{
    LineTo(x, y);
    SpPolygonArch(polygon).last_x = x;
    SpPolygonArch(polygon).last_y = y;
    SpPolygonArch(polygon).points[polygon->num_point].h = x;
    SpPolygonArch(polygon).points[polygon->num_point].v = y;
    
    return SP_TRUE;
}

spBool spAddPolygonRelativePointArch(spPolygon polygon, int dx, int dy)
{
    Line(dx, dy);
    SpPolygonArch(polygon).last_x += dx;
    SpPolygonArch(polygon).last_y += dy;
    SpPolygonArch(polygon).points[polygon->num_point].h
	= polygon->px + dx;
    SpPolygonArch(polygon).points[polygon->num_point].v
	= polygon->py + dy;
    
    return SP_TRUE;
}

spBool spEndPolygonPointArch(spPolygon polygon)
{
    if (SpPolygonArch(polygon).last_x != SpPolygonArch(polygon).start_x
	|| SpPolygonArch(polygon).last_y != SpPolygonArch(polygon).start_y) {
	LineTo(SpPolygonArch(polygon).start_x, SpPolygonArch(polygon).start_y);
	SpPolygonArch(polygon).points[polygon->num_point].h = SpPolygonArch(polygon).points[0].h;
	SpPolygonArch(polygon).points[polygon->num_point].v = SpPolygonArch(polygon).points[0].v;
    }
    
    ClosePoly();
    return SP_TRUE;
}

spBool spGetPolygonPointArch(spPolygon polygon, int index, int *x, int *y)
{
    if (x != NULL) *x = SpPolygonArch(polygon).points[index].h;
    if (y != NULL) *y = SpPolygonArch(polygon).points[index].v;
    
    return SP_TRUE;
}

spBool spMovePolygonArch(spPolygon polygon, int dx, int dy)
{
    int i;
    
    OffsetPoly(SpPolygonArch(polygon).hpoly, dx, dy);
    
    for (i = 0; i < polygon->num_point; i++) {
	SpPolygonArch(polygon).points[i].h += dx;
	SpPolygonArch(polygon).points[i].v += dy;
    }
    
    return SP_TRUE;
}

spBool spFillPolygonArch(spComponent component, spGraphics graphics, spPolygon polygon)
{
    PixMapHandle current_pixmap;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    
    if (spBeginDrawMac(component, graphics, &current_pixmap, &save_gwp, &save_gdh) == SP_TRUE) {
	PaintPoly(SpPolygonArch(polygon).hpoly);
	spEndDrawMac(component, graphics, current_pixmap, save_gwp, save_gdh);

	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

spBool spDrawPolygonArch(spComponent component, spGraphics graphics, spPolygon polygon)
{
    PixMapHandle current_pixmap;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    
    if (spBeginDrawMac(component, graphics, &current_pixmap, &save_gwp, &save_gdh) == SP_TRUE) {
	FramePoly(SpPolygonArch(polygon).hpoly);
	spEndDrawMac(component, graphics, current_pixmap, save_gwp, save_gdh);

	return SP_TRUE;
    } else {
	return SP_FALSE;
    }
}

unsigned long spGetDrawingCapsArch(spComponent component)
{
    return SP_DRAWING_CAPS_COPY_IMAGE
	| SP_DRAWING_CAPS_FILL_RECTANGLE | SP_DRAWING_CAPS_DRAW_RECTANGLE
	| SP_DRAWING_CAPS_FILL_ARC | SP_DRAWING_CAPS_DRAW_ARC
	| SP_DRAWING_CAPS_FILL_POLYGON | SP_DRAWING_CAPS_DRAW_POLYGON
	| SP_DRAWING_CAPS_DRAW_LINE
	| SP_DRAWING_CAPS_STRING_EXTENT | SP_DRAWING_CAPS_STRING_EXTENT_ACCURATE
	| SP_DRAWING_CAPS_DRAW_WIDE_LINE | SP_DRAWING_CAPS_DRAW_STYLED_LINE | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_LINE
	| SP_DRAWING_CAPS_DRAW_WIDE_RECTANGLE | SP_DRAWING_CAPS_DRAW_STYLED_RECTANGLE | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_RECTANGLE
	| SP_DRAWING_CAPS_DRAW_WIDE_POLYGON 
	| SP_DRAWING_CAPS_DRAW_WIDE_ARC | SP_DRAWING_CAPS_DRAW_STYLED_ARC | SP_DRAWING_CAPS_DRAW_WIDE_STYLED_ARC
	;
}
