/*
 *	spFrame_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spDialog.h>
#include <sp/spContainer.h>

#include <sp/spFrameP.h>
#include <sp/spTopLevelP.h>
#include <sp/spStatusBarP.h>

#include <gdk/gdkkeysyms.h>

extern spTopLevel sp_toplevel;

spBool sp_grabbed_loop = SP_FALSE;
static GtkWidget *sp_grabbed_widget = NULL;

static void postMessage(spComponent component, char *call_name)
{
    long k;
    long component_id;
    long num_callback;
    
    if ((component_id = SpGetComponentId(component)) == spGetDestroyedComponentId()) {
	/* now destroying */
	component_id = 0;
    }
    num_callback = SpPrimitivePart(component).num_callback;
    
    for (k = 0; k < num_callback; k++) {
	if (streq(SpPrimitivePart(component).callbacks[k].call_name, call_name)
	    && SpPrimitivePart(component).callbacks[k].component != NULL) {
	    SpPrimitiveArch(SpPrimitivePart(component).callbacks[k].component).call_name = call_name;
	    spComponentCallbackFunc(&SpPrimitivePart(component).callbacks[k]);

	    if (component_id == spGetDestroyedComponentId()) {
		break;
	    }
	}
    }

    return;
}

void spFramePartInitArch(spComponent component)
{
    SpFrameArch(component).toplevel = NULL;
    SpFrameArch(component).accel_group = NULL;
#if !GTK_CHECK_VERSION(3,0,0)
    SpFrameArch(component).tooltips = NULL;
#endif
    SpFrameArch(component).prev_width = 0;
    SpFrameArch(component).prev_height = 0;

    SpFrameArch(component).drop_func_id = 0;
    SpFrameArch(component).drop_call_func = NULL;
    SpFrameArch(component).drop_call_data = NULL;
    
#if defined(HAVE_APP_INDICATOR)
    SpFrameArch(component).indicator = NULL;
#elif GTK_CHECK_VERSION(2,10,0)
    SpFrameArch(component).status_icon = NULL;
#endif
    
    return;
}

void spFramePartFreeArch(spComponent component)
{
    if (SpFrameArch(component).accel_group != NULL) {
#if GTK_CHECK_VERSION(1,2,0)
	if (SpFrameArch(component).toplevel != NULL) {
	    gtk_window_remove_accel_group(GTK_WINDOW(SpFrameArch(component).toplevel),
					  SpFrameArch(component).accel_group);
	}
#if !GTK_CHECK_VERSION(3,0,0)
	gtk_accel_group_unref(SpFrameArch(component).accel_group);
#endif
#else
	if (SpFrameArch(component).toplevel != NULL) {
	    gtk_window_remove_accelerator_table(GTK_WINDOW(SpFrameArch(component).toplevel),
						SpFrameArch(component).accel_group);
	}
	gtk_accelerator_table_unref(SpFrameArch(component).accel_group);
#endif
	SpFrameArch(component).accel_group = NULL;
    }
	
#if !GTK_CHECK_VERSION(3,0,0)
    if (SpFrameArch(component).tooltips != NULL) {
	gtk_object_destroy(GTK_OBJECT(SpFrameArch(component).tooltips));
	SpFrameArch(component).tooltips = NULL;
    }
#endif
    
    return;
}

static gboolean adjustComponentSizeCB(GtkWidget *widget, GdkEventConfigure *event,
				      gpointer data)
{
    spComponent component = (spComponent)data;
    
    spDebug(30, "adjustComponentSizeCB", "event: (%d, %d), window prev: (%d, %d), window current: (%d, %d)\n",
	    event->width, event->height,
	    SpFrameArch(component).prev_width, SpFrameArch(component).prev_height,
	    SpComponentPart(component).current_width, SpComponentPart(component).current_height);

#if GTK_CHECK_VERSION(3,0,0)
    if (SpComponentPart(component).current_width != event->width
        || SpComponentPart(component).current_height != event->height) {
        SpFrameArch(component).prev_width = SpComponentPart(component).current_width;
        SpFrameArch(component).prev_height = SpComponentPart(component).current_height;
        SpComponentPart(component).current_width = event->width;
        SpComponentPart(component).current_height = event->height;
        spDebug(30, "adjustComponentSizeCB", "call gtk_window_resize\n");
        gtk_window_resize(GTK_WINDOW(SpFrameArch(component).toplevel),
                          SpComponentPart(component).current_width, SpComponentPart(component).current_height);
    }
#else
#if !GTK_CHECK_VERSION(2,0,0)
    if (SpFramePart(component).window_type == SP_FULLSCREEN_WINDOW
	|| SpFramePart(component).window_type == SP_TRANSIENT_WINDOW) {
        spDebug(30, "adjustComponentSizeCB", "event->width = %d, event->height = %d\n",
	        event->width, event->height);
	gdk_window_set_decorations(spGetGdkWindowGtk(SpFrameArch(component).toplevel), 0);
    }
#endif
    
    if (event->width != SpFrameArch(component).prev_width
	|| event->height != SpFrameArch(component).prev_height) {
	spDebug(30, "adjustComponentSizeCB", "******** size updated ********\n");
	
	gtk_widget_set_usize(SpFrameArch(component).toplevel, event->width, event->height);
	SpFrameArch(component).toplevel->allocation.width = event->width;
	SpFrameArch(component).toplevel->allocation.height = event->height;

	if (SpFrameArch(component).prev_width > 0 && SpFrameArch(component).prev_height > 0) {
	    /* resize event */
	    SpFrameArch(component).prev_width = event->width;
	    SpFrameArch(component).prev_height = event->height;
	
	    spAdjustComponentSize(component);
	} else {
            gdk_window_move(spGetGdkWindowGtk(SpFrameArch(component).toplevel), SpComponentPart(component).x, SpComponentPart(component).y);
    
	    /* first size update */
	    SpFrameArch(component).prev_width = event->width;
	    SpFrameArch(component).prev_height = event->height;
	}
	
	SpPrimitiveArch(component).set_width = -1;
	SpPrimitiveArch(component).set_height = -1;
    }
#endif
    
    spDebug(30, "adjustComponentSizeCB", "done\n");
    
    return /*TRUE*/FALSE;
}

static gint deleteEventCB(GtkWidget *widget, GdkEvent *event,
			  gpointer data)
{
    gint flag = TRUE;
    spComponent component = (spComponent)data;

    if (sp_grabbed_widget != NULL && sp_grabbed_widget != widget) {
	/* do nothing, if a grabbed widget exists. */
	return TRUE;
    }

    if (component != NULL && SpFrameArch(component).toplevel != NULL) {
	if (SpFramePart(component).close_style == SP_DESTROY_CLOSE) {
	    spDebug(30, "deleteEventCB", "destroy window\n");
	    
	    if (spIsSubClass(component, SpMainFrame) == SP_TRUE) {
		spQuit(0);
		flag = TRUE;
	    } else {
		long component_id;
		
		component_id = SpGetComponentId(component);
		SpFrameArch(component).toplevel = NULL;
		SpPrimitiveArch(component).top_widget = NULL;
		SpPrimitiveArch(component).widget = NULL;
	    
		spDebug(30, "deleteEventCB", "call gtk_widget_destroy\n");
		gtk_widget_destroy(widget);

		if (component_id != spGetDestroyedComponentId()) {
		    spDestroyComponent(component);
		}
		
		flag = TRUE;
	    }
	} else if (SpFramePart(component).close_style == SP_CALLBACK_CLOSE) {
	    postMessage(component, SppCloseCallback);
	    flag = TRUE;
	} else if (SpFramePart(component).close_style == SP_UNMAP_CLOSE) {
	    spPopdownWindow(component);
	    postMessage(component, SppCloseCallback);
	    flag = TRUE;
	}
    }
    
    return flag;
}

static gint keyPressEventCB(GtkWidget *widget, GdkEvent *event,
			    gpointer data)
{
    gint flag = FALSE;
    GtkWidget *focus_widget;
    spComponent component = (spComponent)data;
    
    if (event->type == GDK_KEY_PRESS) {
	if (event->key.keyval ==
#if GTK_CHECK_VERSION(3,0,0)
	    GDK_KEY_Escape
#else
	    GDK_Escape
#endif
	    && SpFramePart(component).cancel_button != NULL) {
	    gtk_widget_activate(SpPrimitiveArch(SpFramePart(component).cancel_button).widget);  
	} else {
#if GTK_CHECK_VERSION(2,0,0)
	    if ((focus_widget = gtk_window_get_focus(GTK_WINDOW(widget))) != NULL) {
		spDebug(40, "keyPressEventCB", "focus widget found\n");
		if (GTK_IS_EDITABLE(focus_widget)) {
		    spDebug(40, "keyPressEventCB", "focus widget is editable, return TRUE\n");
		    gtk_window_propagate_key_event(GTK_WINDOW(widget), &event->key);
		    return TRUE;
		}
	    }
#endif
	}
    }

    return flag;
}

#if GTK_CHECK_VERSION(2,10,0)
spBool spIsTaskTrayIconCreatedGtk(spComponent component)
{
#if defined(HAVE_APP_INDICATOR)
    if (SpFrameArch(component).indicator != NULL) {
	return SP_TRUE;
    }
#endif
    if (SpFrameArch(component).status_icon != NULL) {
	return SP_TRUE;
    }

    return SP_FALSE;
}

spBool spCreateTaskTrayIconGtk(spComponent component)
{
    spBool flag = SP_TRUE;
    char *icon_name;

#if defined(HAVE_APP_INDICATOR)
    if (SpTopLevelArch(sp_toplevel).unity_used == SP_TRUE) {
	if (!strnone(SpFramePart(component).task_tray_icon)) {
	    icon_name = SpFramePart(component).task_tray_icon;
	} else if (!strnone(SpTopLevelPart(sp_toplevel).icon_name)) {
	    icon_name = SpTopLevelPart(sp_toplevel).icon_name;
	} else {
	    icon_name = "indicator-messages";
	}
	spDebug(30, "spCreateTaskTrayIconGtk", "indicator icon_name = %s\n", icon_name);

	SpFrameArch(component).indicator = app_indicator_new(spGetName(component),
							     icon_name,
							     APP_INDICATOR_CATEGORY_APPLICATION_STATUS);
	if (SpFrameArch(component).indicator == NULL) {
	    flag = SP_FALSE;
	}
    } else
#endif
    {
	if (!strnone(SpFramePart(component).task_tray_icon)) {
	    icon_name = SpFramePart(component).task_tray_icon;
	} else if (!strnone(SpTopLevelPart(sp_toplevel).icon_name)) {
	    icon_name = SpTopLevelPart(sp_toplevel).icon_name;
	} else {
	    icon_name = NULL;
	}
	    
	if (icon_name != NULL) {
	    spDebug(30, "spCreateTaskTrayIconGtk", "status icon icon_name = %s\n", icon_name);
	    SpFrameArch(component).status_icon = gtk_status_icon_new_from_file(icon_name);
	} else {
	    SpFrameArch(component).status_icon = gtk_status_icon_new_from_stock(GTK_STOCK_FILE);
	}

	if (SpFrameArch(component).status_icon == NULL) {
	    flag = SP_FALSE;
	} else {
	    if (!strnone(SpFramePart(component).task_tray_tip)) {
#if GTK_CHECK_VERSION(3,0,0)
		gtk_status_icon_set_tooltip_text(SpFrameArch(component).status_icon,
						 SpFramePart(component).task_tray_tip);
#else
		gtk_status_icon_set_tooltip(SpFrameArch(component).status_icon,
					    SpFramePart(component).task_tray_tip);
#endif
	    }
	}
    }

    return flag;
}
#endif

static spBool sp_main_frame_created = SP_FALSE;

void spFrameCreateArch(spComponent component)
{
    char *title = NULL, *icon_name = NULL;
    
    if (!strnone(SpFramePart(component).icon_name)) {
	icon_name = SpFramePart(component).icon_name;
    }
    if (strnone(icon_name)) {
	icon_name = (!strnone(SpGetName(component)) ? SpGetName(component) : "");
    }
    
    if (!strnone(SpComponentPart(component).title)) {
	title = SpComponentPart(component).title;
    }
    if (strnone(title)) {
	title = (!strnone(SpGetName(component)) ? SpGetName(component) : "");
    }

    SpFrameArch(component).toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);

#if GTK_CHECK_VERSION(2,0,0)
    if (SpFramePart(component).window_type == SP_FULLSCREEN_WINDOW) {
	gtk_window_set_decorated(GTK_WINDOW(SpFrameArch(component).toplevel), FALSE);
	gtk_window_fullscreen(GTK_WINDOW(SpFrameArch(component).toplevel));
    } else if (SpFramePart(component).window_type == SP_TRANSIENT_WINDOW) {
	gtk_window_set_decorated(GTK_WINDOW(SpFrameArch(component).toplevel), FALSE);
    }
#else
    if (SpFramePart(component).window_type == SP_FULLSCREEN_WINDOW) {
	gtk_widget_set_usize(SpFrameArch(component).toplevel, gdk_screen_width(), gdk_screen_height());
    }
#endif
    gtk_window_set_title(GTK_WINDOW(SpFrameArch(component).toplevel), title);
    
#if GTK_CHECK_VERSION(3,0,0)
    gtk_window_set_has_resize_grip(GTK_WINDOW(SpFrameArch(component).toplevel),
				   SpFramePart(component).resize_flag);
#else
    if (SpFramePart(component).resize_flag == SP_FALSE) {
	gtk_window_set_policy(GTK_WINDOW(SpFrameArch(component).toplevel),
			      FALSE, FALSE, FALSE);
    } else {
	gtk_window_set_policy(GTK_WINDOW(SpFrameArch(component).toplevel),
			      TRUE, TRUE, FALSE);
    }
#endif

#if GTK_CHECK_VERSION(3,0,0)
    g_signal_connect(SpFrameArch(component).toplevel,
		     "configure_event", G_CALLBACK(adjustComponentSizeCB),
		     component);
#else    
    gtk_signal_connect(GTK_OBJECT(SpFrameArch(component).toplevel),
		       "configure_event", GTK_SIGNAL_FUNC(adjustComponentSizeCB),
		       component);
#endif
    spSignalConnectEventGtk(SpFrameArch(component).toplevel, "delete_event",
			    deleteEventCB, (gpointer)component);
    spSignalConnectEventGtk(SpFrameArch(component).toplevel, "key_press_event",
			    keyPressEventCB, (gpointer)component);
    
    if (spIsSubClass(component, SpMainFrame) == SP_TRUE
	&& sp_main_frame_created == SP_FALSE) {
	sp_main_frame_created = SP_TRUE;
    }

#if GTK_CHECK_VERSION(3,0,0)
    SpPrimitiveArch(component).top_widget = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
#else
    SpPrimitiveArch(component).top_widget = gtk_vbox_new(FALSE, 0);
#endif
    gtk_container_add(GTK_CONTAINER(SpFrameArch(component).toplevel),
		      SpPrimitiveArch(component).top_widget);
    gtk_widget_show(SpPrimitiveArch(component).top_widget);

#if GTK_CHECK_VERSION(3,0,0) && defined(SP_USE_GTK_GRID)
    SpPrimitiveArch(component).widget =	gtk_grid_new();
    gtk_grid_set_row_spacing(GTK_GRID(SpPrimitiveArch(component).widget),
			     /*SpComponentPart(component).spacing*/0);
    gtk_grid_set_column_spacing(GTK_GRID(SpPrimitiveArch(component).widget),
				/*SpComponentPart(component).spacing*/0);
#else
    SpPrimitiveArch(component).widget = gtk_fixed_new();
#if GTK_CHECK_VERSION(3,0,0)
    gtk_widget_set_has_window(SpPrimitiveArch(component).widget, TRUE);
#elif GTK_CHECK_VERSION(2,0,0)
    gtk_fixed_set_has_window(GTK_FIXED(SpPrimitiveArch(component).widget), TRUE);
#endif
#endif
    
#if GTK_CHECK_VERSION(2,4,0)
    if (SpFramePart(component).topmost_flag == SP_TRUE
	|| SpFramePart(component).window_type == SP_FULLSCREEN_WINDOW) {
	gtk_window_set_keep_above(GTK_WINDOW(SpFrameArch(component).toplevel), TRUE);
    }
#endif
#if GTK_CHECK_VERSION(2,10,0)
    if (SpFramePart(component).close_style == SP_NO_CLOSE) {
	gtk_window_set_deletable(GTK_WINDOW(SpFrameArch(component).toplevel), FALSE);
    }
#endif
    
    spSetStyleGtk(component, SpPrimitiveArch(component).widget);
    gtk_widget_show(SpPrimitiveArch(component).widget);
	    
    return;
}

void spFrameDestroyArch(spComponent component)
{
    postMessage(component, SppDestroyCallback);
    return;
}
    
void spFrameSetParamsArch(spComponent component)
{
    if (SpFramePart(SpOldObject(component)).icon_name != SpFramePart(component).icon_name) {
	/* do nothing */
    }

    if (SpComponentPart(SpOldObject(component)).title != SpComponentPart(component).title) {
	gtk_window_set_title(GTK_WINDOW(SpFrameArch(component).toplevel),
			     SpComponentPart(component).title);
    }
	
#if GTK_CHECK_VERSION(2,4,0)
    if (SpFramePart(SpOldObject(component)).topmost_flag != SpFramePart(component).topmost_flag) {
	gtk_window_set_keep_above(GTK_WINDOW(SpFrameArch(component).toplevel),
				  SpFramePart(component).topmost_flag);
    }
#endif
    
#if GTK_CHECK_VERSION(2,10,0)
    if (SpFramePart(SpOldObject(component)).close_style != SpFramePart(component).close_style) {
	if (SpFramePart(component).close_style == SP_NO_CLOSE) {
	    gtk_window_set_deletable(GTK_WINDOW(SpFrameArch(component).toplevel), FALSE);
	} else {
	    gtk_window_set_deletable(GTK_WINDOW(SpFrameArch(component).toplevel), TRUE);
	}
    }
#endif
    
    return;
}

void *spGetArchFrame(spComponent component)
{
    return (void *)SpFrameArch(component).toplevel;
}

spBool spIsTaskTraySupported(void)
{
#if GTK_CHECK_VERSION(2,10,0)
    spDebug(100, "spIsTaskTraySupported", "supported\n");
    return SP_TRUE;
#else
    return SP_FALSE;
#endif
}

spBool spShowTaskTrayArch(spComponent window, spBool flag)
{
#if GTK_CHECK_VERSION(2,10,0)
    if (spIsTaskTrayIconCreatedGtk(window) == SP_FALSE) {
	spCreateTaskTrayIconGtk(window);
    }
    
#if defined(HAVE_APP_INDICATOR)
    if (SpFrameArch(window).indicator != NULL) {
	spDebug(80, "spShowTaskTrayArch", "for indicator\n");
	if (flag) {
	    app_indicator_set_status(SpFrameArch(window).indicator, APP_INDICATOR_STATUS_ACTIVE);
	} else {
	    app_indicator_set_status(SpFrameArch(window).indicator, APP_INDICATOR_STATUS_PASSIVE);
	}
	
	return SP_TRUE;
    }
#endif
    if (SpFrameArch(window).status_icon != NULL) {
	spDebug(80, "spShowTaskTrayArch", "for status icon\n");
	gtk_status_icon_set_visible(SpFrameArch(window).status_icon, flag);	
	return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}
    
GtkWidget *spAddGrabGtk(GtkWidget *widget)
{
    GtkWidget *prev = NULL;

    if (sp_grabbed_widget != NULL) {
	if (spWidgetHasGrabGtk(sp_grabbed_widget)) {
	    gtk_grab_remove(sp_grabbed_widget);
	}
	prev = sp_grabbed_widget;
    }

    if (widget != NULL) {
	sp_grabbed_widget = widget;
	if (!spWidgetHasGrabGtk(sp_grabbed_widget)) {
	    gtk_grab_add(sp_grabbed_widget);
	}
    }
    
    spDebug(20, "spAddGrabGtk", "done\n");
    
    return prev;
}

void spRemoveGrabGtk(GtkWidget *widget)
{
    if (sp_grabbed_widget != NULL
	&& (widget == NULL || sp_grabbed_widget == widget)) {
	if (spWidgetHasGrabGtk(sp_grabbed_widget)) {
	    gtk_grab_remove(sp_grabbed_widget);
	}
	sp_grabbed_widget = NULL;
    }
    
    spDebug(20, "spRemoveGrabGtk", "done\n");
    
    return;
}

void spPopupFrameArch(spComponent window, spBool move_cursor)
{
    GtkWidget *widget;

    spDebug(30, "spPopupFrameArch", "in\n");
    
    widget = SpFrameArch(window).toplevel;
    
    if (!spIsWidgetMappedGtk(widget)) {
	if (!spIsWidgetRealizedGtk(widget)) {
	    spDebug(30, "spPopupFrameArch", "1st map\n");

	    gtk_box_pack_start(GTK_BOX(SpPrimitiveArch(window).top_widget),
			       SpPrimitiveArch(window).widget, TRUE, TRUE, 0);

	    spAdjustComponentSize(window);
	    spDebug(30, "spPopupFrameArch", "spAdjustComponentSize done\n");
	}
    
#if GTK_CHECK_VERSION(2,0,0)
	gtk_window_move(GTK_WINDOW(widget), SpComponentPart(window).x, SpComponentPart(window).y);
#endif
    
	gtk_widget_show(widget);
    } else {
#if !GTK_CHECK_VERSION(2,0,0)
	gdk_window_raise(spGetGdkWindowGtk(widget));
#if 0	
#include <gdk/gdkprivate.h>

	if (move_cursor == SP_TRUE) {
	    GdkWindowPrivate *private = (GdkWindowPrivate *)widget->window;
	    if (!private->destroyed) {
		/* move pointer */
		XWarpPointer(private->xdisplay, None, private->xwindow,
			     0, 0, 0, 0, 0, 0);
	    }
	}
#endif
#else
	gtk_window_present(GTK_WINDOW(widget));
#endif
    }

#if GTK_CHECK_VERSION(2,10,0)
    if (SpFramePart(window).task_tray_visible == SP_TRUE
	&& spIsTaskTrayIconCreatedGtk(window) == SP_FALSE) {
	spShowTaskTrayArch(window, SpFramePart(window).task_tray_visible);
    }
#endif
    
    if (SpFramePart(window).popup_style == SP_MODAL_POPUP) {
	spBool prev_loop;
	GtkWidget *prev_grabbed;
	
	prev_grabbed = spAddGrabGtk(widget);
	prev_loop = sp_grabbed_loop;
	sp_grabbed_loop = SP_TRUE;
	
	spMainLoopGtk(SpPrimitiveArch(window).top_widget, SP_FALSE);

	spAddGrabGtk(prev_grabbed);
	sp_grabbed_loop = prev_loop;
    }
    
    spDebug(30, "spPopupFrameArch", "done\n");
    
    return;
}

void spPopdownFrameArch(spComponent window)
{
    GtkWidget *widget;

    spDebug(30, "spPopdownFrameArch", "in\n");
    
    widget = SpFrameArch(window).toplevel;

    spGetFramePositionArch(window, &SpComponentPart(window).x, &SpComponentPart(window).y);
    
    if (sp_grabbed_widget != NULL) {
	spDebug(30, "spPopdownFrameArch", "call spRemoveGrabGtk\n");
	spRemoveGrabGtk(sp_grabbed_widget);
	if (sp_grabbed_loop == SP_TRUE) {
	    spMainQuitGtk();
	}
    }

    if (widget != NULL) {
	gtk_widget_hide(widget);
    }
    
    spDebug(30, "spPopdownFrameArch", "done\n");
    
    return;
}

void spRaiseFrameArch(spComponent window)
{
#if !GTK_CHECK_VERSION(2,0,0)
    if (spIsWidgetMappedGtk(SpFrameArch(window).toplevel)) {
	gdk_window_raise(spGetGdkWindowGtk(SpFrameArch(window).toplevel));
    }
#else
    gtk_window_present(GTK_WINDOW(SpFrameArch(window).toplevel));
#endif
    return;
}

void spLowerFrameArch(spComponent window)
{
    if (spIsWidgetMappedGtk(SpFrameArch(window).toplevel)) {
	gdk_window_lower(spGetGdkWindowGtk(SpFrameArch(window).toplevel));
    }
    
    return;
}

spBool spSetFramePositionArch(spComponent window, int x, int y)
{
    spDebug(30, "spSetFramePositionArch", "x = %d, y = %d\n", x, y);
#if GTK_CHECK_VERSION(2,0,0)
    gtk_window_move(GTK_WINDOW(SpFrameArch(window).toplevel), x, y);
#else
    if (spIsWidgetMappedGtk(SpFrameArch(window).toplevel)) {
	gdk_window_move(spGetGdkWindowGtk(SpFrameArch(window).toplevel), x, y);
    }
#endif
    
    return SP_TRUE;
}

spBool spGetFramePositionArch(spComponent window, int *x, int *y)
{
    gint lx, ly;
    
    if (spIsWidgetMappedGtk(SpFrameArch(window).toplevel)) {
#if GTK_CHECK_VERSION(2,0,0)
	gtk_window_get_position(GTK_WINDOW(SpFrameArch(window).toplevel), &lx, &ly);
#else
	gdk_window_get_position(spGetGdkWindowGtk(SpFrameArch(window).toplevel), &lx, &ly);
#endif
    } else {
	lx = SpComponentPart(window).x;
	ly = SpComponentPart(window).y;
    }
    spDebug(30, "spGetFramePositionArch", "x = %d, y = %d\n",lx, ly);
    
    *x = lx;
    *y = ly;
    
    return SP_TRUE;
}

spBool spGetFrameModifierKeyMaskArch(spComponent window, spModifierMask *mask)
{
    int x, y;
    GdkModifierType gdkmask;
    
    if (!spIsWidgetMappedGtk(SpFrameArch(window).toplevel)) {
	return SP_FALSE;
    }
    
#if GTK_CHECK_VERSION(3,0,0)
    {
	GdkDeviceManager *manager;
	GdkDevice *pointer;

	manager = gdk_display_get_device_manager(gdk_display_get_default());
	pointer = gdk_device_manager_get_client_pointer(manager);
    
	if (gdk_window_get_device_position(spGetGdkWindowGtk(SpFrameArch(window).toplevel), pointer, &x, &y, &gdkmask) == NULL) {
	    return SP_FALSE;
	}
    }
#else
    if (gdk_window_get_pointer(spGetGdkWindowGtk(SpFrameArch(window).toplevel), &x, &y, &gdkmask) == NULL) {
	return SP_FALSE;
    }
#endif

    *mask = gdkmask;
    
    return SP_TRUE;
}

static gboolean timeoutCB(gpointer data)
{
    gboolean flag;
    spTimerList timer = (spTimerList)data;

    /* required even in 3.0.0+ */
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) gdk_threads_enter();
    
    if (timer->call_func(timer->frame, timer->call_data) == SP_FALSE) {
	timer->kill_flag = SP_TRUE;
	
	/* 2009/10/1 added */
	spRemoveTimerList(timer->frame, timer);
	flag = FALSE;
    } else {
	flag = TRUE;
    }
    
    /* required even in 3.0.0+ */
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) gdk_threads_leave();
    
    return flag;
}

spTimerId spSetTimerArch(spTimerList timer)
{
    guint timer_id;
    
#if GTK_CHECK_VERSION(2,0,0)
    timer_id = g_timeout_add(timer->interval_ms, (GSourceFunc)timeoutCB, timer);
#else
    timer_id = gtk_timeout_add(timer->interval_ms, (GtkFunction)timeoutCB, timer);
#endif
    
    return (spTimerId)timer_id;
}

spBool spKillTimerArch(spTimerList timer)
{
#if GTK_CHECK_VERSION(2,0,0)
    if (!g_source_remove((guint)timer->timer_id)) {
	return SP_FALSE;
    }
#else
    gtk_timeout_remove((guint)timer->timer_id);
#endif
    timer->kill_flag = SP_TRUE;
    
    /* 2009/10/1 added */
    spRemoveTimerList(timer->frame, timer);
    
    return SP_TRUE;
}

#if GTK_CHECK_VERSION(1,2,0)
void spWindowDragDataReceivedGtk(GtkWidget *widget, GdkDragContext *context,
				 gint x, gint y, GtkSelectionData *selection_data,
				 guint info, guint time, gpointer data)
{
    int i;
    int num_line;
    int argc;
    char **argv;
    char *string;
    spComponent component = (spComponent)data;
    
    spDebug(10, "spWindowDragDataReceivedGtk", "in\n");
	    
    if (
#if GTK_CHECK_VERSION(2,14,0)
	gtk_selection_data_get_length(selection_data) > 0 && gtk_selection_data_get_format(selection_data) == 8
#else
	selection_data->length > 0 && selection_data->format == 8
#endif
	) {
	spDebug(10, "spWindowDragDataReceivedGtk", "received \"%s\"\n",
#if GTK_CHECK_VERSION(2,14,0)
		gtk_selection_data_get_data(selection_data)
#else
		(char *)selection_data->data
#endif
	    );
	
	if (spIsFrame(component) == SP_TRUE && SpFrameArch(component).drop_call_func != NULL) {
#if GTK_CHECK_VERSION(2,14,0)
            gchar **uri_list;
            
            argc = 0;
            argv = NULL;
            
	    string = (char *)gtk_selection_data_get_data(selection_data);

            if ((uri_list = g_uri_list_extract_uris(string)) != NULL) {
                for (num_line = 0; uri_list[num_line] != NULL; num_line++);
                spDebug(10, "spWindowDragDataReceivedGtk", "num_line = %d\n", num_line);
                
                argv = xalloc(num_line, char *);
                for (i = 0; uri_list[i] != NULL; i++) {
                    spDebug(10, "spWindowDragDataReceivedGtk", "uri_list[%d] = %s\n", i, uri_list[i]);
                    argv[i] = g_filename_from_uri(uri_list[i], NULL, NULL);
#if 0
                    if (argv[i] == NULL) {
                        argv[i] = g_strdup(uri_list[i]);
                    }
#endif
                    
                    if (argv[i] != NULL) {
                        spDebug(10, "spWindowDragDataReceivedGtk", "argv[%d] = %s\n", i, argv[i]);
                        ++argc;
                    }
                }

                if (argc <= 0) {
                    spDebug(10, "spWindowDragDataReceivedGtk", "error: argc = %d\n", argc);
                    xfree(argv);
                    argv = NULL;
                }
                
                g_strfreev(uri_list);
            }
#else
            int len;
            char *p;
            char *path;
            
	    /* count number of lines */
	    string = (char *)selection_data->data;
	    num_line = 0;
	    while (1) {
		if ((string = strchr(string, '\n')) == NULL)
		    break;

		string++;
		num_line++;
	    }
	    spDebug(10, "spWindowDragDataReceivedGtk", "num_line = %d\n", num_line);

            if (num_line <= 0) {
                argc = 0;
                argv = NULL;
            } else {
                argv = xalloc(num_line, char *);

                /* get file paths */
                string = (char *)selection_data->data;
                argc = 0;
                while (1) {
                    if ((p = strchr(string, '\n')) == NULL)
                        break;

                    if ((path = strchr(string, ':')) != NULL) {
                        path++;
                        len = p - path - 1;
                        if (len >= 3
                            && path[0] == '/' && path[1] == '/' && path[2] == '/') {
                            path += 2;
                            len -= 2;
                        }

                        if (len > 0) {
                            argv[argc] = xalloc(len + 1, char);
                            strncpy(argv[argc], path, len);
                            argv[argc][len] = NUL;
			
                            spDebug(10, "spWindowDragDataReceivedGtk",
                                    "len = %d, file = %s\n", len, argv[argc]);
			
                            argc++;
                        }
                    }

                    string = p + 1;
                }
            }
#endif

	    spDebug(10, "spWindowDragDataReceivedGtk", "argc = %d\n", argc);
	    
            if (argv != NULL) {
                SpFrameArch(component).drop_call_func(component, SpFrameArch(component).drop_call_data,
                                                      argc, argv);

                for (i = 0; i < argc; i++) {
#if GTK_CHECK_VERSION(2,14,0)
                    g_free(argv[i]);
#else
                    xfree(argv[i]);
#endif
                }
                xfree(argv);
            }
	}
	
	if (context != NULL) {
	    gtk_drag_finish(context, TRUE, FALSE, time);
	}
    } else {
	if (context != NULL) {
	    gtk_drag_finish(context, FALSE, FALSE, time);
	}
    }

    spDebug(10, "spWindowDragDataReceivedGtk", "done\n");
    
    return;
}
#endif

spBool spAddDropCallbackArch(spComponent component, spDropCallbackFunc call_func, void *call_data)
{
#if GTK_CHECK_VERSION(1,2,0)
    static GtkTargetEntry target_table[] = {
	{"text/uri-list", 0, 0},
    };

    spDebug(10, "spAddDropCallbackArch", "in\n");
    
    if (call_func == NULL) {
	if (SpFrameArch(component).drop_func_id > 0) {
	    gtk_drag_dest_unset(SpFrameArch(component).toplevel);
#if GTK_CHECK_VERSION(3,0,0)
	    g_signal_handler_disconnect(SpFrameArch(component).toplevel,
					SpFrameArch(component).drop_func_id);
#else
	    gtk_signal_disconnect(GTK_OBJECT(SpFrameArch(component).toplevel),
				  SpFrameArch(component).drop_func_id);
#endif
	    return SP_TRUE;
	}
    } else {
	SpFrameArch(component).drop_call_func = call_func;
	SpFrameArch(component).drop_call_data = call_data;

	gtk_drag_dest_set(SpFrameArch(component).toplevel, GTK_DEST_DEFAULT_ALL,
			  target_table, 1, GDK_ACTION_COPY | GDK_ACTION_MOVE);
	
	SpFrameArch(component).drop_func_id = 
#if GTK_CHECK_VERSION(3,0,0)
	    g_signal_connect(SpFrameArch(component).toplevel, "drag_data_received",
			     G_CALLBACK(spWindowDragDataReceivedGtk), (void *)component);
#else
	    gtk_signal_connect(GTK_OBJECT(SpFrameArch(component).toplevel), "drag_data_received",
			       GTK_SIGNAL_FUNC(spWindowDragDataReceivedGtk), (void *)component);
#endif
	return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}

#if GTK_CHECK_VERSION(1,2,0)
static spBool sp_argv_send_done = SP_FALSE;
static GdkAtom sp_atom_app_exec = GDK_NONE;
static GdkAtom sp_atom_app_argv = GDK_NONE;
static gint sp_selection_data_length = -1;
static char sp_exec_atom_name[SP_MAX_LINE] = "";
static char sp_argv_atom_name[SP_MAX_LINE] = "";

static void spSignalSelectionGetGtk(GtkWidget *widget, GtkSelectionData *data,
				    guint info, guint time, gpointer user_data)
{
    GdkAtom atom;

#if GTK_CHECK_VERSION(2,16,0)
    atom = gtk_selection_data_get_selection(data);
#else
    atom = data->selection;
#endif
    
    spDebug(30, "spSignalSelectionGetGtk", "info = %d, atom = %ld\n", info, (long)atom);
    
    if (atom == sp_atom_app_exec && info == 0) {
	gtk_selection_data_set(data, GDK_SELECTION_TYPE_STRING,
			       8 * sizeof (gchar), (guchar *)SpTopLevelPart(sp_toplevel).unique_id,
			       strlen(SpTopLevelPart(sp_toplevel).unique_id));
	gtk_selection_convert(widget, sp_atom_app_argv,
			      GDK_SELECTION_TYPE_STRING, GDK_CURRENT_TIME);
    } else if (atom == sp_atom_app_argv && info == 0) {
	spDebug(30, "spSignalSelectionGetGtk", "send argv data\n");
	
	if (SpTopLevelPart(sp_toplevel).argcp != NULL &&
	    *SpTopLevelPart(sp_toplevel).argcp >= 2) {
	    int i;
	    int argc;
	    char **argv;
	    int num_file;
	    int list_len;
	    char *list;
	    char *exactname;

	    argc = *SpTopLevelPart(sp_toplevel).argcp;
	    argv = *SpTopLevelPart(sp_toplevel).argvp;
	    spDebug(30, "spSignalSelectionGetGtk", "argc = %d\n", argc);
	    
	    num_file = 0;
	    list_len = 0;
	    list = NULL;
	    
	    for (i = 1; i < argc; i++) {
		spDebug(80, "spSignalSelectionGetGtk", "argv[%d] = %s\n", i, argv[i]);
		
		if (spIsFile(argv[i]) == SP_TRUE
		    && (exactname = xspGetExactName(argv[i])) != NULL) {
		    spDebug(50, "spSignalSelectionGetGtk", "i = %d, exactname = %s\n", i, exactname);
		    
		    list_len += strlen(exactname) + 7;
		    if (i == 1) {
			list = xalloc(list_len + 3, char);
			spStrCopy(list, list_len + 3, "file://");
		    } else {
			list_len += 2;
			list = xrealloc(list, list_len + 3, char);
			spStrCat(list, list_len + 3, "\r\nfile://");
		    }
		    spStrCat(list, list_len + 3, exactname);

		    xfree(exactname);
		    
		    ++num_file;
		}
	    }

	    spDebug(30, "spSignalSelectionGetGtk", "num_file = %d\n", num_file);
	    
	    if (list != NULL) {
		list_len += 2;
		spStrCat(list, list_len + 1, "\r\n");
		gtk_selection_data_set(data, GDK_SELECTION_TYPE_STRING,
				       8 * sizeof(char), (guchar *)list, list_len);
		xfree(list);
	    }
	}
	sp_argv_send_done = SP_TRUE;
    }

    spDebug(30, "spSignalSelectionGetGtk", "done\n");
    
    return;
}

static void spSignalSelectionReceivedGtk(GtkWidget *widget, GtkSelectionData *selection_data, guint time, gpointer user_data)
{
    GdkAtom atom;

    spDebug(30, "spSignalSelectionReceivedGtk", "in\n");
    
#if GTK_CHECK_VERSION(2,16,0)
    atom = gtk_selection_data_get_selection(selection_data);
#else
    atom = selection_data->selection;
#endif
    spDebug(30, "spSignalSelectionReceivedGtk", "atom = %ld\n", (long)atom);
    
    if (atom == sp_atom_app_exec) {
	sp_selection_data_length = MAX(
#if GTK_CHECK_VERSION(2,14,0)
	    gtk_selection_data_get_length(selection_data)
#else
	    selection_data->length
#endif
	    , 0);
	spDebug(30, "spSignalSelectionReceivedGtk", "sp_atom_app_exec: sp_selection_data_length = %d\n", sp_selection_data_length);
    } else if (atom == sp_atom_app_argv) {
	gboolean flag;
	spComponent window, next_window;
	spComponent last_visible_window;
	spComponent component = (spComponent)user_data;

	if (component == NULL) {
	    component = spGetWindow(spGetChild(NULL));
	    window = component;
	    last_visible_window = NULL;
	    
	    while (1) {
		if (spIsFrame(window) == SP_TRUE) {
		    spDebug(30, "spSignalSelectionReceivedGtk", "window = %s\n", spGetName(window));
#if GTK_CHECK_VERSION(2,4,0)
		    flag = gtk_window_has_toplevel_focus(GTK_WINDOW(SpFrameArch(window).toplevel));
#else
		    flag = FALSE;
#endif
		    if (flag) {
			spDebug(30, "spSignalSelectionReceivedGtk", "found focus window\n");
			component = window;
			break;
		    }
		    
		    if (spIsWidgetMappedGtk(SpFrameArch(window).toplevel)) {
			last_visible_window = window;
		    }
		}

		next_window = spGetNextWindow(window, SP_TRUE);
		spDebug(100, "spSignalSelectionReceivedGtk", "window = %ld, next_window = %ld\n",
			(long)window, (long)next_window);
		
		if (next_window == NULL || next_window == component || next_window == window) {
		    if (last_visible_window != NULL) {
			spDebug(30, "spSignalSelectionReceivedGtk", "use last visible window\n");
			component = last_visible_window;
		    } else {
			component = NULL;
		    }
		    break;
		}

		window = next_window;
	    }
	}

	if (component != NULL) {
	    spWindowDragDataReceivedGtk(widget, NULL, 0, 0, selection_data, 0, time, component);
	} else {
	    spDebug(30, "spSignalSelectionReceivedGtk", "target window is not found\n");
	}
    }

    spDebug(30, "spSignalSelectionReceivedGtk", "done\n");
    
    return;
}

void spAddAtomSelectionCallbackGtk(GtkWidget *widget, void *user_data)
{
#if GTK_CHECK_VERSION(3,0,0)
    g_signal_connect(widget, "selection-get",
		     G_CALLBACK(spSignalSelectionGetGtk), user_data);
    g_signal_connect(widget, "selection-received",
		     G_CALLBACK(spSignalSelectionReceivedGtk), user_data);
#else
    gtk_signal_connect(GTK_OBJECT(widget), "selection_get",
		       GTK_SIGNAL_FUNC(spSignalSelectionGetGtk), user_data);
    gtk_signal_connect(GTK_OBJECT(widget), "selection_received",
		       GTK_SIGNAL_FUNC(spSignalSelectionReceivedGtk), user_data);
#endif
    return;
}

void spAddAtomSelectionTargetGtk(GtkWidget *widget, char *unique_id)
{
    static GtkTargetEntry target_table[] = {{"STRING", 0, 0}};
  
    spDebug(30, "spAddAtomSelectionTargetGtk", "in: unique_id = %s\n", unique_id);

    if (sp_atom_app_exec == GDK_NONE) {
	sprintf(sp_exec_atom_name, "%s_0", unique_id);
	sp_atom_app_exec = gdk_atom_intern(sp_exec_atom_name, FALSE);
	sprintf(sp_argv_atom_name, "%s_1", unique_id);
	sp_atom_app_argv = gdk_atom_intern(sp_argv_atom_name, FALSE);
	spDebug(30, "spAddAtomSelectionTargetGtk", "sp_atom_app_exec = %ld (%s), sp_atom_app_argv = %ld (%s)\n",
		(long)sp_atom_app_exec, sp_exec_atom_name, (long)sp_atom_app_argv, sp_argv_atom_name);
    }
	
    gtk_selection_add_targets(widget, sp_atom_app_exec,
			      target_table, /*G_N_ELEMENTS(target_table)*/1);
    gtk_selection_add_targets(widget, sp_atom_app_argv,
			      target_table, /*G_N_ELEMENTS(target_table)*/1);

    return;
}

void spFindExcutedApplicationGtk(GtkWidget *widget)
{
    /* required even in 3.0.0+ */
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) gdk_threads_enter();
    
    gtk_selection_owner_set(widget, sp_atom_app_argv, GDK_CURRENT_TIME);

    /* "selection-received" signal will be generated */
    gtk_selection_convert(widget, sp_atom_app_exec, GDK_SELECTION_TYPE_STRING, GDK_CURRENT_TIME);

    spDebug(100, "spFindExcutedApplicationGtk", "before loop\n");
    while (sp_selection_data_length < 0) {
	spDebug(100, "spFindExcutedApplicationGtk", "in loop\n");
	gtk_main_iteration_do(TRUE);
    }
    
    /* required even in 3.0.0+ */
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) gdk_threads_leave();
    
    spDebug(100, "spFindExcutedApplicationGtk", "after loop: sp_selection_data_length = %d\n", sp_selection_data_length);
    if (sp_selection_data_length > 0) {
	/* found already executed application */

        /* required even in 3.0.0+ */
	if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) gdk_threads_enter();
    
	while (sp_argv_send_done == SP_FALSE) {
	    spDebug(100, "spFindExcutedApplicationGtk", "in client loop\n");
	    gtk_main_iteration_do(TRUE);
	}
	
        /* required even in 3.0.0+ */
	if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) gdk_threads_leave();
	
	spQuit(0);
    }
    
    spDebug(30, "spFindExcutedApplicationGtk", "done\n");
    
    return;
}

void spSetAtomSelectionOwnerGtk(GtkWidget *widget)
{
    if (sp_atom_app_exec != GDK_NONE) {
	gtk_selection_owner_set(widget, sp_atom_app_exec, GDK_CURRENT_TIME);
    }

    return;
}
#endif
