/*
 *	spGraphics_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spDraw.h>

#include <sp/spTopLevelP.h>
#include <sp/spPrimitiveP.h>
#include <sp/spGraphicsP.h>

extern spTopLevel sp_toplevel;

#if GTK_CHECK_VERSION(3,0,0)
static spBool getLineDash(spGraphics graphics, double **dash_list, int *ndash)
{
    int n = 0;
    double *list = NULL;
    static double dash_line[] = {9.0, 3.0};
    static double dot_line[] = {2.0, 4.0};
    static double dash_dot_line[] = {9.0, 3.0, 2.0, 3.0};
    static double dash_dot_dot_line[] = {9.0, 3.0, 2.0, 3.0, 2.0, 3.0};

    if (graphics == NULL) return SP_FALSE;

    switch (SpGraphicsPart(graphics).line_type) {
      case SP_LINE_DASH:
	list = dash_line;
	n = 2;
	break;
      case SP_LINE_DOT:
	list = dot_line;
	n = 2;
	break;
      case SP_LINE_DASH_DOT:
	list = dash_dot_line;
	n = 4;
	break;
      case SP_LINE_DASH_DOT_DOT:
	list = dash_dot_dot_line;
	n = 6;
	break;
      default:
	list = NULL;
	n = 0;
	break;
    }

    if (dash_list != NULL) *dash_list = list;
    if (ndash != NULL) *ndash = n;

    return SP_TRUE;
}

spBool spUpdateLineContextCairo(cairo_t *cr, spGraphics graphics)
{
    int num_dashes = 0;
    double *dashes = NULL;

    if (getLineDash(graphics, &dashes, &num_dashes) == SP_FALSE) {
	return SP_FALSE;
    }
    
    spDebug(80, "spUpdateLineContextCairo", "in\n");

    /*cairo_set_line_cap(cr, CAIRO_LINE_CAP_SQUARE);*/
    cairo_set_dash(cr, dashes, num_dashes, 0.0);
    cairo_set_line_width(cr, (double)SpGraphicsPart(graphics).line_width);

#if GTK_CHECK_VERSION(3,4,0)
    gdk_cairo_set_source_rgba(cr, &SpGraphicsArch(graphics).rgba);
#else
    gdk_cairo_set_source_color(cr, &SpGraphicsArch(graphics).color);
#endif
    spDebug(80, "spUpdateLineContextCairo", "call cairo_set_operator of %d\n",
	    SpGraphicsArch(graphics).op);
    cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
    cairo_set_operator(cr, SpGraphicsArch(graphics).op);

    spDebug(80, "spUpdateLineContextCairo", "done\n");
    
    return SP_TRUE;
}

spBool spUpdateColorCairo(cairo_t *cr, spGraphics graphics)
{
    spDebug(80, "spUpdateColorCairo", "in: graphics = %ld\n", (long)graphics);
    
#if GTK_CHECK_VERSION(3,4,0)
    gdk_cairo_set_source_rgba(cr, &SpGraphicsArch(graphics).rgba);
#else
    gdk_cairo_set_source_color(cr, &SpGraphicsArch(graphics).color);
#endif
    spDebug(80, "spUpdateColorCairo", "call cairo_set_operator of %d\n",
	    SpGraphicsArch(graphics).op);
    cairo_set_antialias(cr, CAIRO_ANTIALIAS_NONE);
    cairo_set_operator(cr, SpGraphicsArch(graphics).op);

    spDebug(80, "spUpdateColorCairo", "done\n");
    
    return SP_TRUE;
}
#endif

#if GTK_CHECK_VERSION(3,4,0)
spBool spEqColorGtk(GdkRGBA *color1, GdkRGBA *color2)
{
    if (gdk_rgba_equal(color1, color2)) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spGetColorValueGtk(const char *color_name, GdkRGBA *color)
{
    if (strcaseeq(color_name, "max")) {
	color->red = 1.0;
	color->green = 1.0;
	color->blue = 1.0;
	color->alpha = 1.0;
    } else if (strcaseeq(color_name, "min")) {
	color->red = 0.0;
	color->green = 0.0;
	color->blue = 0.0;
	color->alpha = 1.0;
    } else {
	if (!gdk_rgba_parse(color, color_name)) {
            return SP_FALSE;
        }
    }

    return SP_TRUE;
}
#else
spBool spEqColorGtk(GdkColor *color1, GdkColor *color2)
{
    if (color1->red == color2->red
	&& color1->green == color2->green
	&& color1->blue == color2->blue) {
	return SP_TRUE;
    }

    return SP_FALSE;
}

spBool spGetColorValueGtk(const char *color_name, GdkColor *color)
{
    spDebug(80, "spGetColorValueGtk", "in: color_name = %s\n", color_name);

    if (strcaseeq(color_name, "max")) {
	color->red = 65535;
	color->green = 65535;
	color->blue = 65535;
    } else if (strcaseeq(color_name, "min")) {
	color->red = 0;
	color->green = 0;
	color->blue = 0;
    } else {
        if (!gdk_color_parse(color_name, color)) {
            return SP_FALSE;
        }
    }
#if !GTK_CHECK_VERSION(3,0,0)
    gdk_color_alloc(SpTopLevelArch(sp_toplevel).colormap, color);
#endif
    spDebug(100, "spGetColorValueGtk", "r = %d, g = %d, b = %d\n", color->red, color->green, color->blue);
    
    return SP_TRUE;
}
#endif

spPixel spGetColorPixel(const char *color_name)
{
    spPixel pixel = 0L;
#if GTK_CHECK_VERSION(3,4,0)
    GdkRGBA rgba;
#else
    GdkColor color;
#endif

    if (strnone(color_name)) return 0L;
    
#if GTK_CHECK_VERSION(3,4,0)
    spGetColorValueGtk(color_name, &rgba);
    pixel = spRGB(spRound(rgba.red * 255.0), spRound(rgba.green * 255.0), spRound(rgba.blue * 255.0));
#else
    spGetColorValueGtk(color_name, &color);
    pixel = spRGB(color.red/256, color.green/256, color.blue/256);
#endif
    
    return pixel;
}

spBool spGetCursorArch(spCursor cursor)
{
    guint shape;
    spCursorType cursor_type = cursor->type;
    
    switch (cursor_type) {
      case SP_CURSOR_TEXT:
	shape = GDK_XTERM;
	break;
      case SP_CURSOR_WAIT:
	shape = GDK_WATCH;
	break;
      case SP_CURSOR_CROSS:
	shape = GDK_CROSSHAIR;
	break;
      case SP_CURSOR_HAND:
	shape = GDK_HAND2;
	break;
      case SP_CURSOR_MOVE:
	shape = GDK_FLEUR;
	break;
      case SP_CURSOR_SIZE:
	shape = GDK_SIZING;
	break;
      case SP_CURSOR_SW_RESIZE:
	shape = GDK_TOP_LEFT_CORNER;
	break;
      case SP_CURSOR_SE_RESIZE:
	shape = GDK_TOP_RIGHT_CORNER;
	break;
      case SP_CURSOR_NW_RESIZE:
	shape = GDK_BOTTOM_LEFT_CORNER;
	break;
      case SP_CURSOR_NE_RESIZE:
	shape = GDK_BOTTOM_RIGHT_CORNER;
	break;
      case SP_CURSOR_N_RESIZE:
	shape = GDK_TOP_SIDE;
	break;
      case SP_CURSOR_S_RESIZE:
	shape = GDK_BOTTOM_SIDE;
	break;
      case SP_CURSOR_W_RESIZE:
	shape = GDK_LEFT_SIDE;
	break;
      case SP_CURSOR_E_RESIZE:
	shape = GDK_RIGHT_SIDE;
	break;
      default:
	shape = GDK_TOP_LEFT_ARROW;
	break;
    }

    if ((SpCursorArch(cursor).cursor = gdk_cursor_new(shape)) == NULL) {
	return SP_FALSE;
    }

    return SP_TRUE;
}

void spDestroyCursorArch(spCursor cursor)
{
#if GTK_CHECK_VERSION(3,0,0)
    /*gdk_cursor_unref(SpCursorArch(cursor).cursor);*/
    g_object_unref(SpCursorArch(cursor).cursor);
#else
    gdk_cursor_destroy(SpCursorArch(cursor).cursor);
#endif
    return;
}

#if !GTK_CHECK_VERSION(1,2,0)
#include <gdk/gdkprivate.h>
#endif

#if GTK_CHECK_VERSION(3,0,0)
spBool spSetGraphicsModeArch(spGraphics graphics, spGraphicsMode mode)
{
    switch (mode) {
      case SP_GM_COPY:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_SOURCE;
	break;
      case SP_GM_SOURCE_OVER:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_OVER;
	break;
      case SP_GM_SOURCE_IN:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_IN;
	break;
      case SP_GM_SOURCE_OUT:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_OUT;
	break;
      case SP_GM_SOURCE_ATOP:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_ATOP;
	break;
      case SP_GM_DEST_OVER:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_DEST_OVER;
	break;
      case SP_GM_DEST_IN:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_DEST_IN;
	break;
      case SP_GM_DEST_OUT:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_DEST_OUT;
	break;
      case SP_GM_DEST_ATOP:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_DEST_ATOP;
	break;
      case SP_GM_PORTER_DUFF_XOR:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_XOR;
	break;
      case SP_GM_PLUS_LIGHTER:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_ADD;
	break;

      case SP_GM_SATURATE:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_SATURATE;
	break;
      case SP_GM_MULTIPLY:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_MULTIPLY;
	break;
      case SP_GM_SCREEN:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_SCREEN;
	break;
      case SP_GM_OVERLAY:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_OVERLAY;
	break;
      case SP_GM_DARKEN:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_DARKEN;
	break;
      case SP_GM_LIGHTEN:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_LIGHTEN;
	break;

      case SP_GM_COLOR_DODGE:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_COLOR_DODGE;
	break;
      case SP_GM_COLOR_BURN:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_COLOR_BURN;
	break;
      case SP_GM_HARD_LIGHT:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_HARD_LIGHT;
	break;
      case SP_GM_SOFT_LIGHT:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_SOFT_LIGHT;
	break;
      case SP_GM_DIFFERENCE:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_DIFFERENCE;
	break;
      case SP_GM_EXCLUSION:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_EXCLUSION;
	break;
	
      case SP_GM_HSL_HUE:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_HSL_HUE;
	break;
      case SP_GM_HSL_SATURATION:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_HSL_SATURATION;
	break;
      case SP_GM_HSL_COLOR:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_HSL_COLOR;
	break;
      case SP_GM_HSL_LUMINOSITY:
	SpGraphicsArch(graphics).op = CAIRO_OPERATOR_HSL_LUMINOSITY;
	break;
      default:
	return SP_FALSE;
    }
    return SP_TRUE;
}
#else
spBool spSetGraphicsModeArch(spGraphics graphics, spGraphicsMode mode)
{
    GdkFunction function = GDK_COPY;

    switch (mode) {
      case SP_GM_COPY:
	function = GDK_COPY;
	break;
      case SP_GM_XOR:
	function = GDK_XOR;
	break;
      case SP_GM_INVERT:
	function = GDK_INVERT;
	break;
#if GTK_CHECK_VERSION(1,2,0)
      case SP_GM_AND:
	function = GDK_AND;
	break;
      case SP_GM_OR:
	function = GDK_OR;
	break;
#else
      case SP_GM_AND:
	{
	    GdkGCPrivate *private = (GdkGCPrivate*)SpGraphicsArch(graphics).gc;
	    XSetFunction (private->xdisplay, private->xgc, GXand);
	}
	return SP_TRUE;
      case SP_GM_OR:
	{
	    GdkGCPrivate *private = (GdkGCPrivate*)SpGraphicsArch(graphics).gc;
	    XSetFunction (private->xdisplay, private->xgc, GXor);
	}
	return SP_TRUE;
#endif
      default:
	return SP_FALSE;
    }

    gdk_gc_set_function(SpGraphicsArch(graphics).gc, function);
    
    return SP_TRUE;
}
#endif

void spGraphicsPartInitArch(spGraphics graphics)
{
#if GTK_CHECK_VERSION(2,0,0)
    SpGraphicsArch(graphics).font_desc = NULL;
#else
    SpGraphicsArch(graphics).font = NULL;
#endif

#if GTK_CHECK_VERSION(3,0,0)
#if GTK_CHECK_VERSION(3,4,0)
    memset(&SpGraphicsArch(graphics).rgba, 0, sizeof(GdkRGBA));
    spGetColorValueGtk("black", &SpGraphicsArch(graphics).rgba);
#else
    memset(&SpGraphicsArch(graphics).color, 0, sizeof(GdkColor));
    spGetColorValueGtk("black", &SpGraphicsArch(graphics).color);
#endif
    SpGraphicsArch(graphics).op = CAIRO_OPERATOR_OVER;
#else
    {
	GdkGCValues gc_values;
	GdkGCValuesMask gc_values_mask;

	spGetColorValueGtk("black", &gc_values.foreground);
	spGetColorValueGtk("white", &gc_values.background);
	gc_values_mask = GDK_GC_FOREGROUND | GDK_GC_BACKGROUND;
	SpGraphicsArch(graphics).gc = gdk_gc_new_with_values(SpTopLevelArch(sp_toplevel).pixmap,
							     &gc_values,
							     gc_values_mask);
    }
#endif
    
    return;
}

void spGraphicsPartFreeArch(spGraphics graphics)
{
#if GTK_CHECK_VERSION(2,0,0)
    if (SpGraphicsArch(graphics).font_desc != NULL) {
	pango_font_description_free(SpGraphicsArch(graphics).font_desc);
	SpGraphicsArch(graphics).font_desc = NULL;
    }
#else
    if (SpGraphicsArch(graphics).font != NULL) {
	gdk_font_unref(SpGraphicsArch(graphics).font);
	SpGraphicsArch(graphics).font = NULL;
    }
#endif

#if GTK_CHECK_VERSION(3,0,0)
#else
    if (SpGraphicsArch(graphics).gc != NULL) {
	gdk_gc_unref(SpGraphicsArch(graphics).gc);
	SpGraphicsArch(graphics).gc = NULL;
    }
#endif
    
    return;
}

#if !GTK_CHECK_VERSION(3,0,0)
static GdkLineStyle getLineType(spGraphics graphics, char **dash_list, int *ndash)
{
    int n = 0;
    GdkLineStyle style = GDK_LINE_SOLID;
    char *list = NULL;
    static char dash_line[] = {9, 3};
    static char dot_line[] = {2, 4};
    static char dash_dot_line[] = {9, 3, 2, 3};
    static char dash_dot_dot_line[] = {9, 3, 2, 3, 2, 3};

    if (graphics == NULL) return style;

    switch (SpGraphicsPart(graphics).line_type) {
      case SP_LINE_DASH:
	style = GDK_LINE_ON_OFF_DASH;
	list = dash_line;
	n = 2;
	break;
      case SP_LINE_DOT:
	style = GDK_LINE_ON_OFF_DASH;
	list = dot_line;
	n = 2;
	break;
      case SP_LINE_DASH_DOT:
	style = GDK_LINE_ON_OFF_DASH;
	list = dash_dot_line;
	n = 4;
	break;
      case SP_LINE_DASH_DOT_DOT:
	style = GDK_LINE_ON_OFF_DASH;
	list = dash_dot_dot_line;
	n = 6;
	break;
      default:
	style = GDK_LINE_SOLID;
	list = NULL;
	n = 0;
	break;
    }

    if (dash_list != NULL) *dash_list = list;
    if (ndash != NULL) *ndash = n;

    return style;
}
#endif

void spGraphicsCreateArch(spGraphics graphics)
{
#if !GTK_CHECK_VERSION(3,0,0)
    GdkLineStyle line_style;
    int ndash = 0;
    char *dash_list = NULL;
    
    line_style = getLineType(graphics, &dash_list, &ndash);
    
    gdk_gc_set_line_attributes(SpGraphicsArch(graphics).gc,
			       SpGraphicsPart(graphics).line_width,
			       line_style,
			       GDK_CAP_BUTT,
			       GDK_JOIN_MITER);

    if (line_style != GDK_LINE_SOLID && dash_list != NULL) {
#if GTK_CHECK_VERSION(1,2,0)
	gdk_gc_set_dashes(SpGraphicsArch(graphics).gc, 0, (gint8 *)dash_list, ndash);
#else
	GdkGCPrivate *private = (GdkGCPrivate*)SpGraphicsArch(graphics).gc;
	
	XSetDashes(private->xdisplay, private->xgc, 0, dash_list, ndash);
#endif
    }
#endif
    
    spSetGraphicsModeArch(graphics, SpGraphicsPart(graphics).mode);
    
    spSetForegroundPixelArch(graphics);
    spSetBackgroundPixelArch(graphics);
    
    return;
}

spBool spSetFontArch(spGraphics graphics)
{
#if GTK_CHECK_VERSION(2,0,0)
    if (SpGraphicsArch(graphics).font_desc != NULL) {
	pango_font_description_free(SpGraphicsArch(graphics).font_desc);
    }
    SpGraphicsArch(graphics).font_desc = pango_font_description_from_string(SP_GTK_DEFAULT_FONT_NAME);
    if (!strnone(SpGraphicsPart(graphics).font_name)) {
	spDebug(10, "spSetFontArch", "font = %s\n",
		SpGraphicsPart(graphics).font_name);

	spFontNameToFontDescriptionGtk(SpGraphicsPart(graphics).font_name,
				       SpGraphicsArch(graphics).font_desc);
    }
#else
    if (SpGraphicsArch(graphics).font != NULL) {
	gdk_font_unref(SpGraphicsArch(graphics).font);
    }

    if (!strnone(SpGraphicsPart(graphics).font_name)) {
	spDebug(10, "spSetFontArch", "font = %s\n",
		SpGraphicsPart(graphics).font_name);

	if ((SpGraphicsArch(graphics).font =
	     gdk_fontset_load(SpGraphicsPart(graphics).font_name)) == NULL) {
	    SpGraphicsArch(graphics).font =
		gdk_font_load(SpGraphicsPart(graphics).font_name);
	}
    } else {
	SpGraphicsArch(graphics).font =
	    gdk_fontset_load(SP_GTK_DEFAULT_FONT_NAME);
    }
    
    if (SpGraphicsArch(graphics).font == NULL) {
	SpGraphicsArch(graphics).font =
	    gdk_fontset_load("-*-*-*-*-*--*-*-*-*-*-*-*-*");
    }
#endif
    
    return SP_TRUE;
}

#if GTK_CHECK_VERSION(3,0,0)
static void color_shade(const GdkRGBA *color,
			gdouble factor,
			GdkRGBA *color_return)
{
    GtkSymbolicColor *literal, *shade;

    spDebug(80, "color_shade", "factor = %f\n", factor);
    
    literal = gtk_symbolic_color_new_literal(color);
    shade = gtk_symbolic_color_new_shade(literal, factor);
    gtk_symbolic_color_unref(literal);

    gtk_symbolic_color_resolve(shade, NULL, color_return);
    gtk_symbolic_color_unref(shade);

    spDebug(80, "color_shade", "done\n");
    
    return;
}
#endif

spBool spGetSystemColorPixelArch(spSystemColorType type, spPixel *pixel)
{
#if GTK_CHECK_VERSION(1,2,0)
    spComponent window;
    
    if ((window = spGetChild(NULL)) != NULL) {
#if GTK_CHECK_VERSION(3,0,0)
	GtkStyleContext *context;
	GtkStateFlags flags = GTK_STATE_FLAG_NORMAL;
	GdkRGBA color;
	unsigned char r, g, b;

	context = gtk_widget_get_style_context(SpPrimitiveArch(window).widget);
	spDebug(80, "spGetSystemColorPixelArch", "gtk_widget_get_style_context done\n");
	
	if (type == SP_SYSTEM_COLOR_TOP_SHADOW || type == SP_SYSTEM_COLOR_BOTTOM_SHADOW
	    || type == SP_SYSTEM_COLOR_HIGHLIGHT || type == SP_SYSTEM_COLOR_DARK_SHADOW) {
	    int i;
	    GdkRGBA *colors[2];

	    colors[0] = colors[1] = NULL;
	    
	    gtk_style_context_get(context, GTK_STATE_FLAG_NORMAL,
				  "border-top-color", &colors[0],
				  "border-bottom-color", &colors[1],
				  NULL);
	    spDebug(80, "spGetSystemColorPixelArch", "colors[0] = %ld, colors[1] = %ld\n",
		    (long)colors[0], (long)colors[1]);
	    
	    if (type == SP_SYSTEM_COLOR_HIGHLIGHT) {
		color = *colors[0];
		color_shade(&color, /*1.3*/1.6, &color);
	    } else if (type == SP_SYSTEM_COLOR_DARK_SHADOW) {
		color = *colors[1];
		color_shade(&color, /*0.7*/0.8, &color);
	    } else if (type == SP_SYSTEM_COLOR_TOP_SHADOW) {
		color = *colors[0];
		color_shade(&color, 1.4, &color);
	    } else if (type == SP_SYSTEM_COLOR_BOTTOM_SHADOW) {
		color = *colors[1];
		color_shade(&color, 0.9, &color);
	    }
	    
	    for (i = 0; i < 2; i++) {
		if (colors[i] != NULL) {
		    gdk_rgba_free(colors[i]);
		}
	    }
	} else if (type == SP_SYSTEM_COLOR_BACKGROUND
		   || type == SP_SYSTEM_COLOR_TAB_BACKGROUND) {
	    gtk_style_context_get_background_color(context, flags, &color);
	} else if (type == SP_SYSTEM_COLOR_FOCUS) {
	    gtk_style_context_get_color(context, flags, &color);
	} else if (type == SP_SYSTEM_COLOR_WHITE) {
	    gdk_rgba_parse(&color, "black");
	} else if (type == SP_SYSTEM_COLOR_BLACK) {
	    gdk_rgba_parse(&color, "white");
	} else {
	    return SP_FALSE;
	}

	r = (unsigned char)MIN(color.red * 255.0, 255.0);
	g = (unsigned char)MIN(color.green * 255.0, 255.0);
	b = (unsigned char)MIN(color.blue * 255.0, 255.0);
	*pixel = spRGB(r, g, b);
#else
	GdkColor color;

	switch (type) {
	  case SP_SYSTEM_COLOR_BACKGROUND:
	  case SP_SYSTEM_COLOR_TAB_BACKGROUND:
	    color = SpPrimitiveArch(window).widget->style->bg[GTK_STATE_NORMAL];
	    break;
	  case SP_SYSTEM_COLOR_TOP_SHADOW:
	    color = SpPrimitiveArch(window).widget->style->light[GTK_STATE_NORMAL];
	    break;
	  case SP_SYSTEM_COLOR_BOTTOM_SHADOW:
	    color = SpPrimitiveArch(window).widget->style->dark[GTK_STATE_NORMAL];
	    break;
	  case SP_SYSTEM_COLOR_HIGHLIGHT:
	    color = SpPrimitiveArch(window).widget->style->light[GTK_STATE_NORMAL];
	    break;
	  case SP_SYSTEM_COLOR_DARK_SHADOW:
	    color = SpPrimitiveArch(window).widget->style->black;
	    break;
	  case SP_SYSTEM_COLOR_BLACK:
	    color = SpPrimitiveArch(window).widget->style->black;
	    break;
	  case SP_SYSTEM_COLOR_WHITE:
	    color = SpPrimitiveArch(window).widget->style->white;
	    break;
	  case SP_SYSTEM_COLOR_FOCUS:
	    color = SpPrimitiveArch(window).widget->style->black;
	    break;
	  default:
	    return SP_FALSE;
	}
	
	gdk_color_alloc(SpTopLevelArch(sp_toplevel).colormap, &color);
	*pixel = spRGB(color.red/256, color.green/256, color.blue/256);
#endif
	
	return SP_TRUE;
    }
#endif
    
    return spGetSystemColorPixelNormal(type, pixel);
}

#if GTK_CHECK_VERSION(3,4,0)
spBool spPixelToGdkRGBA(spPixel pixel, GdkRGBA *color)
{
    color->red = (double)spGetRValue(pixel) / 255.0;
    color->green = (double)spGetGValue(pixel) / 255.0;
    color->blue = (double)spGetBValue(pixel) / 255.0;
    color->alpha = 1.0;

    return SP_TRUE;
}
#else
spBool spPixelToGdkColor(spPixel pixel, GdkColor *color)
{
    color->red = 256 * (unsigned short)spGetRValue(pixel);
    color->green = 256 * (unsigned short)spGetGValue(pixel);
    color->blue = 256 * (unsigned short)spGetBValue(pixel);
#if !GTK_CHECK_VERSION(3,0,0)
    if (!gdk_color_alloc(SpTopLevelArch(sp_toplevel).colormap, color)) {
	spDebug(80, "spPixelToGdkColor", "gdk_color_alloc failed\n");
	return SP_FALSE;
    }
#else
    color->pixel = pixel;
#endif

    return SP_TRUE;
}
#endif

spBool spSetForegroundPixelArch(spGraphics graphics)
{
#if GTK_CHECK_VERSION(3,4,0)
    GdkRGBA rgba;
#else
    GdkColor color;
#endif

#if GTK_CHECK_VERSION(3,4,0)
    if (spPixelToGdkRGBA(SpGraphicsPart(graphics).fg_pixel, &rgba) == SP_TRUE) {
	SpGraphicsArch(graphics).rgba = rgba;
	return SP_TRUE;
    }
#else
    if (spPixelToGdkColor(SpGraphicsPart(graphics).fg_pixel, &color) == SP_TRUE) {
#if GTK_CHECK_VERSION(3,0,0)
	SpGraphicsArch(graphics).color = color;
#else
	gdk_gc_set_foreground(SpGraphicsArch(graphics).gc, &color);
#endif
	return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}

spBool spSetBackgroundPixelArch(spGraphics graphics)
{
#if !GTK_CHECK_VERSION(3,0,0)
    GdkColor color;

    if (spPixelToGdkColor(SpGraphicsPart(graphics).bg_pixel, &color) == SP_TRUE) {
	gdk_gc_set_background(SpGraphicsArch(graphics).gc, &color);
	return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}

const char **spGetFontFamilyListArch(void)
{
/*#if GTK_CHECK_VERSION(3,0,0)*/
#if GTK_CHECK_VERSION(1,2,0)
    return NULL;
#else
    return spGetFontFamilyListX11(GDK_DISPLAY());
#endif
}

spBool spIsFontSizeSupportedArch(int family, unsigned long style, int size)
{
/*#if GTK_CHECK_VERSION(3,0,0)*/
#if GTK_CHECK_VERSION(1,2,0)
    return SP_FALSE;
#else
    return spIsFontSizeSupportedX11(GDK_DISPLAY(), family, style, size);
#endif
}

#if GTK_CHECK_VERSION(2,0,0)
spBool spFontNameToFontDescriptionGtk(const char *font_name, PangoFontDescription *font_desc)
{
    int size;
    unsigned long fstyle;
    PangoWeight weight = PANGO_WEIGHT_NORMAL;
    PangoStyle style = PANGO_STYLE_NORMAL;
    PangoStretch stretch = PANGO_STRETCH_NORMAL;
    char name[SP_MAX_FONTNAME];

    if (font_desc == NULL) return SP_FALSE;
    
    if (spParseFontName(font_name, name, &fstyle, &size) == SP_TRUE) {
	spDebug(80, "spFontNameToFontDescriptionGtk", "size = %d, name = %s\n", size, name);
	if (streq(name, "*")) {
	    pango_font_description_set_family(font_desc, "Sans");
	} else {
	    pango_font_description_set_family(font_desc, name);
	}
	pango_font_description_set_size(font_desc, size * PANGO_SCALE);
	
	if (fstyle & SP_FONT_WEIGHT_BLACK) {
	    weight = PANGO_WEIGHT_ULTRABOLD;
	} else if (fstyle & SP_FONT_WEIGHT_BOLD) {
	    weight = PANGO_WEIGHT_BOLD;
	} else if (fstyle & SP_FONT_WEIGHT_LIGHT) {
	    weight = PANGO_WEIGHT_LIGHT;
	}
	pango_font_description_set_weight(font_desc, weight);
	
	if (fstyle & SP_FONT_SLANT_ITALIC) {
	    style = PANGO_STYLE_ITALIC;
	} else if (fstyle & SP_FONT_SLANT_OBLIQUE) {
	    style = PANGO_STYLE_OBLIQUE;
	}
	pango_font_description_set_style(font_desc, style);
	
	if (fstyle & SP_FONT_WIDTH_CONDENSED || fstyle & SP_FONT_WIDTH_NARROW) {
	    stretch = PANGO_STRETCH_CONDENSED;
	} else if (fstyle & SP_FONT_WIDTH_EXTENDED) {
	    stretch = PANGO_STRETCH_EXPANDED;
	}
	pango_font_description_set_stretch(font_desc, stretch);
    } else {
	PangoFontDescription *tdesc;

	spDebug(80, "spFontNameToFontDescriptionGtk", "spParseFontName failed: font_name = %s\n", font_name);
	
	if ((tdesc = pango_font_description_from_string(font_name)) == NULL) {
	    return SP_FALSE;
	}
	pango_font_description_merge(font_desc, tdesc, TRUE);
	pango_font_description_free(tdesc);
    }
    
    return SP_TRUE;
}

spBool spFontDescriptionToFontNameGtk(PangoFontDescription *font_desc, char *font_name, int buf_size)
{
    int size;
    unsigned long fstyle = 0;
    char family_name[SP_MAX_FONTNAME];
    PangoWeight weight;
    PangoStyle style;
    PangoStretch stretch;
    char font_name_buf[SP_MAX_FONTNAME];
    
    spStrCopy(family_name, sizeof(family_name), pango_font_description_get_family(font_desc));
    size = PANGO_PIXELS(pango_font_description_get_size(font_desc));
    
    weight = pango_font_description_get_weight(font_desc);
    if (weight <= PANGO_WEIGHT_LIGHT) {
	fstyle |= SP_FONT_WEIGHT_LIGHT;
    } else if (weight >= PANGO_WEIGHT_ULTRABOLD) {
	fstyle |= SP_FONT_WEIGHT_BLACK;
    } else if (weight >= PANGO_WEIGHT_BOLD) {
	fstyle |= SP_FONT_WEIGHT_BOLD;
    }
    
    style = pango_font_description_get_style(font_desc);
    if (style == PANGO_STYLE_OBLIQUE) {
	fstyle |= SP_FONT_SLANT_OBLIQUE;
    } else if (style == PANGO_STYLE_ITALIC) {
	fstyle |= SP_FONT_SLANT_ITALIC;
    }
    
    stretch = pango_font_description_get_stretch(font_desc);
    if (stretch >= PANGO_STRETCH_EXPANDED) {
	fstyle |= SP_FONT_WIDTH_EXTENDED;
    } else if (stretch <= PANGO_STRETCH_CONDENSED) {
	fstyle |= SP_FONT_WIDTH_CONDENSED;
    }

    spGetFontName(font_name_buf, family_name, fstyle, size);
    spStrCopy(font_name, buf_size, font_name_buf);
    
    return SP_TRUE;
}
#endif

unsigned long spGetGraphicsModeCapsArch(spComponent component)
{
    return 
#if !GTK_CHECK_VERSION(3,0,0)
	SP_GRAPHICS_MODE_CAPS_XOR | SP_GRAPHICS_MODE_CAPS_INVERT
	| SP_GRAPHICS_MODE_CAPS_AND | SP_GRAPHICS_MODE_CAPS_OR
#else
	SP_GRAPHICS_MODE_CAPS_SOURCE_OVER |
	SP_GRAPHICS_MODE_CAPS_SOURCE_IN |
	SP_GRAPHICS_MODE_CAPS_SOURCE_OUT |
	SP_GRAPHICS_MODE_CAPS_SOURCE_ATOP |
	SP_GRAPHICS_MODE_CAPS_DEST_OVER |
	SP_GRAPHICS_MODE_CAPS_DEST_IN |
	SP_GRAPHICS_MODE_CAPS_DEST_OUT |
	SP_GRAPHICS_MODE_CAPS_DEST_ATOP |
	SP_GRAPHICS_MODE_CAPS_PORTER_DUFF_XOR |
	SP_GRAPHICS_MODE_CAPS_PLUS_LIGHTER |
	
	SP_GRAPHICS_MODE_CAPS_SATURATE |       
    
	SP_GRAPHICS_MODE_CAPS_MULTIPLY |       
	SP_GRAPHICS_MODE_CAPS_SCREEN |         
	SP_GRAPHICS_MODE_CAPS_OVERLAY |        
	SP_GRAPHICS_MODE_CAPS_DARKEN |         
	SP_GRAPHICS_MODE_CAPS_LIGHTEN |        

	SP_GRAPHICS_MODE_CAPS_COLOR_DODGE |    
	SP_GRAPHICS_MODE_CAPS_COLOR_BURN |     
	SP_GRAPHICS_MODE_CAPS_HARD_LIGHT |     
	SP_GRAPHICS_MODE_CAPS_SOFT_LIGHT |     
	SP_GRAPHICS_MODE_CAPS_DIFFERENCE |     
	SP_GRAPHICS_MODE_CAPS_EXCLUSION |      

	SP_GRAPHICS_MODE_CAPS_HSL_HUE |        
	SP_GRAPHICS_MODE_CAPS_HSL_SATURATION |  
	SP_GRAPHICS_MODE_CAPS_HSL_COLOR |      
	SP_GRAPHICS_MODE_CAPS_HSL_LUMINOSITY
#endif
	;
}
