/*
 *	spImage_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spString.h>
#include <sp/spMemory.h>
#include <sp/spCanvas.h>

#include <sp/spLocale.h>
#include <sp/spFrameP.h>
#include <sp/spDrawP.h>
#include <sp/spImageP.h>
#include <sp/spGraphicsP.h>

extern spTopLevel sp_toplevel;

void spImageCreateArch(spComponent component)
{
    return;
}

spBool spImageLoadArch(spComponent component, const char *filename)
{
    char filename2[SP_MAX_PATHNAME];

    if (spEqSuffix(filename, ".xpm")) {
	strcpy(filename2, filename);
#if GTK_CHECK_VERSION(3,0,0)
    } else if (spGetSuffix(filename) != NULL) {
	spStrCopy(filename2, sizeof(filename2), filename);
#endif
    } else {
	spStrCopy(filename2, sizeof(filename2), filename);
	spReplaceSuffix(filename2, ".xpm");
    }
    spDebug(50, "spImageLoadArch", "filename2 = %s\n", filename2);
    
#if GTK_CHECK_VERSION(3,0,0)
    {
	GError *error = NULL;
	GdkPixbuf *pixbuf;
	int w, h;
	
	pixbuf = gdk_pixbuf_new_from_file(filename2, &error);
	if (pixbuf == NULL) {
	    spDebug(50, "spImageLoadArch", "filename = %s, error = %s\n", filename2, error->message);
	    g_error_free (error);
	    return SP_FALSE;
	}
	w = gdk_pixbuf_get_width(pixbuf);
	h = gdk_pixbuf_get_height(pixbuf);
	SpPrimitiveArch(component).pixmap = cairo_image_surface_create(CAIRO_FORMAT_RGB24, w, h);
	if (SpPrimitiveArch(component).pixmap == NULL) {
	    spDebug(50, "spImageLoadArch", "cannot create pixmap\n");
	    return SP_FALSE;
	}
	SpPrimitiveArch(component).pixmap_cr = cairo_create(SpPrimitiveArch(component).pixmap);
	if (SpPrimitiveArch(component).pixmap_cr == NULL) {
	    spDebug(50, "spImageLoadArch", "cannot create cairo context\n");
	    return SP_FALSE;
	}
	gdk_cairo_set_source_pixbuf(SpPrimitiveArch(component).pixmap_cr, pixbuf, 0.0, 0.0);
	cairo_paint(SpPrimitiveArch(component).pixmap_cr);
	g_object_unref(pixbuf);
	
	SpComponentPart(component).width = w;
	SpComponentPart(component).height = h;
    }
#else
    {
	GdkColor color;
	gint width, height;
	
	spGetColorValueGtk("#c0c0c0", &color);

	SpPrimitiveArch(component).pixmap = 
	    gdk_pixmap_colormap_create_from_xpm(NULL,
						gdk_colormap_get_system(),
						NULL, &color,
						filename2);
	
	if (SpPrimitiveArch(component).pixmap == NULL) {
	    spDebug(50, "spImageLoadArch", "cannot open file: %s\n", filename2);
	    return SP_FALSE;
	}
    
	gdk_window_get_size(SpPrimitiveArch(component).pixmap, &width, &height);
	SpComponentPart(component).width = width;
	SpComponentPart(component).height = height;
    }
#endif
    
    spDebug(50, "spImageLoadArch", "done\n");
    
    return SP_TRUE;
}

void spDrawImageArch(spComponent component)
{
    spDebug(80, "spDrawImageArch", "in\n");
    
#if GTK_CHECK_VERSION(1,2,0)
    if (SpPrimitiveArch(component).widget != NULL) {
#if GTK_CHECK_VERSION(3,0,0)
	GtkStyleContext *context;

	if (SpPrimitiveArch(component).pixmap_cr == NULL) return;
	
	context = gtk_widget_get_style_context(SpPrimitiveArch(component).widget);
	gtk_render_background(context, SpPrimitiveArch(component).pixmap_cr, 0.0, 0.0,
			      (gdouble)SpComponentPart(component).client_width,
			      (gdouble)SpComponentPart(component).client_height);
#elif 1
	gtk_style_apply_default_background(SpPrimitiveArch(component).widget->style,
					   SpPrimitiveArch(component).pixmap,
					   FALSE,
					   GTK_WIDGET_STATE(SpPrimitiveArch(component).widget),
					   NULL, 0, 0,
					   -1, -1);
#else
	/* I gave up the following because flickering may happen. */
	
	/* FIXME: I don't know how to draw theme background in a pixmap */
	/* so, copying the background of a window to the pixmap */
	gdk_window_clear(SpPrimitiveArch(component).widget->window);
	gdk_window_copy_area(SpPrimitiveArch(component).pixmap,
			     SpTopLevelArch(sp_toplevel).fg_gc,
			     0, 0,
			     SpPrimitiveArch(component).widget->window,
			     0, 0,
			     SpComponentPart(component).client_width,
			     SpComponentPart(component).client_height);
#endif
    }
#else
    gdk_draw_rectangle(SpPrimitiveArch(component).pixmap,
		       SpTopLevelArch(sp_toplevel).bg_gc,
		       TRUE,
		       0, 0,
		       SpComponentPart(component).client_width,
		       SpComponentPart(component).client_height);
#endif
    
    return;
}

void spRedrawImageArch(spComponent component)
{
    if ((spIsSubClass(component, SpImage) == SP_TRUE
	 && !strnone(SpComponentPart(component).description))
	|| spIsSubClass(component, SpPrintCanvas) == SP_TRUE) return;
    
    spDebug(30, "spRedrawImageArch", "width = %d, height = %d\n", 
	    SpComponentPart(component).client_width, 
	    SpComponentPart(component).client_height);

    if (SpPrimitiveArch(component).pixmap != NULL) {
#if GTK_CHECK_VERSION(3,0,0)
	if (SpPrimitiveArch(component).pixmap_cr != NULL) {
	    cairo_destroy(SpPrimitiveArch(component).pixmap_cr);
	}
	SpPrimitiveArch(component).pixmap_cr = NULL;
	
	cairo_surface_destroy(SpPrimitiveArch(component).pixmap);
#else
	gdk_pixmap_unref(SpPrimitiveArch(component).pixmap);
#endif
	SpPrimitiveArch(component).pixmap = NULL;
    }
#if GTK_CHECK_VERSION(2,0,0)
    if (SpPrimitiveArch(component).layout != NULL) {
	g_object_unref(SpPrimitiveArch(component).layout);
	SpPrimitiveArch(component).layout = NULL;
    }
#endif

    /* create pixmap */
#if GTK_CHECK_VERSION(3,0,0)
    SpPrimitiveArch(component).pixmap = cairo_image_surface_create(CAIRO_FORMAT_RGB24, 
								   SpComponentPart(component).client_width,
								   SpComponentPart(component).client_height);
#else
    SpPrimitiveArch(component).pixmap = gdk_pixmap_new(NULL,
						       SpComponentPart(component).client_width,
						       SpComponentPart(component).client_height,
						       SpTopLevelArch(sp_toplevel).root_depth);
#endif
    
    if (SpPrimitiveArch(component).pixmap == NULL) {
	spDebug(10, "spRedrawImageArch", "can't create pixmap\n");
    } else {
#if GTK_CHECK_VERSION(3,0,0)
	SpPrimitiveArch(component).pixmap_cr = cairo_create(SpPrimitiveArch(component).pixmap);
	if (SpPrimitiveArch(component).pixmap_cr == NULL) {
	    spDebug(10, "spRedrawImageArch", "can't create cairo context\n");
	    return;
	}
#endif
#if GTK_CHECK_VERSION(2,0,0)
	if (SpPrimitiveArch(component).widget != NULL) {
	    SpPrimitiveArch(component).layout = gtk_widget_create_pango_layout(SpPrimitiveArch(component).widget, NULL);
	} else {
	    SpPrimitiveArch(component).layout = pango_layout_new(SpTopLevelArch(sp_toplevel).context);
	}
#endif	
    }
    
    return;
}

