/*
 *	spDraw_Mac.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spPrimitiveP.h>
#include <sp/spCanvasP.h>
#include <sp/spDrawP.h>
#include <sp/spImageP.h>
#include <sp/spGraphicsP.h>

void spImageCreateArch(spComponent component)
{
    return;
}

spBool spImageLoadArch(spComponent component, const char *filename)
{
    Rect rect;
    
    SpPrimitiveArch(component).gworld = spLoadPictToGWorldMac(filename, &rect);

    if (SpPrimitiveArch(component).gworld == nil) return SP_FALSE;
    
    SpComponentPart(component).width = rect.right;
    SpComponentPart(component).height = rect.bottom;
    
    return SP_TRUE;
}

void spDrawImageArch(spComponent component)
{
    Rect rect;
    PixMapHandle current_pixmap;
    GWorldPtr save_gwp;
    GDHandle save_gdh;
    
    /* erase */
    if (spBeginDrawMac(component, NULL, &current_pixmap, &save_gwp, &save_gdh) == SP_TRUE) {
#if 1
	if (spGetAppearanceVersionMac() >= 0x00000101) {
	    int depth;
	
#if TARGET_API_MAC_CARBON
	    depth = GetPixDepth(current_pixmap);
#else
	    depth = (*current_pixmap)->pixelSize;
#endif
	    SetThemeBackground(kThemeBrushDialogBackgroundActive, depth, true);
	}
#endif
	
	SetRect(&rect, 0, 0, SpComponentPart(component).client_width,
		SpComponentPart(component).client_height);
	EraseRect(&rect);
	
	spEndDrawMac(component, NULL, current_pixmap, save_gwp, save_gdh);
    }
    
    return;
}

void spRedrawImageArch(spComponent component)
{
    Rect rect;
#if defined(SP_GWORLD_PIXMAP_ALWAYS_LOCK)
    PixMapHandle pixmap;
#endif
#if TARGET_API_MAC_CARBON
    GWorldFlags flags = 0;
#else
    GWorldFlags flags = useTempMem;
#endif
    
    if (spIsSubClass(component, SpImage) == SP_TRUE
	&& !strnone(SpComponentPart(component).description)) return;
    
    spLockControlMutexMac();
    spLockDrawMutexMac();
    spLockPortMutexMac();

    if (SpPrimitiveArch(component).gworld != NULL) {
#if defined(SP_GWORLD_PIXMAP_ALWAYS_LOCK)
	if ((pixmap = GetGWorldPixMap(SpPrimitiveArch(component).gworld)) != nil) {
	    UnlockPixels(pixmap);
	}
#if TARGET_API_MAC_CARBON
	UnlockPortBits(SpPrimitiveArch(component).gworld);
#endif
#endif
    }

    spDebug(10, "spRedrawImageArch", "client size = %d, %d\n",
	    SpComponentPart(component).client_width, SpComponentPart(component).client_height);
    
    SetRect(&rect, 0, 0, SpComponentPart(component).client_width,
	    SpComponentPart(component).client_height);

    if (SpPrimitiveArch(component).gworld != NULL) {
	if (UpdateGWorld(&SpPrimitiveArch(component).gworld,
			 0, &rect, NULL, NULL, clipPix) == gwFlagErr) {
	    spDebug(10, "spRedrawImageArch", "UpdateGWorld error\n");
	}
    } else {
	if (noErr != NewGWorld(&SpPrimitiveArch(component).gworld,
			       0, &rect, NULL, NULL, flags)) {
	    spDebug(10, "spRedrawImageArch", "NewGWorld error\n");
	    SpPrimitiveArch(component).gworld = NULL;
	}
    }
    
#if defined(SP_GWORLD_PIXMAP_ALWAYS_LOCK)
    if (SpPrimitiveArch(component).gworld != NULL) {
#if TARGET_API_MAC_CARBON
	LockPortBits(SpPrimitiveArch(component).gworld);
#endif
	if ((pixmap = GetGWorldPixMap(SpPrimitiveArch(component).gworld)) != nil) {
	    LockPixels(pixmap);
	}
    }
#endif
    
    spUnlockPortMutexMac();
    spUnlockDrawMutexMac();
    spUnlockControlMutexMac();
    
    return;
}

