/*
 *	spList_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spText.h>

#include <sp/spComboBoxP.h>
#include <sp/spListP.h>

#if GTK_CHECK_VERSION(3,0,0)
static void listChangedCB(GtkTreeSelection *selection, gpointer user_data)
{
    spDebug(50, "listChangedCB", "in\n");
    
    return;
}
#endif

void spListCreateArch(spComponent component)
{
    SpPrimitiveArch(component).top_widget = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (SpPrimitiveArch(component).top_widget),
				    GTK_POLICY_AUTOMATIC, 
				    GTK_POLICY_AUTOMATIC);
    spContainerAddGtk(component, SpPrimitiveArch(component).top_widget);
    if (SpComponentPart(component).visible_flag == SP_TRUE) {
	gtk_widget_show(SpPrimitiveArch(component).top_widget);
    }

#if GTK_CHECK_VERSION(3,0,0)
    SpListArch(component).model = GTK_TREE_MODEL(gtk_list_store_new(1, G_TYPE_STRING));
    SpPrimitiveArch(component).widget = gtk_tree_view_new_with_model(SpListArch(component).model);
    gtk_tree_view_set_rules_hint(GTK_TREE_VIEW(SpPrimitiveArch(component).widget), TRUE);
    gtk_tree_view_set_search_column(GTK_TREE_VIEW(SpPrimitiveArch(component).widget), 0);
    gtk_tree_view_set_headers_visible(GTK_TREE_VIEW(SpPrimitiveArch(component).widget), FALSE);
    
    {
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	GtkTreeSelection *selection;

	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes(NULL, renderer, "text", 0, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(SpPrimitiveArch(component).widget), column);
	
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(SpPrimitiveArch(component).widget));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);
	g_signal_connect(selection, "changed", G_CALLBACK(listChangedCB), component);
    }
#else
    SpPrimitiveArch(component).widget = gtk_list_new();
    gtk_list_set_selection_mode(GTK_LIST(SpPrimitiveArch(component).widget), GTK_SELECTION_BROWSE);
    gtk_widget_set_events(SpPrimitiveArch(component).widget, GDK_BUTTON_PRESS_MASK);
#endif
    
#if GTK_CHECK_VERSION(1,2,0) && !GTK_CHECK_VERSION(3,0,0)
    gtk_scrolled_window_add_with_viewport(GTK_SCROLLED_WINDOW(SpPrimitiveArch(component).top_widget),
		      SpPrimitiveArch(component).widget);
#else
    gtk_container_add(GTK_CONTAINER(SpPrimitiveArch(component).top_widget),
		      SpPrimitiveArch(component).widget);
#endif
    gtk_widget_show(SpPrimitiveArch(component).widget);
    
    return;
}

void spListDestroyArch(spComponent component)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpListArch(component).model != NULL) {
	g_object_unref(SpListArch(component).model);
    }
#endif
    
    return;
}

void spListSetParamsArch(spComponent component)
{
    return;
}

void spListGetParamsArch(spComponent component)
{
    return;
}

#if GTK_CHECK_VERSION(3,0,0)
static GtkTreeModel *getListModelGtk(spComponent component)
{
    GtkTreeModel *model;

    if (spIsComboBox(component) == SP_TRUE) {
	model = SpComboBoxArch(component).model;
    } else {
	model = SpListArch(component).model;
    }
    
    return model;
}

static int findModelIndexGtk(GtkTreeModel *model, const char *item)
{
    int i;
    int index = -1;
    GtkTreeIter iter;
    gchar *gstr;
    
    for (i = 0;; i++) {
	if (!gtk_tree_model_iter_nth_child(model, &iter, NULL, i)) {
	    break;
	}
    
	gtk_tree_model_get(model, &iter, 0, &gstr, -1);

	if (streq(item, gstr)) {
	    index = i;
	    g_free(gstr);
	    break;
	}
	
	g_free(gstr);
    }

    return index;
}

int spAddListIndexArch(spComponent component, const char *item, int index)
{
    GtkTreeModel *model;
    GtkListStore *store;
    GtkTreeIter iter;
    
    spDebug(50, "spAddListIndexArch", "in: index = %d, item = %s\n", index, item);
    
    model = getListModelGtk(component);
    
    store = GTK_LIST_STORE(model);

    if (index < 0) {
	gtk_list_store_append(store, &iter);
    } else {
	gtk_list_store_insert(store, &iter, index);
    }
    gtk_list_store_set(store, &iter, 0, item, -1);

    {
	spDebug(50, "spAddListIndexArch", "n_children = %d\n", 
		gtk_tree_model_iter_n_children(model, NULL));
    }
    
    spDebug(50, "spAddListIndexArch", "done: index = %d\n", index);
    
    return index;
}

int spAddListItemArch(spComponent component, const char *item)
{
    spDebug(80, "spAddListItemArch", "item = %s\n", item);
    
    return spAddListIndexArch(component, item, -1);
}

int spDeleteListItemArch(spComponent component, const char *item)
{
    int index;

    if ((index = spFindListItemArch(component, item)) < 0) {
	return -1;
    }

    return spDeleteListIndexArch(component, index);
}

int spDeleteListIndexArch(spComponent component, int index)
{
    GtkTreeModel *model;
    GtkTreeIter iter;

    model = getListModelGtk(component);
    if (index < 0) {
	index = gtk_tree_model_iter_n_children(model, NULL) - 1;
    }
    
    if (!gtk_tree_model_iter_nth_child(model, &iter, NULL, index)) {
	return -1;
    }
    
    if (gtk_list_store_remove(GTK_LIST_STORE(model), &iter)) {
	return index;
    } else {
	return -1;
    }
}

spBool spClearListArch(spComponent component)
{
    GtkTreeModel *model;
    
    model = getListModelGtk(component);
    
    gtk_list_store_clear(GTK_LIST_STORE(model));
    
    return SP_TRUE;
}

int spFindListItemArch(spComponent component, const char *item)
{
    GtkTreeModel *model;

    model = getListModelGtk(component);
    
    return findModelIndexGtk(model, item);
}

int spSelectListItemArch(spComponent component, const char *item)
{
    int index;

    if ((index = spFindListItemArch(component, item)) < 0) {
	return -1;
    }
    
    spDebug(1, "spSelectListItemArch", "item = %s, index = %d\n", item, index);

    return spSelectListIndexArch(component, index);
}

int spSelectListIndexArch(spComponent component, int index)
{
    if (spIsComboBox(component) == SP_TRUE) {
	gtk_combo_box_set_active(GTK_COMBO_BOX(SpPrimitiveArch(component).widget), index);
    } else {
	GtkTreeModel *model;
	GtkTreeIter iter;
	GtkTreeSelection *selection;
    
	model = getListModelGtk(component);
	if (index < 0) {
	    index = gtk_tree_model_iter_n_children(model, NULL) - 1;
	}
    
	if (!gtk_tree_model_iter_nth_child(model, &iter, NULL, index)) {
	    return -1;
	}

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(SpPrimitiveArch(component).widget));
	gtk_tree_selection_select_iter(selection, &iter);
    }
    
    return index;
}

char *xspGetSelectedListItemArch(spComponent component)
{
    int index;

    if ((index = spGetSelectedListIndexArch(component)) < 0) {
	return NULL;
    }

    return xspGetListItemArch(component, index);
}

int spGetSelectedListIndexArch(spComponent component)
{
    int index;
    
    if (spIsComboBox(component) == SP_TRUE) {
	index = gtk_combo_box_get_active(GTK_COMBO_BOX(SpPrimitiveArch(component).widget));
    } else {
	GtkTreeModel *model;
	GtkTreeSelection *selection;
	GtkTreeIter iter;
	gchar *gstr;
    
	model = getListModelGtk(component);
    
	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(SpPrimitiveArch(component).widget));

	if (!gtk_tree_selection_get_selected(selection, &model, &iter)) {
	    return -1;
	}

	gtk_tree_model_get(model, &iter, 0, &gstr, -1);
	index = findModelIndexGtk(model, gstr);
	g_free(gstr);
    }

    return index; 
}

char *xspGetListItemArch(spComponent component, int index)
{
    GtkTreeModel *model;
    GtkTreeIter iter;
    gchar *gstr;
    char *item;
    
    model = getListModelGtk(component);
    if (index < 0) {
	index = gtk_tree_model_iter_n_children(model, NULL) - 1;
    }
    
    if (!gtk_tree_model_iter_nth_child(model, &iter, NULL, index)) {
	return NULL;
    }
    
    gtk_tree_model_get(model, &iter, 0, &gstr, -1);
    item = strclone(gstr);
    g_free(gstr);
    
    return item;
}

int spGetNumListItemArch(spComponent component)
{
    GtkTreeModel *model;
	
    model = getListModelGtk(component);

    return gtk_tree_model_iter_n_children(model, NULL);
}
#else
static char *getListString(GtkWidget *list_item)
{
    GList *children;
    GtkWidget *label;
    char *string = NULL;

    children = gtk_container_children(GTK_CONTAINER(list_item));
    label = (GtkWidget *)g_list_nth_data(children, 0);
    gtk_label_get(GTK_LABEL(label), &string);
    
    return string;
}

int spAddListIndexArch(spComponent component, const char *item, int index)
{
    GtkWidget *widget;
    GtkWidget *list_item;
    GList *dlist = NULL;

    if (spIsComboBox(component) == SP_TRUE) {
	widget = GTK_COMBO(SpPrimitiveArch(component).widget)->list;
    } else {
	widget = SpPrimitiveArch(component).widget;
    }

    list_item = gtk_list_item_new_with_label(item);
    dlist = g_list_prepend(dlist, list_item);
    gtk_widget_show(list_item);

    if (index < 0) {
	index = g_list_length(GTK_LIST(widget)->children);
	gtk_list_append_items(GTK_LIST(widget), dlist);
    } else {
	gtk_list_insert_items(GTK_LIST(widget), dlist, index);
    }

    return index;
}

int spAddListItemArch(spComponent component, const char *item)
{
    return spAddListIndexArch(component, item, -1);
}

static GtkWidget *findListItem(GtkWidget *widget, const char *item, int *index)
{
    int i;
    char *string;
    GtkWidget *list_item;
    GList *children;

    children = GTK_LIST(widget)->children;

    for (i = 0; children != NULL; i++) {
	if (children->data == NULL) {
	    break;
	}
	
	list_item = GTK_WIDGET(children->data);
	
	if ((string = getListString(list_item)) != NULL) {
	    spDebug(10, "findListItem", "index = %d, item = %s, string = %s\n",
		    i, item, string);
	    
	    if (streq(item, string)) {
		if (index != NULL) *index = i;
	    
		return list_item;
	    }
	}

	children = children->next;
    }

    return NULL;
}

static int findListIndex(GtkWidget *widget, const char *item)
{
    int index;

    if (findListItem(widget, item, &index) != NULL) {
	return index;
    } else {
	return -1;
    }
}

static GtkWidget *getListItem(GtkWidget *widget, int index)
{
    int i;
    GtkWidget *list_item;
    GList *children;

    children = GTK_LIST(widget)->children;

    for (i = 0; children != NULL; i++) {
	if (children->data == NULL) {
	    break;
	}
	
	if (i == index) {
	    list_item = GTK_WIDGET(children->data);
	    return list_item;
	}

	children = children->next;
    }

    return NULL;
}

int spDeleteListItemArch(spComponent component, const char *item)
{
    int index = -1;
    GtkWidget *widget;
    GtkWidget *list_item;
    GList *dlist = NULL;

    if (spIsComboBox(component) == SP_TRUE) {
	widget = GTK_COMBO(SpPrimitiveArch(component).widget)->list;
    } else {
	widget = SpPrimitiveArch(component).widget;
    }

    if ((list_item = findListItem(widget, item, &index)) != NULL) {
	dlist = g_list_prepend(dlist, list_item);
	gtk_list_remove_items(GTK_LIST(widget), dlist);
	g_list_free(dlist);
    }

    return index;
}

int spDeleteListIndexArch(spComponent component, int index)
{
    GtkWidget *widget;
    GtkWidget *list_item;
    GList *dlist = NULL;

    if (spIsComboBox(component) == SP_TRUE) {
	widget = GTK_COMBO(SpPrimitiveArch(component).widget)->list;
    } else {
	widget = SpPrimitiveArch(component).widget;
    }

    if ((list_item = getListItem(widget, index)) != NULL) {
	dlist = g_list_prepend(dlist, list_item);
	gtk_list_remove_items(GTK_LIST(widget), dlist);
	g_list_free(dlist);

	return index;
    }

    return -1;
}

spBool spClearListArch(spComponent component)
{
    return spClearListDefault(component);
}

int spFindListItemArch(spComponent component, const char *item)
{
    GtkWidget *widget;

    if (spIsComboBox(component) == SP_TRUE) {
	widget = GTK_COMBO(SpPrimitiveArch(component).widget)->list;
    } else {
	widget = SpPrimitiveArch(component).widget;
    }

    return findListIndex(widget, item);
}

int spSelectListItemArch(spComponent component, const char *item)
{
    int index = -1;
    GtkWidget *widget;

    if (spIsComboBox(component) == SP_TRUE) {
	widget = GTK_COMBO(SpPrimitiveArch(component).widget)->list;
    } else {
	widget = SpPrimitiveArch(component).widget;
    }

    if ((index = findListIndex(widget, item)) >= 0) {
	gtk_list_select_item(GTK_LIST(widget), index);
    }
    spDebug(1, "spSelectListItemArch", "item = %s, index = %d\n", item, index);

    return index;
}

int spSelectListIndexArch(spComponent component, int index)
{
    GtkWidget *widget;

    if (spIsComboBox(component) == SP_TRUE) {
	widget = GTK_COMBO(SpPrimitiveArch(component).widget)->list;
    } else {
	widget = SpPrimitiveArch(component).widget;
    }

    gtk_list_select_item(GTK_LIST(widget), index);
    
    return index;
}

char *xspGetSelectedListItemArch(spComponent component)
{
    char *item = NULL;
    GtkWidget *widget;
    GtkWidget *list_item;
    GList *selection;
    
    if (spIsComboBox(component) == SP_TRUE) {
	widget = GTK_COMBO(SpPrimitiveArch(component).widget)->list;
    } else {
	widget = SpPrimitiveArch(component).widget;
    }

    selection = GTK_LIST(widget)->selection;

    if (selection != NULL) {
	list_item = GTK_WIDGET(selection->data);
	spDebug(50, "xspGetSelectedListItemArch", "selection found\n");

	if ((item = getListString(list_item)) != NULL) {
	    return strclone(item);
	}
    }
    
    return NULL; 
}

int spGetSelectedListIndexArch(spComponent component)
{
    int i;
    GtkWidget *widget;
    GtkWidget *list_item;
    GList *selection;
    GList *children;

    if (spIsComboBox(component) == SP_TRUE) {
	widget = GTK_COMBO(SpPrimitiveArch(component).widget)->list;
    } else {
	widget = SpPrimitiveArch(component).widget;
    }

    selection = GTK_LIST(widget)->selection;

    if (selection != NULL) {
	list_item = GTK_WIDGET(selection->data);
	spDebug(50, "spGetSelectedListIndexArch", "selection found\n");

	children = GTK_LIST(widget)->children;

	for (i = 0; children != NULL; i++) {
	    if (children->data == NULL) {
		break;
	    }
	
	    if (list_item == GTK_WIDGET(children->data)) {
		return i;
	    }
	    children = children->next;
	}
    }

    return -1; 
}

char *xspGetListItemArch(spComponent component, int index)
{
    char *item = NULL;
    GtkWidget *widget;
    GtkWidget *list_item;
    
    if (spIsComboBox(component) == SP_TRUE) {
	widget = GTK_COMBO(SpPrimitiveArch(component).widget)->list;
    } else {
	widget = SpPrimitiveArch(component).widget;
    }
    
    if ((list_item = getListItem(widget, index)) != NULL) {
	if ((item = getListString(list_item)) != NULL) {
	    return strclone(item);
	}
    }
    
    return NULL;
}

int spGetNumListItemArch(spComponent component)
{
    int i;
    GtkWidget *widget;
    GList *children;

    spDebug(80, "spGetNumListItemArch", "in\n");
    
    if (spIsComboBox(component) == SP_TRUE) {
	widget = GTK_COMBO(SpPrimitiveArch(component).widget)->list;
    } else {
	widget = SpPrimitiveArch(component).widget;
    }
    
    children = GTK_LIST(widget)->children;

    for (i = 0;; i++) {
	if (children == NULL || children->data == NULL) {
	    break;
	}
	children = children->next;
    }
	
    spDebug(80, "spGetNumListItemArch", "done: i = %d\n", i);
    
    return i;
}
#endif
