/*
 *	spParamField.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spComponent.h>
#include <sp/spFrame.h>
#include <sp/spContainer.h>
#include <sp/spButton.h>
#include <sp/spDialog.h>
#include <sp/spSlider.h>
#include <sp/spProgressBar.h>

#include <sp/spParamFieldP.h>

#if defined(MACOS)
#pragma import on
#endif

extern spTopLevel sp_toplevel;

#if defined(MACOS)
#pragma import off
#endif

static spParamTable sp_param_field_param_tables[] = {
    {SppFieldType, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.field_type), SP_FIELD_TYPE_TEXT_STRING},
    {SppSet, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.set), SP_TRUE_STRING},
    {SppBorderOn, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.border_on), SP_FALSE_STRING},
    {SppSize, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.size), "0"},
    {SppEditable, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.editable), SP_TRUE_STRING},
    {SppUseCheckBox, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.use_check_box), SP_FALSE_STRING},
    {SppUseDimensionButton, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.use_dim_button), SP_FALSE_STRING},
    {SppFieldOffset, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.field_offset), "0"},
    {SppFieldSize, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.field_size), "0"},
    {SppDimensionSize, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.dim_size), "0"},
    {SppFieldString, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.string), NULL},
    {SppFieldStrings, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.strings), NULL},
    {SppFieldSelectedIndex, SP_CREATE_ACCESS/* | SP_SET_ACCESS | SP_GET_ACCESS*/,
	 spOffset(spParamField, param_field.index), "-1"},
    {SppDimension, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.dim_string), NULL},

    {SppShowScale, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.show_scale), SP_FALSE_STRING},
    {SppShowValue, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.show_value), SP_FALSE_STRING},
    {SppTrackCallbackOn, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.track_call_on), SP_FALSE_STRING},
    {SppValue, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.value), "0"},
    {SppMinimum, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.minimum), "0"},
    {SppMaximum, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.maximum), "100"},
    {SppPageSize, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.page_size), "100"},
    {SppIncrement, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.increment), "10"},
    {SppPageIncrement, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.page_increment), "100"},
    {SppDecimalPoints, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.decimal_points), "0"},
    
    {SppFileFilters, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.file_filters), NULL},
    {SppFileTypes, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.file_types), NULL},
    {SppFileFilterIndex, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.filter_index), NULL},
    {SppInitialDir, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.initial_dir), NULL},
    {SppInitialFont, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.initial_font), NULL},
    {SppInitialColor, SP_CREATE_ACCESS | SP_SET_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.initial_color), NULL},
	
    {SppDisableShortcut, SP_CREATE_ACCESS /*| SP_SET_ACCESS*/ | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.disable_shortcut), SP_FALSE_STRING},

    {SppSelectOnly, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.select_only), SP_FALSE_STRING},
    
    {SppKeyboardType, SP_CREATE_ACCESS | SP_GET_ACCESS,
	 spOffset(spParamField, param_field.keyboard_type), SP_KEYBOARD_TYPE_DEFAULT_STRING},
};

spParamFieldClassRec SpParamFieldClassRec = {
    /* spObjectClassPart */
    {
	SpParamField,
	(spObjectClass)&SpComponentClassRec,
	sizeof(spParamFieldRec),
	spArraySize(sp_param_field_param_tables),
	sp_param_field_param_tables,
	spParamFieldPartInit,
	spParamFieldPartFree,
	SP_FALSE,
	NULL,
	NULL,
	spParamFieldCreate,
	spParamFieldDestroy,
	spParamFieldSetParams,
	spParamFieldGetParams,
    },
    /* spComponentClassPart */
    {
	SP_FALSE,
	SP_FALSE,
	SP_TRUE,
	spParamFieldMap,
	spParamFieldUnmap,
	spParamFieldSetSize,
	spParamFieldGetSize,
	spParamFieldGetClientSize,
	spParamFieldSetSensitive,
	spParamFieldIsComponentType,

	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
    },
    /* spParamFieldClassPart */
    {
	0,
    },
};

spComponentClass SpParamFieldClass = (spComponentClass)&SpParamFieldClassRec;

void spParamFieldPartInit(spObject object)
{
    spComponent component = (spComponent)object;
    
    SpParamFieldPart(component).box = NULL;
    SpParamFieldPart(component).title_label = NULL;
    SpParamFieldPart(component).field = NULL;
    SpParamFieldPart(component).sub_field = NULL;
    SpParamFieldPart(component).dim_label = NULL;

    SpComponentPart(component).orientation = SP_HORIZONTAL;
    SpComponentPart(component).margin_width = SP_DEFAULT_MARGIN;
    SpComponentPart(component).margin_height = SP_DEFAULT_MARGIN;
    SpComponentPart(component).spacing = SP_DEFAULT_SPACING;
    SpComponentPart(component).spacing_flag = SP_TRUE;

    return;
}

void spParamFieldPartFree(spObject object)
{
    return;
}

static void setSensitiveByCheckBox(spComponent param_field)
{
    spSetSensitive(SpParamFieldPart(param_field).field,
		   SpParamFieldPart(param_field).set);
    if (SpParamFieldPart(param_field).dim_label != NULL) {
	spSetSensitive(SpParamFieldPart(param_field).dim_label,
		       SpParamFieldPart(param_field).set);
    }
    if (SpParamFieldPart(param_field).sub_field != NULL) {
	spSetSensitive(SpParamFieldPart(param_field).sub_field,
		       SpParamFieldPart(param_field).set);
    }
    return;
}

static void checkBoxCB(spComponent component, spComponent param_field)
{
    if (spIsCreated(param_field) == SP_FALSE) return;
    
    if (spGetToggleState(component, &SpParamFieldPart(param_field).set) == SP_TRUE) {
	setSensitiveByCheckBox(param_field);
    }
    
    return;
}

static void pushButtonCB(spComponent component, spComponent param_field)
{
    char *string = NULL;
    char initial_string[SP_MAX_PATHNAME];
    
    if (spIsCreated(param_field) == SP_FALSE) return;

    if ((SpParamFieldPart(param_field).use_dim_button == SP_TRUE
	 && !strnone(SpParamFieldPart(param_field).dim_string))) {
	if (SpComponentPart(param_field).call_func != NULL) {
	    SpComponentPart(param_field).call_reason = SP_CR_ACTIVATE;
	    SpComponentPart(param_field).call_func(param_field,
						   SpComponentPart(param_field).call_data);
	    SpComponentPart(param_field).call_reason = SP_CR_NONE;
	}
	return;
    }

    strcpy(initial_string, "");
	
    if (SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_FONT_TEXT) {
	string = SpParamFieldPart(param_field).initial_font;
    } else if (SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_COLOR_TEXT) {
	string = SpParamFieldPart(param_field).initial_color;
    } else {
	string = SpParamFieldPart(param_field).initial_dir;
    }
	
    if (!strnone(string)) {
	spStrCopy(initial_string, SP_MAX_PATHNAME, string);
    } else {
	if (1 || strnone(SpParamFieldPart(param_field).string)) {
	    if (SpParamFieldPart(param_field).string != NULL)
		xfree(SpParamFieldPart(param_field).string);

	    SpParamFieldPart(param_field).string
		= xspGetTextString(SpParamFieldPart(param_field).field);
	    spStrCopy(initial_string, sizeof(initial_string), SpParamFieldPart(param_field).string);
	}
	
	if (SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_OPEN_FILE_TEXT
	    || SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_SAVE_FILE_TEXT) {
	    if (spIsDir(initial_string) == SP_FALSE) {
		spGetDirName(initial_string);
	    }
	}
    }

    if (SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_OPEN_FILE_TEXT) {
	string = xspGetOpenFileName(component, NULL,
				    SppPathMustExist, SP_TRUE,
				    SppFileMustExist, SP_TRUE,
				    SppFileFilters, SpParamFieldPart(param_field).file_filters,
				    SppFileTypes, SpParamFieldPart(param_field).file_types,
				    SppFileFilterIndex, SpParamFieldPart(param_field).filter_index,
				    SppInitialDir, initial_string,
				    NULL);
    } else if (SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_SAVE_FILE_TEXT) {
	string = xspGetSaveFileName(component, NULL,
				    SppPathMustExist, SP_TRUE,
				    SppFileMustExist, SP_FALSE,
				    SppOverwritePrompt, SP_TRUE,
				    SppFileFilters, SpParamFieldPart(param_field).file_filters,
				    SppFileTypes, SpParamFieldPart(param_field).file_types,
				    SppFileFilterIndex, SpParamFieldPart(param_field).filter_index,
				    SppInitialDir, initial_string,
				    SppInitialFileName, SpParamFieldPart(param_field).string,
				    NULL);
    } else if (SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_FONT_TEXT) {
	string = xspChooseFont(component, NULL,
			       SppInitialFont, initial_string,
			       NULL);
    } else if (SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_COLOR_TEXT) {
	string = xspChooseColor(component, NULL,
				SppInitialColor, initial_string,
				NULL);
    } else {
	string = xspGetSelectedDirName(component, NULL,
				       SppPathMustExist, SP_TRUE,
				       SppInitialDir, initial_string,
				       NULL);
    }

    if (string != NULL) {
	if (SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_FONT_TEXT
	    || SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_COLOR_TEXT) {
	    spSetTextString(SpParamFieldPart(param_field).field, string);
	} else {
	    char *readable;
	    if ((readable = xspGetReadablePath(string)) != NULL) {
		spSetTextString(SpParamFieldPart(param_field).field, readable);
		xfree(readable);
	    }
	}
	xfree(string);
	
	if (SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_FONT_TEXT) {
	    xfree(SpParamFieldPart(param_field).initial_font);
	    SpParamFieldPart(param_field).initial_font = NULL;
	} else if (SpParamFieldPart(param_field).field_type == SP_FIELD_TYPE_COLOR_TEXT) {
	    xfree(SpParamFieldPart(param_field).initial_color);
	    SpParamFieldPart(param_field).initial_color = NULL;
	} else {
	    xfree(SpParamFieldPart(param_field).initial_dir);
	    SpParamFieldPart(param_field).initial_dir = NULL;
	}
    }

    return;
}

static void getTrackBarTextString(spComponent param_field, int value, char *buf)
{
    double valued;
    
    if (SpParamFieldPart(param_field).decimal_points > 0) {
	char format[SP_MAX_LINE];
    
	sprintf(format, "%%.%df", SpParamFieldPart(param_field).decimal_points);
	valued = (double)value / pow(10.0, (double)SpParamFieldPart(param_field).decimal_points);
	sprintf(buf, format, valued);
    } else {
	sprintf(buf, "%d", value);
    }
    spDebug(80, "getTrackBarTextString", "buf = %s\n", buf);
    
    return;
}

static int trackBarTextToValue(const char *string, int decimal_points)
{
    int value;
    double valued;
    
    valued = atof(string) * pow(10.0, (double)decimal_points);
    value = (int)spRound(valued);

    return value;
}

static void trackBarTextCB(spComponent component, spComponent param_field)
{
    int value;
    char *string;

    if ((string = xspGetTextString(component)) != NULL) {
	spDebug(80, "trackBarTextCB", "string = %s\n", string);
	
	value = trackBarTextToValue(string, SpParamFieldPart(param_field).decimal_points);
	spDebug(80, "trackBarTextCB", "value = %d, SpParamFieldPart(param_field).value = %d\n",
		value, SpParamFieldPart(param_field).value);
	
	if (value != SpParamFieldPart(param_field).value) {
	    spSetSliderValue(SpParamFieldPart(param_field).field, value);
	    SpParamFieldPart(param_field).value = value;
	
	    if (SpComponentPart(param_field).call_func != NULL) {
		SpComponentPart(param_field).call_reason = SP_CR_VALUE_CHANGED;
		SpComponentPart(param_field).call_func(param_field,
						       SpComponentPart(param_field).call_data);
		SpComponentPart(param_field).call_reason = SP_CR_NONE;
	    }
	}
	xfree(string);
    }
    
    return;
}

static void trackBarCBForText(spComponent component, void *data)
{
    int value;
    char buf[SP_MAX_LINE];
    spComponent param_field = (spComponent)data;

    spDebug(80, "trackBarCBForText", "in\n");
    
    if (spGetSliderValue(component, &value) == SP_TRUE) {
	spDebug(80, "trackBarCBForText", "value = %d\n", value);
	getTrackBarTextString(param_field, value, buf);
	spSetTextString(SpParamFieldPart(param_field).sub_field, buf);
    }
    
    spDebug(80, "trackBarCBForText", "done\n");
    
    return;
}

spBool spParamFieldCreate(spObject object)
{
    const char *string;
    char *initial_string;
    int default_height;
    int default_text_height;
    int field_size;
    char track_bar_string[SP_MAX_LINE];
    spComponent component = (spComponent)object;

    if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR) {
	default_height = /*SP_DEFAULT_HORIZONTAL_TRACKBAR_WIDTH*/
	    spGetTrackBarDefaultWidth(SP_HORIZONTAL, SpParamFieldPart(component).show_value);
#if !defined(IPHONE)
    } else if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_PROGRESS_BAR) {
	default_height = SP_DEFAULT_PROGRESS_BAR_HEIGHT;
#endif
    } else {
	default_height = MAX(SP_DEFAULT_TEXT_HEIGHT, SP_DEFAULT_BUTTON_HEIGHT);
    }
    default_text_height = SP_DEFAULT_TEXT_HEIGHT;
    spDebug(80, "spParamFieldCreate", "default_height = %d\n", default_height);
    
    SpComponentPart(component).x = 0;
    SpComponentPart(component).y = 0;
    if (SpParentComponentPart(component).orientation == SP_HORIZONTAL) {
	SpComponentPart(component).width = SpParamFieldPart(component).size;
	SpComponentPart(component).height = 0;
    } else {
	SpComponentPart(component).width = 0;
	if (SpParamFieldPart(component).size == 0) {
	    SpParamFieldPart(component).size = default_height;
	    if (SpComponentPart(component).spacing_flag == SP_TRUE) {
		SpParamFieldPart(component).size += (2 * SpComponentPart(component).margin_height);
	    }
	}
	SpComponentPart(component).height = SpParamFieldPart(component).size;
    }
    SpComponentPart(component).current_width = SpComponentPart(component).width;
    SpComponentPart(component).current_height = SpComponentPart(component).height;

    /* create main box */
    SpParamFieldPart(component).box =
	spCreateBox(SpGetParent(component), /*"_ParamField_Box"*/SpGetName(component),
		    SpParamFieldPart(component).size,
		    SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
		    SppOrientation, SP_HORIZONTAL,
		    SppMarginWidth, SpComponentPart(component).margin_width,
		    SppMarginHeight, SpComponentPart(component).margin_height,
		    SppSpacing, SpComponentPart(component).spacing,
		    SppBorderOn, SpParamFieldPart(component).border_on,
		    NULL);

    /* create title label */
    string = (strnone(SpComponentPart(component).title) ?
	      SpGetName(component) : SpComponentPart(component).title);
    spDebug(80, "spParamFieldCreate", "string = %s\n", string);
    if (SpParamFieldPart(component).use_check_box == SP_TRUE) {
	SpParamFieldPart(component).title_label =
	    spCreateCheckBox(SpParamFieldPart(component).box, "_ParamField_TitleLabel",
			     SppTitle, string,
			     SppSet, SpParamFieldPart(component).set,
			     SppWidth, SpParamFieldPart(component).field_offset,
			     SppHeight, default_height,
			     SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
			     SppSpacingOn, SpComponentPart(component).spacing_flag,
			     SppCallbackFunc, checkBoxCB,
			     SppCallbackData, component,
			     NULL);
    } else {
	SpParamFieldPart(component).title_label =
	    spCreateLabel(SpParamFieldPart(component).box, "_ParamField_TitleLabel",
			  SppTitle, string,
			  SppAlignment, SP_ALIGNMENT_BEGINNING,
			  SppWidth, SpParamFieldPart(component).field_offset,
			  SppHeight, default_height,
			  SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
			  SppSpacingOn, SpComponentPart(component).spacing_flag,
			  NULL);
    }
    spDebug(80, "spParamFieldCreate", "create label done\n");
	
    /* create field */
    if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_COMBO_BOX) {
	spDebug(80, "spParamFieldCreate", "combo: field_size = %d\n",
		SpParamFieldPart(component).field_size);
	SpParamFieldPart(component).field =
	    spCreateComboBox(SpParamFieldPart(component).box, "_ParamField_ComboBox",
			     SppWidth, SpParamFieldPart(component).field_size,
			     SppHeight, default_text_height,
			     SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
			     SppSpacingOn, SpComponentPart(component).spacing_flag,
			     SppEditable, SpParamFieldPart(component).editable,
			     SppDisableShortcut, SpParamFieldPart(component).disable_shortcut,
			     SppTextString, SpParamFieldPart(component).string,
			     SppListStrings, SpParamFieldPart(component).strings,
			     SppSelectedListIndex, SpParamFieldPart(component).index,
			     SppSelectOnly, SpParamFieldPart(component).select_only,
                             SppKeyboardType, SpParamFieldPart(component).keyboard_type,
			     SppDescription, SpComponentPart(component).description,
			     SppCallbackFunc, SpComponentPart(component).call_func,
			     SppCallbackData, SpComponentPart(component).call_data,
			     SppUserData, SpComponentPart(component).user_data,
			     SppHelpPath, SpComponentPart(component).help_path,
			     NULL);
    } else if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR
	       || SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_TEXT
	       || SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_COMBO_BOX) {
	int text_width = 0;
	spBool show_value;
	
	if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_TEXT
	    || SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_COMBO_BOX) {
	    field_size = SpParamFieldPart(component).field_size;
	    if (field_size < 0 && SpParamFieldPart(component).dim_size == 0) {
		text_width = -field_size;
		if (SpComponentPart(component).spacing_flag) {
		    text_width -= 2 * SpComponentPart(component).spacing;
		}
	    } else {
		text_width = SpParamFieldPart(component).dim_size;
                if (text_width == 0) {
                    text_width = (2 * SP_DEFAULT_BUTTON_WIDTH) / 3;
                }
	    }
	    show_value = SP_FALSE;
	    spDebug(80, "spParamFieldCreate", "track bar with text: field_size = %d, text_width = %d\n",
		    field_size, text_width);

	    if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_COMBO_BOX) {
		if (SpParamFieldPart(component).strings != NULL
		    && SpParamFieldPart(component).index >= 0) {
		    SpParamFieldPart(component).value
			= trackBarTextToValue(SpParamFieldPart(component).strings[SpParamFieldPart(component).index],
					      SpParamFieldPart(component).decimal_points);
		    spSetSliderValue(SpParamFieldPart(component).field, SpParamFieldPart(component).value);
		}
	    }
	    getTrackBarTextString(component, SpParamFieldPart(component).value, track_bar_string);
	} else {
	    field_size = SpParamFieldPart(component).field_size;
	    show_value = SpParamFieldPart(component).show_value;
	}
	spDebug(80, "spParamFieldCreate", "track bar: field_size = %d, value = %d\n",
		field_size, SpParamFieldPart(component).value);
	SpParamFieldPart(component).field =
	    spCreateTrackBar(SpParamFieldPart(component).box, "_ParamField_TrackBar",
			     SppWidth, field_size,
			     SppHeight, default_height,
			     SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
			     SppSpacingOn, SpComponentPart(component).spacing_flag,
			     SppEditable, SpParamFieldPart(component).editable,
			     SppShowScale, SpParamFieldPart(component).show_scale,
			     SppShowValue, show_value,
			     SppTrackCallbackOn, SpParamFieldPart(component).track_call_on,
			     SppValue, SpParamFieldPart(component).value,
			     SppMinimum, SpParamFieldPart(component).minimum,
			     SppMaximum, SpParamFieldPart(component).maximum,
			     SppPageSize, SpParamFieldPart(component).page_size,
			     SppIncrement, SpParamFieldPart(component).increment,
			     SppPageIncrement, SpParamFieldPart(component).page_increment,
			     SppDecimalPoints, SpParamFieldPart(component).decimal_points,
			     SppDescription, SpComponentPart(component).description,
			     SppCallbackFunc, SpComponentPart(component).call_func,
			     SppCallbackData, SpComponentPart(component).call_data,
			     SppUserData, SpComponentPart(component).user_data,
			     SppHelpPath, SpComponentPart(component).help_path,
			     NULL);
	
	if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_TEXT
	    || SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_COMBO_BOX) {
	    if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_COMBO_BOX) {
		SpParamFieldPart(component).sub_field =
		    spCreateComboBox(SpParamFieldPart(component).box, "_ParamField_TrackBarComboBox",
				     SppWidth, text_width,
				     SppHeight, default_text_height,
				     SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
				     SppSpacingOn, SpComponentPart(component).spacing_flag,
				     SppDisableShortcut, SpParamFieldPart(component).disable_shortcut,
                                     SppKeyboardType, SP_KEYBOARD_TYPE_NUMBER_PAD,
				     SppTextString, track_bar_string,
				     SppDescription, SpComponentPart(component).description,
				     SppHelpPath, SpComponentPart(component).help_path,
				     SppCallbackFunc, trackBarTextCB,
				     SppCallbackData, component,
				     SppListStrings, SpParamFieldPart(component).strings,
				     SppSelectedListIndex, SpParamFieldPart(component).index,
				     NULL);
	    } else {
		SpParamFieldPart(component).sub_field =
		    spCreateTextField(SpParamFieldPart(component).box, "_ParamField_TrackBarTextField",
				      SppWidth, text_width,
				      SppHeight, default_text_height,
				      SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
				      SppSpacingOn, SpComponentPart(component).spacing_flag,
				      SppDisableShortcut, SpParamFieldPart(component).disable_shortcut,
                                      SppKeyboardType, SP_KEYBOARD_TYPE_NUMBER_PAD,
				      SppTextString, track_bar_string,
				      SppDescription, SpComponentPart(component).description,
				      SppHelpPath, SpComponentPart(component).help_path,
				      SppCallbackFunc, trackBarTextCB,
				      SppCallbackData, component,
				      NULL);
	    }

	    spAddCallback(SpParamFieldPart(component).field, SP_VALUE_CHANGED_CALLBACK,
			  trackBarCBForText, component);
	}
    } else if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_PROGRESS_BAR) {
	SpParamFieldPart(component).field = spCreateProgressBar(SpParamFieldPart(component).box, "_ParamField_ProgressBar",
								SppWidth, SpParamFieldPart(component).field_size,
								SppHeight, default_height,
								SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
								SppSpacingOn, SpComponentPart(component).spacing_flag,
								SppDescription, SpComponentPart(component).description,
								SppValue, SpParamFieldPart(component).value,
								SppMinimum, SpParamFieldPart(component).minimum,
								SppMaximum, SpParamFieldPart(component).maximum,
								SppIncrement, SpParamFieldPart(component).increment,
								NULL);
    } else {
	if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_FONT_TEXT) {
	    initial_string = SpParamFieldPart(component).initial_font;
	} else if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_COLOR_TEXT) {
	    initial_string = SpParamFieldPart(component).initial_color;
	} else if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_DIR_TEXT
		   && SpParamFieldPart(component).string == NULL)  {
	    initial_string = SpParamFieldPart(component).initial_dir;
	} else {
	    initial_string = SpParamFieldPart(component).string;
	}
	spDebug(80, "spParamFieldCreate", "text: field_size = %d\n",
		SpParamFieldPart(component).field_size);
	SpParamFieldPart(component).field =
	    spCreateTextField(SpParamFieldPart(component).box, "_ParamField_TextField",
			      SppWidth, SpParamFieldPart(component).field_size,
			      SppHeight, default_text_height,
			      SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
			      SppSpacingOn, SpComponentPart(component).spacing_flag,
			      SppEditable, SpParamFieldPart(component).editable,
			      SppDisableShortcut, SpParamFieldPart(component).disable_shortcut,
                              SppKeyboardType, SpParamFieldPart(component).keyboard_type,
			      SppTextString, initial_string,
			      SppDescription, SpComponentPart(component).description,
			      SppCallbackFunc, SpComponentPart(component).call_func,
			      SppCallbackData, SpComponentPart(component).call_data,
			      SppUserData, SpComponentPart(component).user_data,
			      SppHelpPath, SpComponentPart(component).help_path,
			      NULL);
    }
    spDebug(80, "spParamFieldCreate", "create field done\n");

    if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_DIR_TEXT
	|| SpParamFieldPart(component).field_type == SP_FIELD_TYPE_OPEN_FILE_TEXT
	|| SpParamFieldPart(component).field_type == SP_FIELD_TYPE_SAVE_FILE_TEXT
	|| SpParamFieldPart(component).field_type == SP_FIELD_TYPE_FONT_TEXT
	|| SpParamFieldPart(component).field_type == SP_FIELD_TYPE_COLOR_TEXT
	|| (SpParamFieldPart(component).use_dim_button == SP_TRUE
	    && !strnone(SpParamFieldPart(component).dim_string))) {
	if (!strnone(SpParamFieldPart(component).dim_string)) {
	    string = SpParamFieldPart(component).dim_string;
	} else {
	    if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_DIR_TEXT) {
		string = SP_DIR_SELECTION_BUTTON;
	    } else {
		string = SP_FILE_SELECTION_BUTTON;
	    }
	}
	if (SpParamFieldPart(component).dim_size == 0) {
	    SpParamFieldPart(component).dim_size = (2 * SP_DEFAULT_BUTTON_WIDTH) / 3;
	}
	
	/* create selection button */
	SpParamFieldPart(component).dim_label =
	    spCreatePushButton(SpParamFieldPart(component).box, "_ParamField_DimensionLabel",
			       SppTitle, string,
			       SppWidth, SpParamFieldPart(component).dim_size,
			       SppHeight, default_height,
			       SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
			       SppSpacingOn, SpComponentPart(component).spacing_flag,
			       SppCallbackFunc, pushButtonCB,
			       SppCallbackData, component,
			       NULL);
    } else if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR
	       || SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_TEXT
	       || SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_COMBO_BOX) {
	/* do nothing */
    } else if (!strnone(SpParamFieldPart(component).dim_string)) {
	/* create dimension label */
	SpParamFieldPart(component).dim_label =
	    spCreateLabel(SpParamFieldPart(component).box, "_ParamField_DimensionLabel",
			  SppTitle, SpParamFieldPart(component).dim_string,
			  SppAlignment, SP_ALIGNMENT_BEGINNING,
			  /*SppWidth, -SpComponentPart(component).margin_width,*/
			  SppWidth, SpParamFieldPart(component).dim_size,
			  SppHeight, default_height,
			  SppDisableInitialScaling, SpComponentPart(component).disable_initial_scaling,
			  SppSpacingOn, SpComponentPart(component).spacing_flag,
			  NULL);
    }
    spDebug(80, "spParamFieldCreate", "create dimension done\n");

    SpComponentPart(component).component = SpParamFieldPart(component).field;

    if (SpGetParent(component) != NULL) {
	SpGetWindow(component) = SpParentComponentPart(component).window;
    } else {
	SpGetWindow(component) = NULL;
    }

    if (SpParamFieldPart(component).use_check_box == SP_TRUE) {
	setSensitiveByCheckBox(component);
    }
	
    return SP_TRUE;
}

spBool spParamFieldDestroy(spObject object)
{
    spComponent component = (spComponent)object;
    
    spDestroyComponent(SpParamFieldPart(component).box);

    return SP_TRUE;
}

spBool spParamFieldSetParams(spObject object)
{
    spComponent component = (spComponent)object;

    if (SpComponentPart(component).title != NULL) {
	spSetParams(SpParamFieldPart(component).title_label,
		    SppTitle, SpComponentPart(component).title,
		    NULL);
    }

    if (SpParamFieldPart(component).string != NULL) {
	spSetParams(SpParamFieldPart(component).field,
		    SppTextString, SpParamFieldPart(component).string,
		    NULL);
    }
    
    if (SpParamFieldPart(component).strings != NULL) {
	spSetParams(SpParamFieldPart(component).field,
		    SppListStrings, SpParamFieldPart(component).strings,
		    NULL);
    }

    if (SpParamFieldPart(component).dim_string != NULL
	&& SpParamFieldPart(component).dim_label != NULL) {
	spSetParams(SpParamFieldPart(component).dim_label,
		    SppTitle, SpParamFieldPart(component).dim_string,
		    NULL);
    }
    
    if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_DIR_TEXT
	&& SpParamFieldPart(component).initial_dir != NULL) {
	spSetParams(SpParamFieldPart(component).field,
		    SppTextString, SpParamFieldPart(component).initial_dir,
		    NULL);
    }
    
    if (SpParamFieldPart(component).use_check_box == SP_TRUE) {
	if (SpParamFieldPart(SpOldObject(component)).set != SpParamFieldPart(component).set) {
	    spSetToggleState(SpParamFieldPart(component).title_label, SpParamFieldPart(component).set);
	    setSensitiveByCheckBox(component);
	}
    }
    
    if (SpParamFieldPart(SpOldObject(component)).editable != SpParamFieldPart(component).editable) {
	spSetParams(SpParamFieldPart(component).field,
		    SppEditable, SpParamFieldPart(component).editable,
		    NULL);
    }
    if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR
	|| SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_TEXT
	|| SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_COMBO_BOX) {
	spSetParams(SpParamFieldPart(component).field,
		    SppValue, SpParamFieldPart(component).value,
		    SppTrackCallbackOn, SpParamFieldPart(component).track_call_on,
		    SppMinimum, SpParamFieldPart(component).minimum,
		    SppMaximum, SpParamFieldPart(component).maximum,
		    SppPageSize, SpParamFieldPart(component).page_size,
		    SppIncrement, SpParamFieldPart(component).increment,
		    SppPageIncrement, SpParamFieldPart(component).page_increment,
		    SppDecimalPoints, SpParamFieldPart(component).decimal_points,
		    NULL);
	
	if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_TEXT
	    || SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_COMBO_BOX) {
	    char buf[SP_MAX_LINE];

	    getTrackBarTextString(component, SpParamFieldPart(component).value, buf);
	    spSetTextString(SpParamFieldPart(component).sub_field, buf);
	}
    } else if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_PROGRESS_BAR) {
	if (SpParamFieldPart(SpOldObject(component)).minimum != SpParamFieldPart(component).minimum
	    || SpParamFieldPart(SpOldObject(component)).maximum != SpParamFieldPart(component).maximum
	    || SpParamFieldPart(SpOldObject(component)).increment != SpParamFieldPart(component).increment) {
	    spSetParams(SpParamFieldPart(component).field,
			SppMinimum, SpParamFieldPart(component).minimum,
			SppMaximum, SpParamFieldPart(component).maximum,
			SppIncrement, SpParamFieldPart(component).increment,
			NULL);
	}
	if (SpParamFieldPart(SpOldObject(component)).value != SpParamFieldPart(component).value) {
	    spSetProgressBarValue(SpParamFieldPart(component).field, SpParamFieldPart(component).value);
	}
    }
    
    return SP_TRUE;
}

spBool spParamFieldGetParams(spObject object)
{
    spComponent component = (spComponent)object;

    if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR
	|| SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_TEXT
	|| SpParamFieldPart(component).field_type == SP_FIELD_TYPE_TRACK_BAR_WITH_COMBO_BOX) {
	spGetSliderValue(SpParamFieldPart(component).field, &SpParamFieldPart(component).value);
    } else if (SpParamFieldPart(component).field_type == SP_FIELD_TYPE_PROGRESS_BAR) {
	spGetProgressBarValue(SpParamFieldPart(component).field, &SpParamFieldPart(component).value);
    } else {
	SpParamFieldPart(component).string =
	    xspGetTextString(SpParamFieldPart(component).field);
    }
    
    return SP_TRUE;
}

void spParamFieldMap(spComponent component)
{
    spMapComponent(SpParamFieldPart(component).box);
    
    return;
}

void spParamFieldUnmap(spComponent component)
{
    spUnmapComponent(SpParamFieldPart(component).box);
    
    return;
}

spBool spParamFieldSetSize(spComponent component, int width, int height)
{
    return spSetSize(SpParamFieldPart(component).box, width, height);
}

spBool spParamFieldGetSize(spComponent component, int *width, int *height)
{
    return spGetSize(SpParamFieldPart(component).box, width, height);
}

spBool spParamFieldGetClientSize(spComponent component, int *width, int *height)
{
    return spGetClientSize(SpParamFieldPart(component).box, width, height);
}

spBool spParamFieldSetSensitive(spComponent component, spBool flag)
{
    spBool ret;
    
    ret = spSetSensitive(SpParamFieldPart(component).box, flag);

    if (SpParamFieldPart(component).use_check_box == SP_TRUE) {
	if (flag == SP_TRUE) {
	    setSensitiveByCheckBox(component);
	}
    }
    
    return ret;
}

spBool spParamFieldIsComponentType(spComponent component, const char *class_name)
{
    return SP_FALSE;
}

spBool spIsParamField(spComponent component)
{
    return spIsSubClass(component, SpParamField);
}

spComponent spCreateParamField(spComponent parent, const char *name, int size, ...)
{
    int num_arg = 0;
    spArg args[SP_MAX_NUM_ARG];
    va_list argp;
    
    if (spIsCreated(parent) == SP_FALSE || spIsContainer(parent) == SP_FALSE)
	return NULL;
    
    va_start(argp, size);
    spGetArgs(argp, args, num_arg);
    va_end(argp);

    spSetArg(args[num_arg], SppSize, size); num_arg++;
    
    return spCreateComponentArg(SpParamFieldClass, NULL, name, parent, args, num_arg);
}
