/*
 *        spPrimitive.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spThread.h>

#include <sp/spTopLevelP.h>
#include <sp/spPrimitiveP.h>
#include <sp/spMenuP.h>
#include <sp/spMenuItemP.h>
#include <sp/spContainerP.h>
#include <sp/spFrameP.h>
#include <sp/spLabelP.h>
#include <sp/spButtonP.h>
#include <sp/spTextP.h>
#include <sp/spComboBoxP.h>
#include <sp/spListP.h>
#include <sp/spSliderP.h>
#include <sp/spDrawP.h>
#include <sp/spImageP.h>
#include <sp/spCanvasP.h>
#include <sp/spProgressBarP.h>

#if defined(MACOS)
#pragma import on
#endif
extern spTopLevel sp_toplevel;
#if defined(MACOS)
#pragma import off
#endif

spPrimitiveClassRec SpPrimitiveClassRec = {
    /* spObjectClassPart */
    {
        SpPrimitive,
        (spObjectClass)&SpComponentClassRec,
        sizeof(spPrimitiveRec),
        0,
        NULL,
        spPrimitivePartInit,
        spPrimitivePartFree,
        SP_FALSE,
        NULL,
        NULL,
        spPrimitiveCreate,
        spPrimitiveDestroy,
        spPrimitiveCreate,
        NULL,
    },
    /* spComponentClassPart */
    {
        SP_FALSE,
        SP_FALSE,
        SP_FALSE,
        spPrimitiveMap,
        spPrimitiveUnmap,
        spPrimitiveSetSize,
        spPrimitiveGetSize,
        spPrimitiveGetClientSize,
        spPrimitiveSetSensitive,
        NULL,

        spPrimitiveAddCallback,
        spPrimitiveRemoveCallback,
        spPrimitiveCallbackFunc,
        spPrimitiveGetCallbackMousePosition,
        spPrimitiveGetCallbackKeySym,
        spPrimitiveGetCallbackKeyString,
    },
    /* spPrimitiveClassPart */
    {
        0,
    },
};

spComponentClass SpPrimitiveClass = (spComponentClass)&SpPrimitiveClassRec;

void spAllocCallbacks(spComponent component)
{
    long k;
    
    if (component == NULL ||
        SpPrimitivePart(component).num_callback < SpPrimitivePart(component).num_callback_buffer) return;

    if (SpPrimitivePart(component).callbacks == NULL) {
        SpPrimitivePart(component).num_callback_buffer = SP_CALLBACK_BUFFER;
        SpPrimitivePart(component).callbacks =
            xalloc(SpPrimitivePart(component).num_callback_buffer, spCallback);
        SpPrimitivePart(component).num_callback = 0;
    } else {
        SpPrimitivePart(component).num_callback_buffer += SP_CALLBACK_BUFFER;
        SpPrimitivePart(component).callbacks =
            xrealloc(SpPrimitivePart(component).callbacks,
                     SpPrimitivePart(component).num_callback_buffer, spCallback);
    }
    
    for (k = SpPrimitivePart(component).num_callback; k < SpPrimitivePart(component).num_callback_buffer; k++) {
        spInitCallbackArch(&SpPrimitivePart(component).callbacks[k]);
    }

    return;
}

void spFreeCallbacks(spComponent component)
{
    long k;
    
    if (SpPrimitivePart(component).callbacks != NULL) {
        for (k = 0; k < SpPrimitivePart(component).num_callback; k++) {
            spInitCallbackArch(&SpPrimitivePart(component).callbacks[k]);
        }
        xfree(SpPrimitivePart(component).callbacks);
        SpPrimitivePart(component).callbacks = NULL;
    }

    return;
}

void spPrimitivePartInit(spObject object)
{
    spComponent component = (spComponent)object;
    
    spPrimitivePartInitArch(component);
    SpPrimitivePart(component).index = -1;
    SpPrimitivePart(component).draw_func = NULL;
    SpPrimitivePart(component).draw_data = NULL;
    SpPrimitivePart(component).num_callback = 0;
    SpPrimitivePart(component).num_callback_buffer = 0;
    SpPrimitivePart(component).callbacks = NULL;

#if 0
    SpPrimitivePart(component).event_call_func = NULL;
    SpPrimitivePart(component).event_call_data = NULL;
#else
    SpPrimitivePart(component).event_func_list = NULL;
#endif
    
    spDebug(60, "spPrimitivePartInit", "init done\n");

    return;
}

void spPrimitivePartFree(spObject object)
{
    spComponent component = (spComponent)object;

    spFreeCallbacks(component);

    spDestroyEventCallbackList(&SpPrimitivePart(component).event_func_list);
    
    return;
}

spBool spPrimitiveCreate(spObject object)
{
    spComponent component = (spComponent)object;
        
    if (component == NULL) return SP_FALSE;

    if (spIsSubClass(component, SpMenuBar) == SP_TRUE) {
        spMenuBarCreate(component);
    } else if (spIsSubClass(component, SpMenuSeparator) == SP_TRUE) {
        spMenuSeparatorCreate(component);
    } else if (spIsSubClass(component, SpImage) == SP_TRUE) {
        /*spPrimitiveSetDefaultSize(component);*/
        if (spImageCreate(component) == SP_FALSE) {
            return SP_FALSE;
        }
#if defined(SP_SUPPORT_PRINT_CANVAS)
    } else if (spIsSubClass(component, SpPrintCanvas) == SP_TRUE) {
        spPrintCanvasCreate(component);
#endif
    }

    return SP_TRUE;
}

spBool spPrimitiveDestroy(spObject object)
{
    spComponent component = (spComponent)object;
    
    if (component == NULL) return SP_FALSE;
    
    spPrimitiveDestroyArch(component);
    
    return SP_TRUE;
}

void spPrimitiveMap(spComponent component)
{
    if (component == NULL) return;

    if (spIsWindow(component) == SP_TRUE) {
        spMapWindow(component);
    } else {
        spPrimitiveMapArch(component);
    }

    return;
}

void spPrimitiveUnmap(spComponent component)
{
    if (component == NULL) return;

    if (spIsWindow(component) == SP_TRUE) {
        spUnmapWindow(component);
    } else {
        spPrimitiveUnmapArch(component);
    }

    return;
}

spBool spPrimitiveSetSize(spComponent component, int width, int height)
{
    if (spIsCreated(component) == SP_TRUE) {
        return spPrimitiveSetSizeArch(component, width, height);
    } else {
        return SP_FALSE;
    }
}

spBool spPrimitiveGetSize(spComponent component, int *width, int *height)
{
    if (spIsCreated(component) == SP_TRUE) {
        return spPrimitiveGetSizeArch(component, width, height);
    } else {
        return SP_FALSE;
    }
}

spBool spPrimitiveGetClientSize(spComponent component, int *width, int *height)
{
    if (spIsCreated(component) == SP_TRUE) {
        return spPrimitiveGetClientSizeArch(component, width, height);
    } else {
        spDebug(30, "spPrimitiveGetClientSize", "failed: not created\n");
        return SP_FALSE;
    }
}

spBool spPrimitiveSetSensitive(spComponent component, spBool flag)
{
    if (component == NULL) return SP_FALSE;
    
    return spPrimitiveSetSensitiveArch(component, flag);
}

#if defined(SP_USE_DEFAULT_CALLBACK_STRUCT)
void spInitCallbackArch(spCallback *callback)
{
    callback->component = NULL;
    callback->call_func = NULL;
    callback->call_data = NULL;
    callback->propagate = SP_FALSE;

    return;
}

static spCallbackType getCallbackType(spComponent component, spCallbackType call_type, int *count)
{
    int i;
    int num_call_type = SP_NUM_CALLBACK_TYPE;
    spCallbackType current_type = SP_NO_CALLBACK;

    if (call_type == SP_NO_CALLBACK) return SP_NO_CALLBACK;

    if (*count < num_call_type) {
        for (i = *count; i < num_call_type; i++) {
            if (((spCallbackType)(1L << i)) & call_type) {
                *count = i + 1;
                current_type = (spCallbackType)(1L << i);
                break;
            }
        }
        if (i >= num_call_type) {
            *count = num_call_type;
        }
    }

    return current_type;
}

static spBool addCallback(spComponent component, spBool propagate, spCallbackType call_type,
                          spCallbackFunc call_func, void *call_data)
{
    long k;
    
    if (spIsPrimitive(component) == SP_TRUE && call_type != SP_NO_CALLBACK &&
        call_func != NULL) {
        for (k = 0; k < SpPrimitivePart(component).num_callback; k++) {
            if (SpPrimitivePart(component).callbacks[k].component == component &&
                SpPrimitivePart(component).callbacks[k].call_func == call_func &&
                SpPrimitivePart(component).callbacks[k].call_data == call_data &&
                SpPrimitivePart(component).callbacks[k].call_type == call_type) {
                spDebug(80, "addCallback", "already added callback\n");
                return SP_FALSE;
            } else if (SpPrimitivePart(component).callbacks[k].call_func == NULL) {
                break;
            }
        }
        
        if (k >= SpPrimitivePart(component).num_callback) {
            spAllocCallbacks(component);
            SpPrimitivePart(component).num_callback++;
        }
        
        spDebug(80, "addCallback", "add new callback: k = %ld, num_callback = %ld\n",
                k, SpPrimitivePart(component).num_callback);
        SpPrimitivePart(component).callbacks[k].component = component;
        SpPrimitivePart(component).callbacks[k].call_func = call_func;
        SpPrimitivePart(component).callbacks[k].call_data = call_data;
        SpPrimitivePart(component).callbacks[k].propagate = propagate;
        SpPrimitivePart(component).callbacks[k].call_type = call_type;
    }

    return SP_TRUE;
}

spBool spPrimitiveAddCallbackArch(spComponent component, spBool propagate, spCallbackType call_type,
                                  spCallbackFunc call_func, void *call_data)
{
    int count = 0, added_count = 0;
    spCallbackType current_type;
    
    while ((current_type = getCallbackType(component, call_type, &count)) != SP_NO_CALLBACK) {
        spDebug(80, "spPrimitiveAddCallbackArch", "count = %d\n", count);
        if (addCallback(component, propagate, current_type, call_func, call_data) == SP_TRUE) {
            ++added_count;
        }
    }

    spDebug(80, "spPrimitiveAddCallbackArch", "done: count = %d, added_count = %d\n", count, added_count);

    if (added_count > 0) {
#if defined(COCOA)
        spUpdateAcceptsMouseMovedEventsCocoa(component);
#endif
        
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

static spBool removeCallback(spComponent component, spCallbackType call_type,
                             spCallbackFunc call_func, void *call_data)
{
    long k;
    spBool flag = SP_FALSE;
    
    if (call_type != SP_NO_CALLBACK && call_func != NULL) {
        for (k = 0; k < SpPrimitivePart(component).num_callback; k++) {
            if (SpPrimitivePart(component).callbacks[k].component == component &&
                SpPrimitivePart(component).callbacks[k].call_func == call_func &&
                SpPrimitivePart(component).callbacks[k].call_data == call_data &&
                SpPrimitivePart(component).callbacks[k].call_type == call_type) {
                SpPrimitivePart(component).callbacks[k].component = NULL;
                SpPrimitivePart(component).callbacks[k].call_func = NULL;
                SpPrimitivePart(component).callbacks[k].call_data = NULL;
                SpPrimitivePart(component).callbacks[k].propagate = SP_FALSE;
                SpPrimitivePart(component).callbacks[k].call_type = SP_NO_CALLBACK;
                flag = SP_TRUE;
                spDebug(80, "removeCallback", "callback removed: k = %ld\n", k);
                break;
            }
        }
    }

    spDebug(80, "removeCallback", "flag = %d\n", flag);
    
    return flag;
}

spBool spPrimitiveRemoveCallbackArch(spComponent component, spCallbackType call_type,
                                     spCallbackFunc call_func, void *call_data)
{
    int count = 0, removed_count = 0;
    spCallbackType current_type;
    
    while ((current_type = getCallbackType(component, call_type, &count)) != SP_NO_CALLBACK) {
        if (removeCallback(component, current_type, call_func, call_data) == SP_TRUE) {
            ++removed_count;
        }
    }

    spDebug(80, "spPrimitiveRemoveCallbackArch", "done: count = %d, removed_count = %d\n", count, removed_count);
    
    if (removed_count > 0) {
#if defined(COCOA)
        spUpdateAcceptsMouseMovedEventsCocoa(component);
#endif
        
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spPostMessageDefault(spComponent component, spCallbackType call_type)
{
    long k;
    long component_id;
    long num_callback;
    spBool flag = SP_FALSE;

    if (component == NULL || call_type == SP_NO_CALLBACK) {
        return SP_FALSE;
    }
    
    if ((component_id = SpGetComponentId(component)) == spGetDestroyedComponentId()) {
        /* now destroying */
        component_id = 0;
    }
    num_callback = SpPrimitivePart(component).num_callback;
    spDebug(50, "spPostMessageDefault", "num_callback = %ld\n", num_callback);
    
    for (k = 0; k < num_callback; k++) {
        if (SpPrimitivePart(component).callbacks[k].call_type & call_type) {
            spDebug(50, "spPostMessageDefault", "callback found: k = %ld\n", k);
            
            spComponentCallbackFunc(&SpPrimitivePart(component).callbacks[k]);
            flag = SP_TRUE;

            spDebug(50, "spPostMessageDefault", "k = %ld: spComponentCallbackFunc done, sp_toplevel = %ld\n",
                    k, (long)sp_toplevel);
            if (sp_toplevel == NULL || component_id == spGetDestroyedComponentId()) {
                spDebug(50, "spPostMessageDefault", "destroy: component_id = %ld\n", component_id);
                return flag;
            }
        }
    }
    spDebug(50, "spPostMessageDefault", "flag = %d\n", flag);
    
    return flag;
}
#endif

spBool spPrimitiveAddCallback(spComponent component, spBool propagate, spCallbackType call_type,
                              spCallbackFunc call_func, void *call_data)
{
    if (call_func != NULL && call_type != SP_NO_CALLBACK) {
        return spPrimitiveAddCallbackArch(component, propagate, call_type, call_func, call_data);
    }

    return SP_FALSE;
}

spBool spPrimitiveRemoveCallback(spComponent component, spCallbackType call_type,
                                 spCallbackFunc call_func, void *call_data)
{
    if (call_func != NULL && call_type != SP_NO_CALLBACK) {
        return spPrimitiveRemoveCallbackArch(component, call_type, call_func, call_data);
    }

    return SP_FALSE;
}

spCallbackReason spPrimitiveCallbackFunc(spComponent component)
{
    spCallbackReason call_reason = SP_CR_UNKNOWN;

    call_reason = spPrimitiveGetCallbackReasonArch(component);
    spDebug(100, "spPrimitiveCallbackFunc", "call_reason = %d\n", call_reason);

    if (call_reason == SP_CR_ACTIVATE) {
        if (streq(SpGetName(component), "_OkButton")) {
            call_reason = SP_CR_OK;
        } else if (streq(SpGetName(component), "_CancelButton")) {
            call_reason = SP_CR_CANCEL;
        } else if (streq(SpGetName(component), "_ApplyButton")) {
            call_reason = SP_CR_APPLY;
        } else if (streq(SpGetName(component), "_HelpButton")) {
            call_reason = SP_CR_HELP;
        } else if (streq(SpGetName(component), "_YesButton")) {
            call_reason = SP_CR_YES;
        } else if (streq(SpGetName(component), "_NoButton")) {
            call_reason = SP_CR_NO;
        } else if (streq(SpGetName(component), "_RetryButton")) {
            call_reason = SP_CR_RETRY;
        }
    }
    
    return call_reason;
}

spBool spPrimitiveGetCallbackMousePosition(spComponent component, int *x, int *y)
{
    if ((x != NULL || y != NULL)) {
        return spPrimitiveGetCallbackMousePositionArch(component, x, y);
    }
    
    return SP_FALSE;
}

spBool spPrimitiveGetCallbackKeySym(spComponent component, spKeySym *key_sym)
{
    return spPrimitiveGetCallbackKeySymArch(component, key_sym);
}

int spPrimitiveGetCallbackKeyString(spComponent component,
                                    char *buf, int buf_size, spBool *overflow)
{
    return spPrimitiveGetCallbackKeyStringArch(component, buf, buf_size, overflow);
}

void spPrimitiveSetDefaultSize(spComponent component)
{
    int default_width = 0;
    int default_height = 0;
    
    if (component == NULL
        || spIsCreated(component) == SP_TRUE) return;

    spDebug(60, "spPrimitiveSetDefaultSize", "in\n");

    if (spIsSubClass(component, SpText) == SP_TRUE) {
        if (spIsSubClass(component, SpTextArea) == SP_TRUE) {
            default_width = SP_DEFAULT_TEXT_AREA_WIDTH;
            default_height = SP_DEFAULT_TEXT_AREA_HEIGHT;
        } else {
            default_width = SP_DEFAULT_TEXT_WIDTH;
            default_height = SP_DEFAULT_TEXT_HEIGHT;
        }
    } else if (spIsSubClass(component, SpList) == SP_TRUE) {
        default_width = SP_DEFAULT_LIST_WIDTH;
        default_height = SP_DEFAULT_LIST_HEIGHT;
    } else if (spIsSubClass(component, SpButton) == SP_TRUE) {
        default_width = SP_DEFAULT_BUTTON_WIDTH;
        default_height = SP_DEFAULT_BUTTON_HEIGHT;
    } else if (spIsSubClass(component, SpLabel) == SP_TRUE) {
        default_width = SP_DEFAULT_LABEL_WIDTH;
        default_height = SP_DEFAULT_LABEL_HEIGHT;
    } else if (spIsSubClass(component, SpProgressBar) == SP_TRUE) {
        default_width = SP_DEFAULT_PROGRESS_BAR_WIDTH;
        default_height = SP_DEFAULT_PROGRESS_BAR_HEIGHT;
    } else {
        return;
    }
    
    SpComponentPart(component).x = 0;
    SpComponentPart(component).y = 0;
    if (spIsSubClass(component, SpTextArea) == SP_FALSE
        && spIsSubClass(component, SpList) == SP_FALSE
        && SpComponentPart(component).width == 0
        && (SpParentComponentPart(component).orientation == SP_HORIZONTAL
            && SpComponentPart(component).percentage_width < 0)) {
        SpComponentPart(component).width = default_width;
    }
    if (SpComponentPart(component).require_width <= 0) {
        SpComponentPart(component).require_width
            = MAX(SpComponentPart(component).width, default_width);
    }
    if (spIsSubClass(component, SpTextArea) == SP_FALSE
        && spIsSubClass(component, SpList) == SP_FALSE
        && SpComponentPart(component).height == 0
        && (SpParentComponentPart(component).orientation == SP_VERTICAL
            && SpComponentPart(component).percentage_height < 0)) {
        SpComponentPart(component).height = default_height;
    }
    if (SpComponentPart(component).require_height <= 0) {
        SpComponentPart(component).require_height
            = MAX(SpComponentPart(component).height, default_height);
    }
    SpComponentPart(component).current_width = SpComponentPart(component).width;
    SpComponentPart(component).current_height = SpComponentPart(component).height;
    
    spDebug(60, "spPrimitiveSetDefaultSize", "width = %d, height = %d, current_width = %d, current_height = %d\n",
            SpComponentPart(component).width, SpComponentPart(component).height,
            SpComponentPart(component).current_width, SpComponentPart(component).current_height);
    
    return;
}

spBool spIsPrimitive(spComponent component)
{
    return spIsSubClass(component, SpPrimitive);
}

spBool spShowToolTip(spComponent component)
{
    if (spIsPrimitive(component) == SP_FALSE
        || strnone(SpComponentPart(component).description)) return SP_FALSE;
    
    return spShowToolTipArch(component);
}

spBool spAddEventCallbackList(spEventCallbackList *first_list, spEventCallbackFunc call_func)
{
    spEventCallbackList new_list, list;

    if (first_list == NULL || call_func == NULL) return SP_FALSE;

    list = *first_list;
    while (list != NULL) {
        if (list->call_func == call_func) {
            return SP_FALSE;
        }
        list = list->next;
    }
    
    new_list = xalloc(1, struct _spEventCallbackList);
    new_list->call_func = call_func;
    new_list->prev = NULL;
    new_list->next = NULL;
    
    if (*first_list == NULL) {
        *first_list = new_list;
    } else {
        list = *first_list;
        while (list->next != NULL) {
            list = list->next;
        }
        list->next = new_list;
        new_list->prev = list;
    }

    return SP_TRUE;
}

spBool spRemoveEventCallbackList(spEventCallbackList *first_list, spEventCallbackFunc call_func)
{
    spEventCallbackList list, next;
    spBool flag = SP_FALSE;

    if (first_list == NULL || call_func == NULL) return SP_FALSE;
    
    list = *first_list;
    while (list != NULL) {
        next = list->next;
        if (list->call_func == call_func) {
            if (list == *first_list) {
                *first_list = next;
            }
            if (list->prev != NULL) {
                list->prev->next = next;
            }
            if (next != NULL) {
                next->prev = list->prev;
            }

            xfree(list);

            flag = SP_TRUE;
        }
        list = next;
    }

    return flag;
}

spBool spDestroyEventCallbackList(spEventCallbackList *first_list)
{
    if (first_list != NULL && *first_list != NULL) {
        spEventCallbackList list, next;
        
        list = *first_list;
        while (list != NULL) {
            next = list->next;
            xfree(list);
            list = next;
        }
        *first_list = NULL;
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spCallEventCallbackList(spEventCallbackList *first_list, spComponent component, spEventType event_type, void *call_data)
{
    spBool flag = SP_TRUE;
    spEventCallbackList list;
        
    list = *first_list;
    while (list != NULL) {
        if (list->call_func(component, event_type, call_data) == SP_FALSE) {
            flag = SP_FALSE;
        }
        list = list->next;
    }

    return flag;
}

spBool spExistsPrimitiveEventCallback(spComponent primitive)
{
    if (primitive == NULL)  return SP_FALSE;
    
    if (SpPrimitivePart(primitive).event_func_list != NULL) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spPrimitiveCallEventCallback(spComponent primitive, spEventType event_type, void *call_data)
{
    spBool flag;

    if (primitive == NULL) return SP_FALSE;

    flag = SP_TRUE;
#if 0
    flag = SpPrimitivePart(primitive).event_call_func(primitive,
                                                      event_type, call_data);
#else
    flag = spCallEventCallbackList(&SpPrimitivePart(primitive).event_func_list, primitive, event_type, call_data);
#endif

    return flag;
}

spBool spAddEventCallback(spComponent component, spEventCallbackFunc call_func)
{
    if (component == NULL) return SP_FALSE;

    if (spIsPrimitive(component) == SP_FALSE) {
        if (spIsPrimitive(spGetComponent(component)) == SP_TRUE) {
            component = spGetComponent(component);
        } else {
            return SP_FALSE;
        }
    }

    if (SpPrimitivePart(component).event_func_list == NULL) {
        if (spPrepareEventCallbackArch(component) == SP_FALSE) {
            return SP_FALSE;
        }
    }

    return spAddEventCallbackList(&SpPrimitivePart(component).event_func_list, call_func);
}

spBool spRemoveEventCallback(spComponent component, spEventCallbackFunc call_func)
{
    if (component == NULL) return SP_FALSE;

    if (spIsPrimitive(component) == SP_FALSE) {
        if (spIsPrimitive(spGetComponent(component)) == SP_TRUE) {
            component = spGetComponent(component);
        } else {
            return SP_FALSE;
        }
    }

    if (spRemoveEventCallbackList(&SpPrimitivePart(component).event_func_list, call_func) == SP_TRUE) {
        if (SpPrimitivePart(component).event_func_list == NULL) {
            spUnprepareEventCallbackArch(component);
        }
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spPostUserEvent(spComponent component, spEventType event_type, void *call_data)
{
    if (component == NULL) return SP_FALSE;

    if (spIsPrimitive(component) == SP_FALSE) {
        if (spIsPrimitive(spGetComponent(component)) == SP_TRUE) {
            component = spGetComponent(component);
        } else {
            return SP_FALSE;
        }
    }

    if (spExistsPrimitiveEventCallback(component) == SP_FALSE) {
        return SP_FALSE;
    }
    
    return spPostUserEventArch(component, event_type, call_data);
}

#if defined(SP_USE_DEFAULT_USER_EVENT)
struct _spUserEventDefaultRec {
    spComponent component;
    void *call_data;
    spEventType event_type;
    long id;

    spUserEventDefault prev;
    spUserEventDefault next;
};
    
spBool spAllocUserEventDefault(spUserEventDefault *first_event, long count)
{
    long k;
    spUserEventDefault event, prev;
    
    if (first_event == NULL) return SP_FALSE;

    prev = NULL;
    for (k = 0; k < count; k++) {
        event = xalloc(1, spUserEventDefaultRec);
        memset(event, 0, sizeof(struct _spUserEventDefaultRec));
        event->id = -1;
        if (prev != NULL) {
            prev->next = event;
            event->prev = prev;
        }
        
        if (*first_event == NULL) {
            *first_event = event;
        }

        prev = event;
    }

    return SP_TRUE;
}

spUserEventDefault spAppendUserEventDefault(spUserEventDefault *first_event,
                                            spComponent component, spEventType event_type, void *call_data)
{
    spUserEventDefault event, next, last;
    long new_id;

    if (first_event == NULL
        || spIsPrimitive(component) == SP_FALSE
        || spExistsPrimitiveEventCallback(component) == SP_FALSE) return NULL;
    
    spDebug(50, "spAppendUserEventDefault", "event_type = %lx\n", event_type);

#if defined(MACOS9)
    ThreadBeginCritical();
#endif
    
    new_id = 0;
    event = last = NULL;
    next = *first_event;

    while (next != NULL) {
        spDebug(50, "spAppendUserEventDefault", "id = %ld\n", next->id);
        if (next->id < 0) {
            event = next;
            spDebug(50, "spAppendUserEventDefault", "found vacant id\n");
            break;
        }
        new_id = MAX(next->id + 1, new_id);
        last = next;
        
        next = next->next;
    }

    if (event == NULL) {
        spDebug(50, "spAppendUserEventDefault", "alloc new event\n");
        event = xalloc(1, spUserEventDefaultRec);
        event->prev = NULL;
        event->next = NULL;
        if (last != NULL) {
            last->next = event;
            event->prev = last;
        }
    }
    if (*first_event == NULL) {
        *first_event = event;
        spDebug(50, "spAppendUserEventDefault", "fist event\n");
    }

    event->component = component;
    event->call_data = call_data;
    event->event_type = event_type;
    event->id = new_id;
    
    spDebug(80, "spAppendUserEventDefault", "done\n");
    
#if defined(MACOS9)
    ThreadEndCritical();
#endif
    
    return event;
}

spBool spDestroyUserEventDefault(spUserEventDefault *first_event, spUserEventDefault event)
{
    spUserEventDefault last_event;
    
    if (first_event == NULL || event == NULL) return SP_FALSE;

#if defined(MACOS9)
    ThreadBeginCritical();
#endif
    
    event->component = NULL;
    event->call_data = NULL;
    event->event_type = 0;
    event->id = -1;

    if ((last_event = spGetLastUserEventDefault(first_event)) != NULL
        && last_event != event) {
        /* move event to last */
        
        if (*first_event == event) {
            *first_event = event->next;
        }
        if (event->prev != NULL) event->prev->next = event->next;
        if (event->next != NULL) event->next->prev = event->prev;
        
        last_event->next = event;
        event->prev = last_event;
        event->next = NULL;
    }
    
#if defined(MACOS9)
    ThreadEndCritical();
#endif
    
    return SP_TRUE;
}

spBool spDestroyAllUserEventDefault(spUserEventDefault *first_event)
{
    spUserEventDefault event, next;
    
    if (first_event == NULL || *first_event == NULL) return SP_FALSE;

    event = *first_event;

    while (event != NULL) {
        next = event->next;
        xfree(event);
        
        event = next;
    }
    *first_event = NULL;
    
    return SP_TRUE;
}

spUserEventDefault spGetLastUserEventDefault(spUserEventDefault *first_event)
{
    spUserEventDefault event;
    
    event = *first_event;

    while (event != NULL) {
        if (event->next == NULL) {
            return event;
        }
        event = event->next;
    }

    return NULL;
}

long spGetNumUserEventDefault(spUserEventDefault *first_event)
{
    long count = 0;
    spUserEventDefault next;

    if (first_event == NULL) return 0;
    
    next = *first_event;
    
    while (next != NULL) {
        if (next->id >= 0) {
            ++count;
        }
        next = next->next;
    }

    return count;
}

spBool spProcessUserEventDefault(spUserEventDefault *first_event, spUserEventDefault event)
{
    spBool flag;
                   
    spDebug(80, "spProcessUserEventDefault", "in\n");
    
    if (event == NULL || spIsPrimitive(event->component) == SP_FALSE
        || spExistsPrimitiveEventCallback(event->component) == SP_FALSE) return SP_FALSE;

    spDebug(80, "spProcessUserEventDefault", "before spPrimitiveCallEventCallback\n");
    flag = spPrimitiveCallEventCallback(event->component, event->event_type, event->call_data);
    spDebug(60, "spProcessUserEventDefault", "event_type = %ld, flag = %d\n", event->event_type, flag);
    
    spDestroyUserEventDefault(first_event, event);
    
    spDebug(80, "spProcessUserEventDefault", "done\n");
    
    return flag;
}

long spProcessPendingUserEventDefault(spUserEventDefault *first_event, long end_count)
{
    long k, count;
    spUserEventDefault event, next_event;
    
    if (first_event == NULL || *first_event == NULL) return -1;

    spDebug(100, "spProcessPendingUserEventDefault", "in: end_count = %ld\n", end_count);

    count = 0;
    event = *first_event;

    for (k = 0; event != NULL; k++) {
        spDebug(100, "spProcessPendingUserEventDefault", "k = %ld, count = %ld, event->id = %ld\n",
                k, count, event->id);
        next_event = event->next;

        if (event->id >= 0) {
            spProcessUserEventDefault(first_event, event);
            spDebug(100, "spProcessPendingUserEventDefault", "process done\n");
        
            ++count;
        
            spDebug(100, "spProcessPendingUserEventDefault", "k = %ld, count = %ld / %ld\n", k, count, end_count);
            
            if (end_count > 0 && count >= end_count) {
                break;
            }
        
            if (spIsThreadSafe() == SP_TRUE) {
                spDebug(100, "spProcessPendingUserEventDefault", "before spYieldThread\n");
                spYieldThread();
                spDebug(100, "spProcessPendingUserEventDefault", "after spYieldThread\n");
            }
        }

        event = next_event;
    }

    spDebug(100, "spProcessPendingUserEventDefault", "done: count = %ld\n", count);
    
    return count;
}
#endif
