/*
 *        spPrimitive_Win.c
 */

#include <windows.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spLocale.h>
#include <sp/spComponent.h>
#include <sp/spFrame.h>
#include <sp/spComboBox.h>
#include <sp/spToolBar.h>
#include <sp/spToolItem.h>
#include <sp/spStatusBar.h>

#include <sp/spTopLevelP.h>
#include <sp/spFrameP.h>
#include <sp/spMenuP.h>
#include <sp/spMenuItemP.h>
#include <sp/spDialogP.h>
#include <sp/spButtonP.h>
#include <sp/spSliderP.h>
#include <sp/spTabBoxP.h>
#include <sp/spContainerP.h>
#include <sp/spCanvasP.h>
#include <sp/spGraphicsP.h>
#include <sp/spTextP.h>
#include <sp/spDrawP.h>
#include <sp/spImageP.h>
#include <sp/spPrimitiveP.h>

extern spTopLevel sp_toplevel;

static spCallbackTable sp_callback_tables[] = {
    {SP_NO_CALLBACK, NULL, WM_NULL, 0},
    {SP_KEY_PRESS_CALLBACK, SpCanvas, WM_CHAR, 0},
    {SP_KEY_PRESS_CALLBACK, SpCanvas, WM_IME_CHAR, 0},
    {SP_KEY_PRESS_CALLBACK, SpCanvas, WM_KEYDOWN, 0},
    {SP_KEY_RELEASE_CALLBACK, SpCanvas, WM_KEYUP, 0},
    {SP_BUTTON_PRESS_CALLBACK, SpFrame, WM_TASKTRAY, 0},
    {SP_BUTTON_PRESS_CALLBACK, SpPrimitive, WM_LBUTTONDOWN, 0},
    {SP_BUTTON_PRESS_CALLBACK, SpPrimitive, WM_RBUTTONDOWN, 0},
    {SP_BUTTON_PRESS_CALLBACK, SpPrimitive, WM_MBUTTONDOWN, 0},
    {SP_BUTTON_RELEASE_CALLBACK, SpPrimitive, WM_LBUTTONUP, 0},
    {SP_BUTTON_RELEASE_CALLBACK, SpPrimitive, WM_RBUTTONUP, 0},
    {SP_BUTTON_RELEASE_CALLBACK, SpPrimitive, WM_MBUTTONUP, 0},
    {SP_POINTER_MOTION_CALLBACK, SpCanvas, WM_MOUSEMOVE, 0},
    {SP_BUTTON_MOTION_CALLBACK, SpCanvas, WM_MOUSEMOVE,
        MK_LBUTTON|MK_RBUTTON|MK_MBUTTON},
#if 0
    {SP_ENTER_WINDOW_CALLBACK, SpCanvas, WM_NULL, 0},
    {SP_LEAVE_WINDOW_CALLBACK, SpCanvas, WM_NULL, 0},
#endif
    /**/
    {SP_DESTROY_CALLBACK, SpFrame, WM_DESTROY, 0},
    {SP_CLOSE_CALLBACK, SpFrame, WM_CLOSE, 0},
    {SP_ACTIVATE_CALLBACK, SpPushButton, WM_COMMAND, 0},
    {SP_ACTIVATE_CALLBACK, SpMenuItem, WM_COMMAND, 0},
    {SP_ACTIVATE_CALLBACK, SpToolItem, WM_COMMAND, 0},
    {SP_ACTIVATE_CALLBACK, SpTextField, WM_KEYDOWN, 0},
    {SP_VALUE_CHANGED_CALLBACK, SpCheckBox, WM_COMMAND, 0},
    {SP_VALUE_CHANGED_CALLBACK, SpRadioButton, WM_COMMAND, 0},
    {SP_VALUE_CHANGED_CALLBACK, SpCheckBoxMenuItem, WM_COMMAND, 0},
    {SP_VALUE_CHANGED_CALLBACK, SpRadioButtonMenuItem, WM_COMMAND, 0},
    {SP_VALUE_CHANGED_CALLBACK, SpCheckToolItem, WM_COMMAND, 0},
    {SP_VALUE_CHANGED_CALLBACK, SpSlider, WM_VSCROLL, 0},
    {SP_VALUE_CHANGED_CALLBACK, SpSlider, WM_HSCROLL, 0},
    {SP_VALUE_CHANGED_CALLBACK, SpTextField, WM_COMMAND, EN_CHANGE},
    {SP_VALUE_CHANGED_CALLBACK, SpTextArea, WM_COMMAND, EN_CHANGE},
    {SP_VALUE_CHANGED_CALLBACK, SpComboBox, WM_COMMAND, CBN_EDITCHANGE},
    {SP_VALUE_CHANGED_CALLBACK, SpComboBox, WM_COMMAND, CBN_SELCHANGE},
    {SP_VALUE_CHANGED_CALLBACK, SpList, WM_COMMAND, 0},
#if 0
    {SP_MAP_CALLBACK, SpPrimitive, WM_NULL, 0},
    {SP_UNMAP_CALLBACK, SpPrimitive, WM_NULL, 0},
#endif
    {SP_EXPOSE_CALLBACK, SpCanvas, WM_PAINT, 0},
    {SP_EXPOSE_CALLBACK, SpContainer, WM_PAINT, 0},
    {SP_RESIZE_CALLBACK, SpCanvas, WM_SIZE, 0},
    /**/
    {SP_OK_CALLBACK, SpPushButton, WM_COMMAND, IDOK},
    {SP_OK_CALLBACK, SpDialog, WM_COMMAND, IDOK},
    {SP_CANCEL_CALLBACK, SpPushButton, WM_COMMAND, IDCANCEL},
    {SP_CANCEL_CALLBACK, SpDialog, WM_COMMAND, IDCANCEL},
    {SP_HELP_CALLBACK, SpPushButton, WM_COMMAND, IDHELP},
    {SP_HELP_CALLBACK, SpDialog, WM_COMMAND, IDHELP},
#if 0
    {SP_APPLY_CALLBACK, SpPushButton, WM_NULL, 0},
    {SP_APPLY_CALLBACK, SpDialog, WM_NULL, 0},
#endif
    /* +0.6.13 */
    {SP_SCROLL_CALLBACK, SpCanvas, WM_HSCROLL, 0},
    {SP_SCROLL_CALLBACK, SpCanvas, WM_VSCROLL, 0},
    {SP_ZOOM_CALLBACK, SpCanvas, 0x020A/*WM_MOUSEWHEEL*/, 1},
    {SP_WHEEL_CALLBACK, SpCanvas, 0x020A/*WM_MOUSEWHEEL*/, 0},
    {SP_WHEEL_CALLBACK, SpCanvas, 0x020E/*WM_MOUSEHWHEEL*/, 0},
    {SP_FOCUS_CALLBACK, SpCanvas, WM_SETFOCUS, 0},
    {SP_FOCUS_CALLBACK, SpCanvas, WM_KILLFOCUS, 0},
};
static int sp_num_callback_table = 0;

extern spTopLevel sp_toplevel;

DWORD spGetVisibleStyleWin(spComponent component)
{
    if (spIsVisible(component) == SP_TRUE) {
        return WS_VISIBLE;
    } else {
        return 0L;
    }
}

static spComponent getCurrentMenu(spComponent component, HMENU hmenu,
                                  UINT uitem, UINT uflags)
{
    spComponent child;
    spComponent current = NULL;
    spComponent current_menu = NULL;

    if (hmenu == NULL
        || (uflags & MF_SYSMENU) != 0) return NULL;
    
    if (component == NULL) {
        child = spGetChild(NULL);
    } else {
        if (spIsCreated(component) == SP_FALSE) {
            return NULL;
        } else {
            child = SpGetChild(component);
        }

        if (spIsMenu(component) == SP_TRUE
            && SpPrimitiveArch(component).hmenu == hmenu) {
            current = component;
        }
    }

    if (current == NULL) {
        while (child != NULL) {
            if (spIsMenu(child) == SP_TRUE
                && SpPrimitiveArch(child).hmenu == hmenu) {
                current = child;
                break;
            } else if ((current_menu = getCurrentMenu(child, hmenu, uitem, uflags)) != NULL) {
                return current_menu;
            }
            child = SpGetNextComponent(child);
        }
    }

    if (current != NULL) {
        UINT i;
        
        child = SpGetChild(current);
        
        for (i = 0; child != NULL; i++) {
            if (uflags & MF_POPUP) {
                if (i == uitem) {
                    current_menu = child;
                    break;
                }
            } else {
                if ((UINT)SpGetComponentId(child) == uitem) {
                    current_menu = child;
                    break;
                }
            }

            child = SpGetNextComponent(child);
        }
    }

    return current_menu;
}

static spComponent getCurrentComponent(spComponent component, HWND hwnd)
{
    spComponent child;
    spComponent current = NULL;

    if (hwnd == NULL) return NULL;
    
    if (component == NULL) {
        child = spGetChild(NULL);
    } else {
        child = SpGetChild(component);

        if (SpPrimitiveArch(component).hwnd == hwnd) {
            current = component;
        }
    }

    if (current == NULL) {
        while (child != NULL) {
            if (SpPrimitiveArch(child).hwnd == hwnd) {
                current = child;
                break;
            } else if ((current = getCurrentComponent(child, hwnd)) != NULL) {
                break;
            }
            child = SpGetNextComponent(child);
        }
    }

    return current;
}

static spComponent getCurrentWindow(HWND hwnd, UINT message,
                                    WPARAM wParam, LPARAM lParam)
{
    spComponent window = NULL;

    if ((window = getCurrentComponent(NULL, hwnd)) != NULL) {
        SpPrimitiveArch(window).message = message;
        SpPrimitiveArch(window).wParam = wParam;
        SpPrimitiveArch(window).lParam = lParam;
    } else {
        spDebug(100, "getCurrentWindow", "window not found: hwnd = %lx\n", (long)hwnd);
    }

    return window;
}

static UINT getCallbackMessage(spComponent component, spCallbackType call_type, UINT *command_id, int *count)
{
    int i;
    UINT message = WM_NULL;
    
    if (call_type == SP_NO_CALLBACK) {
        return message;
    }

    if (sp_num_callback_table <= 0) {
        sp_num_callback_table = spArraySize(sp_callback_tables);
    }

    if (*count < sp_num_callback_table) {
        for (i = *count; i < sp_num_callback_table; i++) {
            if (sp_callback_tables[i].component_type == NULL
                || spIsSubClass(component, sp_callback_tables[i].component_type) == SP_TRUE) {
                if (call_type & sp_callback_tables[i].call_type &&
                    sp_callback_tables[i].message != WM_NULL) {
                    message = sp_callback_tables[i].message;
                    if (command_id != NULL) {
                        *command_id = sp_callback_tables[i].command_id;
                    }
                    *count = i + 1;
                    break;
                }
            }
        }
        if (i >= sp_num_callback_table) {
            *count = sp_num_callback_table;
        }
    }
    
    return message;
}

void spInitCallbackArch(spCallback *callback)
{
    callback->component = NULL;
    callback->call_func = NULL;
    callback->call_data = NULL;
    callback->propagate = SP_FALSE;

    callback->component_id = -1;
    callback->message = WM_NULL;
    callback->command_id = 0;
    
    return;
}

static spBool addCallback(spComponent component, spBool propagate, UINT message, UINT command_id,
                          spCallbackFunc call_func, void *call_data)
{
    long k;
    spComponent window = NULL;
    
    if (spIsPrimitive(component) == SP_TRUE && message != WM_NULL &&
        call_func != NULL) {
        window = SpGetWindow(component);
        if (spIsPrimitive(window) == SP_FALSE) {
            return SP_FALSE;
        }
        
        for (k = 0; k < SpPrimitivePart(window).num_callback; k++) {
            if (SpPrimitivePart(window).callbacks[k].component == component &&
                SpPrimitivePart(window).callbacks[k].component_id == SpComponentPart(component).component_id &&
                SpPrimitivePart(window).callbacks[k].message == message &&
                SpPrimitivePart(window).callbacks[k].command_id == command_id &&
                SpPrimitivePart(window).callbacks[k].call_func == call_func &&
                SpPrimitivePart(window).callbacks[k].call_data == call_data) {
                spDebug(80, "addCallback", "already added callback\n");
                return SP_FALSE;
            } else if (SpPrimitivePart(window).callbacks[k].call_func == NULL) {
                break;
            }
        }
        
        if (k >= SpPrimitivePart(window).num_callback) {
            spAllocCallbacks(window);
            SpPrimitivePart(window).num_callback++;
        }
        
        SpPrimitivePart(window).callbacks[k].component = component;
        SpPrimitivePart(window).callbacks[k].component_id = SpComponentPart(component).component_id;
        SpPrimitivePart(window).callbacks[k].message = message;
        SpPrimitivePart(window).callbacks[k].command_id = command_id;
        SpPrimitivePart(window).callbacks[k].call_func = call_func;
        SpPrimitivePart(window).callbacks[k].call_data = call_data;
        SpPrimitivePart(window).callbacks[k].propagate = propagate;

        spDebug(80, "addCallback", "message = %d\n", message);
    }

    return SP_TRUE;
}

static spBool removeCallbackMain(spComponent component, UINT message, UINT command_id,
                                 spCallbackFunc call_func, void *call_data, spBool all_flag)
{
    long k;
    spBool flag = SP_FALSE;
    spComponent window = NULL;
    
    if (spIsPrimitive(component) == SP_TRUE &&
        (all_flag == SP_TRUE ||        (message != WM_NULL && call_func != NULL))) {
        window = SpGetWindow(component);
        if (spIsPrimitive(window) == SP_FALSE) {
            return SP_FALSE;
        }
        spDebug(80, "removeCallback", "%s: all_flag = %d, num_callback = %ld\n", spGetName(component), all_flag, SpPrimitivePart(window).num_callback);
        
        for (k = 0; k < SpPrimitivePart(window).num_callback; k++) {
            spDebug(80, "removeCallback", "callbacks[%ld].component_id = %ld, component_id = %ld\n",
                    k, SpPrimitivePart(window).callbacks[k].component_id, SpComponentPart(component).component_id);
            if ((all_flag == SP_TRUE
                 && (SpPrimitivePart(window).callbacks[k].component == component
                     || spIsDescendant(component, SpPrimitivePart(window).callbacks[k].component) == SP_TRUE))
                || (SpPrimitivePart(window).callbacks[k].component == component &&
                    SpPrimitivePart(window).callbacks[k].component_id == SpComponentPart(component).component_id &&
                    SpPrimitivePart(window).callbacks[k].message == message &&
                    SpPrimitivePart(window).callbacks[k].command_id == command_id &&
                    SpPrimitivePart(window).callbacks[k].call_func == call_func &&
                    SpPrimitivePart(window).callbacks[k].call_data == call_data)) {
                SpPrimitivePart(window).callbacks[k].component = NULL;
                SpPrimitivePart(window).callbacks[k].component_id = -1;
                SpPrimitivePart(window).callbacks[k].message = WM_NULL;
                SpPrimitivePart(window).callbacks[k].command_id = 0;
                SpPrimitivePart(window).callbacks[k].call_func = NULL;
                SpPrimitivePart(window).callbacks[k].call_data = NULL;

                flag = SP_TRUE;
                spDebug(80, "removeCallback", "removed: %s component_id = %ld\n",
                        spGetName(component), SpComponentPart(component).component_id);
                if (all_flag == SP_FALSE) {/* bug fixed: 2009/11/18 */
                    break;
                }
            }
        }
    }

    spDebug(80, "removeCallback", "done: flag = %d\n", flag);
    
    return flag;
}

static spBool removeCallback(spComponent component, UINT message, UINT command_id,
                             spCallbackFunc call_func, void *call_data)
{
    return removeCallbackMain(component, message, command_id, call_func, call_data, SP_FALSE);
}

static spBool removeAllCallback(spComponent component)
{
    spDebug(80, "removeAllCallback", "in\n");
        
    return removeCallbackMain(component, WM_NULL, 0, NULL, NULL, SP_TRUE);
}

spBool spPrimitiveAddCallbackArch(spComponent component, spBool propagate, spCallbackType call_type,
                                  spCallbackFunc call_func, void *call_data)
{
    int count = 0;
    spBool flag = SP_FALSE;
    UINT message;
    UINT command_id = 0;
    
    while ((message = getCallbackMessage(component, call_type, &command_id, &count)) != WM_NULL) {
        flag = addCallback(component, propagate, message, command_id, call_func, call_data);
        command_id = 0;
    }

    return flag;
}

spBool spPrimitiveRemoveCallbackArch(spComponent component, spCallbackType call_type,
                                     spCallbackFunc call_func, void *call_data)
{
    int count = 0;
    spBool flag = SP_FALSE;
    UINT message;
    UINT command_id = 0;
    
    while ((message = getCallbackMessage(component, call_type, &command_id, &count)) != WM_NULL) {
        flag = removeCallback(component, message, command_id, call_func, call_data);
        command_id = 0;
    }

    return flag;
}

spCallbackReason spPrimitiveGetCallbackReasonArch(spComponent component)
{
    spCallbackReason call_reason = SP_CR_NONE;
    UINT sub_message;

    spDebug(100, "spPrimitiveGetCallbackReasonArch", "message = %x\n", SpPrimitiveArch(component).message);
    
    if (SpPrimitiveArch(component).message != WM_NULL) {
        switch (SpPrimitiveArch(component).message) {
          case WM_CUT:
          case WM_PASTE:
          case WM_COPY:
          case WM_CLEAR:
          case WM_UNDO:
            call_reason = SP_CR_VALUE_CHANGED;
            break;
            
          case WM_COMMAND:
            if (spIsSubClass(component, SpList) == SP_TRUE) {
                sub_message = HIWORD(SpPrimitiveArch(component).wParam);
                if (sub_message == LBN_DBLCLK) {
                    call_reason = SP_CR_ACTIVATE;
                } else {
                    call_reason = SP_CR_VALUE_CHANGED;
                }
            } else if (spIsText(component) == SP_TRUE) {
                sub_message = HIWORD(SpPrimitiveArch(component).wParam);
                if (sub_message == EN_SETFOCUS) {
                    call_reason = SP_CR_FOCUS_IN;
                } else if (sub_message == EN_KILLFOCUS) {
                    call_reason = SP_CR_FOCUS_OUT;
                } else {
                    call_reason = SP_CR_VALUE_CHANGED;
                }
            } else if (streq(SpGetName(component), "yesButton")) {
                call_reason = SP_CR_YES;
            } else if (streq(SpGetName(component), "noButton")) {
                call_reason = SP_CR_NO;
            } else if (streq(SpGetName(component), "retryButton")) {
                call_reason = SP_CR_RETRY;
            } else {
                call_reason = SP_CR_ACTIVATE;
            }
            break;
          case WM_KEYDOWN:
            if (spIsSubClass(component, SpTextField) == SP_TRUE) {
                call_reason = SP_CR_ACTIVATE;
                break;
            }
          case WM_CHAR:
          case WM_IME_CHAR:
            call_reason = SP_CR_KEY_PRESS;
            break;
          case WM_KEYUP:
            call_reason = SP_CR_KEY_RELEASE;
            break;
          case WM_LBUTTONDOWN:
            call_reason = SP_CR_LBUTTON_PRESS;
            break;
          case WM_RBUTTONDOWN:
            call_reason = SP_CR_RBUTTON_PRESS;
            break;
          case WM_MBUTTONDOWN:
            call_reason = SP_CR_MBUTTON_PRESS;
            break;
          case WM_LBUTTONUP:
            call_reason = SP_CR_LBUTTON_RELEASE;
            break;
          case WM_RBUTTONUP:
            call_reason = SP_CR_RBUTTON_RELEASE;
            break;
          case WM_MBUTTONUP:
            call_reason = SP_CR_MBUTTON_RELEASE;
            break;
          case WM_MOUSEMOVE:
            if (SpPrimitiveArch(component).wParam & MK_LBUTTON) {
                call_reason = SP_CR_LBUTTON_MOTION;
            } else if (SpPrimitiveArch(component).wParam & MK_RBUTTON) {
                call_reason = SP_CR_RBUTTON_MOTION;
            } else if (SpPrimitiveArch(component).wParam & MK_MBUTTON) {
                call_reason = SP_CR_MBUTTON_MOTION;
            } else {
                call_reason = SP_CR_POINTER_MOTION;
            }
            break;
          case 0x020A/*WM_MOUSEWHEEL*/:
            if (spIsSubClass(component, SpCanvas) == SP_TRUE) {
                if (SpPrimitiveArch(component).wParam & MK_CONTROL) {
                    call_reason = SP_CR_ZOOM;
                } else if (spIsScrollCanvas(component) == SP_TRUE) {
                    call_reason = SP_CR_SCROLL;
                } else {
                    call_reason = SP_CR_WHEEL;
                }
            } else {
                call_reason = SP_CR_VALUE_CHANGED;
            }
            spDebug(100, "spPrimitiveGetCallbackReasonArch", "WM_MOUSEWHEEL: reason = %d\n", call_reason);
            break;
          case 0x020E/*WM_MOUSEHWHEEL*/:
            if (spIsSubClass(component, SpCanvas) == SP_TRUE) {
                if (spIsScrollCanvas(component) == SP_TRUE) {
                    call_reason = SP_CR_SCROLL;
                } else {
                    call_reason = SP_CR_WHEEL;
                }
            } else {
                call_reason = SP_CR_UNKNOWN;
            }
            spDebug(100, "spPrimitiveGetCallbackReasonArch", "WM_MOUSEHWHEEL: reason = %d\n", call_reason);
            break;
          case WM_VSCROLL:
          case WM_HSCROLL:
            if (spIsSubClass(component, SpCanvas) == SP_TRUE) {
                call_reason = SP_CR_SCROLL;
            } else {
                call_reason = SP_CR_VALUE_CHANGED;
            }
            break;
          case WM_PAINT:
            call_reason = SP_CR_EXPOSE;
            break;
          case WM_SIZE:
            call_reason = SP_CR_RESIZE;
            break;
          case WM_TASKTRAY:
            if (SpPrimitiveArch(component).lParam == WM_LBUTTONDOWN) {
                call_reason = SP_CR_LBUTTON_PRESS;
            } else if (SpPrimitiveArch(component).lParam == WM_MBUTTONDOWN) {
                call_reason = SP_CR_MBUTTON_PRESS;
            } else if (SpPrimitiveArch(component).lParam == WM_RBUTTONDOWN) {
                call_reason = SP_CR_RBUTTON_PRESS;
            }
            break;
          case WM_CLOSE:
            call_reason = SP_CR_CLOSE;
            break;
          case WM_DESTROY:
            call_reason = SP_CR_DESTROY;
            break;
          case WM_SETFOCUS:
            call_reason = SP_CR_FOCUS_IN;
            break;
          case WM_KILLFOCUS:
            call_reason = SP_CR_FOCUS_OUT;
            break;
          default:
            call_reason = SP_CR_UNKNOWN;
            break;
        }
    }
    
    return call_reason;
}

spBool spPrimitiveGetCallbackMousePositionArch(spComponent component, int *x, int *y)
{
    short lx = 0, ly = 0;
    POINT point;
    
    if ((x != NULL || y != NULL)) {
        switch (SpPrimitiveArch(component).message) {
          case WM_LBUTTONDOWN:
          case WM_RBUTTONDOWN:
          case WM_MBUTTONDOWN:
          case WM_LBUTTONUP:
          case WM_RBUTTONUP:
          case WM_MBUTTONUP:
          case WM_MOUSEMOVE:
            if (x != NULL) {
                lx = LOWORD(SpPrimitiveArch(component).lParam);
                spDebug(100, "spPrimitiveGetCallbackMousePositionArch",
                        "lx = %d\n", lx);
                *x = lx;
            }
            if (y != NULL) {
                ly = HIWORD(SpPrimitiveArch(component).lParam);
                spDebug(100, "spPrimitiveGetCallbackMousePositionArch",
                        "ly = %d\n", ly);
                *y = ly;
            }
            break;
          case WM_TASKTRAY:
            if (GetCursorPos(&point)) {
                if (x != NULL) {
                    *x = point.x;
                }
                if (y != NULL) {
                    *y = point.y;
                }
            } else {
                return SP_FALSE;
            }
            break;
          default:
            return SP_FALSE;
        }
        
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

spBool spPrimitiveGetCallbackWheelValueArch(spComponent component, double *delta_x, double *delta_y)
{
    short dist;
    double delta;
    spBool shift_key_flag = SP_FALSE;

    if (SpPrimitiveArch(component).message != 0x020A/*WM_MOUSEWHEEL*/
        && SpPrimitiveArch(component).message != 0x020E/*WM_MOUSEHWHEEL*/) {
        return SP_FALSE;
    }
    
    dist = HIWORD(SpPrimitiveArch(component).wParam);
    delta = (double)dist / 120.0;

    if (GetAsyncKeyState(VK_SHIFT) & 0x8000) {
        shift_key_flag = SP_TRUE;
    }
    if (spIsSubClass(component, SpCanvas) == SP_TRUE
        && SpCanvasPart(component).wheel_target_horizontal == SP_TRUE) {
        shift_key_flag = !shift_key_flag;
    }
    
    if (delta_x != NULL) {
        if ((SpPrimitiveArch(component).message == 0x020E && shift_key_flag == SP_FALSE)) {
            *delta_x = delta;
        } else if ((SpPrimitiveArch(component).message == 0x020A && shift_key_flag == SP_TRUE)) {
            *delta_x = -delta;
        } else {
            *delta_x = 0.0;
        }
        if (spIsSubClass(component, SpCanvas) == SP_TRUE
            && SpCanvasPart(component).scroll_left_by_wheel_down == SP_TRUE) {
            *delta_x *= -1.0;
        }
    }
    if (delta_y != NULL) {
        if ((SpPrimitiveArch(component).message == 0x020A && shift_key_flag == SP_FALSE)
            || (SpPrimitiveArch(component).message == 0x020E && shift_key_flag == SP_TRUE)) {
            *delta_y = delta;
        } else {
            *delta_y = 0.0;
        }
    }
    
    return SP_TRUE;
}

static spKeySym convertKeySym(int key_sym)
{
    if ((key_sym >= '0' && key_sym <= '9')
        || (key_sym >= 'A' && key_sym <= 'Z')
        || (key_sym >= 0x60 && key_sym <= 0x6F)
        || key_sym == VK_SPACE) {
        return SPK_Character;
    } else {
        switch (key_sym) {
          case VK_BACK:
            return SPK_BackSpace;
          case VK_TAB:
            return SPK_Tab;
          case VK_CLEAR:
            return SPK_Clear;
          case VK_RETURN:
            return SPK_Return;
          case VK_PAUSE:        
            return SPK_Pause;
          case VK_SCROLL:
            return SPK_ScrollLock;
          case VK_ESCAPE:       
            return SPK_Escape;
          case VK_DELETE:       
            return SPK_Delete;
          case VK_SHIFT:
            return SPK_Shift;
          case VK_CONTROL:      
            return SPK_Control;
          case VK_CAPITAL:
            return SPK_CapsLock;
          case VK_HOME:         
            return SPK_Home;
          case VK_LEFT:         
            return SPK_Left;
          case VK_UP:           
            return SPK_Up;
          case VK_RIGHT:        
            return SPK_Right;
          case VK_DOWN:         
            return SPK_Down;
          case VK_MENU:         
            return SPK_Alt;
          case VK_PRIOR:        
            return SPK_Prior;
          case VK_NEXT:         
            return SPK_Next;
          case VK_END:          
            return SPK_End;
          case VK_SELECT:       
            return SPK_Select;
          case VK_PRINT:        
            return SPK_Print;
          case VK_EXECUTE:      
            return SPK_Execute;
          case VK_INSERT:       
            return SPK_Insert;
          case VK_CANCEL:         
            return SPK_Cancel;
          case VK_HELP:         
            return SPK_Help;
          case VK_NUMLOCK:      
            return SPK_NumLock;
            
          case VK_F1:           
            return SPK_F1;
          case VK_F2:           
            return SPK_F2;
          case VK_F3:           
            return SPK_F3;
          case VK_F4:           
            return SPK_F4;
          case VK_F5:           
            return SPK_F5;
          case VK_F6:           
            return SPK_F6;
          case VK_F7:           
            return SPK_F7;
          case VK_F8:           
            return SPK_F8;
          case VK_F9:           
            return SPK_F9;
          case VK_F10:          
            return SPK_F10;
          case VK_F11:          
            return SPK_F11;
          case VK_F12:         
            return SPK_F12;
          case VK_F13:          
            return SPK_F13;
          case VK_F14:          
            return SPK_F14;
          case VK_F15:          
            return SPK_F15;
          case VK_F16:          
            return SPK_F16;
          case VK_F17:          
            return SPK_F17;
          case VK_F18:          
            return SPK_F18;
          case VK_F19:          
            return SPK_F19;
          case VK_F20:          
            return SPK_F20;
          case VK_F21:          
            return SPK_F21;
          case VK_F22:          
            return SPK_F22;
          case VK_F23:          
            return SPK_F23;
          case VK_F24:          
            return SPK_F24;
          default:
            break;
        }
    }

    return SPK_Unknown;
}

spBool spPrimitiveGetCallbackKeySymArch(spComponent component, spKeySym *key_sym)
{
    switch (SpPrimitiveArch(component).message) {
      case WM_KEYDOWN:
      case WM_KEYUP:
        *key_sym = convertKeySym((int)SpPrimitiveArch(component).wParam);
        spDebug(100, "spPrimitiveGetCallbackKeySymArch", "key_sym = %d\n", *key_sym);
        return SP_TRUE;
        
      case WM_CHAR:
      case WM_IME_CHAR:
        *key_sym = SPK_Character;
        return SP_TRUE;
        
      default:
        break;
    }
    
    return SP_FALSE;
}

int spPrimitiveGetCallbackKeyStringArch(spComponent component,
                                        char *buf, int buf_size, spBool *overflow)
{
    int len;
    unsigned char c1, c2;
    char string[4];
    
    switch (SpPrimitiveArch(component).message) {
      case WM_IME_CHAR:
      case WM_CHAR:
        if (SpPrimitiveArch(component).message == WM_IME_CHAR) {
            c1 = (unsigned char)(0xff &
                                ((WORD)SpPrimitiveArch(component).wParam >> 8));
            c2 = (unsigned char)(0xff & SpPrimitiveArch(component).wParam);

            if (c1 != NUL && c1 > 0x7f) {
                sprintf(string, "%c%c", c1, c2);
            } else {
                sprintf(string, "%c", c2);
            }
        } else {
            sprintf(string, "%c", (unsigned char)SpPrimitiveArch(component).wParam);
        }
        if ((int)strlen(string) >= buf_size) {
            if (overflow != NULL) {
                *overflow = SP_TRUE;
            }
            len = buf_size - 1;
            strncpy(buf, string, len);
        } else {
            if (overflow != NULL) {
                *overflow = SP_FALSE;
            }
            strcpy(buf, string);
            len = (int)strlen(string);
        }
        return len;
      default:
        break;
    }
    
    return -1;
}

void spPrimitivePartInitArch(spComponent component)
{
    if (component != NULL) {
        SpPrimitiveArch(component).hwnd = NULL;
        SpPrimitiveArch(component).hmenu = NULL;
        
        SpPrimitiveArch(component).hdc = NULL;
        SpPrimitiveArch(component).memdc = NULL;
        SpPrimitiveArch(component).hbitmap = NULL;
        SpPrimitiveArch(component).hbrush = NULL;
        SpPrimitiveArch(component).hcursor = NULL;
        
        SpPrimitiveArch(component).message = WM_NULL;
        SpPrimitiveArch(component).wParam = 0;
        SpPrimitiveArch(component).lParam = 0;

        SpPrimitiveArch(component).num_accel = 0;
        SpPrimitiveArch(component).num_accel_buffer = 0;
        SpPrimitiveArch(component).accels = NULL;
        SpPrimitiveArch(component).haccel = NULL;
        
        SpPrimitiveArch(component).realize_flag = SP_FALSE;
        SpPrimitiveArch(component).htooltip = NULL;
    }
    
    return;
}

void spFreeDrawingResourcesWin(spComponent component)
{
    if (SpPrimitiveArch(component).hbrush != NULL) {
        DeleteObject(SpPrimitiveArch(component).hbrush);
        SpPrimitiveArch(component).hbrush = NULL;
    }
    
    if (SpPrimitiveArch(component).memdc != NULL) {
        DeleteDC(SpPrimitiveArch(component).memdc);
        SpPrimitiveArch(component).memdc = NULL;
    }
    if (SpPrimitiveArch(component).hbitmap != NULL) {
        DeleteObject(SpPrimitiveArch(component).hbitmap);
        SpPrimitiveArch(component).hbitmap = NULL;
    }

    return;
}

static void freeResources(spComponent component)
{
    spComponent child;
    
    if (spIsPrimitive(component) == SP_FALSE) return;

    spDebug(80, "freeResources", "in: %s\n", spGetName(component));
    
    child = SpGetChild(component);
    while (child != NULL) {
        freeResources(child);
        child = SpGetNextComponent(child);
    }

    spFreeDrawingResourcesWin(component);

    if (SpPrimitiveArch(component).accels != NULL) {
        xfree(SpPrimitiveArch(component).accels);
        SpPrimitiveArch(component).accels = NULL;
        SpPrimitiveArch(component).num_accel = 0;
        SpPrimitiveArch(component).num_accel_buffer = 0;
    }

    if (SpPrimitiveArch(component).haccel != NULL) {
        DestroyAcceleratorTable(SpPrimitiveArch(component).haccel);
        SpPrimitiveArch(component).haccel = NULL;
    }

    spDebug(80, "freeResources", "done\n");
    
    return;
}

void spPrimitiveDestroyArch(spComponent component)
{
    HWND hwnd;
    HMENU hmenu;
    
    spDebug(10, "spPrimitiveDestroyArch", "in: %s\n", spGetName(component));

    freeResources(component);
    
    hwnd = SpPrimitiveArch(component).hwnd;
    hmenu = SpPrimitiveArch(component).hmenu;

    if (spIsSubClass(component, SpMenuItem) == SP_TRUE
        && spIsPrimitive(SpGetParent(component)) == SP_TRUE
        && SpParentPrimitiveArch(component).hmenu != NULL) {
        spDebug(10, "spPrimitiveDestroyArch", "menu destroy\n");
        DeleteMenu(SpParentPrimitiveArch(component).hmenu,
                   SpComponentPart(component).component_id,
                   MF_BYCOMMAND);
        DrawMenuBar(SpPrimitiveArch(SpGetWindow(component)).hwnd);
    } else if (spIsSubClass(component, SpMenu) == SP_TRUE
               && hmenu != NULL) {
        DestroyMenu(hmenu);
    } else if (spIsSubClass(component, SpImage) == SP_TRUE) {
    } else {
        if (SpPrimitiveArch(component).message != WM_DESTROY) {
            if (hwnd != NULL) {
                spDebug(30, "spPrimitiveDestroyArch", "call DestroyWindow\n");
                /*SpPrimitiveArch(component).hwnd = NULL;*/
                DestroyWindow(hwnd);
                spDebug(30, "spPrimitiveDestroyArch", "call DestroyWindow done\n");
            }
        }
        
    }

    SpPrimitiveArch(component).hwnd = NULL;
    SpPrimitiveArch(component).hmenu = NULL;
    
    removeAllCallback(component);
    
    spDebug(30, "spPrimitiveDestroyArch", "done: %s\n", spGetName(component));
    
    return;
}

void spPrimitiveMapArch(spComponent component)
{
    if (spIsSubClass(component, SpMenuItem) == SP_TRUE
        && spIsPrimitive(SpGetParent(component)) == SP_TRUE
        && SpParentPrimitiveArch(component).hmenu != NULL) {
#if 0
        EnableMenuItem(SpParentPrimitiveArch(component).hmenu,
                       SpComponentPart(component).component_id,
                       MF_BYCOMMAND | MF_ENABLED);
#else
        spCreateMenuItemWin(component);
#endif
        DrawMenuBar(SpPrimitiveArch(SpGetWindow(component)).hwnd);
    } else if (SpPrimitiveArch(component).hwnd != NULL) {
        ShowWindow(SpPrimitiveArch(component).hwnd, SW_RESTORE);
        UpdateWindow(SpPrimitiveArch(component).hwnd);
    }
    
    return;
}

void spPrimitiveUnmapArch(spComponent component)
{
    if (spIsSubClass(component, SpMenuItem) == SP_TRUE
        && spIsPrimitive(SpGetParent(component)) == SP_TRUE
        && SpParentPrimitiveArch(component).hmenu != NULL) {
#if 0
        EnableMenuItem(SpParentPrimitiveArch(component).hmenu,
                       SpComponentPart(component).component_id,
                       MF_BYCOMMAND | MF_GRAYED);
#else
        DeleteMenu(SpParentPrimitiveArch(component).hmenu,
                   SpComponentPart(component).component_id,
                   MF_BYCOMMAND);
#endif
        DrawMenuBar(SpPrimitiveArch(SpGetWindow(component)).hwnd);
    } else if (SpPrimitiveArch(component).hwnd != NULL) {
        ShowWindow(SpPrimitiveArch(component).hwnd, SW_HIDE);
    }
    
    return;
}

void spPrimitiveScaleDefaultSizeArch(spComponent component)
{
    if (SpComponentPart(component).x != 0) SpComponentPart(component).x = spScaleXWin(sp_toplevel, SpComponentPart(component).x);
    if (SpComponentPart(component).y != 0) SpComponentPart(component).y = spScaleYWin(sp_toplevel, SpComponentPart(component).y);
    if (SpComponentPart(component).width != 0) SpComponentPart(component).width = spScaleXWin(sp_toplevel, SpComponentPart(component).width);
    if (SpComponentPart(component).height != 0) SpComponentPart(component).height = spScaleYWin(sp_toplevel, SpComponentPart(component).height);
    if (SpComponentPart(component).current_width != 0) SpComponentPart(component).current_width = spScaleXWin(sp_toplevel, SpComponentPart(component).current_width);
    if (SpComponentPart(component).current_height != 0) SpComponentPart(component).current_height = spScaleYWin(sp_toplevel, SpComponentPart(component).current_height);
    if (SpComponentPart(component).require_width != 0) SpComponentPart(component).require_width = spScaleXWin(sp_toplevel, SpComponentPart(component).require_width);
    if (SpComponentPart(component).require_height != 0) SpComponentPart(component).require_height = spScaleYWin(sp_toplevel, SpComponentPart(component).require_height);
    
    if (SpComponentPart(component).margin_width != 0) SpComponentPart(component).margin_width = spScaleXWin(sp_toplevel, SpComponentPart(component).margin_width);
    if (SpComponentPart(component).margin_height != 0) SpComponentPart(component).margin_height = spScaleYWin(sp_toplevel, SpComponentPart(component).margin_height);

    if (SpComponentPart(component).spacing != 0) SpComponentPart(component).spacing = spScaleXWin(sp_toplevel, SpComponentPart(component).spacing);
    
    return;
}

spBool spPrimitiveSetSizeArch(spComponent component, int width, int height)
{
    RECT rect;
    int hoffset = 0;
    int x, y;
    int window_width, window_height;
    
    if (SpPrimitiveArch(component).hwnd != NULL) {
        spDebug(100, "spPrimitiveSetSizeArch", "in: %s\n", spGetTitle(component));
        
        if (spIsFrame(component) == SP_TRUE) {
            if (spIsVisible(SpFramePart(component).status_bar) == SP_TRUE) {
                int h;
                if (spGetSize(SpFramePart(component).status_bar, NULL, &h) == SP_TRUE) {
                    height += h;
                }
            }
            
            /* get current window size */
            GetWindowRect(SpPrimitiveArch(component).hwnd, &rect);
            window_width = rect.right - rect.left;
            window_height = rect.bottom - rect.top;

            /* set exact window width */
            GetClientRect(SpPrimitiveArch(component).hwnd, &rect);
            window_width = width + window_width - (rect.right - rect.left);
            window_height = height + window_height - (rect.bottom - rect.top);

            /* bRepaint arg must be TRUE, otherwise setting window size in the program results in dirty */
            MoveWindow(SpPrimitiveArch(component).hwnd,
                       SpComponentPart(component).x, SpComponentPart(component).y,
                       window_width, window_height, TRUE/*FALSE*//* 2007/7 modified */);
            
            spDebug(50, "spPritimiveSetSizeArch", "input size: %d, %d, updated size = %d, %d\n",
                    width, height, window_width, window_height);
        } else {
            if (spEqClass(component, SpLabel) == SP_TRUE) {
                hoffset = spScaleYWin(sp_toplevel, 4);
            }
            if (spIsSubClass(component, SpContainer) == SP_TRUE
                && SpContainerPart(component).title_height > 0) {
                hoffset = -SpContainerPart(component).title_height;
            }
            if (spIsSubClass(SpGetParent(component), SpContainer) == SP_TRUE
                && SpContainerPart(SpGetParent(component)).title_height > 0) {
                hoffset = SpContainerPart(SpGetParent(component)).title_height;
                height += hoffset; /* for MoveWindow arg of `height - hoffset' */
            }
            
            if (spIsTabBox(SpGetParent(component)) == SP_TRUE) {
                x = SpComponentPart(component).x + SpParentComponentPart(component).x;
                y = SpComponentPart(component).y + hoffset + SpParentComponentPart(component).y;
            } else {
                x = SpComponentPart(component).x;
                y = SpComponentPart(component).y + hoffset;
            }
            
            spDebug(80, "spPrimitiveSetSizeArch", "before MoveWindow\n");
            
            if (spIsSubClass(component, SpComboBox) == SP_TRUE) {
                /* get current window size */
                GetWindowRect(SpPrimitiveArch(component).hwnd, &rect);
                window_width = rect.right - rect.left;
                window_height = rect.bottom - rect.top;
                
                MoveWindow(SpPrimitiveArch(component).hwnd,
                           x, y, width, window_height, TRUE);
            } else if (spIsToolBar(component) == SP_TRUE || spIsToolItem(component) == SP_TRUE) {
                /* set window size of tool bar and tool item without repainting to prevent from flickering */
                MoveWindow(SpPrimitiveArch(component).hwnd,
                           x, y, width, height - hoffset, FALSE);
            } else {
                /* set window size */
                MoveWindow(SpPrimitiveArch(component).hwnd,
                           x, y, width, height - hoffset, TRUE);
            }

            spDebug(80, "spPrimitiveSetSizeArch", "after MoveWindow\n");
            
            if (spIsVisible(component) == SP_TRUE
                && SpPrimitiveArch(SpGetWindow(component)).realize_flag == SP_TRUE) {
                spDebug(10, "spPrimitiveSetSizeArch", "refresh canvas\n");

                if (spIsCanvas(component) == SP_TRUE) {
                    spGetClientSize(component, NULL, NULL);
                    spRedrawCanvas(component);
                } else if (spIsToolBar(component) == SP_TRUE || spIsToolItem(component) == SP_TRUE) {
                    /* do nothing */
                } else if (SpPrimitiveArch(component).hwnd != NULL) {
                    InvalidateRect(SpPrimitiveArch(component).hwnd, NULL, TRUE);
                }
            }
        }
        
        spDebug(100, "spPrimitiveSetSizeArch", "done: %s\n", spGetTitle(component));
        
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spPrimitiveGetSizeArch(spComponent component, int *width, int *height)
{
    RECT rect;
    
    if (SpPrimitiveArch(component).hwnd != NULL) {
        GetWindowRect(SpPrimitiveArch(component).hwnd, &rect);
        *width = rect.right - rect.left;
        if (spIsSubClass(component, SpComboBox) == SP_TRUE) {
            *height = SP_DEFAULT_TEXT_HEIGHT;
        } else {
            *height = rect.bottom - rect.top;

            if (spIsWindow(component) == SP_TRUE) {
                int h;

                if (spIsVisible(SpFramePart(component).status_bar) == SP_TRUE) {
                    if (spGetSize(SpFramePart(component).status_bar, NULL, &h) == SP_TRUE) {
                        *height -= h;
                    }
                }
            } else if (spIsSubClass(component, SpContainer) == SP_TRUE
                       && SpComponentPart(component).top_offset > 0) {
                *height -= SpComponentPart(component).top_offset;
            }
        }
        
        spDebug(100, "spPrimitiveGetSizeArch", "%s: size = %d, %d\n", spGetName(component), *width, *height);
    
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

spBool spPrimitiveGetClientSizeArch(spComponent component, int *width, int *height)
{
    RECT rect;
    
    if (SpPrimitiveArch(component).hwnd != NULL) {
        GetClientRect(SpPrimitiveArch(component).hwnd, &rect);
        *width = rect.right - rect.left;
        *height = rect.bottom - rect.top;

        if (spIsWindow(component) == SP_TRUE) {
            int h;
            if (spIsVisible(SpFramePart(component).status_bar) == SP_TRUE) {
                if (spGetSize(SpFramePart(component).status_bar, NULL, &h) == SP_TRUE) {
                    *height -= h;
                }
            }
        } else if (spIsSubClass(component, SpContainer) == SP_TRUE
                   && SpComponentPart(component).top_offset > 0) {
            *height -= SpComponentPart(component).top_offset;
        }
        
        spDebug(100, "spPrimitiveGetClientArch", "%s: size = %d, %d\n", spGetName(component), *width, *height);
    
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

static void setContainerSensitive(spComponent component, spBool flag)
{
    spComponent next;

    next = SpGetChild(component);
    while (next != NULL) {
        if (spIsCreated(next) == SP_TRUE
            && spIsPrimitive(next) == SP_TRUE) {
            if (flag == SP_FALSE) {
                spPrimitiveSetSensitiveArch(next, SP_FALSE);
            } else {
                spPrimitiveSetSensitiveArch(next, spIsSensitive(next));
            }

        }
        next = SpGetNextComponent(next);
    }

    return;
}

spBool spPrimitiveSetSensitiveArch(spComponent component, spBool flag)
{
    if (spIsSubClass(component, SpMenuItem) == SP_TRUE) {
        if (spIsCreated(SpGetParent(component)) == SP_TRUE
            && SpParentPrimitiveArch(component).hmenu != NULL) {
            EnableMenuItem(SpParentPrimitiveArch(component).hmenu,
                           SpComponentPart(component).component_id,
                           MF_BYCOMMAND | (flag == SP_TRUE ? MF_ENABLED : MF_GRAYED));
#if 0 /* 1999/6/16 */
            if (SpPrimitiveArch(SpGetWindow(component)).realize_flag == SP_TRUE) {
                DrawMenuBar(SpPrimitiveArch(SpGetWindow(component)).hwnd);
            }
#endif
            return SP_TRUE;
        }
    } else if (spIsSubClass(component, SpToolItem) == SP_TRUE) {
        if (SpParentPrimitiveArch(component).hwnd != NULL) {
            SendMessage(SpParentPrimitiveArch(component).hwnd, TB_ENABLEBUTTON,
                        (WPARAM)SpComponentPart(component).component_id,
                        (LPARAM)MAKELONG((flag == SP_TRUE ? TRUE : FALSE), 0));
            return SP_TRUE;
        }
    } else {
        if (spIsSubClass(component, SpBox) == SP_TRUE || spIsTabBox(component) == SP_TRUE) {
            setContainerSensitive(component, flag);
        }
        
        if (SpPrimitiveArch(component).hwnd != NULL) {
            EnableWindow(SpPrimitiveArch(component).hwnd, (flag == SP_TRUE ? TRUE : FALSE));
            if (spIsSlider(component) == SP_TRUE && flag == SP_TRUE) {
                spSetSliderValueArch(component);
            }
            return SP_TRUE;
        }
    }
    
    return SP_FALSE;
}

spBool spBeginPaintWin(spComponent component, PAINTSTRUCT *paintstruct)
{
    if (BeginPaint(SpPrimitiveArch(component).hwnd, paintstruct) != NULL) {
        return SP_TRUE;
    }

    return SP_FALSE;
}

spBool spEndPaintWin(spComponent component, PAINTSTRUCT *paintstruct)
{
    if (EndPaint(SpPrimitiveArch(component).hwnd, paintstruct)) {
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool processTimerMessage(spComponent window, spTimerId timer_id)
{
    spTimerList timer;
    
    if (spIsFrame(window) == SP_FALSE) return SP_FALSE;
    
    spDebug(80, "processCallback", "received WM_TIMER message\n");

    if ((timer = spFindTimerList(window, timer_id)) == NULL) {
        return SP_FALSE;
    }

    if (timer->kill_flag == SP_FALSE
        && timer->call_func(timer->frame, timer->call_data) == SP_FALSE) {
        timer->kill_flag = SP_TRUE;
    }

    if (timer->kill_flag == SP_TRUE) {
        spRemoveTimerWin(timer);
        spRemoveTimerList(timer->frame, timer);
    }
    
    return SP_TRUE;
}

static spBool processHelpMessage(spComponent window, LPHELPINFO lphi)
{
    spComponent menu, child;
    
    if (lphi->iContextType == HELPINFO_MENUITEM) {
        spDebug(10, "processCallback",
                "WM_HELP for menu received: iCtrlId = %d, hItemHandle = %ld, dwContextId = %ld\n",
                lphi->iCtrlId, lphi->hItemHandle, lphi->dwContextId);
        if ((menu = getCurrentMenu(window, (HMENU)lphi->hItemHandle, (UINT)lphi->iCtrlId, 0)) != NULL) {
            spDebug(10, "processCallback", "found menu\n");
            spDisplayHelpArch(menu, NULL);
        }
        return SP_TRUE;
    } else if (lphi->iContextType == HELPINFO_WINDOW) {
        spDebug(10, "processCallback",
                "WM_HELP for window received: iCtrlId = %d, hItemHandle = %ld, dwContextId = %ld\n",
                lphi->iCtrlId, lphi->hItemHandle, lphi->dwContextId);
        if ((child = getCurrentComponent(window, (HWND)lphi->hItemHandle)) != NULL) {
            spDebug(10, "processCallback",        "found child\n");
            spDisplayHelpArch(child, NULL);
        }
        return SP_TRUE;
    }

    return SP_FALSE;
}

static void processMenuSelectMessage(spComponent window, HMENU hmenu, WPARAM wParam)
{
    spComponent menu;
    
    menu = getCurrentMenu(window, hmenu, LOWORD(wParam), HIWORD(wParam));
        
    if (menu != NULL && SpGetDescription(menu) != NULL) {
        spSetHelpStatusText(window, SpGetDescription(menu));
    } else {
        spSetHelpStatusText(window, "");
    }

    return;
}

static void processSysColorChangeMessage(spComponent window)
{
    spDebug(10, "processCallback", "WM_SYSCOLORCHANGE\n");
    if (spIsCanvas(window) == SP_FALSE) {
        HWND hnext;
        spComponent canvas;
            
        spUpdateSystemGraphics();

        hnext = FindWindowEx(SpPrimitiveArch(window).hwnd, NULL, SP_CANVAS_CLASS_NAME, NULL);
        while (hnext != NULL) {
            canvas = getCurrentComponent(window, hnext);
            if (spIsCanvas(canvas) == SP_TRUE) {
                spDebug(10, "processCallback", "WM_SYSCOLORCHANGE, call spRedrawCanvas\n");
                spRedrawCanvas(canvas);
            }
            hnext = FindWindowEx(SpPrimitiveArch(window).hwnd, hnext, SP_CANVAS_CLASS_NAME, NULL);
        }
    }

    return;
}

static spBool processDropFilesMessage(spComponent window, HDROP hdrop)
{
    if (spIsFrame(window) == SP_TRUE && SpFrameArch(window).drop_call_func != NULL) {
        int i;
        int argc;
        char **argv;
        
        SetForegroundWindow(SpPrimitiveArch(window).hwnd);
            
        argc = DragQueryFile(hdrop, 0xFFFFFFFF, NULL, 0);
        argv = xalloc(argc, char *);
        for (i = 0; i < argc; i++) {
            argv[i] = xalloc(SP_MAX_PATHNAME, char);
            DragQueryFile(hdrop, i, argv[i], SP_MAX_PATHNAME);
        }
            
        SpFrameArch(window).drop_call_func(window, SpFrameArch(window).drop_call_data,
                                           argc, argv);

        for (i = 0; i < argc; i++) xfree(argv[i]);
        xfree(argv);
        DragFinish(hdrop);
        
        return SP_TRUE;
    }
    
    return SP_FALSE;
}

static spBool processCopyDataMessage(spComponent window, PCOPYDATASTRUCT pcds)
{
    if (pcds->dwData == SP_FIND_EXECUTED_APPLICATION_COPYDATA_ID) {
        spDebug(50, "processCallback", "WM_COPYDATA: SP_FIND_EXECUTED_APPLICATION_COPYDATA_ID received\n");
            
        if (pcds->lpData != NULL && spIsFrame(window) == SP_TRUE && SpFrameArch(window).drop_call_func != NULL) {
            int argc;
            char **argv;

            if ((argv = spConvertArgsToArgv(pcds->lpData, NULL, &argc)) != NULL) {
                spDebug(50, "processCallback", "WM_COPYDATA: argc = %d, argv = %s\n", argc, pcds->lpData);
                SpFrameArch(window).drop_call_func(window, SpFrameArch(window).drop_call_data,
                                                   argc, argv);
                spFreeArgv(argc, argv);
            }
        }
        return SP_TRUE;
    }

    return SP_FALSE;
}

static spBool processNotifyMessage(spComponent window, LPARAM lParam)
{
    int index;
    spComponent component;
    spComponent child;
    NMHDR *hdr;
    spBool flag = SP_FALSE;

    hdr = (NMHDR *)lParam;
    
    component = getCurrentComponent(window, hdr->hwndFrom);

    if (spIsTabBox(component) == SP_TRUE) {
        spDebug(80, "processCallback", "WM_NOTIFY: tab, code = %d\n",
                hdr->code);
        if (hdr->code == (UINT)TCN_SELCHANGE) {
            index = spGetSelectedTabIndex(component);
            spMapTabItem(component, index);
        }
        flag = SP_TRUE;
    } else if (hdr->code == (UINT)TTN_NEEDTEXT
               && spIsFrame(window) == SP_TRUE
               && SpFramePart(window).tool_bar != NULL) {
        LPTOOLTIPTEXT ttext = (LPTOOLTIPTEXT)lParam;
            
        child = SpGetChild(SpFramePart(window).tool_bar);
            
        while (child != NULL) {
            if (ttext->hdr.idFrom == (UINT)SpComponentPart(child).component_id) {
                if (!strnone(SpGetDescription(child))) {
                    ttext->lpszText = SpGetDescription(child);
                }
                break;
            }
            child = SpGetNextComponent(child);
        }
        flag = SP_TRUE;
    }

    return flag;
}

static spBool processSetCursorMessage(spComponent window)
{
#if 0
    if (SpPrimitiveArch(window).hcursor != NULL) {
        SetCursor(SpPrimitiveArch(window).hcursor);
    } else {
        SetCursor(LoadCursor(NULL, IDC_ARROW));
    }
    return SP_TRUE;
#else
    if (SpPrimitiveArch(window).hcursor != NULL) {
        SetCursor(SpPrimitiveArch(window).hcursor);
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
#endif
}

static spBool preprocessCanvasWheelMessage(spComponent canvas, UINT message, WPARAM wParam)
{
    int key_sym;
    int value;
    short dist;
    double delta;
    short key;
    double pos;
    HWND hfocus;

#if 0
    if (spIsCanvas(canvas) == SP_FALSE) {
        return SP_FALSE;
    }
#endif

    hfocus = SpPrimitiveArch(canvas).hwnd;
    dist = (short)HIWORD(wParam);
    delta = (double)dist / 120.0;
    key = (short)LOWORD(wParam);
    spDebug(80, "preprocessCanvasWheelMessage",
            "WM_MOUSEWHEEL: dist = %d, delta = %f, key = %d\n", dist, delta, key);
                
    if (spIsScrollCanvas(canvas) == SP_TRUE) {
        if (key == 0 || key == MK_SHIFT) {
            int horizontal_flag;
            message = 0;
            horizontal_flag = ((message == 0x020E/*WM_MOUSEHWHEEL*/ && key == 0)
                               || (message == 0x020A/*WM_MOUSEWHEEL*/ && key == MK_SHIFT));
            if (SpCanvasPart(canvas).wheel_target_horizontal == SP_TRUE) {
                spDebug(80, "preprocessCanvasWheelMessage", "WHEEL_TARGET IS HORIZONTAL\n");
                horizontal_flag = !horizontal_flag;
            }
                        
            if (horizontal_flag) {
                if (message == 0x020A/*WM_MOUSEWHEEL*/) {
                    if (SpCanvasPart(canvas).scroll_left_by_wheel_down == SP_TRUE) {
                        spDebug(80, "preprocessCanvasWheelMessage", "scroll_left_by_wheel_down is TRUE\n");
                    } else {
                        delta *= -1.0;
                    }
                }
                        
                if (SpCanvasPart(canvas).h_scroll == SP_TRUE) {
                    message = WM_HSCROLL;

                    value = SpCanvasPart(canvas).h_value;
                    spMoveCanvasScrollBarValue(&value, delta,
                                               SpCanvasPart(canvas).h_minimum,
                                               SpCanvasPart(canvas).h_maximum,
                                               SpCanvasPart(canvas).h_page_size,
                                               SpCanvasPart(canvas).h_wheel_increment);
                                
                    pos = (double)value / (double)SpCanvasPart(canvas).h_scroll_coef;
                    spDebug(80, "preprocessCanvasWheelMessage",
                            "WM_MOUSEWHEEL: h_value = %d, value = %d, pos = %f\n",
                            SpCanvasPart(canvas).h_value, value, pos);
                    wParam = MAKEWPARAM(SB_THUMBPOSITION, (short)spRound(pos));
                }
            } else {
                if (SpCanvasPart(canvas).v_scroll == SP_TRUE) {
                    message = WM_VSCROLL;

                    value = SpCanvasPart(canvas).v_value;
                    spMoveCanvasScrollBarValue(&value, -delta,
                                               SpCanvasPart(canvas).v_minimum,
                                               SpCanvasPart(canvas).v_maximum,
                                               SpCanvasPart(canvas).v_page_size,
                                               SpCanvasPart(canvas).v_wheel_increment);
                                
                    pos = (double)value / (double)SpCanvasPart(canvas).v_scroll_coef;
                    spDebug(80, "preprocessCanvasWheelMessage",
                            "WM_MOUSEWHEEL: v_value = %d, value = %d, pos = %f\n",
                            SpCanvasPart(canvas).v_value, value, pos);
                    wParam = MAKEWPARAM(SB_THUMBPOSITION, (short)spRound(pos));
                }
            }

            if (message) {
                PostMessage(hfocus, message, wParam, (LPARAM)SpPrimitiveArch(canvas).hwnd);
                    
                return SP_TRUE;
            }
        } else if (!(key & MK_CONTROL)) {
            return SP_TRUE;
        }
    } else if (message == 0x020A/*WM_MOUSEWHEEL*/
               && SpCanvasPart(canvas).accept_wheel_event == SP_FALSE) {
        if (dist > 0) {
            key_sym = VK_PRIOR;
        } else {
            key_sym = VK_NEXT;
        }
        spDebug(80, "preprocessCanvasWheelMessage",
                "WM_MOUSEWHEEL: wheel event is converted into key event, keysym = %d\n", key_sym);
        PostMessage(hfocus, WM_KEYDOWN, key_sym, 0L);
        PostMessage(hfocus, WM_KEYUP, key_sym, 0L);
        return SP_TRUE;
    }

    return SP_FALSE;
}

static spBool processScrollMessage(spComponent scroll_bar, WPARAM wParam)
{
    UINT sub_message = 0;
    
    spDebug(80, "processScrollMessage", "in\n");
    
    if (spIsSubClass(scroll_bar, SpTrackBar) == SP_TRUE) {
        if (SpPrimitiveArch(scroll_bar).message == WM_HSCROLL) {
            sub_message = LOWORD(wParam);
            switch (sub_message) {
              case TB_LINEUP:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value
                        += SpSliderPart(scroll_bar).increment;
                } else {
                    SpSliderPart(scroll_bar).value
                        -= SpSliderPart(scroll_bar).increment;
                }
                break;
              case TB_LINEDOWN:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value
                        -= SpSliderPart(scroll_bar).increment;
                } else {
                    SpSliderPart(scroll_bar).value
                        += SpSliderPart(scroll_bar).increment;
                }
                break;
              case TB_PAGEUP:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value
                        += SpSliderPart(scroll_bar).page_increment;
                } else {
                    SpSliderPart(scroll_bar).value
                        -= SpSliderPart(scroll_bar).page_increment;
                }
                break;
              case TB_PAGEDOWN:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value
                        -= SpSliderPart(scroll_bar).page_increment;
                } else {
                    SpSliderPart(scroll_bar).value
                        += SpSliderPart(scroll_bar).page_increment;
                }
                break;
              case TB_THUMBPOSITION:
              case TB_THUMBTRACK:
                if (sub_message == TB_THUMBTRACK
                    && SpSliderPart(scroll_bar).track_call_on == SP_FALSE) {
                    return SP_TRUE;
                }

                /* get current value */
                SpSliderPart(scroll_bar).value
                    = (int)SendMessage(SpPrimitiveArch(scroll_bar).hwnd,
                                       TBM_GETPOS, (WPARAM)0, (LPARAM)0) * SpSliderPart(scroll_bar).scroll_coef;
                SpSliderPart(scroll_bar).value = spReverseSliderValue(scroll_bar, SpSliderPart(scroll_bar).value);
                break;
              case TB_TOP:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value = SpSliderPart(scroll_bar).maximum;
                } else {
                    SpSliderPart(scroll_bar).value = SpSliderPart(scroll_bar).minimum;
                }
                break;
              case TB_BOTTOM:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value = SpSliderPart(scroll_bar).minimum;
                } else {
                    SpSliderPart(scroll_bar).value = SpSliderPart(scroll_bar).maximum;
                }
                break;
              default:
                return SP_TRUE;
            }
        } else {
            sub_message = LOWORD(wParam);
            switch (sub_message) {
              case TB_LINEUP:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value
                        += SpSliderPart(scroll_bar).increment;
                } else {
                    SpSliderPart(scroll_bar).value
                        -= SpSliderPart(scroll_bar).increment;
                }
                break;
              case TB_LINEDOWN:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value
                        -= SpSliderPart(scroll_bar).increment;
                } else {
                    SpSliderPart(scroll_bar).value
                        += SpSliderPart(scroll_bar).increment;
                }
                break;
              case TB_PAGEUP:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value
                        += SpSliderPart(scroll_bar).page_increment;
                } else {
                    SpSliderPart(scroll_bar).value
                        -= SpSliderPart(scroll_bar).page_increment;
                }
                break;
              case TB_PAGEDOWN:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value
                        -= SpSliderPart(scroll_bar).page_increment;
                } else {
                    SpSliderPart(scroll_bar).value
                        += SpSliderPart(scroll_bar).page_increment;
                }
                break;
              case TB_THUMBPOSITION:
              case TB_THUMBTRACK:
                if (sub_message == TB_THUMBTRACK
                    && SpSliderPart(scroll_bar).track_call_on == SP_FALSE) {
                    return SP_TRUE;
                }

                /* get current value */
                SpSliderPart(scroll_bar).value
                    = (int)SendMessage(SpPrimitiveArch(scroll_bar).hwnd,
                                       TBM_GETPOS, (WPARAM)0, (LPARAM)0) * SpSliderPart(scroll_bar).scroll_coef;
                SpSliderPart(scroll_bar).value = spReverseSliderValue(scroll_bar, SpSliderPart(scroll_bar).value);
                break;
              case TB_TOP:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value = SpSliderPart(scroll_bar).maximum;
                } else {
                    SpSliderPart(scroll_bar).value = SpSliderPart(scroll_bar).minimum;
                }
                break;
              case TB_BOTTOM:
                if (SpSliderPart(scroll_bar).reversed == SP_TRUE) {
                    SpSliderPart(scroll_bar).value = SpSliderPart(scroll_bar).minimum;
                } else {
                    SpSliderPart(scroll_bar).value = SpSliderPart(scroll_bar).maximum;
                }
                break;
              default:
                return SP_TRUE;
            }
        }
        spSetTrackBarValueWin(scroll_bar);
    } else if (spIsSubClass(scroll_bar, SpCanvas) == SP_TRUE) {
        spDebug(80, "processScrollMessage", "in canvas: message = %x\n", SpPrimitiveArch(scroll_bar).message);
        if (SpPrimitiveArch(scroll_bar).message == WM_HSCROLL) {
            if (spUpdateHScrollValueWin(&SpCanvasPart(scroll_bar).h_value, wParam,
                                        SpCanvasPart(scroll_bar).h_minimum, SpCanvasPart(scroll_bar).h_maximum, 
                                        SpCanvasPart(scroll_bar).h_page_size, SpCanvasPart(scroll_bar).h_increment,
                                        SpCanvasPart(scroll_bar).h_page_increment, SpCanvasPart(scroll_bar).h_scroll_coef,
                                        SpCanvasPart(scroll_bar).h_content_factor, SP_TRUE) == SP_TRUE) {
                return SP_TRUE;
            }
            spSetCanvasHScrollBarValueWin(scroll_bar);
        } else {
            if (spUpdateVScrollValueWin(&SpCanvasPart(scroll_bar).v_value, wParam,
                                        SpCanvasPart(scroll_bar).v_minimum, SpCanvasPart(scroll_bar).v_maximum, 
                                        SpCanvasPart(scroll_bar).v_page_size, SpCanvasPart(scroll_bar).v_increment,
                                        SpCanvasPart(scroll_bar).v_page_increment, SpCanvasPart(scroll_bar).v_scroll_coef,
                                        SpCanvasPart(scroll_bar).v_content_factor, SP_TRUE) == SP_TRUE) {
                return SP_TRUE;
            }
            spSetCanvasVScrollBarValueWin(scroll_bar);
        }
        SpComponentPart(scroll_bar).call_reason = SP_CR_SCROLL;
        spRedrawCanvas(scroll_bar);
    } else if (spIsSubClass(scroll_bar, SpScrollBar) == SP_TRUE) {
        if (spUpdateScrollValueWin(&SpSliderPart(scroll_bar).value, SpPrimitiveArch(scroll_bar).message, wParam,
                                   SpSliderPart(scroll_bar).minimum, SpSliderPart(scroll_bar).maximum, 
                                   SpSliderPart(scroll_bar).page_size, SpSliderPart(scroll_bar).increment,
                                   SpSliderPart(scroll_bar).page_increment, SpSliderPart(scroll_bar).scroll_coef,
                                   1.0, SpSliderPart(scroll_bar).track_call_on) == SP_TRUE) {
            return SP_TRUE;
        }
        spSetSliderValueWin(scroll_bar);
    }

    spDebug(80, "processScrollMessage", "done\n");
    
    return SP_FALSE;
}

static spBool processCallback(spComponent window)
{
    long k;
    long component_id;
    long num_callback;
    spCallback *callbacks;
    spBool flag = SP_FALSE;
    spBool need_capture = SP_FALSE;
    spComponent child;
    spComponent parent_window;
    spComponent target_component;
    spComponent scroll_bar = NULL;
    RECT rect;
    POINT point;
    UINT message;
    UINT sub_message = 0;
    WPARAM wParam;
    LPARAM lParam;
    spCloseStyle window_close_style = SP_NO_CLOSE;
    static spComponent capture_window = NULL;
    static int capture_dx = -1, capture_dy = -1;

    if (window == NULL) return SP_FALSE;

    message = SpPrimitiveArch(window).message;
    wParam = SpPrimitiveArch(window).wParam;
    lParam = SpPrimitiveArch(window).lParam;
    
    /*spDebug(100, "processCallback", "message = %d\n", message);*/
    
    if (message == WM_SIZE) {
        spDebug(80, "processCallback", "WM_SIZE: %s\n", SpGetClassName(window));
        if (SpPrimitiveArch(window).hwnd == SpPrimitiveArch(SpGetWindow(window)).hwnd) {
            spAdjustComponentSize(window);
            DrawMenuBar(SpPrimitiveArch(window).hwnd);
            return SP_FALSE;
        } else if (spIsCanvas(window) == SP_TRUE && SpCanvasPart(window).exposed == SP_FALSE) {
            spDebug(80, "processCallback", "WM_SIZE before canvas exposure: %s\n", SpGetClassName(window));
            return SP_TRUE;
        }
    } else if (message == WM_GETMINMAXINFO) {
        LPMINMAXINFO lpmm;
        lpmm = (LPMINMAXINFO)lParam;
        spDebug(100, "processCallback", "receive WM_GETMINMAXINFO, ptMinTrackSize = (%d, %d)\n",
                lpmm->ptMinTrackSize.x, lpmm->ptMinTrackSize.y);
        /* for Windows 8 ptMinTrackSize bug (ptMinTrackSize is calculated by using high-dpi based metrics) */
        lpmm->ptMinTrackSize.x = 0;
        lpmm->ptMinTrackSize.y = 0;
        return SP_TRUE;
    } else if (message == WM_DESTROY) {
        spDebug(10, "processCallback", "%s: receive WM_DESTROY, window = %s\n",
                spGetName(window), spGetName(SpGetWindow(window)));
        if (SpPrimitiveArch(SpGetWindow(window)).hwnd == NULL) {
        /*if (SpPrimitiveArch(window).hwnd == NULL) {*/
            spDebug(10, "processCallback", "already hwnd == NULL\n");
            return SP_TRUE;
        } else if (spIsSubClass(SpGetWindow(window), SpFrame) == SP_TRUE && SpFramePart(SpGetWindow(window)).system_window != NULL) {
            spDebug(10, "processCallback", "system window exists\n");
            return SP_TRUE;
        }
        spDebug(10, "processCallback", "destroy window\n");
        SpPrimitiveArch(SpGetWindow(window)).hwnd = NULL;
        /*SpPrimitiveArch(window).hwnd = NULL;*/
    } else if (message == WM_CLOSE && spIsFrame(window) == SP_TRUE) {
        window_close_style = SpFramePart(window).close_style;
        spDebug(10, "processCallback", "close window\n");
        if (window_close_style == SP_NO_CLOSE) {
            return SP_TRUE;
        }
    } else if (message >= WM_USER && message <= 0x7FFF) {
        spDebug(10, "processCallback", "received user message\n");
        child = (spComponent)wParam;
        if (child != NULL && spExistsPrimitiveEventCallback(child) == SP_TRUE) {
#if 0
            return SpPrimitivePart(child).event_call_func(child,
                                                          (spEventType)(message - WM_USER) + SP_EVENT_USER,
                                                          (void *)lParam);
#else
            return spPrimitiveCallEventCallback(child,
                                                (spEventType)(message - WM_USER) + SP_EVENT_USER,
                                                (void *)lParam);
#endif
        }
    } else if (message == WM_TIMER) {
        return processTimerMessage(window, (spTimerId)wParam);
    } else if (message == WM_HELP) {
        if (processHelpMessage(window, (LPHELPINFO)lParam) == SP_TRUE) {
            return SP_TRUE;
        }
    } else if (message == WM_TASKTRAY) {
        if (lParam != WM_LBUTTONDOWN && lParam != WM_RBUTTONDOWN
            && lParam != WM_MBUTTONDOWN) {
            return SP_TRUE;
        }
    } else if (message == WM_MENUSELECT) {
        processMenuSelectMessage(window, (HMENU)lParam, wParam);
        return SP_TRUE;
    } else if (message == WM_SYSCOLORCHANGE) {
        processSysColorChangeMessage(window);
        return SP_FALSE;
    } else if (message == WM_SETFOCUS || message == WM_KILLFOCUS) {
#if 0
        if (spIsCanvas(window) == SP_TRUE
            && SpCanvasPart(window).focusable == SP_TRUE) {
            spDebug(10, "processCallback", "canvas focus changed\n");
            spRedrawCanvas(window);
            flag = SP_TRUE;
        }
#endif
    } else if (message == WM_CHAR) {
        unsigned char c = (unsigned char)wParam;
        spDebug(10, "processCallback", "WM_CHAR: %c\n", c);
        
        if (c == '\r' || c == '\b' || c == '\n'
            || c == '\t' || c == 0x01b/*ESC*/) {
            return SP_TRUE;
        }
    } else if (message == WM_KEYDOWN) {
        spKeySym key_sym;
        
        if (spPrimitiveGetCallbackKeySym(window, &key_sym) == SP_TRUE
            && key_sym == SPK_Character) {
            spDebug(10, "processCallback", "WM_KEYDOWN: character\n");
            return SP_FALSE;
        } else {
            spDebug(10, "processCallback", "WM_KEYDOWN: wParam = %lx\n", wParam);
            switch ((int)wParam) {
              case VK_MENU:
              case VK_CONTROL:
              case VK_SHIFT:
              case VK_ESCAPE:
              case VK_CAPITAL:
                spDebug(10, "processCallback", "WM_KEYDOWN: num repeat = %d\n",
                        LOWORD(lParam));
                if (LOWORD(lParam) > 1) {
                    return SP_TRUE;
                }
                break;
              case VK_APPS:
                if (GetCursorPos(&point) && ScreenToClient(SpPrimitiveArch(window).hwnd, &point)) {
                    SetFocus(SpPrimitiveArch(window).hwnd);
                    PostMessage(SpPrimitiveArch(window).hwnd,
                                WM_RBUTTONDOWN, (WPARAM)MK_RBUTTON, MAKELPARAM(point.x, point.y));
                    return SP_TRUE;
                }
              default:
                break;
            }
        }
    } else if (message == WM_KEYUP) {
        if (spIsPrimitive(window) == SP_TRUE) {
            switch ((int)wParam) {
              case VK_APPS:
                if (GetCursorPos(&point) && ScreenToClient(SpPrimitiveArch(window).hwnd, &point)) {
                    PostMessage(SpPrimitiveArch(window).hwnd,
                                WM_RBUTTONUP, (WPARAM)MK_RBUTTON, MAKELPARAM(point.x, point.y));
                    return SP_TRUE;
                }
              default:
                break;
            }
        } else {
            return SP_TRUE;
        }
    } else if (message == WM_DROPFILES) {
        processDropFilesMessage(window, (HDROP)wParam);
        return SP_TRUE;
    } else if (message == WM_COPYDATA) {
        if (processCopyDataMessage(window, (PCOPYDATASTRUCT)lParam) == SP_TRUE) {
            return SP_TRUE;
        }
    } else if (message == WM_NOTIFY) {
        flag = processNotifyMessage(window, lParam);
    } else if (message == WM_LBUTTONDOWN || message == WM_RBUTTONDOWN ||
               message == WM_MBUTTONDOWN) {
        capture_window = window;
        point.x = LOWORD(lParam); point.y = HIWORD(lParam);
        spDebug(50, "processCallback", "BUTTONDOWN: x = %d, y = %d\n", point.x, point.y);
        
        ClientToScreen(SpPrimitiveArch(window).hwnd, &point);
        capture_dx = point.x; capture_dy = point.y;
        
        GetWindowRect(SpPrimitiveArch(SpGetWindow(window)).hwnd, &rect);
        point.x = rect.left; point.y = rect.top;
        capture_dx -= point.x; capture_dy -= point.y;

        SetFocus(SpPrimitiveArch(window).hwnd);
        SetCapture(SpPrimitiveArch(window).hwnd);
        spDebug(10, "processCallback", "capture started\n");
        need_capture = SP_TRUE;
        flag = SP_TRUE;
    } else if (message == WM_MOUSEMOVE &&
               (wParam & (MK_LBUTTON | MK_RBUTTON | MK_MBUTTON))) {
        if (window != capture_window) {
            return SP_FALSE;
        }
        if (spIsCanvas(window) == SP_TRUE
            && SpCanvasPart(window).capture_flag == SP_TRUE) {
            point.x = LOWORD(lParam); point.y = HIWORD(lParam);
            ClientToScreen(SpPrimitiveArch(window).hwnd, &point);
            point.x -= capture_dx;
            point.y -= capture_dy;
            
            GetWindowRect(SpPrimitiveArch(SpGetWindow(window)).hwnd, &rect);
            SpComponentPart(SpGetWindow(window)).x = point.x;
            SpComponentPart(SpGetWindow(window)).y = point.y;
            MoveWindow(SpPrimitiveArch(SpGetWindow(window)).hwnd,
                       SpComponentPart(SpGetWindow(window)).x,
                       SpComponentPart(SpGetWindow(window)).y,
                       rect.right - rect.left, rect.bottom - rect.top, TRUE);
            
            flag = SP_TRUE;
        }
    } else if (message == WM_LBUTTONUP || message == WM_RBUTTONUP ||
               message == WM_MBUTTONUP) {
        capture_dx = -1; capture_dy = -1;
        spDebug(10, "processCallback", "call ReleaseCapture\n");
        ReleaseCapture();
        flag = SP_TRUE;
        if (window != capture_window) {
            capture_window = NULL;
            return SP_FALSE;
        } else {
            capture_window = NULL;
        }
    } else if (message == WM_PAINT) {
        spDebug(50, "processCallback", "WM_PAINT: %s\n", SpGetClassName(window));
        if (SpTopLevelArch(sp_toplevel).getting_message == SP_FALSE) {
            spDebug(50, "processCallback", "WM_PAINT before realization\n");
            return SP_TRUE;
        }
        memset(&SpPrimitiveArch(window).paintstruct, 0, sizeof(PAINTSTRUCT));
    } else if (message == WM_SETCURSOR) {
        return processSetCursorMessage(window);
    } else if (message == WM_COMMAND) {
        sub_message = HIWORD(wParam);
        spDebug(80, "processCallback", "WM_COMMAND: sub_message = %x, wParam = %lx\n", sub_message, wParam);
        if (sub_message == EN_UPDATE) {
            spDebug(80, "processCallback", "WM_COMMAND / EN_UPDATE\n");
            return SP_FALSE;
        } else if (sub_message == EN_SETFOCUS) {
            spDebug(50, "processCallback", "EN_SETFOCUS\n");
            SpFrameArch(SpGetWindow(window)).hfocus_text = (HWND)lParam;
        } else if (sub_message == EN_KILLFOCUS) {
            spDebug(50, "processCallback", "EN_KILLFOCUS\n");
            SpFrameArch(SpGetWindow(window)).hfocus_text = NULL;
        }
    } else if (message == WM_VSCROLL || message == WM_HSCROLL) {
        if ((scroll_bar = getCurrentComponent(window, (HWND)lParam)) == NULL) {
            scroll_bar = window;
        }
        SpPrimitiveArch(scroll_bar).message = message;

        if (processScrollMessage(scroll_bar, wParam) == SP_TRUE) {
            spDebug(50, "processCallback", "processScrollMessage returns SP_TRUE\n");
            return SP_TRUE;
        }
        flag = SP_TRUE;
    } else if (message == 0x020A/*WM_MOUSEWHEEL*/ || message == 0x020E/*WM_MOUSEHWHEEL*/) {
        spDebug(50, "processCallback", "WM_MOUSEWHEEL or WM_MOUSEHWHEEL (%x)\n", message);
        if (spIsCanvas(window) == SP_TRUE) {
            spDebug(50, "processCallback", "WM_MOUSEWHEEL or WM_MOUSEHWHEEL in canvas\n");
            if (preprocessCanvasWheelMessage(window, message, wParam) == SP_TRUE) {
                spDebug(50, "processCallback", "preprocessCanvasWheelMessage returns SP_TRUE\n");
                return SP_TRUE;
            }
            if (wParam & MK_CONTROL) {
                short dist;
                double delta;
                double relative_factor;
            
                spGetClientSize(window, NULL, NULL);

                dist = HIWORD(wParam);
                spDebug(50, "processCallback", "dist = %d\n", dist);
                delta = (double)dist / 120.0;
                relative_factor = pow(2.0, delta);
                spDebug(50, "processCallback", "delta = %f, relative_factor = %f\n", delta, relative_factor);
                
                if (spIsScrollCanvas(window) == SP_TRUE) {
                    flag = spZoomScrollCanvas(window, relative_factor);
                } else {
                    flag = spUpdateCanvasContentFactor(window, relative_factor, SP_TRUE);
                }
            } else if (SpCanvasPart(window).accept_wheel_event == SP_FALSE) {
                spDebug(50, "processCallback", "WM_MOUSEWHEEL: canvas doesn't accept wheel event\n");
                return SP_FALSE;
            }
        } else {
            spDebug(50, "processCallback", "WM_MOUSEWHEEL or WM_MOUSEHWHEEL in non-canvas: %s\n", SpGetClassName(window));
            return SP_FALSE;
        }
    }

    component_id = SpGetComponentId(window);
    if (SpGetWindow(window) != NULL) {
        parent_window = SpGetWindow(window);
    } else {
        parent_window = window;
    }
    num_callback = SpPrimitivePart(parent_window).num_callback;
                
    if (SpPrimitivePart(parent_window).callbacks != NULL) {
        for (k = 0; k < num_callback; k++) {
            /* Since callbacks can be reallocated in a callback function,
             * the variable `callbacks' must be updated each time in this loop */
            callbacks = SpPrimitivePart(parent_window).callbacks;
            target_component = callbacks[k].component;
            spDebug(100, "processCallback", "callbacks[%ld].component_id = %ld\n", k, callbacks[k].component_id);
            
            if (target_component == NULL
                || spIsPrimitive(target_component) == SP_FALSE
                || (spIsCanvas(target_component) == SP_TRUE
                    && SpPrimitiveArch(window).hwnd != SpPrimitiveArch(target_component).hwnd)) {
                spDebug(80, "processCallback", "continue\n");
                continue;
            }
            
            if (callbacks[k].message == WM_COMMAND &&
                (message == WM_COMMAND || message == WM_CUT || message == WM_PASTE ||
                 message == WM_COPY || message == WM_CLEAR || message == WM_UNDO)) {
                sub_message = HIWORD(wParam);
                spDebug(80, "processCallback", "message = %x (WM_COMMAND family), sub_message = %x\n",
                        message, sub_message);
                if (callbacks[k].component_id == LOWORD(wParam)
                    && (callbacks[k].command_id == 0 || callbacks[k].command_id == sub_message)) {
                    SpPrimitiveArch(target_component).message = message;
                    SpPrimitiveArch(target_component).wParam = wParam;
                    SpPrimitiveArch(target_component).lParam = lParam;
                    spComponentCallbackFunc(&callbacks[k]);
                    SpPrimitiveArch(target_component).message = WM_NULL;
                    SpPrimitiveArch(target_component).wParam = 0;
                    SpPrimitiveArch(target_component).lParam = 0;
                    flag = SP_TRUE;
                }
            } else if (callbacks[k].message == message) {
                spDebug(100, "processCallback", "message = %d\n", message);
                if ((message == WM_VSCROLL || message == WM_HSCROLL) &&
                    scroll_bar == target_component) {
                    if (1        /*(sub_message != SB_THUMBTRACK && sub_message != TB_THUMBTRACK)
                                  || callbacks[k].command_id == SB_THUMBTRACK*/) {
                        SpPrimitiveArch(target_component).message = message;
                        SpPrimitiveArch(target_component).wParam = wParam;
                        SpPrimitiveArch(target_component).lParam = lParam;
                        spComponentCallbackFunc(&callbacks[k]);
                        SpPrimitiveArch(target_component).message = WM_NULL;
                        SpPrimitiveArch(target_component).wParam = 0;
                        SpPrimitiveArch(target_component).lParam = 0;
                        spDebug(80, "processCallback", "scroll callback: %d\n",
                                (int)sub_message);
                        flag = SP_TRUE;
                    }
                } else if (message == WM_PAINT &&
                           SpPrimitiveArch(window).hwnd != SpPrimitiveArch(target_component).hwnd) {
                    continue;
                } else if (message == WM_MOUSEMOVE && callbacks[k].command_id != 0 &&
                           !(wParam & callbacks[k].command_id)) {
                    continue;
                } else if ((message == 0x020A/*WM_MOUSEWHEEL*/ || message == 0x020E/*WM_MOUSEHWHEEL*/)
                           && (((wParam & MK_CONTROL) && callbacks[k].command_id == 0)
                               || (!(wParam & MK_CONTROL) && callbacks[k].command_id == 1))) {
                    spDebug(50, "processCallback",
                            "WM_MOUSEWHEEL: different wheel event (control key flag = %d), continue, k = %ld, command_id = %d\n",
                            !!(wParam & MK_CONTROL), k, callbacks[k].command_id);
                    continue;
                } else {
                    switch (message) {
                      case WM_KEYDOWN:
                        if (spIsSubClass(target_component, SpTextField) == SP_TRUE
                            && ((int)wParam != VK_RETURN
                                || SpPrimitiveArch(target_component).hwnd != GetFocus())) {
                            continue;
                        }
                      case WM_CHAR:
                      case WM_IME_CHAR:
                      case WM_KEYUP:
                      case WM_LBUTTONDOWN:
                      case WM_RBUTTONDOWN:
                      case WM_MBUTTONDOWN:
                      case WM_LBUTTONUP:
                      case WM_RBUTTONUP:
                      case WM_MBUTTONUP:
                      case WM_MOUSEMOVE:
                      case WM_SETFOCUS:
                      case WM_KILLFOCUS:
                      case 0x020A/*WM_MOUSEWHEEL*/:
                      case 0x020E/*WM_MOUSEHWHEEL*/:
                        if (SpComponentPart(target_component).sensitive_flag == SP_FALSE) {
                            continue;
                        }
                      case WM_PAINT:
                      case WM_SIZE:
                      case WM_TASKTRAY:
                      case WM_CLOSE:
                      case WM_DESTROY:
                        SpPrimitiveArch(target_component).message = message;
                        SpPrimitiveArch(target_component).wParam = wParam;
                        SpPrimitiveArch(target_component).lParam = lParam;
                        spComponentCallbackFunc(&callbacks[k]);
                        if (!(message == WM_CLOSE || message == WM_DESTROY)) {
                            SpPrimitiveArch(target_component).message = WM_NULL;
                            SpPrimitiveArch(target_component).wParam = 0;
                            SpPrimitiveArch(target_component).lParam = 0;
                        }
                        flag = SP_TRUE;
                        break;
                      default:
                        break;
                    }

                    if (message == WM_DESTROY
                        || (message == WM_CLOSE &&
                            window_close_style == SP_DESTROY_CLOSE)) {
                        flag = SP_FALSE;
                    }

                    if ((message == WM_PAINT || message == WM_SIZE)
                        && spIsCanvas(target_component) == SP_TRUE) {
                        /* processed in the first callback function */
                        break;
                    }
                }
            }
    
            if (message != WM_DESTROY
                && component_id == spGetDestroyedComponentId()) {
                flag = SP_TRUE;
                break;
            }
        }
    }

    if (component_id != spGetDestroyedComponentId()) {
        if (message == WM_CLOSE && window_close_style == SP_UNMAP_CLOSE) {
            spPopdownWindow(window);
            flag = SP_TRUE;
        }
        if (need_capture == SP_TRUE) {
            /* cursor change doesn't work before capturing */
            if (spIsCanvas(window) == SP_TRUE
                && SpPrimitiveArch(window).hcursor != NULL) {
                SetCursor(SpPrimitiveArch(window).hcursor);
            }
        }
    }
    spDebug(100, "processCallback", "done: message = %x, flag = %d\n", message, flag);
    
    return flag;
}

LRESULT CALLBACK spWindowFunc(HWND hwnd, UINT message,
                              WPARAM wParam, LPARAM lParam)
{
    LRESULT lresult;
    RECT rect;
    HBRUSH hbrush;
    char class_name[128];
    spComponent tab_box, tab_item;
    spComponent component;
    spComponent window = NULL;

    if ((window = getCurrentWindow(hwnd, message, wParam, lParam)) != NULL) {
        spDebug(100, "spWindowFunc", "message received: %x\n", message);
        if (message == WM_CTLCOLORSTATIC) {
            if (GetClassName((HWND)lParam, class_name, sizeof(class_name))) {
                spDebug(100, "spWindowFunc", "receive WM_CTLCOLORSTATIC: %s\n", class_name);
                
                if (1) {        /* All controls are OK? */
                    hbrush = NULL;
                    
                    if (spIsOSWinXPLater() == SP_TRUE
                        && (component = getCurrentComponent(window, (HWND)lParam)) != NULL) {
                        spDebug(100, "spWindowFunc", "component: %s\n", spGetTitle(component));
                        if ((tab_item = spGetAncestorTabItem(component)) != NULL
                            && spIsVisible(tab_item) == SP_TRUE) {
                            spDebug(100, "spWindowFunc", "WM_CTLCOLORSTATIC: spGetAncestorTabBox: tab_item found\n");
                            tab_box = SpGetParent(tab_item);
                            
                            if (spRedrawTabThemeBackgroundWin(tab_box) == SP_TRUE
                                || SpPrimitiveArch(component).hbrush == NULL) {
                                spUpdateTabThemeBackgroundBrushWin(tab_box, component);
                            } else {
                                spDebug(100, "spWindowFunc", "WM_CTLCOLORSTATIC: brush not updated\n");
                            }
                            hbrush = SpPrimitiveArch(component).hbrush;
                            spDebug(100, "spWindowFunc", "WM_CTLCOLORSTATIC: use hbrush: %ld\n", (long)hbrush);
                            
                            if (hbrush == NULL) {
                                spDebug(100, "spWindowFunc", "WM_CTLCOLORSTATIC: use white brush\n");
                                hbrush = GetStockObject(WHITE_BRUSH);
                            }
                        }
                    }

                    if (hbrush == NULL) {
                        spDebug(100, "spWindowFunc", "WM_CTLCOLORSTATIC: use default brush\n");
                        hbrush = GetSysColorBrush(COLOR_BTNFACE);
                    }
                    
                    SetBkMode((HDC)wParam, TRANSPARENT);
                    return (LRESULT)hbrush;
                }
            }
        } else if (message == WM_ERASEBKGND) {
            if (spIsCanvas(window) == SP_TRUE) {
                /* drawing in WM_PAINT */
                spDebug(100, "spWindowFunc", "WM_ERASEBKGND in canvas\n");
                return 0;
            }

            spDebug(100, "spWindowFunc", "WM_ERASEBKGND in frame or container: %s\n", spGetTitle(window));
            hbrush = NULL;
            GetClientRect(hwnd, &rect);
            spDebug(100, "spWindowFunc", "rect: left = %d, top = %d, right = %d, bottm = %d\n",
                    rect.left, rect.top, rect.right, rect.bottom);
            
            if (spIsOSWinXPLater() == SP_TRUE
                && (tab_item = spGetAncestorTabItem(window)) != NULL
                && spIsVisible(tab_item) == SP_TRUE) {
                spDebug(100, "spWindowFunc", "spGetAncestorTabBox: tab_item found\n");
                tab_box = SpGetParent(tab_item);
                
                if (spRedrawTabThemeBackgroundWin(tab_box) == SP_TRUE
                    || SpPrimitiveArch(window).hbrush == NULL) {
                    spUpdateTabThemeBackgroundBrushWin(tab_box, window);
                }
                hbrush = SpPrimitiveArch(window).hbrush;
            }

            if (hbrush == NULL) {
                hbrush = SpTopLevelArch(sp_toplevel).bg_brush;
            }
            spDebug(100, "spWindowFunc", "hbrush = %ld\n", (long)hbrush);
            
            FillRect((HDC)wParam, &rect, hbrush);
            return 1;
        } else if (processCallback(window) == SP_TRUE) {
            return 0;
        } else if (message == WM_DESTROY) {
            spDebug(10, "spWindowFunc", "destroy component\n");
            /*spDestroyWindow(window);*/
            spDestroyComponent(window);
            spDebug(10, "spWindowFunc", "destroy component done\n");
            return 0;
        }
    }

    spDebug(100, "spWindowFunc", "call DefWindowProc\n");
    spThreadEnterArch(sp_toplevel);
    lresult = DefWindowProc(hwnd, message, wParam, lParam);
    spThreadLeaveArch(sp_toplevel);

    return lresult;
}

BOOL CALLBACK spDialogFunc(HWND hwnd, UINT message,
                           WPARAM wParam, LPARAM lParam)
{
    spComponent window;
    
    window = getCurrentWindow(hwnd, message, wParam, lParam);
    if (processCallback(window) == SP_TRUE) {
        return 1;
    } else {
        return 0;
    }
}

HWND spGetChildCanvasFromPointWin(HWND hwnd, POINT *screenpt)
{
    HWND hwndparent, hwndchild;
    POINT point;
    char class_name[128];

    spDebug(100, "spGetChildCanvasFromPointWin", "in\n");
    
    hwndchild = hwnd;
    
    while (hwndchild != NULL) {
        spDebug(100, "spGetChildCanvasFromPointWin", "hwndchild = %lx\n", (long)hwndchild);
        hwndparent = hwndchild;
        point = *screenpt;
        ScreenToClient(hwndparent, &point);
        if ((hwndchild = ChildWindowFromPointEx(hwndparent, point,
                                                CWP_SKIPINVISIBLE | CWP_SKIPDISABLED | CWP_SKIPTRANSPARENT)) != NULL) {
            if (hwndchild == hwndparent) {
                hwndchild = NULL;
                break;
            }
            
            if (GetClassName(hwndchild, class_name, sizeof(class_name))) {
                if (streq(class_name, SP_CANVAS_CLASS_NAME)) {
                    break;
                }
            }
        }
    } 
    
    spDebug(100, "spGetChildCanvasFromPointWin", "done\n");
    
    return hwndchild;
}

int spTranslateAllAcceleratorWin(MSG *msg)
{
    spComponent window;
    spComponent component;
    spComponent grabbed_window;
    HWND hwnd;
    HWND hfocus;
    char class_name[128];

    hwnd = GetActiveWindow();
    window = getCurrentComponent(NULL, hwnd);
    
    if (spIsPrimitive(window) == SP_TRUE) {
        grabbed_window = spGetCurrentGrabWin();
        if (grabbed_window != NULL && grabbed_window != window && spIsAncestor(window, grabbed_window) == SP_FALSE) {
            /* shortcut will be ignored when another grabbed window exists */
            spDebug(80, "spTranslateAllAcceleratorWin", "grabbed window exists\n");
            return 0;
        }
        
        spDebug(80, "spTranslateAllAcceleratorWin", "%s: message = %x\n", SpGetClassName(window), msg->message);

        if (GetFocus() == NULL) {
            spDebug(100, "spTranslateAllAcceleratorWin", "GetFocus = NULL, hwnd = %lx\n", (long)hwnd);
            SetFocus(hwnd);
        }
    
        if (msg->message == WM_KEYDOWN && spIsFrame(window) == SP_TRUE
            && (msg->wParam == VK_ESCAPE || msg->wParam == VK_RETURN)) {
            spDebug(80, "spTranslateAllAcceleratorWin", "escape or return key pressed\n");
            if (msg->wParam == VK_ESCAPE
                && spIsButton(SpFramePart(window).cancel_button) == SP_TRUE) {
                PostMessage(SpPrimitiveArch(SpFramePart(window).cancel_button).hwnd,
                            BM_CLICK, 0, 0);
                return 1;
            } else if (msg->wParam == VK_RETURN) {
                spBool edit_control = SP_FALSE;
                spBool multi_line = SP_FALSE;

                strcpy(class_name, "");
                if ((hfocus = msg->hwnd) != NULL
                    && GetClassName(hfocus, class_name, sizeof(class_name))) {
                    if (strcaseeq(class_name, "LISTBOX")) {
                        PostMessage(hwnd, WM_COMMAND,
                                    MAKELONG(GetWindowLong(hfocus, GWL_ID),
                                             LBN_DBLCLK), (LPARAM)hfocus);
                        return 1;
                    } else if (strcaseeq(class_name, "EDIT")) {
                        edit_control = SP_TRUE;
                        if ((GetWindowLong(hfocus, GWL_STYLE) & ES_MULTILINE)) {
                            multi_line = SP_TRUE;
                        }
                    }
                }

                if (multi_line == SP_FALSE) {
                    if (hfocus == NULL || !strcaseeq(class_name, "BUTTON")) {
                        if (spIsButton(SpFramePart(window).default_button) == SP_TRUE) {
                            hfocus = SpPrimitiveArch(SpFramePart(window).default_button).hwnd;
                        }
                    }

                    if (hfocus != NULL && (GetWindowLong(hfocus, GWL_STYLE) & BS_DEFPUSHBUTTON)) {
                        PostMessage(hfocus, BM_CLICK, 0, 0);
                        return 1;
                    }
                }
                
                if (SpFramePart(window).window_type != SP_DIALOG_WINDOW
                    && edit_control == SP_TRUE && multi_line == SP_FALSE
                    && !(GetWindowLong(hfocus, GWL_STYLE) & ES_READONLY)) {
                    /* VK_RETURN in single line text */
                    spDebug(60, "spTranslateAllAcceleratorWin", "VK_RETURN in text field\n");
                    PostMessage(hwnd, msg->message, msg->wParam, msg->lParam);
                    return 1;
                }

            }
        } else if (msg->message == WM_KEYDOWN
                   && (component = getCurrentComponent(NULL, msg->hwnd)) != NULL
                   && spIsText(component) == SP_TRUE
                   && SpTextPart(component).disable_shortcut == SP_TRUE) {
            /* disable shortcut key */
            return 0;
        } else if (msg->message == WM_KEYDOWN && msg->wParam == VK_TAB) {
            spDebug(50, "spTranslateAllAcceleratorWin", "tab key pressed\n");
            if ((hfocus = msg->hwnd) != NULL
                && GetClassName(hfocus, class_name, sizeof(class_name))
                && strcaseeq(class_name, "EDIT")
                && (GetWindowLong(hfocus, GWL_STYLE) & ES_MULTILINE)) {
                spDebug(50, "spTranslateAllAcceleratorWin", "EM_REPLACESEL\n");
                if (0x8000 & GetKeyState(VK_SHIFT)) {
                    SetFocus(GetNextDlgTabItem(hwnd, hfocus, TRUE));
                } else {
                    PostMessage(hfocus, WM_CHAR, '\t', 0L);
                }
                return 1;
            } else if (IsWindow(hwnd) && IsDialogMessage(hwnd, msg)) {
                return 1;
            }
        } else if (msg->message == 0x020A/*WM_MOUSEWHEEL*/ || msg->message == 0x020E/*WM_MOUSEHWHEEL*/) {
            spDebug(100, "spTranslateAllAcceleratorWin", "WM_MOUSEWHEEL or WM_MOUSEHWHEEL received\n");
            
            /*hfocus = msg->hwnd;*/
            hfocus = hwnd;
            
            if (spIsFrame(window) == SP_TRUE) {
                HWND hwndchild;
                
                spDebug(100, "spTranslateAllAcceleratorWin", "WM_MOUSEWHEEL or WM_MOUSEHWHEEL in frame\n");

                if ((hwndchild = spGetChildCanvasFromPointWin(hfocus, &msg->pt)) != NULL) {
                    spComponent canvas;
                    
                    spDebug(50, "spTranslateAllAcceleratorWin", "found child window by ChildWindowFromPointEx\n");
                    if ((canvas = getCurrentWindow(hwndchild, msg->message, msg->wParam, msg->lParam)) != NULL) {
                        spDebug(50, "spTranslateAllAcceleratorWin", "current canvas found\n");
                        processCallback(canvas);
                    } else {
                        spDebug(50, "spTranslateAllAcceleratorWin", "cannot find current canvas\n");
                    }
                    return 1;
                }
            } else {
                spDebug(100, "spTranslateAllAcceleratorWin", "WM_MOUSEWHEEL or WM_MOUSEHWHEEL in non-frame\n");
            }
        } else if (SpPrimitiveArch(window).haccel != NULL) {
            if (TranslateAccelerator(hwnd, SpPrimitiveArch(window).haccel, msg)) {
                spDebug(50, "spTranslateAllAcceleratorWin", "TranslateAccelerator returns TRUE\n");
                return 1;
            }
        }
    } else {
        spDebug(100, "spTranslateAllAcceleratorWin", "non-primitive window\n");
    }
    spDebug(80, "spTranslateAllAcceleratorWin", "done\n");
    
    return 0;
}

static BOOL CALLBACK EnumDropdownChild(HWND hwnd, LPARAM lparam)
{
    char class_name[128];
    HWND *htext = (HWND *)lparam;
    
    if (GetClassName(hwnd, class_name, sizeof(class_name))) {
        spDebug(80, "EnumDropdownChild", "class_name = %s\n", class_name);
        if (strcaseeq(class_name, "Edit")) {
            *htext = hwnd;
            spDebug(80, "EnumDropdownChild", "Edit window found\n");
            return FALSE;
        }
    }
    
    return TRUE;
}

static HWND getToolTipWindow(spComponent component)
{
    HWND htext = NULL;

    if (spIsComboBox(component) == SP_TRUE && SpTextPart(component).editable == SP_TRUE) {
        spDebug(80, "getToolTipWindow", "call EnumChildWindows\n");
        EnumChildWindows(SpPrimitiveArch(component).hwnd, (WNDENUMPROC)EnumDropdownChild, (LPARAM)&htext);
        if (htext != NULL) {
            return htext;
        }
    }
    return SpPrimitiveArch(component).hwnd;
}

spBool spShowToolTipArch(spComponent component)
{
    TOOLINFO ti;
    spComponent window;

    if (SpPrimitiveArch(component).hwnd == NULL) return SP_FALSE;
    
    spDebug(30, "spShowToolTipArch", "description = %s\n",
            SpComponentPart(component).description);
    
    window = SpGetWindow(component);
    
    if (SpPrimitiveArch(window).htooltip == NULL) {
        SpPrimitiveArch(window).htooltip =
            CreateWindowEx(WS_EX_TOOLWINDOW, TOOLTIPS_CLASS, (LPSTR)NULL,
                           TTS_ALWAYSTIP,
                           CW_USEDEFAULT, CW_USEDEFAULT,
                           CW_USEDEFAULT, CW_USEDEFAULT, 
                           SpPrimitiveArch(window).hwnd, (HMENU)NULL,
                           SpTopLevelArch(sp_toplevel).hThisInst, NULL);
    }
 
    ti.cbSize = sizeof(TOOLINFO); 
    ti.uFlags = TTF_IDISHWND | TTF_SUBCLASS;
    ti.hwnd = SpPrimitiveArch(window).hwnd;
    ti.hinst = SpTopLevelArch(sp_toplevel).hThisInst;
    ti.uId = (UINT_PTR)getToolTipWindow(component);
    ti.lpszText = (LPSTR)SpComponentPart(component).description;
    ti.rect.left = 0;         ti.rect.top = 0;
    ti.rect.right = 0;         ti.rect.bottom = 0;
    
    if (!SendMessage(SpPrimitiveArch(window).htooltip, TTM_ADDTOOL,
                     0, (LPARAM)(LPTOOLINFO)&ti)) {
        spDebug(30, "spShowToolTipArch", "TTM_ADDTOOL error\n");
    }
                
    return SP_TRUE;
}

spBool spPrepareEventCallbackArch(spComponent component)
{
    return SP_TRUE;
}

spBool spUnprepareEventCallbackArch(spComponent component)
{
    return SP_TRUE;
}

spBool spPostUserEventArch(spComponent component, spEventType event_type, void *call_data)
{
    spComponent window;
    UINT msg;

    window = SpGetWindow(component);
    
    if (SpPrimitiveArch(window).hwnd == NULL
        || event_type < SP_EVENT_USER) return SP_FALSE;

    msg = WM_USER + (UINT)(event_type - SP_EVENT_USER);
    
    if (PostMessage(SpPrimitiveArch(window).hwnd,
                    msg, (WPARAM)component, (LPARAM)call_data)) {
        return SP_TRUE;
    } else {
        return SP_FALSE;
    }
}

void *spGetPrimitiveHandleArch(spComponent component)
{
    return SpPrimitiveArch(component).hwnd;
}
