/*
 *	spSlider_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spSliderP.h>

void spSetAdjustmentParamsGtk(GtkAdjustment *adjustment, gdouble minimum, gdouble maximum, gdouble value,
			      gdouble increment, gdouble page_increment, gdouble page_size,
			      gdouble factor, spBool track_bar_flag)
{
    if (factor != 1.0) {
	minimum *= factor;
	maximum *= factor;
	value *= factor;
	page_size *= factor;
    }
    
#if GTK_CHECK_VERSION(3,0,0)
    gtk_adjustment_set_lower(adjustment, minimum);
    if (track_bar_flag == SP_TRUE) {
	gtk_adjustment_set_upper(adjustment, 1.0 + maximum);
    } else {
	gtk_adjustment_set_upper(adjustment, maximum);
    }
    gtk_adjustment_set_value(adjustment, value);
    gtk_adjustment_set_step_increment(adjustment, increment);
    gtk_adjustment_set_page_increment(adjustment, page_increment);
    if (track_bar_flag == SP_TRUE) {
	gtk_adjustment_set_page_size(adjustment, 1.0);
    } else {
	gtk_adjustment_set_page_size(adjustment, page_size);
    }
#else
    adjustment->lower = minimum;
    adjustment->upper = maximum;
    if (track_bar_flag == SP_TRUE) {
	adjustment->upper += 1.0;
    }
    adjustment->value = value;
    adjustment->step_increment = increment;
    adjustment->page_increment = page_increment;
    if (track_bar_flag == SP_TRUE) {
	adjustment->page_size = 1.0;
    } else {
	adjustment->page_size = page_size;
    }
#endif

    return;
}

static gfloat valueToScaleValue(spComponent component, int value)
{
    return (gfloat)((double)value / pow(10.0, (double)SpSliderPart(component).decimal_points));
}
    
static int scaleValueToValue(spComponent component, gfloat svalue)
{
    return (int)spRound((double)svalue * pow(10.0, (double)SpSliderPart(component).decimal_points));
}
    
#if GTK_CHECK_VERSION(3,0,0)
spBool spGetMouseModifierMaskGtk(spComponent component, GdkModifierType *mask)
{
    GdkWindow *gdkwin;
    GdkDisplay *display;
    GdkDevice *pointer;
    GdkDeviceManager *manager;

    if ((gdkwin = gtk_widget_get_window(GTK_WIDGET(SpPrimitiveArch(component).widget))) == NULL) {
	return SP_FALSE;
    }

    if ((display = gdk_window_get_display(gdkwin)) != NULL) {
	if ((manager = gdk_display_get_device_manager(display)) != NULL) {
	    if ((pointer = gdk_device_manager_get_client_pointer(manager)) != NULL) {
		gdk_device_get_state(pointer, gdkwin, NULL, mask);
		spDebug(30, "spGetMouseModifierMaskGtk", "mask = %x\n", (int)*mask);
		return SP_TRUE;
	    }
	}
    }

    return SP_FALSE;
}
#endif

static void setUpdatePolicy(spComponent component)
{
#if !GTK_CHECK_VERSION(3,0,0)
    if (SpSliderPart(component).track_call_on == SP_FALSE) {
	gtk_range_set_update_policy(GTK_RANGE(SpPrimitiveArch(component).widget),
				    GTK_UPDATE_DISCONTINUOUS);
    } else {
	gtk_range_set_update_policy(GTK_RANGE(SpPrimitiveArch(component).widget),
				    GTK_UPDATE_CONTINUOUS);
    }
#endif

    return;
}

void spSliderCreateArch(spComponent component)
{
    SpPrimitiveArch(component).adjustment =
	gtk_adjustment_new (valueToScaleValue(component, SpSliderPart(component).value),
			    valueToScaleValue(component, SpSliderPart(component).minimum),
			    valueToScaleValue(component, (SpSliderPart(component).maximum)),
			    valueToScaleValue(component, SpSliderPart(component).increment),
			    valueToScaleValue(component, SpSliderPart(component).page_increment),
			    valueToScaleValue(component, SpSliderPart(component).page_size));
    
#if GTK_CHECK_VERSION(3,0,0)
    SpPrimitiveArch(component).widget
	= gtk_scrollbar_new((SpComponentPart(component).orientation == SP_HORIZONTAL
			     ? GTK_ORIENTATION_HORIZONTAL : GTK_ORIENTATION_VERTICAL),
			    GTK_ADJUSTMENT(SpPrimitiveArch(component).adjustment));
#else
    if (SpComponentPart(component).orientation == SP_HORIZONTAL) {
	SpPrimitiveArch(component).widget =
	    gtk_hscrollbar_new(GTK_ADJUSTMENT(SpPrimitiveArch(component).adjustment));
    } else {
	SpPrimitiveArch(component).widget =
	    gtk_vscrollbar_new(GTK_ADJUSTMENT(SpPrimitiveArch(component).adjustment));
    }
#endif
    
    spPrimitiveContainerAddGtk(component);

    setUpdatePolicy(component);

    return;
}

void spTrackBarCreateArch(spComponent component)
{
    SpPrimitiveArch(component).adjustment =
	gtk_adjustment_new (valueToScaleValue(component, SpSliderPart(component).value),
			    valueToScaleValue(component, SpSliderPart(component).minimum),
			    valueToScaleValue(component, (SpSliderPart(component).maximum)) + 1.0,
			    valueToScaleValue(component, SpSliderPart(component).increment),
			    valueToScaleValue(component, SpSliderPart(component).page_increment),
			    1.0);
    
#if GTK_CHECK_VERSION(3,0,0)
    SpPrimitiveArch(component).widget
	= gtk_scale_new((SpComponentPart(component).orientation == SP_HORIZONTAL
			 ? GTK_ORIENTATION_HORIZONTAL : GTK_ORIENTATION_VERTICAL),
			GTK_ADJUSTMENT(SpPrimitiveArch(component).adjustment));
#else
    if (SpComponentPart(component).orientation == SP_HORIZONTAL) {
	SpPrimitiveArch(component).widget =
	    gtk_hscale_new(GTK_ADJUSTMENT(SpPrimitiveArch(component).adjustment));
	gtk_scale_set_value_pos (GTK_SCALE(SpPrimitiveArch(component).widget), GTK_POS_TOP);
    } else {
	SpPrimitiveArch(component).widget =
	    gtk_vscale_new(GTK_ADJUSTMENT(SpPrimitiveArch(component).adjustment));
	gtk_scale_set_value_pos (GTK_SCALE(SpPrimitiveArch(component).widget), GTK_POS_LEFT);
    }
#endif
	
    if (SpSliderPart(component).show_value == SP_TRUE) {
	gtk_scale_set_draw_value (GTK_SCALE(SpPrimitiveArch(component).widget), TRUE);
    } else {
	gtk_scale_set_draw_value (GTK_SCALE(SpPrimitiveArch(component).widget), FALSE);
    }
    gtk_scale_set_digits(GTK_SCALE(SpPrimitiveArch(component).widget),
			 (gint)SpSliderPart(component).decimal_points);
    
    spPrimitiveContainerAddGtk(component);

    setUpdatePolicy(component);
    
    return;
}

void spSliderSetParamsArch(spComponent component)
{
    GtkAdjustment *adjustment;

    if (spIsSubClass(component, SpTrackBar) == SP_TRUE) {
	if (SpSliderPart(component).show_value == SP_TRUE) {
	    gtk_scale_set_draw_value(GTK_SCALE(SpPrimitiveArch(component).widget), TRUE);
	}
	gtk_scale_set_digits(GTK_SCALE(SpPrimitiveArch(component).widget),
			     (gint)SpSliderPart(component).decimal_points);
    }

#if !GTK_CHECK_VERSION(3,0,0)
    setUpdatePolicy(component);
#endif

#if GTK_CHECK_VERSION(3,0,0)
    adjustment = SpPrimitiveArch(component).adjustment;
#else
    adjustment = GTK_ADJUSTMENT(SpPrimitiveArch(component).adjustment);
#endif
    
    spSetAdjustmentParamsGtk(adjustment,
			     valueToScaleValue(component, SpSliderPart(component).minimum),
			     valueToScaleValue(component, SpSliderPart(component).maximum),
			     valueToScaleValue(component, SpSliderPart(component).value),
			     valueToScaleValue(component, SpSliderPart(component).increment),
			     valueToScaleValue(component, SpSliderPart(component).page_increment),
			     valueToScaleValue(component, SpSliderPart(component).page_size),
			     1.0, spIsSubClass(component, SpTrackBar));
    
#if GTK_CHECK_VERSION(3,0,0)
    g_signal_emit_by_name(SpPrimitiveArch(component).adjustment, "changed");
#else
    gtk_signal_emit_by_name(GTK_OBJECT(SpPrimitiveArch(component).adjustment), "changed");
#endif
    
    return;
}

void spTrackBarSetParamsArch(spComponent component)
{
    spSliderSetParamsArch(component);
    return;
}

void spSetSliderValueArch(spComponent component)
{
    gtk_adjustment_set_value(GTK_ADJUSTMENT(SpPrimitiveArch(component).adjustment),
			     valueToScaleValue(component, SpSliderPart(component).value));
    return;
}

int spGetSliderValueArch(spComponent component)
{
#if GTK_CHECK_VERSION(3,0,0)
    return scaleValueToValue(component, gtk_adjustment_get_value(SpPrimitiveArch(component).adjustment));
#else
    return scaleValueToValue(component, GTK_ADJUSTMENT(SpPrimitiveArch(component).adjustment)->value);
#endif
}

void spCorrectAdjustmentValueGtk(GtkAdjustment *adjustment, gdouble *value)
{
#if GTK_CHECK_VERSION(3,0,0)
    *value = MIN(*value,
		 gtk_adjustment_get_upper(adjustment) - gtk_adjustment_get_page_size(adjustment));
    *value = MAX(*value, gtk_adjustment_get_lower(adjustment));
#else
    *value = MIN(*value, adjustment->upper - adjustment->page_size);
    *value = MAX(*value, adjustment->lower);
#endif
    
    return;
}
