/*
 *	spText_Gtk.c
 */

/*#define GTK_ENABLE_BROKEN*/
#undef GTK_ENABLE_BROKEN

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spMemory.h>

#include <sp/spComboBoxP.h>
#include <sp/spTextP.h>
#include <sp/spGraphicsP.h>

long spGetTextFieldTextLengthGtk(spComponent component)
{
#if GTK_CHECK_VERSION(3,0,0)
    return gtk_entry_get_text_length(GTK_ENTRY(SpPrimitiveArch(component).widget));
#else
    return GTK_ENTRY(SpPrimitiveArch(component).widget)->text_length;
#endif
}

void spTextPartInitArch(spComponent component)
{
    SpTextArch(component).vscroll_bar = NULL;
    SpTextArch(component).hscroll_bar = NULL;
#if GTK_CHECK_VERSION(2,0,0)
    SpTextArch(component).font_desc = NULL;
#else
    SpTextArch(component).font = NULL;
#endif
    SpTextArch(component).suspend_callback = SP_FALSE;
    
    return;
}

void spTextPartFreeArch(spComponent component)
{
#if GTK_CHECK_VERSION(2,0,0)
    if (SpTextArch(component).font_desc != NULL) {
	pango_font_description_free(SpTextArch(component).font_desc);
    }
#else
    if (SpTextArch(component).font != NULL) {
	gdk_font_unref(SpTextArch(component).font);
    }
#endif
    
    return;
}

static void createTextFont(spComponent component)
{
    if (!strnone(SpTextPart(component).font_name)) {
#if GTK_CHECK_VERSION(2,0,0)
	if (SpTextArch(component).font_desc != NULL) {
	    pango_font_description_free(SpTextArch(component).font_desc);
	}
	
	spDebug(30, "createTextFont", "font_name = %s\n", SpTextPart(component).font_name);
    
	SpTextArch(component).font_desc = pango_font_description_from_string(SP_GTK_DEFAULT_FONT_NAME);
	spFontNameToFontDescriptionGtk(SpTextPart(component).font_name, SpTextArch(component).font_desc);

#if GTK_CHECK_VERSION(3,0,0)
	gtk_widget_override_font(SpPrimitiveArch(component).widget, SpTextArch(component).font_desc);
#else
	gtk_widget_modify_font(SpPrimitiveArch(component).widget, SpTextArch(component).font_desc);
#endif
#else
	GtkStyle *style;
	
	if (SpTextArch(component).font != NULL) {
	    gdk_font_unref(SpTextArch(component).font);
	}
	
	spDebug(30, "createTextFont", "font_name = %s\n", SpTextPart(component).font_name);
    
	if ((SpTextArch(component).font =
	     gdk_fontset_load(SpTextPart(component).font_name)) == NULL) {
	    SpTextArch(component).font =
		gdk_font_load(SpTextPart(component).font_name);
	}
	
	style = gtk_style_copy(gtk_widget_get_style(SpPrimitiveArch(component).widget));
#if GTK_CHECK_VERSION(2,0,0)
	gtk_style_set_font(style, SpTextArch(component).font);
#else
	gdk_font_unref(style->font);
	style->font = SpTextArch(component).font;
	gdk_font_ref(style->font);
#endif
	gtk_widget_set_style(SpPrimitiveArch(component).widget, style);
#endif
    }

    return;
}

#if !GTK_CHECK_VERSION(2,0,0)
static void textRealizeCB(GtkWidget *widget, spComponent component)
{
    if (!strnone(SpTextPart(component).string)) {
	gtk_text_freeze(GTK_TEXT(SpPrimitiveArch(component).widget));
	gtk_text_insert(GTK_TEXT(SpPrimitiveArch(component).widget), NULL, NULL, NULL,
			SpTextPart(component).string, -1);
	gtk_text_thaw(GTK_TEXT(SpPrimitiveArch(component).widget));
    }

    return;
}
#endif

void spTextCreateArch(spComponent component)
{
    if (spIsSubClass(component, SpTextField) == SP_TRUE) {
	SpPrimitiveArch(component).widget = gtk_entry_new();
	
#if GTK_CHECK_VERSION(3,0,0)
	gtk_editable_set_editable(GTK_EDITABLE(SpPrimitiveArch(component).widget),
				  (SpTextPart(component).editable == SP_TRUE ? TRUE : FALSE));
	
	gtk_entry_set_width_chars(GTK_ENTRY(SpPrimitiveArch(component).widget), 4);
#else
	gtk_entry_set_editable(GTK_ENTRY(SpPrimitiveArch(component).widget),
			       (SpTextPart(component).editable == SP_TRUE ? TRUE : FALSE));
#endif
	
	spPrimitiveContainerAddGtk(component);
    } else {
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0)
	SpPrimitiveArch(component).top_widget = gtk_scrolled_window_new(NULL, NULL);
#else
	SpPrimitiveArch(component).top_widget = gtk_table_new(2, 2, FALSE);
	gtk_table_set_row_spacing(GTK_TABLE(SpPrimitiveArch(component).top_widget), 0, 2);
	gtk_table_set_col_spacing(GTK_TABLE(SpPrimitiveArch(component).top_widget), 0, 2);
#endif	
	spContainerAddGtk(component, SpPrimitiveArch(component).top_widget);
	if (SpComponentPart(component).visible_flag == SP_TRUE) {
	    gtk_widget_show(SpPrimitiveArch(component).top_widget);
	}
	
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0)
	SpPrimitiveArch(component).widget = gtk_text_view_new();
	gtk_container_add(GTK_CONTAINER(SpPrimitiveArch(component).top_widget),
			  SpPrimitiveArch(component).widget);
#else
	SpPrimitiveArch(component).widget = gtk_text_new(NULL, NULL);
	gtk_table_attach(GTK_TABLE(SpPrimitiveArch(component).top_widget),
			 SpPrimitiveArch(component).widget, 0, 1, 0, 1,
			 GTK_EXPAND | GTK_SHRINK | GTK_FILL,
			 GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
#endif
	gtk_widget_show(SpPrimitiveArch(component).widget);

#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0)
#else
	SpTextArch(component).vscroll_bar =
	    gtk_vscrollbar_new(GTK_TEXT(SpPrimitiveArch(component).widget)->vadj);
	gtk_table_attach(GTK_TABLE(SpPrimitiveArch(component).top_widget),
			 SpTextArch(component).vscroll_bar, 1, 2, 0, 1,
			 GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
	gtk_widget_show(SpTextArch(component).vscroll_bar);
#endif
	
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0)
	gtk_text_view_set_editable(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget),
				   (SpTextPart(component).editable == SP_TRUE ? TRUE : FALSE));
#else
	gtk_text_set_editable(GTK_TEXT(SpPrimitiveArch(component).widget),
			      (SpTextPart(component).editable == SP_TRUE ? TRUE : FALSE));
	
	SpPrimitivePart(component).index =
	    gtk_signal_connect(GTK_OBJECT(SpPrimitiveArch(component).widget),
			       "realize", GTK_SIGNAL_FUNC(textRealizeCB),
			       (gpointer)component);
#endif
    }

    createTextFont(component);
    
    return;
}

void spTextDestroyArch(spComponent component)
{
    return;
}

void spTextSetParamsArch(spComponent component)
{
    if (SpTextPart(SpOldObject(component)).editable != SpTextPart(component).editable) {
	if (spIsSubClass(component, SpTextField) == SP_TRUE) {
#if GTK_CHECK_VERSION(3,0,0)
	    gtk_editable_set_editable(GTK_EDITABLE(SpPrimitiveArch(component).widget),
				      (SpTextPart(component).editable == SP_TRUE ? TRUE : FALSE));
#else
	    gtk_entry_set_editable(GTK_ENTRY(SpPrimitiveArch(component).widget),
				   (SpTextPart(component).editable == SP_TRUE ? TRUE : FALSE));
#endif
	} else {
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0)
	    gtk_text_view_set_editable(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget),
				       (SpTextPart(component).editable == SP_TRUE ? TRUE : FALSE));
#else
	    gtk_text_set_editable(GTK_TEXT(SpPrimitiveArch(component).widget),
				  (SpTextPart(component).editable == SP_TRUE ? TRUE : FALSE));
#endif
	}
    }
    
    if (!streq(SpTextPart(SpOldObject(component)).font_name, SpTextPart(component).font_name)) {
	createTextFont(component);
    }
    
    return;
}

void spSetTextStringArch(spComponent component)
{
    char *string;

    spDebug(30, "spSetTextStringArch", "string = %s\n",
	    SpTextPart(component).string);
    
    SpTextArch(component).suspend_callback = SP_TRUE;
    
#if GTK_CHECK_VERSION(2,0,0)
    string = (char *)g_locale_to_utf8(SpTextPart(component).string, -1, NULL, NULL, NULL);
#else
    string = SpTextPart(component).string;
#endif

    if (spIsComboBox(component) == SP_TRUE) {
	gtk_entry_set_text(GTK_ENTRY(spGetComboBoxEntryGtk(component)),
			   string);
    } else if (spIsSubClass(component, SpTextField) == SP_TRUE) {
	gtk_entry_set_text(GTK_ENTRY(SpPrimitiveArch(component).widget),
			   string);
    } else {
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0)
	GtkTextBuffer *buffer;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget));
	gtk_text_buffer_set_text(buffer, string, -1);
#else
	gint position;
	
	if (GTK_WIDGET_REALIZED(SpPrimitiveArch(component).widget)) {
	    spDebug(30, "spSetTextStringArch", "text string to text area\n");

	    gtk_text_freeze(GTK_TEXT(SpPrimitiveArch(component).widget));
	    
#if 1
	    position = 0;
	    gtk_editable_insert_text(GTK_EDITABLE(SpPrimitiveArch(component).widget),
				     string,
				     strlen(string),
				     &position);
	    spDebug(30, "spSetTextStringArch", "position = %d\n", position);
	    gtk_editable_delete_text(GTK_EDITABLE(SpPrimitiveArch(component).widget),
				     position, -1);
#else
	    /* the following code sometimes doesn't work */
	    gtk_editable_delete_text(GTK_EDITABLE(SpPrimitiveArch(component).widget),
				     0, -1);
	    gtk_text_insert(GTK_TEXT(SpPrimitiveArch(component).widget), NULL, NULL, NULL,
			    string, -1);
#endif
	    
	    gtk_text_thaw(GTK_TEXT(SpPrimitiveArch(component).widget));
	}
#endif
    }
    spSetTextPositionArch(component, -1);
    
#if GTK_CHECK_VERSION(2,0,0)
    g_free(string);
#endif
    
    SpTextArch(component).suspend_callback = SP_FALSE;
    
    spDebug(30, "spSetTextStringArch", "done\n");
    
    return;
}

void spGetTextStringArch(spComponent component)
{
    char *string;
    
    if (spIsComboBox(component) == SP_TRUE) {
	string =
	    gtk_editable_get_chars(GTK_EDITABLE(spGetComboBoxEntryGtk(component)),
				   0, -1);
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0) 
    } else if (spIsSubClass(component, SpTextArea) == SP_TRUE) {
	GtkTextBuffer *buffer;
	GtkTextIter start, end;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget));
	gtk_text_buffer_get_start_iter(buffer, &start);
	gtk_text_buffer_get_end_iter(buffer, &end);
	string = gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
#endif
    } else {
	string = gtk_editable_get_chars(GTK_EDITABLE(SpPrimitiveArch(component).widget),
					0, -1);
    }
#if GTK_CHECK_VERSION(2,0,0)
    {
	gchar *ustring;
	
	ustring = (char *)g_locale_from_utf8(string, -1, NULL, NULL, NULL);
	SpTextPart(component).string = strclone(ustring);
	g_free(ustring);
    }
#else
    SpTextPart(component).string = strclone(string);
#endif
    g_free(string);
    
    return;
}

spBool spSetTextSelectionArch(spComponent component, long start, long end)
{
    if (start < 0) {
	/* is this correct? */
	start = 0;
	end = 0;
    }
    
    if (spIsComboBox(component) == SP_TRUE) {
	if (end > spGetComboBoxTextLengthGtk(component)) {
	    end = spGetComboBoxTextLengthGtk(component);
	}
	gtk_editable_select_region(GTK_EDITABLE(spGetComboBoxEntryGtk(component)),
				   (gint)start, (gint)end);
    } else {
	if (spIsSubClass(component, SpTextField) == SP_TRUE) {
	    if (end > spGetTextFieldTextLengthGtk(component)) {
		end = spGetTextFieldTextLengthGtk(component);
	    }
	    gtk_editable_select_region(GTK_EDITABLE(SpPrimitiveArch(component).widget),
				       (gint)start, (gint)end);
	} else {
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0)
	    GtkTextIter tstart, tend;
	    GtkTextBuffer *buffer;
	
	    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget));
	    gtk_text_buffer_get_iter_at_offset(buffer, &tstart, (gint)start);
	    gtk_text_buffer_get_iter_at_offset(buffer, &tend, (gint)end);
	    
	    gtk_text_buffer_select_range(buffer, &tstart, &tend);
#else
	    if (end > gtk_text_get_length(GTK_TEXT(SpPrimitiveArch(component).widget))) {
		end = gtk_text_get_length(GTK_TEXT(SpPrimitiveArch(component).widget));
	    }
	    gtk_editable_select_region(GTK_EDITABLE(SpPrimitiveArch(component).widget),
				       (gint)start, (gint)end);
#endif
	}
    }
    
    return SP_TRUE;
}

spBool spGetTextSelectionArch(spComponent component, long *start, long *end)
{
#if GTK_CHECK_VERSION(2,0,0)
    gint tstart, tend;

    if (spIsComboBox(component) == SP_TRUE) {
	if (!gtk_editable_get_selection_bounds(GTK_EDITABLE(SpPrimitiveArch(component).widget), &tstart, &tend)) {
	    return SP_FALSE;
	} else {
	    *start = (long)tstart;
	    *end = (long)tend;
	}
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0) 
    } else if (spIsSubClass(component, SpTextArea) == SP_TRUE) {
	GtkTextIter tstart, tend;
	GtkTextBuffer *buffer;
	
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget));
	
	if (!gtk_text_buffer_get_selection_bounds(buffer, &tstart, &tend)) {
	    return SP_FALSE;
	} else {
	    *start = (long)gtk_text_iter_get_offset(&tstart);
	    *end = (long)gtk_text_iter_get_offset(&tend);
	}
#endif
    } else {
	if (!gtk_editable_get_selection_bounds(GTK_EDITABLE(SpPrimitiveArch(component).widget), &tstart, &tend)) {
	    return SP_FALSE;
	} else {
	    *start = (long)tstart;
	    *end = (long)tend;
	}
    }
#else
    if (spIsComboBox(component) == SP_TRUE) {
	if (!GTK_EDITABLE(spGetComboBoxEntryGtk(component))->has_selection) {
	    return SP_FALSE;
	} else {
	    *start = GTK_EDITABLE(spGetComboBoxEntryGtk(component))->selection_start_pos;
	    *end = GTK_EDITABLE(spGetComboBoxEntryGtk(component))->selection_end_pos;
	}
    } else {
	if (!GTK_EDITABLE(SpPrimitiveArch(component).widget)->has_selection) {
	    return SP_FALSE;
	} else {
	    *start = GTK_EDITABLE(SpPrimitiveArch(component).widget)->selection_start_pos;
	    *end = GTK_EDITABLE(SpPrimitiveArch(component).widget)->selection_end_pos;
	}
    }
#endif
    
    return SP_TRUE;
}

spBool spSetTextPositionArch(spComponent component, long position)
{
    if (spIsComboBox(component) == SP_TRUE) {
	if (position == -1
	    || position > spGetComboBoxTextLengthGtk(component)) {
	    position = spGetComboBoxTextLengthGtk(component);
	}
	gtk_editable_set_position(GTK_EDITABLE(spGetComboBoxEntryGtk(component)),
				  (gint)position);
    } else {
	if (spIsSubClass(component, SpTextField) == SP_TRUE) {
	    if (position == -1
		|| position > spGetTextFieldTextLengthGtk(component)) {
		position = spGetTextFieldTextLengthGtk(component);
	    }
	    gtk_editable_set_position(GTK_EDITABLE(SpPrimitiveArch(component).widget),
				      (gint)position);
	} else {
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0)
	    GtkTextBuffer *buffer;
	    GtkTextIter iter;
	    
	    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget));
	    gtk_text_buffer_get_iter_at_offset(buffer, &iter, (gint)position);
	    gtk_text_buffer_place_cursor(buffer, &iter);
#else
	    if (position == -1
		|| position > gtk_text_get_length(GTK_TEXT(SpPrimitiveArch(component).widget))) {
		position = gtk_text_get_length(GTK_TEXT(SpPrimitiveArch(component).widget));
	    }
	    gtk_editable_set_position(GTK_EDITABLE(SpPrimitiveArch(component).widget),
				      (gint)position);
#endif
	}
    }
    
    return SP_TRUE;
}

spBool spGetTextPositionArch(spComponent component, long *position)
{
    if (spIsComboBox(component) == SP_TRUE) {
#if GTK_CHECK_VERSION(2,0,0)
	*position = gtk_editable_get_position(GTK_EDITABLE(spGetComboBoxEntryGtk(component)));
#else
	*position = GTK_EDITABLE(spGetComboBoxEntryGtk(component))->current_pos;
#endif
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0) 
    } else if (spIsSubClass(component, SpTextArea) == SP_TRUE) {
	GtkTextBuffer *buffer;
	GtkTextIter iter;
	GtkTextMark *mark;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget));
	mark = gtk_text_buffer_get_insert(buffer);
	gtk_text_buffer_get_iter_at_mark(buffer, &iter, mark);
	*position = (long)gtk_text_iter_get_offset(&iter);
	spDebug(30, "spGetTextPositionArch", "position = %ld\n", *position);
#endif
    } else {
#if GTK_CHECK_VERSION(2,0,0)
	*position = gtk_editable_get_position(GTK_EDITABLE(SpPrimitiveArch(component).widget));
#else
	*position = GTK_EDITABLE(SpPrimitiveArch(component).widget)->current_pos;
#endif
    }
    
    return SP_TRUE;
}

spBool spCutTextArch(spComponent component)
{
    if (spIsComboBox(component) == SP_TRUE) {
	gtk_editable_cut_clipboard(GTK_EDITABLE(spGetComboBoxEntryGtk(component)));
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0) 
    } else if (spIsSubClass(component, SpTextArea) == SP_TRUE) {
	GtkTextBuffer *buffer;
	GtkClipboard *clipboard;
	
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget));
	clipboard = gtk_widget_get_clipboard(GTK_WIDGET(SpPrimitiveArch(component).widget),
					     GDK_SELECTION_CLIPBOARD);
	gtk_text_buffer_cut_clipboard(buffer, clipboard, SpTextPart(component).editable);
#endif
    } else {
	gtk_editable_cut_clipboard(GTK_EDITABLE(SpPrimitiveArch(component).widget));
    }
    return SP_TRUE;
}

spBool spCopyTextArch(spComponent component)
{
    if (spIsComboBox(component) == SP_TRUE) {
	gtk_editable_copy_clipboard(GTK_EDITABLE(spGetComboBoxEntryGtk(component)));
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0) 
    } else if (spIsSubClass(component, SpTextArea) == SP_TRUE) {
	GtkTextBuffer *buffer;
	GtkClipboard *clipboard;
	
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget));
	clipboard = gtk_widget_get_clipboard(GTK_WIDGET(SpPrimitiveArch(component).widget),
					     GDK_SELECTION_CLIPBOARD);
	gtk_text_buffer_copy_clipboard(buffer, clipboard);
#endif
    } else {
	gtk_editable_copy_clipboard(GTK_EDITABLE(SpPrimitiveArch(component).widget));
    }
    return SP_TRUE;
}

spBool spPasteTextArch(spComponent component)
{
    if (spIsComboBox(component) == SP_TRUE) {
	gtk_editable_paste_clipboard(GTK_EDITABLE(spGetComboBoxEntryGtk(component)));
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0) 
    } else if (spIsSubClass(component, SpTextArea) == SP_TRUE) {
	GtkTextBuffer *buffer;
	GtkClipboard *clipboard;
	
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget));
	clipboard = gtk_widget_get_clipboard(GTK_WIDGET(SpPrimitiveArch(component).widget),
					     GDK_SELECTION_CLIPBOARD);
	gtk_text_buffer_paste_clipboard(buffer, clipboard, NULL, SpTextPart(component).editable);
#endif
    } else {
	gtk_editable_paste_clipboard(GTK_EDITABLE(SpPrimitiveArch(component).widget));
    }
    return SP_TRUE;
}

spBool spClearTextArch(spComponent component)
{
    if (spIsComboBox(component) == SP_TRUE) {
	gtk_editable_delete_selection(GTK_EDITABLE(spGetComboBoxEntryGtk(component)));
#if !defined(GTK_ENABLE_BROKEN) && GTK_CHECK_VERSION(2,0,0) 
    } else if (spIsSubClass(component, SpTextArea) == SP_TRUE) {
	GtkTextBuffer *buffer;
	
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(SpPrimitiveArch(component).widget));
	gtk_text_buffer_delete_selection(buffer, FALSE, SpTextPart(component).editable);
#endif
    } else {
	gtk_editable_delete_selection(GTK_EDITABLE(SpPrimitiveArch(component).widget));
    }
    return SP_TRUE;
}
