/*
 *	spToolItem_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>

#include <sp/spToolBarP.h>
#include <sp/spToolItemP.h>

extern spTopLevel sp_toplevel;

void spToolItemPartInitArch(spComponent component)
{
#if GTK_CHECK_VERSION(3,0,0)
    SpToolItemArch(component).pixbuf = NULL;
    SpToolItemArch(component).icon_set = NULL;
    SpToolItemArch(component).icon_widget = NULL;
#else
    SpToolItemArch(component).bitmap = NULL;
    SpToolItemArch(component).insens_pixmap = NULL;
#endif
    return;
}

void spToolItemPartFreeArch(spComponent component)
{
    return;
}

/*#if !GTK_CHECK_VERSION(3,0,0)*/
#if !GTK_CHECK_VERSION(1,2,0)
static void drawInsensitivePixmap(GdkPixmap *pixmap, int width, int height)
{
    int x, y;
    int size = 2;

    for (y = 0; y < height; y += size) {
	for (x = 0; x < width; x += size) {
	    if (((x / size) % 2 == 0 && (y / size) % 2 == 0)
		|| ((x / size) % 2 == 1 && (y / size) % 2 == 1)) {
		gdk_draw_rectangle(pixmap,
				   SpTopLevelArch(sp_toplevel).bg_gc,
				   TRUE, 
				   x, y, size, size);
	    }
	}
    }

    return;
}
#endif

static void addToolItem(GtkWidget *widget, gpointer data)
{
    spComponent component = (spComponent)data;
	    
#if GTK_CHECK_VERSION(3,0,0)
    if (SpParentToolBarArch(component).pixbuf != NULL) {
	spDebug(50, "addToolItem", "parent's pixbuf exists, bitmap_width = %d, bitmap_height = %d\n",
		SpParentToolBarPart(component).bitmap_width,
		SpParentToolBarPart(component).bitmap_height);
	
	if (spIsSubClass(component, SpToolSeparator) == SP_TRUE) {
	    SpPrimitiveArch(component).widget = GTK_WIDGET(gtk_separator_tool_item_new());
	} else {
	    spDebug(50, "addToolItem", "index = %d\n", SpPrimitivePart(component).index);
	    SpToolItemArch(component).pixbuf
		= gdk_pixbuf_new_subpixbuf(SpParentToolBarArch(component).pixbuf,
					   SpPrimitivePart(component).index * SpParentToolBarPart(component).bitmap_width, 0,
					   SpParentToolBarPart(component).bitmap_width,
					   SpParentToolBarPart(component).bitmap_height);

#if 1	/* The following code is maybe OK (GTK_ICON_SIZE_MENU is not good?). */
	    SpToolItemArch(component).icon_set = gtk_icon_set_new_from_pixbuf(SpToolItemArch(component).pixbuf);

#if 0
	    {
		gint i;
		gint width, height;
		GtkIconSize *sizes;
		gint n_sizes;
		gtk_icon_set_get_sizes(SpToolItemArch(component).icon_set,
				       &sizes, &n_sizes);

		for (i = 0; i < n_sizes; i++) {
		    gtk_icon_size_lookup(sizes[i], &width, &height);
		    spDebug(50, "addToolItem", "sizes[%d] %d: width = %d, height = %d\n",
			    i, sizes[i], width, height);
		}

		g_free(sizes);
	    }
#endif
	    
	    SpToolItemArch(component).icon_widget = gtk_image_new_from_icon_set(SpToolItemArch(component).icon_set,
										GTK_ICON_SIZE_MENU);
#else
	    /* insensitive icon doesn't work */
	    SpToolItemArch(component).icon_widget = gtk_image_new_from_pixbuf(SpToolItemArch(component).pixbuf);
#endif
	    gtk_widget_show(SpToolItemArch(component).icon_widget);
	    
	    if (spIsSubClass(component, SpCheckToolItem)) {
		SpPrimitiveArch(component).widget = GTK_WIDGET(gtk_toggle_tool_button_new());
		gtk_tool_button_set_icon_widget(GTK_TOOL_BUTTON(SpPrimitiveArch(component).widget),
						SpToolItemArch(component).icon_widget);
		gtk_tool_button_set_label(GTK_TOOL_BUTTON(SpPrimitiveArch(component).widget),
					  SpGetTitle(component));
	    } else {
		SpPrimitiveArch(component).widget = GTK_WIDGET(gtk_tool_button_new(SpToolItemArch(component).icon_widget,
										   SpGetTitle(component)));
	    }
	}
	gtk_widget_show(SpPrimitiveArch(component).widget);
	
	gtk_toolbar_insert(GTK_TOOLBAR(SpParentPrimitiveArch(component).widget),
			   GTK_TOOL_ITEM(SpPrimitiveArch(component).widget), -1);
    }
#else
    if (spIsSubClass(component, SpToolSeparator) == SP_TRUE) {
	return;
    }
	
    if (SpParentPrimitiveArch(component).pixmap != NULL) {
	GdkBitmap *bitmap = NULL;
	
	/* create pixmap */
	SpPrimitiveArch(component).pixmap =
	    gdk_pixmap_new(SpParentPrimitiveArch(component).widget->window,
			   SpParentToolBarPart(component).bitmap_width,
			   SpParentToolBarPart(component).bitmap_height,
			   -1);
	gdk_draw_pixmap(SpPrimitiveArch(component).pixmap,
			SpTopLevelArch(sp_toplevel).fg_gc,
			SpParentPrimitiveArch(component).pixmap,
			SpPrimitivePart(component).index * SpParentToolBarPart(component).bitmap_width,
			0,
			0, 0,
			SpParentToolBarPart(component).bitmap_width,
			SpParentToolBarPart(component).bitmap_height);

#if !GTK_CHECK_VERSION(1,2,0)
	/* create insensitive pixmap */
	SpToolItemArch(component).insens_pixmap = 
	    gdk_pixmap_new(SpParentPrimitiveArch(component).widget->window,
			   SpParentToolBarPart(component).bitmap_width,
			   SpParentToolBarPart(component).bitmap_height,
			   -1);
	gdk_draw_pixmap(SpToolItemArch(component).insens_pixmap,
			SpTopLevelArch(sp_toplevel).fg_gc,
			SpParentPrimitiveArch(component).pixmap,
			SpPrimitivePart(component).index * SpParentToolBarPart(component).bitmap_width,
			0,
			0, 0,
			SpParentToolBarPart(component).bitmap_width,
			SpParentToolBarPart(component).bitmap_height);
	drawInsensitivePixmap(SpToolItemArch(component).insens_pixmap,
			      SpParentToolBarPart(component).bitmap_width,
			      SpParentToolBarPart(component).bitmap_height);
#endif

	/* create mask bitmap */
	if (SpParentToolBarArch(component).bitmap != NULL) {
	    GdkGC *gc;
	    
	    SpToolItemArch(component).bitmap =
		gdk_pixmap_new(NULL,
			       SpParentToolBarPart(component).bitmap_width,
			       SpParentToolBarPart(component).bitmap_height,
			       1);

	    gc = gdk_gc_new(SpToolItemArch(component).bitmap);
	    
	    gdk_draw_pixmap(SpToolItemArch(component).bitmap,
			    gc,
			    SpParentToolBarArch(component).bitmap,
			    SpPrimitivePart(component).index * SpParentToolBarPart(component).bitmap_width,
			    0,
			    0, 0,
			    SpParentToolBarPart(component).bitmap_width,
			    SpParentToolBarPart(component).bitmap_height);
	    
	    gdk_gc_destroy(gc);
	    
	    bitmap = SpToolItemArch(component).bitmap;
	}
	
	SpPrimitiveArch(component).sub_widget =
	    gtk_pixmap_new(SpPrimitiveArch(component).pixmap, bitmap);
	
	if (spIsSubClass(component, SpCheckToolItem)) {
	    SpPrimitiveArch(component).widget =
		gtk_toolbar_append_element(GTK_TOOLBAR(SpParentPrimitiveArch(component).widget),
					   GTK_TOOLBAR_CHILD_TOGGLEBUTTON, NULL,
					   SpComponentPart(component).description,
					   SpComponentPart(component).description,
					   NULL,
					   SpPrimitiveArch(component).sub_widget,
					   NULL, NULL);
	} else {
	    SpPrimitiveArch(component).widget =
		gtk_toolbar_append_item(GTK_TOOLBAR(SpParentPrimitiveArch(component).widget),
					SpComponentPart(component).description,
					SpComponentPart(component).description,
					NULL,
					SpPrimitiveArch(component).sub_widget,
					NULL, NULL);
	}
	
	if (spIsSubClass(component, SpCheckToolItem)) {
	    gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(SpPrimitiveArch(component).widget),
					(SpToolItemPart(component).set == SP_ON ? TRUE : FALSE));
	}
    }
#endif
    
    spPrimitiveSetSensitiveArch(component, SpComponentPart(component).sensitive_flag);
    
    return;
}
    
void spToolItemCreateArch(spComponent component)
{
#if GTK_CHECK_VERSION(3,0,0)
    addToolItem(SpParentPrimitiveArch(component).widget, component);
#else
    if (spIsSubClass(component, SpToolSeparator) == SP_FALSE) {
	if (GTK_WIDGET_REALIZED(SpParentPrimitiveArch(component).widget)) {
	    addToolItem(SpParentPrimitiveArch(component).widget, component);
	} else {
	    spSignalConnectCallbackGtk(SpParentPrimitiveArch(component).widget,
				       "realize", addToolItem, (gpointer)component);
	}
    }
#endif
    
    return;
}

void spToolItemSetParamsArch(spComponent component)
{
    return;
}

void spToolItemDestroyArch(spComponent component)
{
#if GTK_CHECK_VERSION(3,0,0)
    if (SpToolItemArch(component).icon_widget != NULL) {
	gtk_widget_destroy(SpToolItemArch(component).icon_widget);
    }
    if (SpToolItemArch(component).icon_set != NULL) {
	gtk_icon_set_unref(SpToolItemArch(component).icon_set);
    }
    if (SpToolItemArch(component).pixbuf != NULL) {
	g_object_unref(SpToolItemArch(component).pixbuf);
    }
#else
    if (SpToolItemArch(component).bitmap != NULL) {
	gdk_bitmap_unref(SpToolItemArch(component).bitmap);
    }
    
#if !GTK_CHECK_VERSION(1,2,0)
    if (SpToolItemArch(component).insens_pixmap != NULL) {
	gdk_pixmap_unref(SpToolItemArch(component).insens_pixmap);
    }
#endif
#endif

    return;
}

void spSetToolItemToggleStateArch(spComponent component)
{
    if (
#if GTK_CHECK_VERSION(3,0,0)
	1
#else
	GTK_WIDGET_REALIZED(SpParentPrimitiveArch(component).widget)
#endif
	) {
#if GTK_CHECK_VERSION(3,0,0)
	gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(SpPrimitiveArch(component).widget),
					  (SpToolItemPart(component).set == SP_ON ? TRUE : FALSE));
#else
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(SpPrimitiveArch(component).widget),
				    (SpToolItemPart(component).set == SP_ON ? TRUE : FALSE));
#endif
    }
    
    return;
}

void spGetToolItemToggleStateArch(spComponent component)
{
    if (
#if GTK_CHECK_VERSION(3,0,0)
	1
#else
	GTK_WIDGET_REALIZED(SpParentPrimitiveArch(component).widget)
#endif
	) {
	if (
#if GTK_CHECK_VERSION(3,0,0)
	    gtk_toggle_tool_button_get_active(GTK_TOGGLE_TOOL_BUTTON(SpPrimitiveArch(component).widget))
#else
	    GTK_TOGGLE_BUTTON(SpPrimitiveArch(component).widget)->active
#endif
	    ) {
	    SpToolItemPart(component).set = SP_TRUE;
	} else {
	    SpToolItemPart(component).set = SP_FALSE;
	}
    }
    
    return;
}
