/*
 *	spTopLevel_Gtk.c
 */

#include <stdio.h>
#include <stdlib.h>

#include <glib.h>
#include <gdk/gdk.h>

#include <sp/spDefs.h>
#include <sp/spBase.h>
#include <sp/spMemory.h>
#include <sp/spThread.h>

#include <sp/spLocaleP.h>
#include <sp/spComponentP.h>
#include <sp/spTopLevelP.h>
#include <sp/spGraphicsP.h>
#include <sp/spPrimitiveP.h>
#include <sp/spFrameP.h>

#if !GTK_CHECK_VERSION(3,0,0)
#include <gdk/gdkx.h>
#endif
#if !GTK_CHECK_VERSION(2,0,0)
#include <gtk/gtkinvisible.h>
#endif

extern spTopLevel sp_toplevel;

#if (defined(HAVE_PTHREAD) || defined(__CYGWIN32__))
#include <pthread.h>

static pthread_t sp_main_thread_id;
#if defined(SP_GTK_USE_DRAW_MUTEX)
static pthread_mutex_t sp_draw_mutex;
#endif
#if defined(SP_GTK_USE_MAIN_LOOP_MUTEX)
static pthread_mutex_t sp_thread_loop_mutex, sp_main_loop_mutex, sp_thread_enter_mutex;
#endif
static volatile spBool sp_gtk_in_modal_loop = SP_FALSE;
#endif

#if GTK_CHECK_VERSION(1,2,0)
GtkWidget *sp_atom_selection_widget = NULL;
#endif

static volatile spBool sp_gtk_main_loop_exist = SP_FALSE;

#if defined(SP_GTK_USE_ASYNC_EVENTS_QUEUE)
static GAsyncQueue *sp_events_queue = NULL;

static gboolean eventsDispatch(GSource *source, GSourceFunc callback, gpointer user_data)
{
    gpointer data;
    gboolean result;

    spDebug(80, "eventsDispatch", "in\n");

    data = g_async_queue_pop(sp_events_queue);
    spDebug(50, "eventsDispatch", "g_async_queue_pop done\n");
    result = callback(data);
    g_free(data);
    
    spDebug(80, "eventsDispatch", "done: result = %d\n", result);
    
    return result;
}


static gboolean eventsPrepare(GSource *source, gint *timeout)
{
    spDebug(80, "eventsPrepare", "in\n");
    *timeout=-1;
    return g_async_queue_length(sp_events_queue)>0;
}


static gboolean eventsCheck(GSource *source)
{
    spDebug(80, "eventsCheck", "in\n");
    return g_async_queue_length(sp_events_queue)>0;
}

static GSourceFuncs spEventsFunctionsGtk = {
    eventsPrepare,
    eventsCheck,
    eventsDispatch,
    NULL,
};
#else
gint spMainIdleFuncGtk(gpointer data)
{
    spTopLevel toplevel = (spTopLevel)data;

    spDebug(100, "spMainIdleFuncGtk", "in\n");
    spProcessPendingUserEventDefault(&SpTopLevelArch(toplevel).user_event, 0);
    spDebug(100, "spMainIdleFuncGtk", "done\n");
    
    return TRUE;
}
#endif

void spTopLevelPartInitArch(spTopLevel toplevel)
{
    return;
}

void spTopLevelPartFreeArch(spTopLevel toplevel)
{
    return;
}

spBool spIsUnityUsedGtk(void)
{
    char *ds;
    
    if ((ds = getenv("DESKTOP_SESSION")) != NULL) {
	spDebug(100, "spIsUnityUsedGtk", "ds = %s\n", ds);
	if (strcaseeq(ds, "ubuntu") || strcaseeq(ds, "ubuntu-2d")) {
	    spDebug(80, "spIsUnityUsedGtk", "unity used!\n");
	    return SP_TRUE;
	}
    }

    spDebug(100, "spIsUnityUsedGtk", "failed\n");

    return SP_FALSE;
}

spBool spIsMenuProxyUsedGtk(void)
{
    char *env;
    
    if ((env = getenv("UBUNTU_MENUPROXY")) != NULL) {
	spDebug(100, "spIsMenuProxyUsedGtk", "env = %s\n", env);
	if (strneq(env, "libappmenu", 10)) {
	    spDebug(80, "spIsMenuProxyUsedGtk", "menu proxy used!\n");
	    return SP_TRUE;
	}
    }

    spDebug(100, "spIsMenuProxyUsedGtk", "failed\n");

    return SP_FALSE;
}

void spTopLevelCreateArch(spTopLevel toplevel)
{
#if (defined(HAVE_PTHREAD) || defined(__CYGWIN32__)) && GTK_CHECK_VERSION(1,2,4) /* menus may not work before 1.2.4. */
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	spDebug(50, "spTopLevelCreateArch", "initialize thread\n");
	g_thread_init(NULL);
	spDebug(50, "spTopLevelCreateArch", "g_thread_init done\n");
	if (g_thread_supported()) {
#if GTK_CHECK_VERSION(2,0,0)
	    gdk_threads_init();
#endif
	    
#if defined(SP_GTK_USE_DRAW_MUTEX)
	    pthread_mutex_init(&sp_draw_mutex, NULL);
#endif
#if defined(SP_GTK_USE_MAIN_LOOP_MUTEX)
	    pthread_mutex_init(&sp_main_loop_mutex, NULL);
	    pthread_mutex_init(&sp_thread_loop_mutex, NULL);
	    pthread_mutex_init(&sp_thread_enter_mutex, NULL);
#endif
	    sp_main_thread_id = pthread_self();
	    spDebug(50, "spTopLevelCreateArch", "main_thread = %ld\n", (long)sp_main_thread_id);
	} else {
	    SpTopLevelPart(toplevel).thread_safe = SP_FALSE;
	}
    }
#else
    SpTopLevelPart(toplevel).thread_safe = SP_FALSE;
#endif
    
    gtk_init(SpTopLevelPart(toplevel).argcp, SpTopLevelPart(toplevel).argvp);
    spDebug(50, "spTopLevelCreateArch", "gtk_init done\n");
    
#if !GTK_CHECK_VERSION(3,0,0)
    {
	char buf[1024];
    
	sprintf(buf, "%s/.gtkrc", getenv("HOME"));
	gtk_rc_parse(buf);
	spDebug(50, "spTopLevelCreateArch", "gtkrc parse done\n");
    }
#endif

#if GTK_CHECK_VERSION(3,0,0)
    SpTopLevelArch(toplevel).visual = gdk_screen_get_rgba_visual(gdk_screen_get_default());
#else
#if GTK_CHECK_VERSION(2,0,0)
    SpTopLevelArch(toplevel).root_window = gdk_window_foreign_new(GDK_ROOT_WINDOW());
    SpTopLevelArch(toplevel).root_depth = gdk_drawable_get_depth(SpTopLevelArch(toplevel).root_window);
#else
    SpTopLevelArch(toplevel).root_window = NULL;
    SpTopLevelArch(toplevel).root_depth = /*gdk_visual_get_best_depth()*/24;
#endif
    spDebug(50, "spTopLevelCreateArch", "root_depth = %d\n", SpTopLevelArch(toplevel).root_depth);
    SpTopLevelArch(toplevel).pixmap = gdk_pixmap_new(NULL, 1, 1,
						     SpTopLevelArch(toplevel).root_depth);
    SpTopLevelArch(toplevel).colormap = gdk_colormap_get_system();
    
    gdk_color_black(SpTopLevelArch(toplevel).colormap,
		    &SpTopLevelArch(toplevel).black);
    gdk_color_white(SpTopLevelArch(toplevel).colormap,
		    &SpTopLevelArch(toplevel).white);
    gdk_color_parse("#d7d7d7", &SpTopLevelArch(toplevel).gray);
    gdk_color_alloc(SpTopLevelArch(toplevel).colormap, &SpTopLevelArch(toplevel).gray);
#endif
    
#if !GTK_CHECK_VERSION(2,0,0)
    if ((SpTopLevelArch(toplevel).font = gdk_fontset_load(SP_GTK_DEFAULT_FONT_NAME)) == NULL) {
	SpTopLevelArch(toplevel).font = gdk_fontset_load("-*-*-*-*-*--*-*-*-*-*-*-*-*");
    }
#else
    SpTopLevelArch(toplevel).font_desc = pango_font_description_from_string(SP_GTK_DEFAULT_FONT_NAME);
#endif
    
#if !GTK_CHECK_VERSION(3,0,0)
    {
	GdkGCValues gc_values;
	GdkGCValuesMask gc_values_mask;
	
	gc_values.foreground = SpTopLevelArch(toplevel).black;
	gc_values.background = SpTopLevelArch(toplevel).white;
	gc_values_mask = GDK_GC_FOREGROUND | GDK_GC_BACKGROUND;
#if !GTK_CHECK_VERSION(2,0,0)
	gc_values.font = SpTopLevelArch(toplevel).font;
	gc_values_mask |= GDK_GC_FONT;
#endif
	SpTopLevelArch(toplevel).fg_gc = gdk_gc_new_with_values(SpTopLevelArch(toplevel).pixmap,
								&gc_values,
								gc_values_mask);

	gc_values.foreground = SpTopLevelArch(toplevel).gray;
	gc_values.background = SpTopLevelArch(toplevel).black;
	gc_values_mask = GDK_GC_FOREGROUND | GDK_GC_BACKGROUND;
#if !GTK_CHECK_VERSION(2,0,0)
	gc_values.font = SpTopLevelArch(toplevel).font;
	gc_values_mask |= GDK_GC_FONT;
#endif
	SpTopLevelArch(toplevel).bg_gc = gdk_gc_new_with_values(SpTopLevelArch(toplevel).pixmap,
								&gc_values,
								gc_values_mask);
    }
#endif

#if GTK_CHECK_VERSION(2,0,0)
    SpTopLevelArch(toplevel).context = gdk_pango_context_get_for_screen(gdk_screen_get_default());
#endif

    SpTopLevelArch(toplevel).unity_used = spIsUnityUsedGtk();
    SpTopLevelArch(toplevel).menu_proxy_used = spIsMenuProxyUsedGtk();
    
#if defined(SP_GTK_USE_ASYNC_EVENTS_QUEUE)
    if (g_thread_supported()) {
	SpTopLevelArch(toplevel).events_queue = g_async_queue_new();
	sp_events_queue = SpTopLevelArch(toplevel).events_queue;
	SpTopLevelArch(toplevel).events_source
	    = g_source_new(&spEventsFunctionsGtk, sizeof(GSource));
	g_source_set_callback(SpTopLevelArch(toplevel).events_source,
			      spEventsCallbackGtk, NULL, NULL);
	g_source_attach(SpTopLevelArch(toplevel).events_source, NULL);
    }
    spSetExitFunc(exit);
#else
    SpTopLevelArch(toplevel).user_event = NULL;
#if GTK_CHECK_VERSION(2,0,0)
    SpTopLevelArch(toplevel).main_idle_func_id = g_idle_add(spMainIdleFuncGtk, toplevel);
#else
    SpTopLevelArch(toplevel).main_idle_func_id = gtk_idle_add(spMainIdleFuncGtk, toplevel);
#endif
    spSetExitFunc(exit);
#endif

#if GTK_CHECK_VERSION(1,2,0)
    if (!strnone(SpTopLevelPart(toplevel).unique_id)) {
	sp_atom_selection_widget = gtk_invisible_new();
	spAddAtomSelectionCallbackGtk(sp_atom_selection_widget, NULL);
	spAddAtomSelectionTargetGtk(sp_atom_selection_widget, SpTopLevelPart(toplevel).unique_id);
	gtk_widget_realize(sp_atom_selection_widget);
	spFindExcutedApplicationGtk(sp_atom_selection_widget);
	spSetAtomSelectionOwnerGtk(sp_atom_selection_widget);
    }
#endif
    
#if GLIB_CHECK_VERSION(2,2,0)
    {
	const char *string;
	string = spGetString(SpGetName(toplevel));
	spDebug(50, "spTopLevelCreateArch", "call g_set_application_name, name = %s, string = %s\n",
		SpGetName(toplevel), string);
	g_set_application_name(string);
    }
#endif
	
    spDebug(50, "spTopLevelCreateArch", "done\n");
    
    return;
}

void spTopLevelSetParamsArch(spTopLevel toplevel)
{
    return;
}
    
void spTopLevelDestroyArch(spTopLevel toplevel)
{
#if GTK_CHECK_VERSION(1,2,0)
    if (sp_atom_selection_widget != NULL) {
	gtk_widget_destroy(sp_atom_selection_widget);
	sp_atom_selection_widget = NULL;
    }
#endif
    
#if !GTK_CHECK_VERSION(3,0,0)
    if (SpTopLevelArch(toplevel).root_window != NULL) {
	gdk_window_unref(SpTopLevelArch(toplevel).root_window);
    }
    if (SpTopLevelArch(toplevel).pixmap != NULL) {
	gdk_pixmap_unref(SpTopLevelArch(toplevel).pixmap);
    }
#endif
#if GTK_CHECK_VERSION(2,0,0)
    if (SpTopLevelArch(toplevel).font_desc != NULL) {
	pango_font_description_free(SpTopLevelArch(toplevel).font_desc);
    }
#else
    if (SpTopLevelArch(toplevel).font != NULL) {
	gdk_font_unref(SpTopLevelArch(toplevel).font);
    }
#endif
#if !GTK_CHECK_VERSION(3,0,0)
    if (SpTopLevelArch(toplevel).fg_gc != NULL) {
	gdk_gc_unref(SpTopLevelArch(toplevel).fg_gc);
    }
    if (SpTopLevelArch(toplevel).bg_gc != NULL) {
	gdk_gc_unref(SpTopLevelArch(toplevel).bg_gc);
    }
#endif
#if defined(HAVE_PTHREAD) || defined(__CYGWIN32__)
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
#if defined(SP_GTK_USE_DRAW_MUTEX)
	pthread_mutex_destroy(&sp_draw_mutex);
#endif
#if defined(SP_GTK_USE_MAIN_LOOP_MUTEX)
	pthread_mutex_destroy(&sp_main_loop_mutex);
	pthread_mutex_destroy(&sp_thread_loop_mutex);
	pthread_mutex_destroy(&sp_thread_enter_mutex);
#endif
    }
#endif

#if GTK_CHECK_VERSION(2,0,0)
    if (SpTopLevelArch(toplevel).context != NULL) {
	g_object_unref(SpTopLevelArch(toplevel).context);
    }
#endif
    
#if defined(SP_GTK_USE_ASYNC_EVENTS_QUEUE)
    if (g_thread_supported()) {
	if (SpTopLevelArch(toplevel).events_source != NULL) {
	    g_source_destroy(SpTopLevelArch(toplevel).events_source);
	}
	if (SpTopLevelArch(toplevel).events_queue != NULL) {
	    g_async_queue_unref(SpTopLevelArch(toplevel).events_queue);
	}
    }
#else
#if GTK_CHECK_VERSION(2,0,0)
    g_source_remove(SpTopLevelArch(toplevel).main_idle_func_id);
#else
    gtk_idle_remove(SpTopLevelArch(toplevel).main_idle_func_id);
#endif
#endif
    
    return;
}

void spQuitArch(int status)
{
    if (sp_gtk_main_loop_exist == SP_TRUE) {
	gtk_main_quit();
    }
    return;
}

void spExitArch(int status)
{
    spExit(status);
    return;
}

int spWaitNextEventArch(spTopLevel toplevel, spBool lock_flag)
{
    gint flag;

    spDebug(50, "spWaitNextEventArch", "in\n");
    
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE
	&& sp_gtk_main_loop_exist == SP_FALSE) gdk_threads_enter();
    
#if GTK_CHECK_VERSION(2,0,0)
    /* disabled for 1.X: because gtk_events_pending of 1.X sometimes returns TRUE endlessly */
    while ((flag = gtk_events_pending()) > 0) {
	spDebug(50, "spWaitNextEventArch", "pending event exists: flag = %d\n", flag);
	gtk_main_iteration_do(TRUE);
    }
    spDebug(50, "spWaitNextEventArch", "pending event processed: flag = %d\n", flag);
#endif
    gtk_main_iteration_do(TRUE);
    
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE
	&& sp_gtk_main_loop_exist == SP_FALSE) gdk_threads_leave();
    
    spDebug(50, "spWaitNextEventArch", "done\n");
    
    return 1;
}

int spDispatchEventArch(spTopLevel toplevel, spBool lock_flag)
{
    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE
	&& sp_gtk_main_loop_exist == SP_FALSE) gdk_threads_enter();

#if GTK_CHECK_VERSION(2,0,0)
    while (gtk_events_pending() > 0) {
        gtk_main_iteration_do(TRUE);
    }
#endif
    gtk_main_iteration_do(FALSE);

    if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE
	&& sp_gtk_main_loop_exist == SP_FALSE) gdk_threads_leave();
    
    return 1;
}

void spMainQuitGtk(void)
{
    sp_gtk_in_modal_loop = SP_FALSE;
#if !GTK_CHECK_VERSION(2,4,0)
    gtk_main_quit();
#endif
    spDebug(50, "spMainQuitGtk", "gtk_main_quit done\n");
    
    return;
}

int spMainLoopGtk(GtkWidget *widget, spBool dialog_flag)
{
    gint result = 0;
    spDebug(50, "spMainLoopGtk", "in\n");
    
#if GTK_CHECK_VERSION(2,4,0)
    if (dialog_flag) {
	if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE
	    && sp_gtk_main_loop_exist == SP_FALSE) gdk_threads_enter();
    
	result = gtk_dialog_run(GTK_DIALOG(widget));
	spDebug(30, "spMainLoopGtk", "gtk_dialog_run result = %d\n", result);

	if (SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE
	    && sp_gtk_main_loop_exist == SP_FALSE) gdk_threads_leave();
    } else {
	sp_gtk_in_modal_loop = SP_TRUE;

	while (sp_gtk_in_modal_loop == SP_TRUE) {
	    spWaitNextEventArch(sp_toplevel, SP_FALSE);
	    spDebug(100, "spMainLoopGtk", "in loop: sp_gtk_in_modal_loop = %d\n", sp_gtk_in_modal_loop);
	}
    }
#else
#if defined(SP_GTK_USE_MAIN_LOOP_MUTEX) && (defined(HAVE_PTHREAD) || defined(__CYGWIN32__))
    if (g_thread_supported()) {
	if (pthread_self() != sp_main_thread_id) {
	    spDebug(30, "spMainLoopGtk", "from non-main thread: main_loop_exist = %d\n", sp_gtk_main_loop_exist);

	    /* prevent from going to next main loop of main thread */	
	    spDebug(30, "spMainLoopGtk", "in thread: before pthread_mutex_lock(&sp_thread_loop_mutex);\n");
	    pthread_mutex_lock(&sp_thread_loop_mutex);
		
	    if (sp_gtk_main_loop_exist == SP_TRUE) {
		gtk_main_quit();
		gdk_flush();		/* flush current events */
		gdk_threads_leave();	/* leave thread for a while to go back to main thread */
		
		/* just waiting for finish of main loop of main thread */
		spDebug(30, "spMainLoopGtk", "in thread: before pthread_mutex_lock(&sp_main_loop_mutex);\n");
		pthread_mutex_lock(&sp_main_loop_mutex);
		pthread_mutex_unlock(&sp_main_loop_mutex);
		
		gdk_threads_enter();	/* enter thread again */
	    }
	    
	    spDebug(30, "spMainLoopGtk", "in thread: main entered\n");
	    gtk_main();
	    spDebug(30, "spMainLoopGtk", "in thread: main finished\n");
	    
	    /* now, go back to next main loop of main thread */
	    pthread_mutex_unlock(&sp_thread_loop_mutex);
	} else {
	    spDebug(30, "spMainLoopGtk", "from main thread: main_loop_exist = %d\n", sp_gtk_main_loop_exist);
	    if (sp_gtk_main_loop_exist == SP_FALSE) {
		gdk_threads_enter();
	    } else {
#if 0
		gdk_threads_leave();
#endif
	    }
	    spDebug(30, "spMainLoopGtk", "main entered\n");
	    gtk_main();
	    spDebug(30, "spMainLoopGtk", "main finished: main_loop_exist = %d\n", sp_gtk_main_loop_exist);
	    if (sp_gtk_main_loop_exist == SP_FALSE) {
		gdk_threads_leave();
	    } else {
		spDebug(30, "spMainLoopGtk", "call gdk_threads_enter\n");
#if 0
		gdk_threads_enter();
#endif
	    }
	}
        spDebug(50, "spMainLoopGtk", "done: result = %d\n", result);

	return result;
    }
#endif
    
    gtk_main();
#endif
    
    spDebug(50, "spMainLoopGtk", "done\n");
    
    return result;
}

int spMainLoopArch(spTopLevel toplevel)
{
    while (1) {
	spDebug(30, "spMainLoopArch", "loop start\n");
#if (defined(HAVE_PTHREAD) || defined(__CYGWIN32__))
	if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
#if defined(SP_GTK_USE_MAIN_LOOP_MUTEX)
	    spDebug(30, "spMainLoopArch", "before pthread_mutex_lock(&sp_thread_enter_mutex);\n");
	    pthread_mutex_lock(&sp_thread_enter_mutex);
	    pthread_mutex_unlock(&sp_thread_enter_mutex);
	    
	    /* waiting for finish of thread loop */
	    spDebug(30, "spMainLoopArch", "before pthread_mutex_lock(&sp_thread_loop_mutex);\n");
	    pthread_mutex_lock(&sp_thread_loop_mutex);
	    pthread_mutex_unlock(&sp_thread_loop_mutex);
	    
	    pthread_mutex_lock(&sp_main_loop_mutex);
#endif
	    gdk_threads_enter();
	    spDebug(30, "spMainLoopArch", "main entered\n");
	    sp_gtk_main_loop_exist = SP_TRUE;
	    gtk_main();
	    sp_gtk_main_loop_exist = SP_FALSE;
	    gdk_threads_leave();
	    spDebug(30, "spMainLoopArch", "main quited\n");

#if defined(SP_GTK_USE_MAIN_LOOP_MUTEX)
	    pthread_mutex_unlock(&sp_main_loop_mutex);
#endif
	} else {
	    sp_gtk_main_loop_exist = SP_TRUE;
	    gtk_main();
	    sp_gtk_main_loop_exist = SP_FALSE;
	}
#else
	sp_gtk_main_loop_exist = SP_TRUE;
	gtk_main();
	sp_gtk_main_loop_exist = SP_FALSE;
#endif
	spDebug(30, "spMainLoopArch", "loop end\n");
    }
    
    return 0;
}

spBool spThreadEnterArch(spTopLevel toplevel)
{
#if (defined(HAVE_PTHREAD) || defined(__CYGWIN32__))
    spDebug(100, "spThreadEnterArch", "in\n");
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	if (pthread_self() != sp_main_thread_id) {
	    spDebug(100, "spThreadEnterArch", "in thread\n");
#if defined(SP_GTK_USE_MAIN_LOOP_MUTEX)
	    pthread_mutex_lock(&sp_thread_enter_mutex);
#endif
	    gdk_threads_enter();
	}
	return SP_TRUE;
    }
#endif
    return SP_FALSE;
}

spBool spThreadLeaveArch(spTopLevel toplevel)
{
#if (defined(HAVE_PTHREAD) || defined(__CYGWIN32__))
    spDebug(100, "spThreadLeaveArch", "in\n");
    if (SpTopLevelPart(toplevel).thread_safe == SP_TRUE) {
	if (pthread_self() != sp_main_thread_id) {
	    spDebug(100, "spThreadLeaveArch", "in thread\n");
	    /*gdk_flush();*/ /* comment out because this flush sometimes causes dead-lock */
	    gdk_threads_leave();
	    
#if defined(SP_GTK_USE_MAIN_LOOP_MUTEX)
	    pthread_mutex_unlock(&sp_thread_enter_mutex);
#else
	    spYieldThread();
#endif
	}
	return SP_TRUE;
    }
#endif
    return SP_FALSE;
}
    
spBool spSuspendMainLockArch(spTopLevel toplevel)
{
#if defined(SP_GTK_USE_MAIN_LOOP_MUTEX)
    {
	if (sp_gtk_main_loop_exist == SP_TRUE) {
#if 0
	    gdk_threads_leave();
#endif
	    return SP_TRUE;
	} else {
	    return SP_FALSE;
	}
    }
#endif
    
    return SP_TRUE;
}

spBool spResumeMainLockArch(spTopLevel toplevel)
{
#if defined(SP_GTK_USE_MAIN_LOOP_MUTEX)
    {
	if (sp_gtk_main_loop_exist == SP_TRUE) {
#if 0
	    gdk_threads_enter();
#endif
	    return SP_TRUE;
	} else {
	    return SP_FALSE;
	}
    }
#endif
    
    return SP_TRUE;
}

spBool spWaitThreadPostDummyEventArch(spTopLevel toplevel)
{
#if defined(SP_GTK_USE_ASYNC_EVENTS_QUEUE)
    spPostUserEventArch(NULL, 0, NULL);
#endif
    return SP_TRUE;
}

long spWaitThreadLoopArch(spTopLevel toplevel, void *handle)
{
    spBool dispatch_flag;

#if defined(SP_GTK_USE_ASYNC_EVENTS_QUEUE)
    dispatch_flag = SP_FALSE;
#else
    dispatch_flag = SP_TRUE;
#endif
    
    return spWaitThreadLoopDefault(toplevel, handle, dispatch_flag);
}

spBool spIsPostEventSupportedArch(spTopLevel toplevel)
{
#if defined(SP_GTK_USE_ASYNC_EVENTS_QUEUE)
    if (SpTopLevelArch(toplevel).events_queue != NULL) {
	return SP_TRUE;
    }
#else
    if (SpTopLevelArch(toplevel).main_idle_func_id > 0) {
	return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}

spBool spLockDrawMutexGtk(void)
{
#if defined(HAVE_PTHREAD) || defined(__CYGWIN32__)
    if (sp_toplevel != NULL
	&& SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) {
#if defined(SP_GTK_USE_DRAW_MUTEX)
	pthread_mutex_lock(&sp_draw_mutex);
#endif
	return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}

spBool spUnlockDrawMutexGtk(void)
{
#if defined(HAVE_PTHREAD) || defined(__CYGWIN32__)
    if (sp_toplevel != NULL
	&& SpTopLevelPart(sp_toplevel).thread_safe == SP_TRUE) {
#if defined(SP_GTK_USE_DRAW_MUTEX)
	pthread_mutex_unlock(&sp_draw_mutex);
#endif
	return SP_TRUE;
    }
#endif
    
    return SP_FALSE;
}
